using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using System;
using System.Text;
using System.Text.RegularExpressions;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.IO;
using MeshVR;
using MVR.FileManagementSecure;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;
using MVR;

// VAMStory v1.17
//
// Dialog / narration system for VAM

namespace VAMStoryPlugin
{
	public partial class VAMStory : MVRScript
    {
	    // Strings! (yes I know you like that)
	    private static string PLUGIN_PATH;
	    private static string BACKUP_PATH = "Saves\\PluginData\\vamstory\\backups";
	    private static string UI_BUNDLE_PATH = "/vamstory_assets.uibundle";
	    private static string SAVE_PREFIX = "Dial_";
	    private static string OBJ_MAINDIALOG_BTN = "VAMS_MainDialogBtn";
	    private static string BACKGROUNDS_PATH = "assets/vamstory/textures/backgrounds/";
	    private static string FONTS_PATH = "assets/vamstory/fonts/";
	    private static string SOUNDS_PATH = "assets/vamstory/sounds/";
	    private static string MATERIALS_PATH = "assets/vamstory/materials/materials_choices/";
	    
	    private string lastFileExplorerPath = "Custom";
	    
		// Atoms, gameobjects, components, assets and assets lists
		protected Camera VAMSCamera;
		protected GameObject VAMSUIGO;

		protected Transform TRStoryUI;
		protected Transform TRMainPanel;
		protected Transform TRAudio;
		protected Transform TRDialogPanel;
		protected Transform TRAvatar;
		protected Transform TRAvatarImg;
		protected Transform TRMainDialogBtn;
		protected Transform TRMainDialogTxt;
		protected Transform TRAnswer01Btn;
		protected Transform TRAnswer01Txt;
		protected Transform TRAnswer02Btn;
		protected Transform TRAnswer02Txt;
		protected Transform TRAnswer03Btn;
		protected Transform TRAnswer03Txt;
		protected Transform TRHideUIGrp;
		protected Transform TRHideUIBtn;
		protected Transform TRShowUIGrp;
		protected Transform TRShowUIBtn;

		protected Canvas CANStoryUI;
		protected AudioSource ASAudio;
		protected CanvasGroup CGDialogPanel;
		protected VerticalLayoutGroup VLGAvatar;
		protected RawImage IMGAvatar;
		protected Button BTNAvatar;
		protected VerticalLayoutGroup VLGMainDialog;
		protected Button BTNMainDialog;
		protected Text TXTMainDialog;
		protected Image IMGMainDialog;
		protected VerticalLayoutGroup VLGAnswer01;
		protected Button BTNAnswer01;
		protected Text TXTAnswer01;
		protected Image IMGAnswer01;
		protected VerticalLayoutGroup VLGAnswer02;
		protected Button BTNAnswer02;
		protected Text TXTAnswer02;
		protected Image IMGAnswer02;
		protected VerticalLayoutGroup VLGAnswer03;
		protected Button BTNAnswer03;
		protected Text TXTAnswer03;
		protected Image IMGAnswer03;
		protected Button BTNHideUI;
		protected RawImage IMGHideUI;
		protected Button BTNShowUI;
		protected RawImage IMGShowUI;

		protected Atom ActionsAtom;
		protected Transform TRStoryActionsUI;
		protected Transform TRStoryActionsMainPanel;
		protected Transform TRStoryActionsHideGroup;
		protected Transform TRStoryActionsShowGroup;
		protected CanvasGroup CGStoryActionsMainPanel;
		
		protected Texture2D tmpTexture;
		protected Dictionary<string, Texture2D> DialogAvatars = new Dictionary<String, Texture2D>();
		protected List<string> backgroundsChoices = new List<String>();
		protected List<Sprite> backgroundsSprites = new List<Sprite>();
		protected List<string> fontsChoices = new List<String>();
		protected List<Font> fontsFiles = new List<Font>();
		protected List<string> soundChoices = new List<String>();
		protected List<AudioClip> soundClips = new List<AudioClip>();
		protected List<string> materialChoices = new List<String>() {"None"};
		protected List<Material> materialList = new List<Material>();

		// Actions
		protected JSONStorableAction nextDialogAction;
		protected JSONStorableAction showDialogUI;
		protected JSONStorableAction hideDialogUI;
		protected JSONStorableAction clearActions;
		protected JSONStorableAction executeStoryTrigger;
		protected JSONStorableStringChooser A_PlayStorySound;
		
		// Static Storables
		protected EventTrigger GlobalOptStoryTrigger;
		protected JSONStorableBool GlobalOptUIFollowCamera;
		protected JSONStorableFloat GlobalOptUIFollowMinDist;
		protected JSONStorableFloat GlobalOptUIFollowMaxDist;
		protected JSONStorableFloat GlobalOptUIFollowVertOffset;
		protected JSONStorableFloat GlobalOptUIFollowHorOffset;
		protected JSONStorableBool GlobalOptUIRotationDisabled;
		protected JSONStorableBool GlobalOptUIConstraintToY;
		protected JSONStorableBool GlobalOptUIHideToggle;
		protected JSONStorableFloat GlobalOptUISize;
		protected JSONStorableStringChooser GlobalOptSoundOnShow;
		protected JSONStorableStringChooser GlobalOptSoundOnClick;
		protected JSONStorableFloat GlobalOptSoundOnShowVolume;
		protected JSONStorableFloat GlobalOptSoundOnClickVolume;
		protected JSONStorableBool GlobalOptLastGroupDialogClose;
		
		// Text options
		protected JSONStorableStringChooser GlobalOptUIAlignment;
		protected JSONStorableFloat GlobalOptFontSize;
		protected JSONStorableFloat GlobalOptLineHeight;
		protected JSONStorableStringChooser GlobalOptFont;
		protected JSONStorableFloat GlobalOptTxtPadT;
		protected JSONStorableFloat GlobalOptTxtPadB;
		protected JSONStorableFloat GlobalOptTxtPadL;
		protected JSONStorableFloat GlobalOptTxtPadR;
		
		// Appearance
		protected JSONStorableStringChooser GlobalOptBg;
		protected JSONStorableColor GlobalOptBgColor;
		protected JSONStorableFloat GlobalOptBgOpacity;
		protected JSONStorableFloat GlobalOptBgTransitionSpeed;
		protected JSONStorableColor GlobalOptBgOverColor;
		protected JSONStorableFloat GlobalOptBgOverOpacity;
		protected JSONStorableColor GlobalOptBgClickColor;
		protected JSONStorableFloat GlobalOptBgClickOpacity;
		protected JSONStorableColor GlobalOptTxtColor;
		protected JSONStorableColor GlobalOptAnswersBgColor;
		protected JSONStorableFloat GlobalOptAnswersBgOpacity;
		protected JSONStorableFloat GlobalOptAnswersBgTransitionSpeed;
		protected JSONStorableColor GlobalOptAnswersBgOverColor;
		protected JSONStorableFloat GlobalOptAnswersBgOverOpacity;
		protected JSONStorableColor GlobalOptAnswersBgClickColor;
		protected JSONStorableFloat GlobalOptAnswersBgClickOpacity;
		protected JSONStorableColor GlobalOptAnswersTxtColor;
		protected JSONStorableStringChooser GlobalOptMaterial;
		protected JSONStorableFloat GlobalOptMatDetSSX;
		protected JSONStorableFloat GlobalOptMatDetSSY;
		protected JSONStorableFloat GlobalOptMatSubDetOpacity;
		protected JSONStorableFloat GlobalOptMatRefraction;
		protected JSONStorableColor GlobalOptMatDetColorMainDialog;
		protected JSONStorableColor GlobalOptMatDetColorAnswers;
		
		// The usual help section
		protected JSONStorableString helpText;
		
		// Dialog UI storables (or input fields)
		public JSONStorableString DialogTextUI;
		protected InputField DialogTextInputFieldUI;
		public JSONStorableFloat DialogDelayUI;
		public JSONStorableStringChooser DialogAvatarPositionUI;
		public JSONStorableUrl DialogAvatarUrlUI;
		
		// Group UI storables (or input fields)
		public JSONStorableString GroupNameUI;
		public InputField GroupNameUIInputField;
		public JSONStorableBool GroupHideMainDialogUI;
		public JSONStorableBool GroupAnswer01EnabledUI;
		public JSONStorableString GroupAnswer01TxtUI;
		public InputField GroupAnswer01TxtUIInputField;
		public JSONStorableBool GroupAnswer02EnabledUI;
		public JSONStorableString GroupAnswer02TxtUI;
		public InputField GroupAnswer02TxtUIInputField;
		public JSONStorableBool GroupAnswer03EnabledUI;
		public JSONStorableString GroupAnswer03TxtUI;
		public InputField GroupAnswer03TxtUIInputField;
		public JSONStorableStringChooser GroupStoryActionsUI;

		// VAMS External values
		public JSONStorableBool VAMSValue_Ready;
		
		// Groups and dialogs
		protected List<StoryGroup> storyGroups = new List<StoryGroup>();
		protected Dictionary<int, string> groupsIdsToNames = new Dictionary<int, string>();
		protected StoryGroup activeGroup
		{
			get
			{
				return storyGroups.ElementAtOrDefault(activeGroupId);
			}
		}
		protected int activeGroupId
		{
			get
			{
				return groupsIdsToNames.FirstOrDefault(el => el.Value == GroupSelector.val).Key;
			}
		}
		protected StoryDialog activeDialog
		{
			get
			{
				return storyGroups.ElementAtOrDefault(activeGroupId).GetDialogID(int.Parse(DialogSelector.val));
			}
		}
		
		protected List<EventTrigger> _timedTriggers = new List<EventTrigger>();
		
		// Plugin data (due to the fact that the UI is instantiated after init, I need to grab my data and restore manually
		protected JSONClass dialogsDatas { get; set; }
		
		// Dynamic UI
		protected List<UIDynamic> globalControlsUIs = new List<UIDynamic>();
		protected List<UIDynamic> globalOptionsUIs = new List<UIDynamic>();
		protected List<UIDynamic> globalTextOptUIs = new List<UIDynamic>();
		protected List<UIDynamic> globalColorsOptUIs = new List<UIDynamic>();
		protected List<UIDynamic> globalOptionsFollowUIs = new List<UIDynamic>();
		protected List<UIDynamic> helpUIs = new List<UIDynamic>();
		
		protected List<UIDynamic> dialogsComponentsUI = new List<UIDynamic>();
		protected List<UIDynamic> groupComponentsUI = new List<UIDynamic>();
		protected List<UIDynamic> ignoreLayoutUI = new List<UIDynamic>();
		
		// Global UI components
		protected JSONStorableStringChooser RightUISelector;
		protected List<string> uisChoices = new List<string> {"Group/Dialog Control", "Group options", "Global options", "Text options", "Appearance options", "Help"};
		protected JSONStorableStringChooser GroupSelector;
		protected List<string> groupsChoices;
		protected JSONStorableStringChooser DialogSelector;
		protected List<string> dialogsChoices;
		protected JSONStorableStringChooser CopySelector;
		protected List<string> copyChoices = new List<String>();
		protected List<string> alignmentChoices = new List<string> {"Left", "Right"};
		protected List<string> avatarPositionChoices = new List<string> {"Left", "Right"};

		protected List<string> actionsChoices = new List<String>() {"None"};

		protected UIDynamicTextField dialogTitle;
		protected UIDynamicTextField RightUIMainTitle;
		protected UIDynamicTextField helpWindow;
		
		protected UIDynamicButton positionSaveButton;
		protected UIDynamicButton positionClearButton;
		
		// A bit of colors
		protected Color unsavedPositionColor = new Color(0.5f,0.5f,0.5f,1f);
		protected Color savedPositionColor = new Color(0.1f,1f,0.1f,1f);
		
		// "Triggers / Actions"
		protected JSONStorableStringChooser StartDialogGroup;
		protected List<string> startDialogChoices;
		
		// Counters for the dialog progress
		protected int CurrentGroupID = 0;
		protected int CurrentDialogID = 0;

		// Assetbundle
		private Request bundleRequest = null;

		// Init, security flags and values
		protected bool _isSceneLoading { get { return SuperController.singleton.isLoading; } }
		public bool isSceneFrozen { get { return SuperController.singleton.freezeAnimation; } }
		
		private bool isBundleLoaded = false;
		private bool isInit = false;
		private bool _isMovingGrpOrDialog = false;
		private bool _allVAMSReady = false;
		private float _vamsReadyCheckPeriod = 0.5f;
		

		public override void Init()
        {			 
            try
            {
                if (containingAtom.type != "Empty")
                {
                    SuperController.LogError("Please add VAMStory on an Empty Atom");
                    return;
                }

                PLUGIN_PATH = GetPluginPath(this);

				VAMSCamera = SuperController.singleton.centerCameraTarget.targetCamera;
				
				// Backup folder
				FileManagerSecure.CreateDirectory(BACKUP_PATH);
				StartCoroutine("AutoBackup");
               
                // ******* CREATING MAIN TRIGGERS/ACTIONS ********
                // They are here to be displayed on TOP of every other JSONStorables
                JSONStorableAction fakeFuncUseBelow = new JSONStorableAction("- - - - Use these functions below ↓ - - - - -", () => {});
                RegisterAction(fakeFuncUseBelow);
                
                startDialogChoices = new List<string>();
                StartDialogGroup = new JSONStorableStringChooser("Start Dialog Group ID", startDialogChoices, "", "Start Dialog Group ID") {isStorable=false,isRestorable=false};
                StartDialogGroup.setCallbackFunction += (val) => { OnStartDialogGroup(val); };
                RegisterStringChooser(StartDialogGroup);
                
                nextDialogAction = new JSONStorableAction("Next Dialog", () => { OnClickMainButton(); });
                RegisterAction(nextDialogAction);
                
                showDialogUI = new JSONStorableAction("Show Dialog UI", () => { ToggleStoryUI(true); });
                RegisterAction(showDialogUI);
                
                hideDialogUI = new JSONStorableAction("Hide Dialog UI", () => { ToggleStoryUI(false); });
                RegisterAction(hideDialogUI);
                
                clearActions = new JSONStorableAction("Clear Actions", () => { ClearActionsUI(); });
                RegisterAction(clearActions);
	                
                executeStoryTrigger = new JSONStorableAction("Execute Story Trigger", () => { GlobalOptStoryTrigger.Trigger(); });
                RegisterAction(executeStoryTrigger);
                
                A_PlayStorySound = new JSONStorableStringChooser("Play Sound", soundChoices, "", "Play Sound") {isStorable=false,isRestorable=false};
                A_PlayStorySound.setCallbackFunction += (val) => { OnPlayStorySound( val ); };
                RegisterStringChooser(A_PlayStorySound);

                VAMSValue_Ready = new JSONStorableBool("VAMSValue_Ready", false) {isStorable=false,isRestorable=false};
                RegisterBool(VAMSValue_Ready);
                
                // ******* CREATING MAIN UI *********************************
                // Temporary vars
                HSVColor hsvcDefault = HSVColorPicker.RGBToHSV(1f, 1f, 1f);
                HSVColor hsvcDefaultOverClick = HSVColorPicker.RGBToHSV(0.55f, 0.55f, 0.55f);
                HSVColor hsvcTxtDefault = HSVColorPicker.RGBToHSV(0.2f, 0.2f, 0.2f);
                UIDynamicTextField tmpTextfield;
                UIDynamicPopup tmpPopup;
                UIDynamicSlider tmpSlider;
                UIDynamicColorPicker tmpColor;
                UIDynamicButton tmpButton;
                UIDynamicToggle tmpToggle;
                UIDynamic tmpSpacer;
                
                // ******* TITLE **********************
                createStaticDescriptionText("VAMStory Title","<color=#000><size=35><b>VAM STORY</b></size></color>",true,40,TextAnchor.MiddleRight);
                
                RightUISelector = new JSONStorableStringChooser("Right UI Selector", uisChoices, "Group/Dialog Control", "Show") {isStorable=false,isRestorable=false};
                RightUISelector.setCallbackFunction += (val) => { OnChangeRightUISelection(); };
                UIDynamicPopup RUISelsp = CreateScrollablePopup(RightUISelector, true);
                RUISelsp.labelWidth = 95f;

                tmpSpacer = CreateSpacer(true);
                setupSpacer(tmpSpacer, 25f);                                
                
                RightUIMainTitle = createStaticDescriptionText("Right UI Main Title","<color=#000><size=35><b>GROUP/DIALOG CONTROLS</b></size></color>",true,55,TextAnchor.MiddleRight);
                
                GlobalOptStoryTrigger = new EventTrigger(this, "Global_Opt_StoryTrigger");
                
                // ****** GLOBAL OPTIONS *********
                tmpButton = CreateButton("Edit story trigger",true);
                tmpButton.button.onClick.AddListener( () => { GlobalOptStoryTrigger.OpenPanelActionStart(); } );
                setButtonColor(tmpButton, new Color(0.1f, 0.5f, 0.8f, 1f));
                setButtonTextColor(tmpButton, new Color(1f, 1f, 1f, 1f));
                globalOptionsUIs.Add((UIDynamic) tmpButton);
                
                GlobalOptUIFollowCamera = new JSONStorableBool("Global_Opt_UIFollow", false);
                tmpToggle = CreateToggle(GlobalOptUIFollowCamera,true);
                GlobalOptUIFollowCamera.setCallbackFunction += (val) => { OnToggleGlobalOptionsFollow(val); };
				tmpToggle.label = "UI follows camera";
				RegisterBool(GlobalOptUIFollowCamera);
				globalOptionsUIs.Add((UIDynamic) tmpToggle);
				
				GlobalOptUIFollowMinDist = new JSONStorableFloat("Global_Opt_UIFollowMinDist", 0.9f, 0.5f, 1f);
				tmpSlider = CreateSlider(GlobalOptUIFollowMinDist, true);
				tmpSlider.valueFormat = "F1";
				tmpSlider.label = "UI follow min distance";
				RegisterFloat(GlobalOptUIFollowMinDist);
				globalOptionsFollowUIs.Add((UIDynamic) tmpSlider);
				
				GlobalOptUIFollowMaxDist = new JSONStorableFloat("Global_Opt_UIFollowMaxDist", 2f, 1f, 3f);
				tmpSlider = CreateSlider(GlobalOptUIFollowMaxDist, true);
				tmpSlider.valueFormat = "F1";
				tmpSlider.label = "UI follow max distance";
				RegisterFloat(GlobalOptUIFollowMaxDist);
				globalOptionsFollowUIs.Add((UIDynamic) tmpSlider);
				
				GlobalOptUIFollowVertOffset = new JSONStorableFloat("Global_Opt_UIFollowVertOffset", 0.05f, -0.15f, 0.15f);
				tmpSlider = CreateSlider(GlobalOptUIFollowVertOffset, true);
				tmpSlider.valueFormat = "F2";
				tmpSlider.label = "UI follow vertical offset";
				RegisterFloat(GlobalOptUIFollowVertOffset);
				globalOptionsFollowUIs.Add((UIDynamic) tmpSlider);
				
				GlobalOptUIFollowHorOffset = new JSONStorableFloat("Global_Opt_UIFollowHorOffset", 0.2f, -0.3f, 0.3f);
				tmpSlider = CreateSlider(GlobalOptUIFollowHorOffset, true);
				tmpSlider.valueFormat = "F2";
				tmpSlider.label = "UI follow horizontal offset";
				RegisterFloat(GlobalOptUIFollowHorOffset);
				globalOptionsFollowUIs.Add((UIDynamic) tmpSlider);
				
				GlobalOptUIRotationDisabled = new JSONStorableBool("Global_Opt_UIRotationDisabled", false);
				tmpToggle = CreateToggle(GlobalOptUIRotationDisabled,true);
				tmpToggle.label = "UI rotation disabled";
				RegisterBool(GlobalOptUIRotationDisabled);
				GlobalOptUIRotationDisabled.setCallbackFunction += (val) => { OnUIRotationDisabledChanged(val); };
				globalOptionsUIs.Add((UIDynamic) tmpToggle);
				
                GlobalOptUIConstraintToY = new JSONStorableBool("Global_Opt_UIConstraintToY", false);
                tmpToggle = CreateToggle(GlobalOptUIConstraintToY,true);
				tmpToggle.label = "UI rotation constrained to Y";
				RegisterBool(GlobalOptUIConstraintToY);
				globalOptionsUIs.Add((UIDynamic) tmpToggle);
				
				GlobalOptUIHideToggle = new JSONStorableBool("Global_Opt_UIHideToggle", false);
				tmpToggle = CreateToggle(GlobalOptUIHideToggle,true);
				tmpToggle.label = "UI hide toggle button";
				RegisterBool(GlobalOptUIHideToggle);
				GlobalOptUIHideToggle.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
				globalOptionsUIs.Add((UIDynamic) tmpToggle);

                GlobalOptUISize = new JSONStorableFloat("Global_Opt_UISize", 1f, 0.5f, 2f);
                tmpSlider = CreateSlider(GlobalOptUISize, true);
                tmpSlider.valueFormat = "F1";
                tmpSlider.label = "UI Size";
                RegisterFloat(GlobalOptUISize);
                GlobalOptUISize.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                globalOptionsUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptSoundOnShow = new JSONStorableStringChooser("Global_Opt_SndShow", soundChoices, "click_05", "Sound On Show");
                RegisterStringChooser(GlobalOptSoundOnShow);
                GlobalOptSoundOnShow.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpPopup = CreateScrollablePopup(GlobalOptSoundOnShow, true);
                tmpPopup.labelWidth = 120f;
                globalOptionsUIs.Add((UIDynamic) tmpPopup);
                
                GlobalOptSoundOnClick = new JSONStorableStringChooser("Global_Opt_SndClick", soundChoices, "click_02", "Sound On Click");
                RegisterStringChooser(GlobalOptSoundOnClick);
                GlobalOptSoundOnClick.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpPopup = CreateScrollablePopup(GlobalOptSoundOnClick, true);
                tmpPopup.labelWidth = 120f;
                globalOptionsUIs.Add((UIDynamic) tmpPopup);
                
                GlobalOptSoundOnShowVolume = new JSONStorableFloat("Global_Opt_SndShowVol", 1f, 0f, 1f);
                tmpSlider = CreateSlider(GlobalOptSoundOnShowVolume, true);
                tmpSlider.valueFormat = "F3";
                tmpSlider.label = "On show sound volume";
                RegisterFloat(GlobalOptSoundOnShowVolume);
                globalOptionsUIs.Add((UIDynamic) tmpSlider);

                GlobalOptSoundOnClickVolume = new JSONStorableFloat("Global_Opt_SndClickVol", 1f, 0f, 1f);
                tmpSlider = CreateSlider(GlobalOptSoundOnClickVolume, true);
                tmpSlider.valueFormat = "F3";
                tmpSlider.label = "On click sound volume";
                RegisterFloat(GlobalOptSoundOnClickVolume);
                globalOptionsUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptLastGroupDialogClose = new JSONStorableBool("Global_Opt_LastGroupDialogClose", false);
                tmpToggle = CreateToggle(GlobalOptLastGroupDialogClose,true);
                tmpToggle.label = "Close when clicking last dialog";
                RegisterBool(GlobalOptLastGroupDialogClose);
                globalOptionsUIs.Add((UIDynamic) tmpToggle);
                
                // Backups
                tmpSpacer = CreateSpacer(true);
                globalOptionsUIs.Add((UIDynamic) tmpSpacer);
                setupSpacer(tmpSpacer, 25f);                                

                tmpTextfield = createStaticDescriptionText("Backups Title","<color=#000><size=27><b>BACKUPS</b></size></color>",true,35,TextAnchor.MiddleRight);
                globalOptionsUIs.Add((UIDynamic) tmpTextfield);
                
                tmpButton = CreateButton("Restore backup",true);
                tmpButton.button.onClick.AddListener(() => { LoadBackup(); } );
                setButtonColor(tmpButton, new Color(0.1f, 0.5f, 0.7f, 1f));
                setButtonTextColor(tmpButton, new Color(1f, 1f, 1f, 1f));
                globalOptionsUIs.Add((UIDynamic) tmpButton);
                
                tmpButton = CreateButton("Save manual backup",true);
                tmpButton.button.onClick.AddListener(() => { SaveBackup(2); } );
                setButtonColor(tmpButton, new Color(0.1f, 0.7f, 0.4f, 1f));
                setButtonTextColor(tmpButton, new Color(1f, 1f, 1f, 1f));
                globalOptionsUIs.Add((UIDynamic) tmpButton);
                
                // ****** TEXT OPTIONS *********
                GlobalOptUIAlignment = new JSONStorableStringChooser("Global_Opt_UIAlignment", alignmentChoices, "Right", "UI Alignment");
                RegisterStringChooser(GlobalOptUIAlignment);
                GlobalOptUIAlignment.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpPopup = CreateScrollablePopup(GlobalOptUIAlignment, true);
                tmpPopup.labelWidth = 220f;
                globalTextOptUIs.Add((UIDynamic) tmpPopup);
                
                GlobalOptFontSize = new JSONStorableFloat("Global_Opt_FontSize", 40f, 30f, 60f, false);
                tmpSlider = CreateSlider(GlobalOptFontSize, true);
                tmpSlider.valueFormat = "F0";
                tmpSlider.label = "Font Size";
                RegisterFloat(GlobalOptFontSize);
                GlobalOptFontSize.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                globalTextOptUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptLineHeight = new JSONStorableFloat("Global_Opt_LineHeight", 1f, 0.1f, 10f);
                tmpSlider = CreateSlider(GlobalOptLineHeight, true);
                tmpSlider.valueFormat = "F2";
                tmpSlider.label = "Line Height";
                RegisterFloat(GlobalOptLineHeight);
                GlobalOptLineHeight.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                globalTextOptUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptFont = new JSONStorableStringChooser("Global_Opt_Font", fontsChoices, "arial", "Font");
                RegisterStringChooser(GlobalOptFont);
                GlobalOptFont.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpPopup = CreateScrollablePopup(GlobalOptFont, true);
                tmpPopup.labelWidth = 120f;
                globalTextOptUIs.Add((UIDynamic) tmpPopup);
                
                GlobalOptTxtPadT = new JSONStorableFloat("Global_Opt_TxtPadT", 15f, 0f, 150f, false);
                tmpSlider = CreateSlider(GlobalOptTxtPadT, true);
                tmpSlider.valueFormat = "F1";
                tmpSlider.label = "Text padding Top";
                RegisterFloat(GlobalOptTxtPadT);
                GlobalOptTxtPadT.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                globalTextOptUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptTxtPadB = new JSONStorableFloat("Global_Opt_TxtPadB", 15f, 0f, 150f, false);
                tmpSlider = CreateSlider(GlobalOptTxtPadB, true);
                tmpSlider.valueFormat = "F1";
                tmpSlider.label = "Text padding Bottom";
                RegisterFloat(GlobalOptTxtPadB);
                GlobalOptTxtPadB.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                globalTextOptUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptTxtPadL = new JSONStorableFloat("Global_Opt_TxtPadL", 15f, 0f, 150f, false);
                tmpSlider = CreateSlider(GlobalOptTxtPadL, true);
                tmpSlider.valueFormat = "F1";
                tmpSlider.label = "Text padding Left";
                RegisterFloat(GlobalOptTxtPadL);
                GlobalOptTxtPadL.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                globalTextOptUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptTxtPadR = new JSONStorableFloat("Global_Opt_TxtPadR", 15f, 0f, 150f, false);
                tmpSlider = CreateSlider(GlobalOptTxtPadR, true);
                tmpSlider.valueFormat = "F1";
                tmpSlider.label = "Text padding Right";
                RegisterFloat(GlobalOptTxtPadR);
                GlobalOptTxtPadR.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                globalTextOptUIs.Add((UIDynamic) tmpSlider);
                
                
                // Colors and materials
                GlobalOptBg = new JSONStorableStringChooser("Global_Opt_Bg", backgroundsChoices, "Default", "Style");
                GlobalOptBg.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpPopup = CreateScrollablePopup(GlobalOptBg, true);
                tmpPopup.labelWidth = 120f;
                RegisterStringChooser(GlobalOptBg);
                globalColorsOptUIs.Add((UIDynamic) tmpPopup);
                
                GlobalOptBgColor = new JSONStorableColor("Global_Opt_BgColor", hsvcDefault);
                GlobalOptBgColor.setJSONCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpColor = CreateColorPicker(GlobalOptBgColor, true);
                tmpColor.label = "Dialog background color";
                RegisterColor(GlobalOptBgColor);
                globalColorsOptUIs.Add((UIDynamic) tmpColor);
				
                GlobalOptBgOpacity = new JSONStorableFloat("Global_Opt_BgOpacity", 1f, 0f, 1f);
                GlobalOptBgOpacity.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpSlider = CreateSlider(GlobalOptBgOpacity, true);
                tmpSlider.label = "Dialog background Opacity";
                RegisterFloat(GlobalOptBgOpacity);
                globalColorsOptUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptBgTransitionSpeed = new JSONStorableFloat("Global_Opt_BgTransitionSpeed", 0.1f, 0f, 0.5f);
                GlobalOptBgTransitionSpeed.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpSlider = CreateSlider(GlobalOptBgTransitionSpeed, true);
                tmpSlider.label = "Colors transition speed";
                RegisterFloat(GlobalOptBgTransitionSpeed);
                globalColorsOptUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptBgOverColor = new JSONStorableColor("Global_Opt_BgOverColor", hsvcDefaultOverClick);
                GlobalOptBgOverColor.setJSONCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpColor = CreateColorPicker(GlobalOptBgOverColor, true);
                tmpColor.label = "Dialog background color (over)";
                RegisterColor(GlobalOptBgOverColor);
                globalColorsOptUIs.Add((UIDynamic) tmpColor);
				
                GlobalOptBgOverOpacity = new JSONStorableFloat("Global_Opt_BgOverOpacity", 1f, 0f, 1f);
                GlobalOptBgOverOpacity.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpSlider = CreateSlider(GlobalOptBgOverOpacity, true);
                tmpSlider.label = "Dialog background opacity (over)";
                RegisterFloat(GlobalOptBgOverOpacity);
                globalColorsOptUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptBgClickColor = new JSONStorableColor("Global_Opt_BgClickColor", hsvcDefaultOverClick);
                GlobalOptBgClickColor.setJSONCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpColor = CreateColorPicker(GlobalOptBgClickColor, true);
                tmpColor.label = "Dialog background color (click)";
                RegisterColor(GlobalOptBgClickColor);
                globalColorsOptUIs.Add((UIDynamic) tmpColor);
				
                GlobalOptBgClickOpacity = new JSONStorableFloat("Global_Opt_BgClickOpacity", 1f, 0f, 1f);
                GlobalOptBgClickOpacity.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpSlider = CreateSlider(GlobalOptBgClickOpacity, true);
                tmpSlider.label = "Dialog background opacity (click)";
                RegisterFloat(GlobalOptBgClickOpacity);
                globalColorsOptUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptTxtColor = new JSONStorableColor("Global_Opt_TxtColor", hsvcTxtDefault);
                GlobalOptTxtColor.setJSONCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpColor = CreateColorPicker(GlobalOptTxtColor, true);
                tmpColor.label = "Dialog text Color";
                RegisterColor(GlobalOptTxtColor);
                globalColorsOptUIs.Add((UIDynamic) tmpColor);
                
                GlobalOptAnswersBgColor = new JSONStorableColor("Global_Opt_AnswersBgColor", hsvcDefault);
                GlobalOptAnswersBgColor.setJSONCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpColor = CreateColorPicker(GlobalOptAnswersBgColor, true);
                tmpColor.label = "Answers background color";
                RegisterColor(GlobalOptAnswersBgColor);
                globalColorsOptUIs.Add((UIDynamic) tmpColor);
				
                GlobalOptAnswersBgOpacity = new JSONStorableFloat("Global_Opt_AnswersBgOpacity", 1f, 0f, 1f);
                GlobalOptAnswersBgOpacity.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpSlider = CreateSlider(GlobalOptAnswersBgOpacity, true);
                tmpSlider.label = "Answers background Opacity";
                RegisterFloat(GlobalOptAnswersBgOpacity);
                globalColorsOptUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptAnswersBgOverColor = new JSONStorableColor("Global_Opt_AnswersBgOverColor", hsvcDefaultOverClick);
                GlobalOptAnswersBgOverColor.setJSONCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpColor = CreateColorPicker(GlobalOptAnswersBgOverColor, true);
                tmpColor.label = "Answers background color (over)";
                RegisterColor(GlobalOptAnswersBgOverColor);
                globalColorsOptUIs.Add((UIDynamic) tmpColor);
				
                GlobalOptAnswersBgOverOpacity = new JSONStorableFloat("Global_Opt_AnswersBgOverOpacity", 1f, 0f, 1f);
                GlobalOptAnswersBgOverOpacity.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpSlider = CreateSlider(GlobalOptAnswersBgOverOpacity, true);
                tmpSlider.label = "Answers background opacity (over)";
                RegisterFloat(GlobalOptAnswersBgOverOpacity);
                globalColorsOptUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptAnswersBgClickColor = new JSONStorableColor("Global_Opt_AnswersBgClickColor", hsvcDefaultOverClick);
                GlobalOptAnswersBgClickColor.setJSONCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpColor = CreateColorPicker(GlobalOptAnswersBgClickColor, true);
                tmpColor.label = "Answers background color (click)";
                RegisterColor(GlobalOptAnswersBgClickColor);
                globalColorsOptUIs.Add((UIDynamic) tmpColor);
				
                GlobalOptAnswersBgClickOpacity = new JSONStorableFloat("Global_Opt_AnswersBgClickOpacity", 1f, 0f, 1f);
                GlobalOptAnswersBgClickOpacity.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpSlider = CreateSlider(GlobalOptAnswersBgClickOpacity, true);
                tmpSlider.label = "Answers background opacity (click)";
                RegisterFloat(GlobalOptAnswersBgClickOpacity);
                globalColorsOptUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptAnswersTxtColor = new JSONStorableColor("Global_Opt_AnswersTxtColor", hsvcTxtDefault);
                GlobalOptAnswersTxtColor.setJSONCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpColor = CreateColorPicker(GlobalOptAnswersTxtColor, true);
                tmpColor.label = "Answers text Color";
                RegisterColor(GlobalOptAnswersTxtColor);
                globalColorsOptUIs.Add((UIDynamic) tmpColor);

                tmpSpacer = CreateSpacer(true);
                globalColorsOptUIs.Add((UIDynamic) tmpSpacer);
                setupSpacer(tmpSpacer, 25f); 
                
                GlobalOptMaterial = new JSONStorableStringChooser("Global_Opt_Material", materialChoices, "None", "Material");
                RegisterStringChooser(GlobalOptMaterial);
                GlobalOptMaterial.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpPopup = CreateScrollablePopup(GlobalOptMaterial, true);
                tmpPopup.labelWidth = 160f;
                globalColorsOptUIs.Add((UIDynamic) tmpPopup);
                
                GlobalOptMatDetSSX = new JSONStorableFloat("Global_Opt_MatDetSSX", 0.05f, -2f, 2f);
                tmpSlider = CreateSlider(GlobalOptMatDetSSX, true);
                tmpSlider.valueFormat = "F2";
                tmpSlider.label = "Details Scroll Speed X";
                RegisterFloat(GlobalOptMatDetSSX);
                GlobalOptMatDetSSX.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                globalColorsOptUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptMatDetSSY = new JSONStorableFloat("Global_Opt_MatDetSSY", 0.05f, -2f, 2f);
                tmpSlider = CreateSlider(GlobalOptMatDetSSY, true);
                tmpSlider.valueFormat = "F2";
                tmpSlider.label = "Details Scroll Speed Y";
                RegisterFloat(GlobalOptMatDetSSY);
                GlobalOptMatDetSSY.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                globalColorsOptUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptMatSubDetOpacity = new JSONStorableFloat("Global_Opt_MatSubDetOpacity", 0f, 0f, 1f);
                tmpSlider = CreateSlider(GlobalOptMatSubDetOpacity, true);
                tmpSlider.valueFormat = "F1";
                tmpSlider.label = "Sub-details Opacity";
                RegisterFloat(GlobalOptMatSubDetOpacity);
                GlobalOptMatSubDetOpacity.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                globalColorsOptUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptMatRefraction = new JSONStorableFloat("Global_Opt_MatRefraction", 0.5f, 0f, 1f);
                tmpSlider = CreateSlider(GlobalOptMatRefraction, true);
                tmpSlider.valueFormat = "F2";
                tmpSlider.label = "Details Refraction";
                RegisterFloat(GlobalOptMatRefraction);
                GlobalOptMatRefraction.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                globalColorsOptUIs.Add((UIDynamic) tmpSlider);

                GlobalOptMatDetColorMainDialog = new JSONStorableColor("Global_Opt_MatDetColorMainDialog", hsvcDefault);
                GlobalOptMatDetColorMainDialog.setJSONCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpColor = CreateColorPicker(GlobalOptMatDetColorMainDialog, true);
                tmpColor.label = "Main dialog details Color";
                RegisterColor(GlobalOptMatDetColorMainDialog);
                globalColorsOptUIs.Add((UIDynamic) tmpColor);
                
                GlobalOptMatDetColorAnswers = new JSONStorableColor("Global_Opt_MatDetColorAnswers", hsvcDefault);
                GlobalOptMatDetColorAnswers.setJSONCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpColor = CreateColorPicker(GlobalOptMatDetColorAnswers, true);
                tmpColor.label = "Answers details color";
                RegisterColor(GlobalOptMatDetColorAnswers);
                globalColorsOptUIs.Add((UIDynamic) tmpColor);
                
				// Fake action	                
                JSONStorableAction fakeFuncDontUseBelow = new JSONStorableAction("- - - - DON'T Use these functions below ↓ - - - - -", () => {});
                RegisterAction(fakeFuncDontUseBelow);
                
                // ******* GROUPS AND DIALOGS CONTROLS **********************
                
                // Test current group button
                UIDynamicButton testCurrentGroupBtn = CreateButton("Test current group", true);
                testCurrentGroupBtn.button.onClick.AddListener( () => { OnTestCurrentGroup(); } );
                setButtonColor(testCurrentGroupBtn, new Color(0.3f, 0.5f, 0.8f, 1f));
                setButtonTextColor(testCurrentGroupBtn, new Color(1f, 1f, 1f, 1f));
                globalControlsUIs.Add((UIDynamic) testCurrentGroupBtn);
                setupButtonWithLayout(testCurrentGroupBtn, 250f);

                // Test displayed dialog
                UIDynamicButton editDisplayedDialogBtn = CreateButton("Edit displayed", true);
                editDisplayedDialogBtn.button.onClick.AddListener( () => { OnEditDisplayedDialog(); } );
                setButtonColor(editDisplayedDialogBtn, new Color(0.1f, 0.6f, 0.7f, 1f));
                setButtonTextColor(editDisplayedDialogBtn, new Color(1f, 1f, 1f, 1f));
                globalControlsUIs.Add((UIDynamic) editDisplayedDialogBtn);
                ignoreLayoutUI.Add((UIDynamic) editDisplayedDialogBtn);
                setupButtonWithoutLayout(editDisplayedDialogBtn, 250f, new Vector2(270f, -315f));

                // Dialog choices
                dialogsChoices = new List<string>();
                DialogSelector = new JSONStorableStringChooser("Dialog Line ID", dialogsChoices, "", "Dialog Line ID") {isStorable=false,isRestorable=false};
                DialogSelector.setCallbackFunction += (val) => { OnToggleDialogId(val); };
                UIDynamicPopup DSelsp = CreateScrollablePopup(DialogSelector, true);
                DSelsp.labelWidth = 250f;
                globalControlsUIs.Add((UIDynamic) DSelsp);
                
                UIDynamicButton newDialogBtn = CreateButton("New dialog", true);
                newDialogBtn.button.onClick.AddListener( () => { OnAddNewDialog(); } );
                setButtonColor(newDialogBtn, new Color(0.3f, 0.6f, 0.3f, 1f));
                setButtonTextColor(newDialogBtn, new Color(1f, 1f, 1f, 1f));
                globalControlsUIs.Add((UIDynamic) newDialogBtn);
                setupButtonWithLayout(newDialogBtn, 190f);
                
                UIDynamicButton deleteDialogBtn = CreateButton("Delete dialog", true);
                deleteDialogBtn.button.onClick.AddListener( () => { OnDeleteCurrentDialog(); } );
                setButtonColor(deleteDialogBtn, new Color(0.6f, 0.3f, 0.3f, 1f));
                setButtonTextColor(deleteDialogBtn, new Color(1f, 1f, 1f, 1f));
                globalControlsUIs.Add((UIDynamic) deleteDialogBtn);
                ignoreLayoutUI.Add((UIDynamic) deleteDialogBtn);
                setupButtonWithoutLayout(deleteDialogBtn, 190f, new Vector2(210, -495));
                
                UIDynamicButton moveUpDialogBtn = CreateButton("↑", true);
                moveUpDialogBtn.button.onClick.AddListener( () => { OnMoveCurrentDialog(false); } );
                setButtonColor(moveUpDialogBtn, new Color(0.1f, 0.5f, 0.6f, 1f));
                setButtonTextColor(moveUpDialogBtn, new Color(1f, 1f, 1f, 1f));
                globalControlsUIs.Add((UIDynamic) moveUpDialogBtn);
                ignoreLayoutUI.Add((UIDynamic) moveUpDialogBtn);
                setupButtonWithoutLayout(moveUpDialogBtn, 50f, new Vector2(410, -495));
                
                UIDynamicButton moveDownDialogBtn = CreateButton("↓", true);
                moveDownDialogBtn.button.onClick.AddListener( () => { OnMoveCurrentDialog(true); } );
                setButtonColor(moveDownDialogBtn, new Color(0.1f, 0.5f, 0.6f, 1f));
                setButtonTextColor(moveDownDialogBtn, new Color(1f, 1f, 1f, 1f));
                globalControlsUIs.Add((UIDynamic) moveDownDialogBtn);
                ignoreLayoutUI.Add((UIDynamic) moveDownDialogBtn);
                setupButtonWithoutLayout(moveDownDialogBtn, 50f, new Vector2(470, -495));

                groupsChoices = new List<string>();
                GroupSelector = new JSONStorableStringChooser("Dialog Group ID", groupsChoices, "", "Dialog Group ID") {isStorable=false,isRestorable=false};
                GroupSelector.setCallbackFunction += (val) => { OnToggleGroupId(val); };
                UIDynamicPopup GSsp = CreateScrollablePopup(GroupSelector, true);
                GSsp.labelWidth = 250f;
                globalControlsUIs.Add((UIDynamic) GSsp);
                
                UIDynamicButton newGroupBtn = CreateButton("New group", true);
                newGroupBtn.button.onClick.AddListener( () => { OnAddNewGroup(); } );
                setButtonColor(newGroupBtn, new Color(0.3f, 0.6f, 0.3f, 1f));
                setButtonTextColor(newGroupBtn, new Color(1f, 1f, 1f, 1f));
                globalControlsUIs.Add((UIDynamic) newGroupBtn);
                setupButtonWithLayout(newGroupBtn, 190f);
                
                UIDynamicButton deleteGroupBtn = CreateButton("Delete group", true);
                deleteGroupBtn.button.onClick.AddListener( () => { OnDeleteCurrentGroup(); } );
                setButtonColor(deleteGroupBtn, new Color(0.6f, 0.3f, 0.3f, 1f));
                setButtonTextColor(deleteGroupBtn, new Color(1f, 1f, 1f, 1f));
                globalControlsUIs.Add((UIDynamic) deleteGroupBtn);
                ignoreLayoutUI.Add((UIDynamic) deleteGroupBtn);
                setupButtonWithoutLayout(deleteGroupBtn, 190f, new Vector2(210, -675));
                
                UIDynamicButton moveUpGrpBtn = CreateButton("↑", true);
                moveUpGrpBtn.button.onClick.AddListener( () => { OnMoveCurrentGroup(false); } );
                setButtonColor(moveUpGrpBtn, new Color(0.1f, 0.5f, 0.6f, 1f));
                setButtonTextColor(moveUpGrpBtn, new Color(1f, 1f, 1f, 1f));
                globalControlsUIs.Add((UIDynamic) moveUpGrpBtn);
                ignoreLayoutUI.Add((UIDynamic) moveUpGrpBtn);
                setupButtonWithoutLayout(moveUpGrpBtn, 50f, new Vector2(410, -675));
                
                UIDynamicButton moveDownGrpBtn = CreateButton("↓", true);
                moveDownGrpBtn.button.onClick.AddListener( () => { OnMoveCurrentGroup(true); } );
                setButtonColor(moveDownGrpBtn, new Color(0.1f, 0.5f, 0.6f, 1f));
                setButtonTextColor(moveDownGrpBtn, new Color(1f, 1f, 1f, 1f));
                globalControlsUIs.Add((UIDynamic) moveDownGrpBtn);
                ignoreLayoutUI.Add((UIDynamic) moveDownGrpBtn);
                setupButtonWithoutLayout(moveDownGrpBtn, 50f, new Vector2(470, -675));
                
                tmpTextfield = createStaticDescriptionText("Dialog copy Title","<color=#000><size=27><b>COPY DIALOG</b></size></color>",true,35,TextAnchor.MiddleRight);
                globalControlsUIs.Add((UIDynamic) tmpTextfield);
                
                CopySelector = new JSONStorableStringChooser("Dialog Source", copyChoices, "", "Source/Target") {isStorable=false,isRestorable=false};
                UIDynamicPopup DSosp = CreateScrollablePopup(CopySelector, true);
                DSosp.labelWidth = 220f;
                globalControlsUIs.Add((UIDynamic) DSosp);
                
                UIDynamicButton copyDialogBtn = CreateButton("Copy", true);
                copyDialogBtn.button.onClick.AddListener(() => { OnCopyDialog(); });
                setButtonColor(copyDialogBtn, new Color(0.8f, 0.8f, 0.1f, 1f));
                globalControlsUIs.Add((UIDynamic) copyDialogBtn);
                setupButtonWithLayout(copyDialogBtn, 250f);
                
                UIDynamicButton transferDialogBtn = CreateButton("Transfer", true);
                transferDialogBtn.button.onClick.AddListener(() => { OnTransferDialog(); });
                setButtonColor(transferDialogBtn, new Color(0.8f, 0.5f, 0f, 1f));
                globalControlsUIs.Add((UIDynamic) transferDialogBtn);
                ignoreLayoutUI.Add((UIDynamic) transferDialogBtn);
                setupButtonWithoutLayout(transferDialogBtn, 250f, new Vector2(270, -905));
                
                // ******* HELP **********************
                helpText = new JSONStorableString("Help",
	                "<size=30><color=#333>" +
	                "This plugin helps you create story-driven scenes through dialogs.\n\n" +
	                "<b>Formatting tips:</b>\n" +
	                "You can format the text of dialogs or answers by using rich text snippets.\n" +
	                "<size=27>" +
	                "- <b>Bold:</b> <<i></i>b>Text</b<i></i>>\n" +
	                "- <b>Italic:</b> <<i></i>i>Text</i<i></i>>\n" +
	                "- <b>Size:</b> <<i></i>size=20>Text</size<i></i>>\n" +
	                "- <b>Color:</b> <<i></i>color=green>Text</color<i></i>>\n" +
	                "- <b>Color:</b> <<i></i>color=#00ffffff>Text</color<i></i>>\n" +
	                "</size>\n\n" +
	                "<b>PLUGIN HELP:</b>\n" +
	                "VAMStory revolves around groups containing one or more dialog lines.\n\n" +
	                "You can create a new group, define it's default position, and display it whenever you want.\n\n" +
	                "You can set a trigger when a group is displayed, a trigger when a dialog shows up and a trigger when the dialog is clicked.\n\n" +
	                "<color=#c81b0d><b>WARNING!: You need to keep VAMStory and all other VAMS plugins atoms always on. See hub's documentation for more information.</b></color>\n\n" +
	                "<b>GLOBAL OPTIONS</b>\n" +
	                "Click on \"Edit global options\" to define the story options. These options are global and define the behavior or appearance of the story system.\n\n" +
	                "You could if you want, change the appearance through dialog triggers by modifying these options on the fly.\n\n" +
	                "<b>Edit story trigger</b> allows you to edit the trigger executed when the story is loaded. It fires when the plugin is ready, or you can if you want, trigger it manually. It can be a way to start / reset your story.\n\n" +
	                "<b>Close when clicking last dialog</b> this will close the story UI when you click the last dialog of a group if enabled.\n\n" +
	                "<b>GROUP/DIALOG CONTROLS</b>\n" +
	                "These are the controls for your story content. It allows you to navigate through your groups and dialogs.\n\n" +
	                "Clicking <b>Test current group</b> will start the currently selected group in <b>Dialog Group ID</b>. Useful to test a dialog group.\n\n" +
	                "Clicking <b>Edit displayed</b> will automatically select the dialog in the plugin UI based on the one displayed in the scene.\n\n" +
	                "Changing the <b>Dialog Group ID</b> will display the selected group (showing up the <b>group options</b> at the bottom of the right column).\n\n" +
	                "Changing the <b>Dialog Line ID</b> will display the selected dialog for the current group (showing up the <b>dialog content</b> in the left column).\n\n" +
	                "<b>COPY/TRANSFER DIALOG</b>\n" +
	                "This is the way to copy or transfer a dialog inside the story.\n\n" +
	                "Select a dialog in the Source/Target dropdown. (the value means GroupID_DialogID)\n\n" +
	                "Then click either <b>Copy</b> or <b>Transfer</b>.\n\n" +
	                "<b>Copy</b> will grab the values of the selected dialog in the dropdown and copy its content to the current dialog. <b>Transfer</b> will send the current dialog values to the selected dialog in the dropdown.\n\n" +
	                "<b>GROUP OPTIONS</b>\n" +
	                "These are the options for the current group.\n\n" +
	                "Click on <b>save position</b> to store the UI position for this group. Everytime you will start the group through a trigger, VAMStory will move to this position. You just need to move the VAMStory atom at the place you want before clicking the button. If the button is green it means the position is already saved. You can click it again to overwrite it.\n\n" +
	                "<b>Edit Group Trigger</b> will allow you to add or modify the trigger executed when a group starts.\n\n" +
	                "You also have the answers options for this group. They will be displayed when the dialog reaches the last dialog of the group.\n\n" +
	                "You can enable up to three answers. Check the answer you want to enabled. Type a text for the answer, and define a trigger.\n\n" +
	                "Know that when a group reach the last dialog, the UI will not hide by itself. You will need to control it yourself by hiding it or starting another group. You can do that either in the answer triggers, or the dialog triggers.\n\n" +
	                "</color></size>"
                );
				
                helpWindow = CreateTextField(helpText, true);
                helpWindow.height = 850.0f;
                helpUIs.Add((UIDynamic) helpWindow);

                // Loading the assetbundle
                Request request = new AssetLoader.AssetBundleFromFileRequest {path = PLUGIN_PATH + UI_BUNDLE_PATH, callback = OnBundleLoaded};
                AssetLoader.QueueLoadAssetBundleFromFile(request);
                
                // Renaming callback
                SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;
            }
            catch(Exception e)
            {
                SuperController.LogError("VAMStory - Exception caught: " + e);
            }
		}
		
		void Start()
        {

        }

		void Update() {
			// Base initialization after assetbundle load
			if (isInit == false && !_isSceneLoading && isBundleLoaded == true)
			{
				try
				{
					// Setting options and ready state for the suite once
					if (isInit == false && VAMSValue_Ready.val == false)
					{
						// Starting to wait for the VAMStory plugins
						StartCoroutine("WaitForVAMSPlugins");
						
						VAMSValue_Ready.val = true;
					}
					// Firing triggers when the whole VAMS "suite" in the scene is ready
					else if ( _allVAMSReady )
					{
						isInit = true;
						
						// Everything is loaded, initialize the global options
						OnGlobalOptionsChanged();
						
						// Fire the story trigger at load
						GlobalOptStoryTrigger.Trigger();
						AddTriggerToTimerPool(GlobalOptStoryTrigger);
					}
				}
				catch(Exception e)
				{
					SuperController.LogError("VAMStory - Exception caught: " + e);
				}
			}
			
			// UI Update
			UpdateUIPosition();
			
			// Updating triggers with timers
			if (isInit == true && _timedTriggers.Count > 0)
			{
				foreach (EventTrigger trig in _timedTriggers)
				{
					trig.Update();
				}
			}
		}

		// **************************
		// Functions
		// **************************
		private void InitComponents()
		{
			if (VAMSUIGO != null)
			{
				// Transforms init
				TRStoryUI = getChildNamed(VAMSUIGO.transform,"VAMStoryUI");
				TRMainPanel = getChildNamed(VAMSUIGO.transform,"MainPanel");
				TRAudio = getChildNamed(VAMSUIGO.transform,"VAMStoryAudio");
				TRDialogPanel = getChildNamed(VAMSUIGO.transform,"DialogPanel");
				TRAvatar = getChildNamed(VAMSUIGO.transform,"Avatar");
				TRAvatarImg = getChildNamed(VAMSUIGO.transform,"AvatarImage");
				TRMainDialogBtn = getChildNamed(VAMSUIGO.transform,"MainDialogButton");
				TRMainDialogTxt = getChildNamed(VAMSUIGO.transform,"MainDialogText");
				TRAnswer01Btn = getChildNamed(VAMSUIGO.transform,"Answer01Btn");
				TRAnswer01Txt = getChildNamed(VAMSUIGO.transform,"Answer01Text");
				TRAnswer02Btn = getChildNamed(VAMSUIGO.transform,"Answer02Btn");
				TRAnswer02Txt = getChildNamed(VAMSUIGO.transform,"Answer02Text");
				TRAnswer03Btn = getChildNamed(VAMSUIGO.transform,"Answer03Btn");
				TRAnswer03Txt = getChildNamed(VAMSUIGO.transform,"Answer03Text");
				TRHideUIGrp = getChildNamed(VAMSUIGO.transform,"HideGroup");
				TRHideUIBtn = getChildNamed(VAMSUIGO.transform,"HideBtn");
				TRShowUIGrp = getChildNamed(VAMSUIGO.transform,"ShowGroup");
				TRShowUIBtn = getChildNamed(VAMSUIGO.transform,"ShowBtn");
				
				// Components init
				CANStoryUI = TRStoryUI.GetComponent<Canvas>();
				ASAudio = TRAudio.GetComponent<AudioSource>();
				CGDialogPanel = TRDialogPanel.GetComponent<CanvasGroup>();
				VLGAvatar = TRAvatar.GetComponent<VerticalLayoutGroup>();
				IMGAvatar = TRAvatarImg.GetComponent<RawImage>();
				BTNAvatar = TRAvatarImg.GetComponent<Button>();
				VLGMainDialog = TRMainDialogBtn.GetComponent<VerticalLayoutGroup>();
				BTNMainDialog = TRMainDialogBtn.GetComponent<Button>();
				TXTMainDialog = TRMainDialogTxt.GetComponent<Text>();
				IMGMainDialog = TRMainDialogBtn.GetComponent<Image>();
				VLGAnswer01 = TRAnswer01Btn.GetComponent<VerticalLayoutGroup>();
				BTNAnswer01 = TRAnswer01Btn.GetComponent<Button>();
				TXTAnswer01 = TRAnswer01Txt.GetComponent<Text>();
				IMGAnswer01 = TRAnswer01Btn.GetComponent<Image>();
				VLGAnswer02 = TRAnswer02Btn.GetComponent<VerticalLayoutGroup>();
				BTNAnswer02 = TRAnswer02Btn.GetComponent<Button>();
				TXTAnswer02 = TRAnswer02Txt.GetComponent<Text>();
				IMGAnswer02 = TRAnswer02Btn.GetComponent<Image>();
				VLGAnswer03 = TRAnswer03Btn.GetComponent<VerticalLayoutGroup>();
				BTNAnswer03 = TRAnswer03Btn.GetComponent<Button>();
				TXTAnswer03 = TRAnswer03Txt.GetComponent<Text>();
				IMGAnswer03 = TRAnswer03Btn.GetComponent<Image>();
				BTNHideUI = TRHideUIBtn.GetComponent<Button>();
				IMGHideUI = TRHideUIBtn.GetComponent<RawImage>();
				BTNShowUI = TRShowUIBtn.GetComponent<Button>();
				IMGShowUI = TRShowUIBtn.GetComponent<RawImage>();
				
				// Adding Canvas to VAM's controller
				TRStoryUI.gameObject.AddComponent<IgnoreCanvas>();
				SuperController.singleton.AddCanvas(CANStoryUI);
				
				// Hiding toggle on UI
				TRShowUIGrp.gameObject.SetActive(false);
				
				// Hiding answers
				TRAnswer01Btn.gameObject.SetActive(false);
				TRAnswer02Btn.gameObject.SetActive(false);
				TRAnswer03Btn.gameObject.SetActive(false);
				
				// Buttons actions
				BTNMainDialog.onClick.AddListener(delegate { OnClickMainButton(); } );
				BTNAvatar.onClick.AddListener(delegate { OnClickMainButton(); } );
				BTNAnswer01.onClick.AddListener(delegate { OnClickAnswer01Button(); } );
				BTNAnswer02.onClick.AddListener(delegate { OnClickAnswer02Button(); } );
				BTNAnswer03.onClick.AddListener(delegate { OnClickAnswer03Button(); } );
				
				BTNHideUI.onClick.AddListener(delegate { OnClickHideUIButton(); } );
				BTNShowUI.onClick.AddListener(delegate { OnClickShowUIButton(); } );
				
				// Hiding UI
				ToggleStoryUI(false);

				// Looking at the empty atoms to see if there is any VAMStoryActions
				foreach(Atom at in SuperController.singleton.GetAtoms() ) {
					if( at.type == "Empty" ) {
						foreach (string stId in at.GetStorableIDs())
						{
							if (stId.Contains("_VAMStoryActionPlugin.VAMStoryAction"))
							{
								actionsChoices.Add(at.name);
							}
						}
					}
				}
			}
		}
		
		private void CreateDefaultDialog()
		{
			StoryGroup newGroup = new StoryGroup(this, 0);
			newGroup.CreateNewDialog();
			storyGroups.Add(newGroup);

			RefreshSelectors(0, 0);
		}

		private void LoadDialogs()
		{
			// The save system is pretty "awkward" as there are not a lot of possibilities to properly create a json file
			// that would help to restore properly the storables natively
			// So
			Dictionary<int, int> dialogList = new Dictionary<int, int>();
			foreach (string paramKey in dialogsDatas.Keys)
			{
				if (paramKey == "id") continue;// Skipping plugin id
				string[] paramSplit = paramKey.Split('-');

				if (paramSplit.Length != 2) continue; // Skipping anything that has not at least 2 entries on its param name
				
				// Only keeping "texts" (it's enough to know what I need to create the UI
				if (paramSplit[1] == "Text")
				{
					string[] idsSplit = paramSplit[0].Split('_');
					if (idsSplit.Length == 3)
					{
						if (!dialogList.ContainsKey(int.Parse(idsSplit[1])))
						{
							dialogList.Add(int.Parse(idsSplit[1]), int.Parse(idsSplit[2]));
						}
						else
						{
							if ( dialogList[ int.Parse(idsSplit[1]) ] < int.Parse(idsSplit[2]) )
							{
								dialogList[int.Parse(idsSplit[1])] = int.Parse(idsSplit[2]);
							}
						}
					}
				}
			}
			
			// Ordering, since the save seem to change depending on the Register system
			// I obviously shouldn't have trusted the json :p
			dialogList = dialogList.OrderBy(x => x.Key).ToDictionary( x => x.Key, x => x.Value );

			// If we did find dialogs in the scene
			if (dialogList.Count > 0)
			{
				foreach (var groups in dialogList)
				{
					StoryGroup newGroup = new StoryGroup(this, groups.Key);
					newGroup.CreateDialogs(groups.Value + 1);
					storyGroups.Add(newGroup);
				}
				
				RefreshSelectors(0, 0);
			}
			// If it's not the case, let's create the default dialog
			else
			{
				CreateDefaultDialog();
			}
		}

		protected void CreateDialogUI()
	    {
		    // Temporary vars
		    UIDynamicTextField tmpTextfield;
		    UIDynamicPopup tmpPopup;
		    UIDynamicSlider tmpSlider;
		    UIDynamicColorPicker tmpColor;
		    UIDynamicButton tmpButton;
		    UIDynamicToggle tmpToggle;
		    UIDynamic tmpSpacer;
		    
		    // Creating Dialog components
			// ******* DIALOG TITLE (ID) ***********
			dialogTitle = createStaticDescriptionText("DialogTitle","<color=#000><size=35><b>Dialog</b></size></color>",false,55,TextAnchor.MiddleLeft);
			dialogsComponentsUI.Add((UIDynamic) dialogTitle);
			
			// ******* DIALOG TEXT ***********
			string newDefaultText = "Type some text...";
			DialogTextUI = new JSONStorableString("DialogTextUI", "_default_"){isStorable=false,isRestorable=false};
			tmpTextfield = CreateTextField(DialogTextUI);
			setupTextField(tmpTextfield, 550f, false, false);
			DialogTextInputFieldUI = tmpTextfield.UItext.gameObject.AddComponent<InputField>();
			DialogTextInputFieldUI.textComponent = tmpTextfield.UItext;
			DialogTextInputFieldUI.lineType = InputField.LineType.MultiLineNewline;
			dialogsComponentsUI.Add((UIDynamic) tmpTextfield);
			DialogTextUI.valNoCallback = newDefaultText;
			DialogTextInputFieldUI.text = newDefaultText;				
			DialogTextInputFieldUI.onValueChanged.AddListener( delegate { OnDialogTextChanged(); } );
			
			// ******* DIALOG START DELAY ***********
			DialogDelayUI = new JSONStorableFloat("DialogDelayUI", 0.2f, 0f, 5f, false){isStorable=false,isRestorable=false};
			DialogDelayUI.setCallbackFunction += (val) => { OnDialogDelayChanged(); };
			tmpSlider = CreateSlider(DialogDelayUI);
			tmpSlider.valueFormat = "F1";
			tmpSlider.label = "Dialog start delay";
			dialogsComponentsUI.Add((UIDynamic) tmpSlider);

			// ******* TRIGGERS ***********
			tmpButton = CreateButton("On show trigger", false);
			dialogsComponentsUI.Add((UIDynamic)tmpButton);
			tmpButton.button.onClick.AddListener(() => { activeDialog.triggerOnShowDialog.OpenPanelActionStart(); } );
			setupButtonWithLayout(tmpButton, 250f);
			
			tmpButton = CreateButton("On click trigger", false);
			dialogsComponentsUI.Add((UIDynamic)tmpButton);
			ignoreLayoutUI.Add((UIDynamic)tmpButton);
			tmpButton.button.onClick.AddListener(() => { activeDialog.triggerOnClickDialog.OpenPanelActionStart(); } );
			setupButtonWithoutLayout(tmpButton, 250f, new Vector2(270f, -805f));

			// ******* AVATAR POSITION ***********
			DialogAvatarPositionUI = new JSONStorableStringChooser("DialogAvatarPositionUI", avatarPositionChoices, "Right", "Avatar position"){isStorable=false,isRestorable=false};
			DialogAvatarPositionUI.setCallbackFunction += (val) => { OnDialogAvatarPositionChanged(); };
			tmpPopup = CreateScrollablePopup(DialogAvatarPositionUI);
			tmpPopup.labelWidth = 220f;
			dialogsComponentsUI.Add((UIDynamic) tmpPopup);
			
			// ******* AVATAR SELECTION ***********
			tmpButton = CreateButton("Select avatar", false);
			setButtonColor(tmpButton, new Color(0.7f,1f,0.7f,1f) );
			dialogsComponentsUI.Add((UIDynamic)tmpButton);
			tmpButton.button.onClick.AddListener(() => { OnSelectDialogAvatar(); } );
			setupButtonWithLayout(tmpButton, 250f);
			
			tmpButton = CreateButton("Remove avatar", false);
			setButtonColor(tmpButton, new Color(1f,0.7f,0.7f,1f) );
			dialogsComponentsUI.Add((UIDynamic)tmpButton);
			ignoreLayoutUI.Add((UIDynamic)tmpButton);
			tmpButton.button.onClick.AddListener(() => { DialogAvatarUrlUI.val = ""; activeDialog.DialogAvatarUrl.val = ""; } );
			setupButtonWithoutLayout(tmpButton, 250f, new Vector2(270f, -985f));
			
			// ******* AVATAR SELECTION ***********
			DialogAvatarUrlUI = new JSONStorableUrl("DialogAvatarUrlUI", "");
			tmpTextfield = CreateTextField(DialogAvatarUrlUI);
			setupTextField(tmpTextfield, 80f);
			dialogsComponentsUI.Add((UIDynamic)tmpTextfield);
	    }

		protected void CreateGroupUI()
		{
			// Temporary vars
			UIDynamicTextField tmpTextfield;
			UIDynamicPopup tmpPopup;
			UIDynamicSlider tmpSlider;
			UIDynamicColorPicker tmpColor;
			UIDynamicButton tmpButton;
			UIDynamicToggle tmpToggle;
			UIDynamic tmpSpacer;
			string newDefaultAnswer = "Answer text...";

			// ******* GROUP NAME ***********
			groupComponentsUI.Add((UIDynamic) createStaticDescriptionText("GroupNameTitleUI","\n<color=#000><size=25><b>Group Name</b></size></color>",true,65,TextAnchor.MiddleRight));

			GroupNameUI = new JSONStorableString("GroupNameUI", ""){isStorable=false,isRestorable=false};
			tmpTextfield = CreateTextField(GroupNameUI, true);
			setupTextField(tmpTextfield, 70, false, false);
			GroupNameUIInputField = tmpTextfield.UItext.gameObject.AddComponent<InputField>();
			GroupNameUIInputField.textComponent = tmpTextfield.UItext;
			GroupNameUIInputField.lineType = InputField.LineType.MultiLineNewline;
			groupComponentsUI.Add((UIDynamic) tmpTextfield);
			GroupNameUIInputField.onValueChanged.AddListener( delegate { OnGroupNameChanged(); } );
			GroupNameUIInputField.onEndEdit.AddListener(delegate { OnGroupNameEdited(); } );
			
			// ******* GROUP POSITION ***********
			positionSaveButton = CreateButton("Save Position", true);
			setButtonColor(positionSaveButton, unsavedPositionColor );
		    groupComponentsUI.Add((UIDynamic)positionSaveButton);
		    positionSaveButton.button.onClick.AddListener(() => { OnSaveGroupPosition(); } );
		    setupButtonWithLayout(positionSaveButton, 250f);
		    
		    positionClearButton = CreateButton("Clear Position", true);
		    setButtonColor(positionClearButton, unsavedPositionColor );
		    groupComponentsUI.Add((UIDynamic)positionClearButton);
		    ignoreLayoutUI.Add((UIDynamic) positionClearButton);
		    positionClearButton.button.onClick.AddListener(() => { OnClearGroupPosition(); } );
		    setButtonColor(positionClearButton, new Color(0.6f, 0.3f, 0.3f, 1f));
		    setButtonTextColor(positionClearButton, new Color(1f, 1f, 1f, 1f));
		    setupButtonWithoutLayout(positionClearButton, 250f, new Vector2(270, -479));
		    
		    // ******* GROUP TRIGGER ***********
		    tmpButton = CreateButton("Edit Group Trigger", true);
		    groupComponentsUI.Add((UIDynamic)tmpButton);
		    tmpButton.button.onClick.AddListener(() => { activeGroup.GroupTrigger.OpenPanelActionStart(); } );
		    
		    // ******* MAIN DIALOG DISABLED ***********
		    GroupHideMainDialogUI = new JSONStorableBool("GroupHideMainDialogUI", false){isStorable=false,isRestorable=false};
		    GroupHideMainDialogUI.setCallbackFunction += (val) => { OnGroupHideMainDialogChanged(); };
		    tmpToggle = CreateToggle(GroupHideMainDialogUI,true);
		    tmpToggle.label = "Hide main dialog at the end";
		    groupComponentsUI.Add((UIDynamic) tmpToggle);
		    
		    // ******* ANSWERS TITLE ***********
		    groupComponentsUI.Add((UIDynamic) createStaticDescriptionText("GroupAnswersTitleUI","\n<color=#000><size=25><b>Answers (displayed with last dialog)</b></size></color>",true,65,TextAnchor.MiddleRight));
		    
		    // ******* ANSWER 01 ***********
		    GroupAnswer01EnabledUI = new JSONStorableBool("Answer01EnabledUI", false){isStorable=false,isRestorable=false};
		    GroupAnswer01EnabledUI.setCallbackFunction += (val) => { OnGroupAnswer01EnabledChanged(); };
		    tmpToggle = CreateToggle(GroupAnswer01EnabledUI,true);
		    tmpToggle.label = "Enable answer 01";
		    groupComponentsUI.Add((UIDynamic) tmpToggle);
		    
		    GroupAnswer01TxtUI = new JSONStorableString("Answer01TextUI", "_default_"){isStorable=false,isRestorable=false};
		    tmpTextfield = CreateTextField(GroupAnswer01TxtUI, true);
		    setupTextField(tmpTextfield, 120, false, false);
		    GroupAnswer01TxtUIInputField = tmpTextfield.UItext.gameObject.AddComponent<InputField>();
		    GroupAnswer01TxtUIInputField.textComponent = tmpTextfield.UItext;
		    GroupAnswer01TxtUIInputField.lineType = InputField.LineType.MultiLineNewline;
		    groupComponentsUI.Add((UIDynamic) tmpTextfield);
		    GroupAnswer01TxtUI.valNoCallback = newDefaultAnswer;
		    GroupAnswer01TxtUIInputField.text = newDefaultAnswer;				
		    GroupAnswer01TxtUIInputField.onValueChanged.AddListener( delegate { OnGroupAnswer01TextChanged(); } );
		    
		    tmpButton = CreateButton("Edit Answer 01 Trigger", true);
		    groupComponentsUI.Add((UIDynamic)tmpButton);
		    tmpButton.button.onClick.AddListener(() => { activeGroup.GroupAnswer01Trigger.OpenPanelActionStart(); } );

		    // ******* ANSWER 02 ***********
		    GroupAnswer02EnabledUI = new JSONStorableBool("Answer02EnabledUI", false){isStorable=false,isRestorable=false};
		    GroupAnswer02EnabledUI.setCallbackFunction += (val) => { OnGroupAnswer02EnabledChanged(); };
		    tmpToggle = CreateToggle(GroupAnswer02EnabledUI,true);
		    tmpToggle.label = "Enable answer 02";
		    groupComponentsUI.Add((UIDynamic) tmpToggle);
		    
		    GroupAnswer02TxtUI = new JSONStorableString("Answer02TextUI", "_default_"){isStorable=false,isRestorable=false};
		    tmpTextfield = CreateTextField(GroupAnswer02TxtUI, true);
		    setupTextField(tmpTextfield, 120, false, false);
		    GroupAnswer02TxtUIInputField = tmpTextfield.UItext.gameObject.AddComponent<InputField>();
		    GroupAnswer02TxtUIInputField.textComponent = tmpTextfield.UItext;
		    GroupAnswer02TxtUIInputField.lineType = InputField.LineType.MultiLineNewline;
		    groupComponentsUI.Add((UIDynamic) tmpTextfield);
		    GroupAnswer02TxtUI.valNoCallback = newDefaultAnswer;
		    GroupAnswer02TxtUIInputField.text = newDefaultAnswer;				
		    GroupAnswer02TxtUIInputField.onValueChanged.AddListener( delegate { OnGroupAnswer02TextChanged(); } );
		    
		    tmpButton = CreateButton("Edit Answer 02 Trigger", true);
		    groupComponentsUI.Add((UIDynamic)tmpButton);
		    tmpButton.button.onClick.AddListener(() => { activeGroup.GroupAnswer02Trigger.OpenPanelActionStart(); } );
		    
		    // ******* ANSWER 03 ***********
		    GroupAnswer03EnabledUI = new JSONStorableBool("Answer03EnabledUI", false){isStorable=false,isRestorable=false};
		    GroupAnswer03EnabledUI.setCallbackFunction += (val) => { OnGroupAnswer03EnabledChanged(); };
		    tmpToggle = CreateToggle(GroupAnswer03EnabledUI,true);
		    tmpToggle.label = "Enable answer 03";
		    groupComponentsUI.Add((UIDynamic) tmpToggle);

		    GroupAnswer03TxtUI = new JSONStorableString("Answer03TextUI", "_default_"){isStorable=false,isRestorable=false};
		    tmpTextfield = CreateTextField(GroupAnswer03TxtUI, true);
		    setupTextField(tmpTextfield, 120, false, false);
		    GroupAnswer03TxtUIInputField = tmpTextfield.UItext.gameObject.AddComponent<InputField>();
		    GroupAnswer03TxtUIInputField.textComponent = tmpTextfield.UItext;
		    GroupAnswer03TxtUIInputField.lineType = InputField.LineType.MultiLineNewline;
		    groupComponentsUI.Add((UIDynamic) tmpTextfield);
		    GroupAnswer03TxtUI.valNoCallback = newDefaultAnswer;
		    GroupAnswer03TxtUIInputField.text = newDefaultAnswer;				
		    GroupAnswer03TxtUIInputField.onValueChanged.AddListener( delegate { OnGroupAnswer03TextChanged(); } );
		    
		    tmpButton = CreateButton("Edit Answer 03 Trigger", true);
		    groupComponentsUI.Add((UIDynamic)tmpButton);
		    tmpButton.button.onClick.AddListener(() => { activeGroup.GroupAnswer03Trigger.OpenPanelActionStart(); } );
		    
		    // ******* GROUP ACTIONS ***********
		    tmpSpacer = CreateSpacer(true);
		    setupSpacer(tmpSpacer, 25f);  
		    groupComponentsUI.Add((UIDynamic) tmpSpacer);
		    
		    GroupStoryActionsUI = new JSONStorableStringChooser("ActionsUI", actionsChoices, "None", "Group Actions"){isStorable=false,isRestorable=false};
		    GroupStoryActionsUI.setCallbackFunction += (val) => { OnGroupStoryActionsUIChanged(); };
		    tmpPopup = CreateScrollablePopup(GroupStoryActionsUI,true);
		    tmpPopup.labelWidth = 220f;
		    groupComponentsUI.Add((UIDynamic) tmpPopup);
		    
		    tmpSpacer = CreateSpacer(true);
		    setupSpacer(tmpSpacer, 155f);  
		    groupComponentsUI.Add((UIDynamic) tmpSpacer);
		}

		protected void RefreshSelectors( int groupId, int dialogId = 0 )
		{
			groupsChoices = new List<string>();
			dialogsChoices = new List<string>();
			copyChoices = new List<string>();
			copyChoices.Add(""); // The empty default one
			groupsIdsToNames.Clear();

			foreach (StoryGroup grp in storyGroups)
			{
				groupsIdsToNames.Add(grp.GroupID, grp.GroupName.val);
				groupsChoices.Add(grp.GroupName.val);
				
				for (var i = 0; i < grp.GetDialogCount(); i++)
				{
					if( grp.GroupID == groupId ) dialogsChoices.Add("" + i); // Only adding dialog for the target group
					copyChoices.Add(i + "_" + grp.GroupName.val);
				}
			}

			// Updating groupe selector and dialog selector
			GroupSelector.valNoCallback = "";
			GroupSelector.choices = groupsChoices;
			GroupSelector.val = groupsIdsToNames[groupId];
			
			DialogSelector.valNoCallback = "";
			DialogSelector.choices = dialogsChoices;
			DialogSelector.val = "" + dialogId;

			// Copy Choices
			CopySelector.choices = copyChoices;
			CopySelector.val = findNewCopySelectorVal(CopySelector.val);

			// Also updating the trigger system to start the dialogs
			StartDialogGroup.valNoCallback = "";
			StartDialogGroup.choices = groupsChoices;
		}

		public void RefreshVAMStoryActionsSelectors()
		{
			// Grabbing all VAMStoryActions
			actionsChoices = new List<string>();
			actionsChoices.Add("None");
			foreach(Atom at in SuperController.singleton.GetAtoms() ) {
				if( at.type == "Empty" ) {
					foreach (string stId in at.GetStorableIDs())
					{
						if (stId.Contains("_VAMStoryActionPlugin.VAMStoryAction"))
						{
							actionsChoices.Add(at.name);
						}
					}
				}
			}
			// Updating groups
			GroupStoryActionsUI.choices = actionsChoices;
		}
		
		protected void ToggleUIDynamicGrp(List<UIDynamic> uiGrp,bool enabled)
		{
			foreach( UIDynamic uidElem in uiGrp )
			{
				ToggleUIDynamic(enabled, uidElem);
			}
		}
		
		protected void ToggleUIDynamic(bool enabled, UIDynamic target)
		{
			if (target == null) return;
			
			LayoutElement tmpLE = target.GetComponent<LayoutElement>();

			if (tmpLE != null)
			{
				if (enabled == true)
				{
					tmpLE.transform.localScale = new Vector3(1, 1, 1);
					if( !ignoreLayoutUI.Contains(target) ) tmpLE.ignoreLayout = false;
				}
				else
				{
					tmpLE.transform.localScale = new Vector3(0, 0, 0);
					tmpLE.ignoreLayout = true;
				}
			}
		}
		
		private void setupTextField( UIDynamicTextField target, float fieldHeight, bool disableBackground = true, bool disableScroll = true ) {
			if( disableBackground ) target.backgroundColor = new Color(1f, 1f, 1f, 0f);
			LayoutElement tfLayout = target.GetComponent<LayoutElement>();
			tfLayout.preferredHeight = tfLayout.minHeight = fieldHeight;
			target.height = fieldHeight;
			if( disableScroll ) disableScrollOnText(target);
		}

		private void setupSpacer(UIDynamic target, float newHeight)
		{
			LayoutElement tmpLE = target.GetComponent<LayoutElement>();
			tmpLE.minHeight = newHeight;
			tmpLE.preferredHeight = newHeight;
		}

		private void setupButtonWithLayout(UIDynamicButton target, float newSize)
		{
			ContentSizeFitter tmpCSF = target.button.transform.gameObject.AddComponent<ContentSizeFitter>();
			tmpCSF.horizontalFit = UnityEngine.UI.ContentSizeFitter.FitMode.MinSize;
			LayoutElement tmpLE = target.button.transform.gameObject.GetComponent<LayoutElement>();
			RectTransform tmpRectT = target.button.transform.GetComponent<RectTransform>();
			Rect tmpRect = tmpRectT.rect;
			tmpRectT.pivot = new Vector2(0f, 0.5f);
			tmpLE.minWidth = newSize;
			tmpLE.preferredWidth = newSize;
		}
		
		private void setupButtonWithoutLayout(UIDynamicButton target, float newSize, Vector2 newPosition)
		{
			ContentSizeFitter tmpCSF = target.button.transform.gameObject.AddComponent<ContentSizeFitter>();
			tmpCSF.horizontalFit = UnityEngine.UI.ContentSizeFitter.FitMode.MinSize;
			LayoutElement tmpLE = target.button.transform.gameObject.GetComponent<LayoutElement>();
			RectTransform tmpRectT = target.button.transform.GetComponent<RectTransform>();
			Rect tmpRect = tmpRectT.rect;
			tmpRectT.pivot = new Vector2(0f, 0.5f);
			tmpLE.minWidth = newSize;
			tmpLE.preferredWidth = newSize;
			tmpLE.ignoreLayout = true;
			tmpRectT.anchoredPosition = newPosition;
		}

		private void setButtonColor(UIDynamicButton targetBtn, Color newColor)
		{
			targetBtn.buttonColor = newColor;
		}
		
		private void setButtonTextColor(UIDynamicButton targetBtn, Color newColor)
		{
			targetBtn.textColor = newColor;
		}
		
		public UIDynamicTextField createStaticDescriptionText( string DescTitle, string DescText, bool rightSide, int fieldHeight, TextAnchor textAlignment = TextAnchor.UpperLeft ) {
			JSONStorableString staticDescString = new JSONStorableString(DescTitle,DescText) {isStorable=false,isRestorable=false};
			staticDescString.hidden = true;
			UIDynamicTextField staticDescStringField = CreateTextField(staticDescString, rightSide);
			staticDescStringField.backgroundColor = new Color(1f, 1f, 1f, 0f);
			staticDescStringField.UItext.alignment = textAlignment;
			LayoutElement sdsfLayout = staticDescStringField.GetComponent<LayoutElement>();
			sdsfLayout.preferredHeight = sdsfLayout.minHeight = fieldHeight;
			staticDescStringField.height = fieldHeight;
			disableScrollOnText(staticDescStringField);

			return staticDescStringField;
		}
		
		private static void disableScrollOnText(UIDynamicTextField target) {
			ScrollRect targetSR = target.UItext.transform.parent.transform.parent.transform.parent.GetComponent<ScrollRect>();
			if( targetSR != null ) {
				targetSR.horizontal = false;
				targetSR.vertical = false;
			}
		}
		
		public static void scrollToTop(UIDynamicTextField target) {
			ScrollRect targetSR = target.UItext.transform.parent.transform.parent.transform.parent.GetComponent<ScrollRect>();
			if( targetSR != null ) {
				targetSR.normalizedPosition = new Vector2(0, 1);
			}
		}
		
		private Transform getChildNamed( Transform parent, string childName ) {
			foreach( Transform child in parent )
			{
				if( child.name == childName ) {
					return child;
				} else {
					Transform childSearch = getChildNamed(child,childName);
					if (childSearch != null) return childSearch;
				}
			}
        
			return null;
		}

		protected void ToggleStoryUI(bool enabled)
		{
			if (TRStoryUI != null)
			{
				TRStoryUI.gameObject.SetActive(enabled);
			}
		}
		
		protected void SetupUIForCurrentDialogID()
		{
			// Clearing actions
			ClearActionsUI();
			
			// Display UI if needed
			ToggleStoryUI(true);
			
			// Execute show trigger
			storyGroups[CurrentGroupID].DoShowTriggerDialogID(CurrentDialogID);
			
			// Play on show sound
			PlayOnShowSound();

			// Displaying the main dialog by default
			TRAvatar.gameObject.SetActive(true);
			TRMainDialogBtn.gameObject.SetActive(true);
			
			// Main Dialog Text
			TXTMainDialog.text = storyGroups[CurrentGroupID].GetDialogID(CurrentDialogID).DialogText.val;
			
			// Main Dialog Avatar
			string avatarUrl = storyGroups[CurrentGroupID].GetDialogID(CurrentDialogID).DialogAvatarUrl.val;
			if( DialogAvatars.ContainsKey(avatarUrl) )
			{
				TRAvatarImg.gameObject.SetActive(true);
				IMGAvatar.texture = DialogAvatars[ avatarUrl ];
			}
			else
			{
				TRAvatarImg.gameObject.SetActive(false);
			}
			
			// Avatar position
			string avatarPos = storyGroups[CurrentGroupID].GetDialogID(CurrentDialogID).DialogAvatarPosition.val;
			VLGAvatar.childAlignment = TextAnchor.LowerRight; // default
			if( avatarPos == "Left" ) VLGAvatar.childAlignment = TextAnchor.LowerLeft;
			
			// Hiding answers by default
			TRAnswer01Btn.gameObject.SetActive(false);
			TRAnswer02Btn.gameObject.SetActive(false);
			TRAnswer03Btn.gameObject.SetActive(false);
			// Showing up answers if needed
			if (storyGroups[CurrentGroupID].IsLastDialogID(CurrentDialogID))
			{
				// On the last dialog we can hide the main dialog to leave only the answers and actions
				bool showMainDialog = !storyGroups[CurrentGroupID].GroupHideMainDialog.val;
				TRAvatar.gameObject.SetActive(showMainDialog);
				TRMainDialogBtn.gameObject.SetActive(showMainDialog);
				
				// Answer 01
				if (storyGroups[CurrentGroupID].GroupAnswer01Enabled.val)
				{
					TRAnswer01Btn.gameObject.SetActive(true);
					TXTAnswer01.text = storyGroups[CurrentGroupID].GroupAnswer01Txt.val;
				}
				
				// Answer 02
				if (storyGroups[CurrentGroupID].GroupAnswer02Enabled.val)
				{
					TRAnswer02Btn.gameObject.SetActive(true);
					TXTAnswer02.text = storyGroups[CurrentGroupID].GroupAnswer02Txt.val;
				}
				
				// Answer 02
				if (storyGroups[CurrentGroupID].GroupAnswer03Enabled.val)
				{
					TRAnswer03Btn.gameObject.SetActive(true);
					TXTAnswer03.text = storyGroups[CurrentGroupID].GroupAnswer03Txt.val;
				}
				
				// Actions!
				if (storyGroups[CurrentGroupID].GroupStoryActions.val != "None")
				{
					AddActionsToUI();
				}
			}
		}

		protected void UpdateUIPosition() {
			if (isInit == false || VAMSUIGO == null) return;
			
			// Handling forced rotation
			// Some Spacedog code and some VRStudy code (even if it is heavily modified), thank you guyz!
			Vector3 followOffset = new Vector3(0,0.15f,0.2f);
			Vector3 followPos = Vector3.zero;
			Vector3 followVelocity = Vector3.zero;

			// If the UI is set to follow the camera
			if( GlobalOptUIFollowCamera.val == true ) {
				if( !IsUIVisible() ) {
					// Immediately updating the position if it's not visible
					// Smooth / interpolated motion is fun, but I want this to be snappy
					// With interactable UI, it would be a pain for the player to wait for the UI to settle
					followPos = VAMSCamera.transform.TransformPoint(new Vector3(GlobalOptUIFollowHorOffset.val, GlobalOptUIFollowVertOffset.val, GlobalOptUIFollowMinDist.val));
					containingAtom.mainController.MoveControl( followPos );
				}
			}

			// UI Rotation
			if (GlobalOptUIRotationDisabled.val != true)
			{
				Vector3 lookPos = VAMSCamera.transform.position - VAMSUIGO.transform.position;
				Quaternion lookRot = Quaternion.LookRotation(lookPos, Vector3.up);
				float eulerX = 0;
				if (GlobalOptUIConstraintToY.val == false)
				{
					eulerX = lookRot.eulerAngles.x;
				}

				float eulerY = lookRot.eulerAngles.y;
				Quaternion rotation = Quaternion.Euler(eulerX, eulerY, 0);
				VAMSUIGO.transform.rotation = rotation;
			}
		}

		protected bool IsUIVisible() {
			Vector3 UIvp = VAMSCamera.WorldToViewportPoint(containingAtom.mainController.control.position);
			
			// Checking if inside the view of the camera with a bit of padding
			bool inView = UIvp.x > 0.05f && UIvp.x < 0.95f && UIvp.y > 0.1f && UIvp.y < 0.9f && UIvp.z > 0;
			// Checking if the UI is close enough
			bool distanceOk = Vector3.Distance(VAMSCamera.transform.position,containingAtom.mainController.control.position) >= GlobalOptUIFollowMinDist.val && Vector3.Distance(VAMSCamera.transform.position,containingAtom.mainController.control.position) <= GlobalOptUIFollowMaxDist.val;
			
			// Both are ok ?
			return inView && distanceOk;	
		}

		protected void AddActionsToUI()
		{
			// If we already have an action atom, then the creator might by trying to restart a group that has already displayed the actions
			// No need to re-initialize it
			if (ActionsAtom != null) return;
			
			// Checking the target atom
			ActionsAtom = GetAtomById(storyGroups[CurrentGroupID].GroupStoryActions.val);
			if (ActionsAtom != null)
			{
				TRStoryActionsUI = getChildNamed(ActionsAtom.reParentObject, "VAMStoryActionsUI");
				if (TRStoryActionsUI != null)
				{
					TRStoryActionsMainPanel = getChildNamed(ActionsAtom.reParentObject, "ActionsMainPanel");
					CGStoryActionsMainPanel = TRStoryActionsMainPanel.GetComponent<CanvasGroup>();
						
					TRStoryActionsHideGroup = getChildNamed(TRStoryActionsMainPanel, "HideGroup");
					TRStoryActionsShowGroup = getChildNamed(TRStoryActionsMainPanel, "ShowGroup");
					
					// Hiding show/hide btns, we don't need them since they're handled by the main UI
					// I'm disabling both, I don't know if a user might have changed the status of them (even tho it should not be the case)
					TRStoryActionsHideGroup.gameObject.SetActive(false);
					TRStoryActionsShowGroup.gameObject.SetActive(false);
					
					// Moving the Action UI to my container
					TRStoryActionsMainPanel.SetParent(TRMainPanel, false);
					TRStoryActionsMainPanel.SetSiblingIndex(1);
				}
			}
		}

		protected void ClearActionsUI()
		{
			// If this var is not null, then we have create actions, we can clear them
			if (ActionsAtom != null)
			{
				// If the actual transform is null, then the action atom might have reload the plugin
				// so, I'm not gonna change the states of the transforms
				if (TRStoryActionsMainPanel != null)
				{
					// Moving back to the original atom
					TRStoryActionsMainPanel.SetParent(TRStoryActionsUI, false);
				}
				
				// Broadcasting to original atom the reset sequence
				ActionsAtom.BroadcastMessage("ResetFromVAMStory");

				// Nulling all vars ( it is meant to be able to know if actions are actually in the layout or not)
				ActionsAtom = null;
				TRStoryActionsUI = null;
				TRStoryActionsMainPanel = null;
				TRStoryActionsHideGroup = null;
				TRStoryActionsShowGroup = null;
				CGStoryActionsMainPanel = null;
			}
		}

		// Saving the datas
		// saveType: 0 = auto, 1 = critical, 2 = manual
		protected void SaveBackup( int saveType = 0 )
		{
			if (SuperController.singleton.gameMode != SuperController.GameMode.Edit) return;
			string fileName = "vams-" + GetCleanFilenameFromString(containingAtom.name);
			if (saveType == 0)
			{
				logDebug("VAMStory: Saving automatic backup");
				string backupPath = CleanAndGetBackupPath( fileName, "auto" );
				SuperController.singleton.SaveJSON( this.GetJSON(), backupPath);
			} else if (saveType == 1)
			{
				string backupPath = CleanAndGetBackupPath( fileName, "critical" );
				SuperController.singleton.SaveJSON( this.GetJSON(), backupPath);
			} else if (saveType == 2)
			{
				string backupPath = BACKUP_PATH + "\\" + fileName + "-manual-" + DateTime.Now.ToString("yyyyMMdd-HHmmss") + ".json";
				SuperController.singleton.SaveJSON( this.GetJSON(), backupPath);
				logDebug("VAMStory: Manual backup saved in:" + backupPath);
			}
		}

		protected string CleanAndGetBackupPath( string backupName, string backupSuffix )
		{
			// I'm keeping 5 files
			for (int i = 5; i > 0; i--)
			{
				string backupPath = BACKUP_PATH + "\\" + backupName + "-" + backupSuffix + "-" + i + ".json";
				if (FileManagerSecure.FileExists(backupPath) == true && i == 5)
				{
					FileManagerSecure.DeleteFile(backupPath);
				}
				else if(FileManagerSecure.FileExists(backupPath) == true)
				{
					FileManagerSecure.MoveFile(BACKUP_PATH + "\\" + backupName + "-" + backupSuffix + "-" + i + ".json", BACKUP_PATH + "\\" + backupName + "-" + backupSuffix + "-" + ( i + 1 ) + ".json",false);
				}
			}

			return BACKUP_PATH + "\\" + backupName + "-" + backupSuffix + "-1.json";
		}
		
		// Load backup
		protected void LoadBackup()
		{
			SuperController.singleton.fileBrowserUI.defaultPath = BACKUP_PATH;
			SuperController.singleton.fileBrowserUI.SetTitle("Select a backup");
			SuperController.singleton.fileBrowserUI.fileFormat = "json";
			SuperController.singleton.fileBrowserUI.showFiles = true;
			SuperController.singleton.fileBrowserUI.showDirs = true;
			SuperController.singleton.fileBrowserUI.SetTextEntry(false);
			SuperController.singleton.fileBrowserUI.hideExtension = false;
			SuperController.singleton.fileBrowserUI.showInstallFolderInDirectoryList = false;
			List<ShortCut> shortCutList = new List<ShortCut>();

			ShortCut shortCutCustom = new ShortCut();
			shortCutCustom.path = BACKUP_PATH;
			shortCutCustom.displayName = "VAMStory Backups";
			shortCutCustom.flatten = false;
			shortCutList.Add(shortCutCustom);

			SuperController.singleton.fileBrowserUI.shortCuts = shortCutList;			
			SuperController.singleton.fileBrowserUI.Show((val) =>
			{
				if (val != "")
				{
					RestoreBackup(val);
				}
			});
		}

		protected void RestoreBackup(string backupPath)
		{
			// Loading the file
			string backupDatas = SuperController.singleton.ReadFileIntoString(backupPath);
			JSONClass backupJSON = JSON.Parse(backupDatas).AsObject;
			
			// Checking if the save is meant for VAMStory
			if ( !backupJSON.HasKey("id") || !backupJSON["id"].Value.Contains("_VAMStoryPlugin.VAMStory") )
			{
				logError("VAMStory: Backup not compatible with VAMStory.");
				return;
			}
			
			logDebug("VAMStory: Restoring " + backupPath);
			
			// Clearing all groups
			foreach (StoryGroup grp in storyGroups)
			{
				grp.Delete();
			}

			storyGroups.Clear();
			
			// Hiding UI (if it's displayed)
			ToggleStoryUI(false);
			
			// Restoring datas
			this.RestoreFromJSON(backupJSON);
			this.LateRestoreFromJSON(backupJSON);
			
			// Loading Dialogs
			LoadDialogs();
			
			// Refreshing UI
			RefreshSelectors(0, 0);
		}

		protected string findNewCopySelectorVal(string originalValue)
		{
			if (CopySelector.choices.Contains(originalValue)) return originalValue;
			return "";
		}
		
		protected void AddTriggerToTimerPool(EventTrigger trig)
		{
			if (trig.HasActionsWithActiveTimers && !_timedTriggers.Contains(trig))
			{
				_timedTriggers.Add(trig);
			}
		}
		
		// **************************************
		// Callbacks & similar ( when dialog or group UI value is changed )
		// **************************************
		protected void OnDialogTextChanged()
		{
			DialogTextUI.val = DialogTextInputFieldUI.text;
			activeDialog.DialogText.val = DialogTextInputFieldUI.text;
		}
		
		protected void OnDialogDelayChanged()
		{
			activeDialog.DialogDelay.val = DialogDelayUI.val;
		}
		
		protected void OnDialogAvatarPositionChanged()
		{
			activeDialog.DialogAvatarPosition.val = DialogAvatarPositionUI.val;
		}

		protected void UpdateLocalUIFromActiveGroup()
		{
			// Updating the local dialog UI based on dialog datas
			StoryGroup grp = activeGroup;
			GroupNameUI.valNoCallback = grp.GroupName.val;
			GroupNameUIInputField.text = grp.GroupName.val;
			
			GroupHideMainDialogUI.valNoCallback = grp.GroupHideMainDialog.val;
			GroupAnswer01EnabledUI.valNoCallback = grp.GroupAnswer01Enabled.val;
			GroupAnswer01TxtUI.valNoCallback = grp.GroupAnswer01Txt.val;
			GroupAnswer01TxtUIInputField.text = grp.GroupAnswer01Txt.val;
			
			GroupAnswer02EnabledUI.valNoCallback = grp.GroupAnswer02Enabled.val;
			GroupAnswer02TxtUI.valNoCallback = grp.GroupAnswer02Txt.val;
			GroupAnswer02TxtUIInputField.text = grp.GroupAnswer02Txt.val;
			
			GroupAnswer03EnabledUI.valNoCallback = grp.GroupAnswer03Enabled.val;
			GroupAnswer03TxtUI.valNoCallback = grp.GroupAnswer03Txt.val;
			GroupAnswer03TxtUIInputField.text = grp.GroupAnswer03Txt.val;
			
			GroupStoryActionsUI.valNoCallback = grp.GroupStoryActions.val;
		}
		
		protected void UpdateLocalUIFromActiveDialog()
		{
			// Updating the local dialog UI based on dialog datas
			DialogTextUI.valNoCallback = activeDialog.DialogText.val;
			DialogTextInputFieldUI.text = activeDialog.DialogText.val;
			DialogDelayUI.valNoCallback = activeDialog.DialogDelay.val;
			DialogAvatarPositionUI.valNoCallback = activeDialog.DialogAvatarPosition.val;
			DialogAvatarUrlUI.valNoCallback = activeDialog.DialogAvatarUrl.val;
			
			// Updating the title
			dialogTitle.text = "<color=#000><size=35><b>GROUP " + GroupSelector.val + " : Dialog " + DialogSelector.val + "</b></size></color>";
		}
		
		protected void OnGroupHideMainDialogChanged()
		{
			activeGroup.GroupHideMainDialog.val = GroupHideMainDialogUI.val;
		}
		
		protected void OnGroupNameChanged()
		{
			GroupNameUI.val = GroupNameUIInputField.text;
			activeGroup.GroupName.val = GroupNameUI.val;
		}

		protected void OnGroupNameEdited()
		{
			string newName = GetCleanAndUniqueGroupName(activeGroup.GroupID);
			GroupNameUIInputField.text = newName;
			UpdateGroupsIdsToNameDict();
			RefreshSelectors(groupsIdsToNames.FirstOrDefault(el => el.Value == newName).Key);
	
		}
		
		protected void UpdateGroupsIdsToNameDict()
		{
			StoryGroup grp = activeGroup;
			if (!groupsIdsToNames.ContainsKey(grp.GroupID))
			{
				groupsIdsToNames.Add(grp.GroupID,GroupNameUI.val);
			}
			else
			{
				groupsIdsToNames[grp.GroupID] = GroupNameUI.val;
			}
		}
		
		protected void OnGroupAnswer01TextChanged()
		{
			GroupAnswer01TxtUI.val = GroupAnswer01TxtUIInputField.text;
			activeGroup.GroupAnswer01Txt.val = GroupAnswer01TxtUI.val;
		}
		
		protected void OnGroupAnswer02TextChanged()
		{
			GroupAnswer02TxtUI.val = GroupAnswer02TxtUIInputField.text;
			activeGroup.GroupAnswer02Txt.val = GroupAnswer02TxtUI.val;
		}
		
		protected void OnGroupAnswer03TextChanged()
		{
			GroupAnswer03TxtUI.val = GroupAnswer03TxtUIInputField.text;
			activeGroup.GroupAnswer03Txt.val = GroupAnswer03TxtUI.val;
		}

		protected void OnGroupAnswer01EnabledChanged()
		{
			activeGroup.GroupAnswer01Enabled.val = GroupAnswer01EnabledUI.val;
		}
		
		protected void OnGroupAnswer02EnabledChanged()
		{
			activeGroup.GroupAnswer02Enabled.val = GroupAnswer02EnabledUI.val;
		}
		
		protected void OnGroupAnswer03EnabledChanged()
		{
			activeGroup.GroupAnswer03Enabled.val = GroupAnswer03EnabledUI.val;
		}
		
		protected void OnGroupStoryActionsUIChanged()
		{
			activeGroup.GroupStoryActions.val = GroupStoryActionsUI.val;
		}

		// **************************************
		// Callbacks & similar ( internal )
		// **************************************
		private void OnBundleLoaded(Request aRequest) {
			try
			{
				// Loading triggers assets
				SimpleTriggerHandler.LoadAssets();
				
				// Reading my UI bundle
				bundleRequest = aRequest;
				GameObject tmp = bundleRequest.assetBundle.LoadAsset<GameObject>("Assets/VAMStory/prefabs/VAMStory.prefab");
				VAMSUIGO = Instantiate(tmp);
				
				// Parenting to my current Empty
				VAMSUIGO.transform.SetParent(containingAtom.reParentObject.GetChild(0).transform, false);
				VAMSUIGO.transform.localPosition = new Vector3(0, 0, 0);
				
				// Grabbing all global assets
				backgroundsChoices = new List<string>();
				backgroundsChoices.Add("Default");
				fontsChoices = new List<string>();
				fontsChoices.Add("arial");
				soundChoices = new List<string>();
				materialChoices = new List<string>();
				materialChoices.Add("None");
				
				// Default assets
				fontsFiles.Add(Resources.GetBuiltinResource(typeof(Font), "Arial.ttf") as Font);
				materialList.Add( bundleRequest.assetBundle.LoadAsset<Material>("Assets/VAMStory/materials/M_VAMS_UI_default.mat") );
				
				// Other assets
				foreach (string bundleFileItem in bundleRequest.assetBundle.GetAllAssetNames())
				{
					if (bundleFileItem.Contains(BACKGROUNDS_PATH))
					{
						backgroundsChoices.Add(bundleFileItem.Replace(BACKGROUNDS_PATH, "").Replace(".png", ""));
						backgroundsSprites.Add( bundleRequest.assetBundle.LoadAsset<Sprite>(bundleFileItem) );
					}
					
					if (bundleFileItem.Contains(FONTS_PATH))
					{
						fontsChoices.Add(bundleFileItem.Replace(FONTS_PATH, "").Replace(".ttf", "").Replace(".otf", ""));
						fontsFiles.Add( bundleRequest.assetBundle.LoadAsset<Font>(bundleFileItem) );
					}
					
					if (bundleFileItem.Contains(SOUNDS_PATH))
					{
						soundChoices.Add(bundleFileItem.Replace(SOUNDS_PATH, "").Replace(".ogg", ""));
						soundClips.Add( bundleRequest.assetBundle.LoadAsset<AudioClip>(bundleFileItem) );
					}
					
					if (bundleFileItem.Contains(MATERIALS_PATH))
					{
						materialChoices.Add(bundleFileItem.Replace(MATERIALS_PATH, "").Replace(".mat", ""));
						materialList.Add( bundleRequest.assetBundle.LoadAsset<Material>(bundleFileItem) );
					}
				}
				
				// Settings global storable
				GlobalOptFont.choices = fontsChoices;
				GlobalOptFont.valNoCallback = fontsChoices.First();
				GlobalOptBg.choices = backgroundsChoices;
				GlobalOptBg.valNoCallback = backgroundsChoices.First();
				GlobalOptSoundOnShow.choices = soundChoices;
				GlobalOptSoundOnShow.valNoCallback = soundChoices.First();
				GlobalOptSoundOnClick.choices = soundChoices;
				GlobalOptSoundOnClick.valNoCallback = soundChoices.First();
				GlobalOptMaterial.choices = materialChoices;
				GlobalOptMaterial.valNoCallback = materialChoices.First();
				
				// Setting actions storables
				A_PlayStorySound.choices = soundChoices;
				
				// Since everything happens after loading the assets
				// I need to restore some of the global options that I just updated
				if (dialogsDatas != null)
				{
					GlobalOptFont.RestoreFromJSON( dialogsDatas );
					GlobalOptBg.RestoreFromJSON( dialogsDatas );
					GlobalOptSoundOnShow.RestoreFromJSON( dialogsDatas );
					GlobalOptSoundOnClick.RestoreFromJSON( dialogsDatas );
					GlobalOptMaterial.RestoreFromJSON( dialogsDatas );
				}
				
				// Initializing UI
				CreateDialogUI();
				CreateGroupUI();
				
				// Refresh actions
				RefreshVAMStoryActionsSelectors();
				
				// Initializing all UI components
				InitComponents();
					
				// If there is no data saved, let's create a default group and a first dialog
				if (dialogsDatas == null)
				{
					CreateDefaultDialog();
				}
				else
				{
					LoadDialogs();
				}
				
				// Selecting the first group (will update the UI)
				OnChangeRightUISelection();
				
				// Updating dialog/group UI
				UpdateLocalUIFromActiveGroup();
				UpdateLocalUIFromActiveDialog();
				
				isBundleLoaded = true;
			}
			catch(Exception e)
			{
				SuperController.LogError("VAMStory - Exception caught: " + e);
			}
		}
		
		protected void OnChangeRightUISelection()
		{
			if (RightUISelector.val == "Group/Dialog Control")
			{
				RightUIMainTitle.text = "<color=#000><size=35><b>GROUP/DIALOG CONTROLS</b></size></color>";

				ToggleUIDynamicGrp(globalControlsUIs, true);
				ToggleUIDynamicGrp(groupComponentsUI, false);
				ToggleUIDynamicGrp(globalOptionsUIs, false);
				ToggleUIDynamicGrp(globalTextOptUIs, false);
				ToggleUIDynamicGrp(globalColorsOptUIs, false);
				ToggleUIDynamicGrp(helpUIs, false);
				OnToggleGlobalOptionsFollow(false);
			} else if (RightUISelector.val == "Group options")
			{
				RightUIMainTitle.text = "<color=#000><size=35><b>GROUP OPTIONS</b></size></color>";
				
				ToggleUIDynamicGrp(globalControlsUIs, false);
				ToggleUIDynamicGrp(groupComponentsUI, true);
				ToggleUIDynamicGrp(globalOptionsUIs, false);
				ToggleUIDynamicGrp(globalTextOptUIs, false);
				ToggleUIDynamicGrp(globalColorsOptUIs, false);
				ToggleUIDynamicGrp(helpUIs, false);
				OnToggleGroupId(GroupSelector.val);
				
				// Swapping position color if needed
				if (activeGroup.hasSavedPosition())
				{
					setButtonColor(positionSaveButton, savedPositionColor);
				}
				else
				{
					setButtonColor(positionSaveButton, unsavedPositionColor);
				}


			} else if (RightUISelector.val == "Text options")
			{
				ToggleUIDynamicGrp(globalControlsUIs, false);
				ToggleUIDynamicGrp(groupComponentsUI, false);
				ToggleUIDynamicGrp(globalOptionsUIs, false);
				ToggleUIDynamicGrp(globalTextOptUIs, true);
				ToggleUIDynamicGrp(globalColorsOptUIs, false);
				ToggleUIDynamicGrp(helpUIs, false);
				RightUIMainTitle.text = "<color=#000><size=35><b>GLOBAL OPTIONS</b></size></color>";
				OnToggleGlobalOptionsFollow(GlobalOptUIFollowCamera.val);
			} else if (RightUISelector.val == "Global options")
			{
				ToggleUIDynamicGrp(globalControlsUIs, false);
				ToggleUIDynamicGrp(groupComponentsUI, false);
				ToggleUIDynamicGrp(globalOptionsUIs, true);
				ToggleUIDynamicGrp(globalTextOptUIs, false);
				ToggleUIDynamicGrp(globalColorsOptUIs, false);
				ToggleUIDynamicGrp(helpUIs, false);
				RightUIMainTitle.text = "<color=#000><size=35><b>GLOBAL OPTIONS</b></size></color>";
				OnToggleGlobalOptionsFollow(GlobalOptUIFollowCamera.val);
			} else if (RightUISelector.val == "Appearance options")
			{
				ToggleUIDynamicGrp(globalControlsUIs, false);
				ToggleUIDynamicGrp(groupComponentsUI, false);
				ToggleUIDynamicGrp(globalOptionsUIs, false);
				ToggleUIDynamicGrp(globalTextOptUIs, false);
				ToggleUIDynamicGrp(globalColorsOptUIs, true);
				ToggleUIDynamicGrp(helpUIs, false);
				RightUIMainTitle.text = "<color=#000><size=35><b>APPEARANCE OPTIONS</b></size></color>";
				OnToggleGlobalOptionsFollow(GlobalOptUIFollowCamera.val);
			} else
			{
				ToggleUIDynamicGrp(globalControlsUIs, false);
				ToggleUIDynamicGrp(groupComponentsUI, false);
				ToggleUIDynamicGrp(globalOptionsUIs, false);
				ToggleUIDynamicGrp(globalTextOptUIs, false);
				ToggleUIDynamicGrp(globalColorsOptUIs, false);
				ToggleUIDynamicGrp(helpUIs, true);
				RightUIMainTitle.text = "<color=#000><size=35><b>HELP</b></size></color>";
				OnToggleGlobalOptionsFollow(GlobalOptUIFollowCamera.val);
				scrollToTop(helpWindow);
			}
		}
		
		protected void OnToggleGlobalOptionsFollow(bool enabled)
		{
			foreach( UIDynamic uidElem in globalOptionsFollowUIs )
			{
				ToggleUIDynamic(enabled, uidElem);
			}
		}
		
		protected void OnGlobalOptionsChanged()
		{
			if ( TXTMainDialog == null || isInit == false ) return;
			
			// Hiding UI toggle
			TRHideUIGrp.gameObject.SetActive(!GlobalOptUIHideToggle.val);
			TRShowUIGrp.gameObject.SetActive(false);
			
			// Changing UI size
			VAMSUIGO.transform.localScale = new Vector3(GlobalOptUISize.val, GlobalOptUISize.val, GlobalOptUISize.val);
			
			// Text alignment value
			TextAnchor txtAlign = GlobalOptUIAlignment.val == "Left" ? TextAnchor.MiddleLeft : TextAnchor.MiddleRight;
			
			// Changing texts alignement
			TXTMainDialog.alignment = txtAlign;
			TXTAnswer01.alignment = txtAlign;
			TXTAnswer02.alignment = txtAlign;
			TXTAnswer03.alignment = txtAlign;
			
			// Changing font size
			TXTMainDialog.fontSize = (int)GlobalOptFontSize.val;
			TXTAnswer01.fontSize = (int)GlobalOptFontSize.val;
			TXTAnswer02.fontSize = (int)GlobalOptFontSize.val;
			TXTAnswer03.fontSize = (int)GlobalOptFontSize.val;
			
			// Changing padding
			VLGMainDialog.padding = new RectOffset((int)GlobalOptTxtPadL.val,(int)GlobalOptTxtPadR.val,(int)GlobalOptTxtPadT.val,(int)GlobalOptTxtPadB.val);
			VLGAnswer01.padding = new RectOffset((int)GlobalOptTxtPadL.val,(int)GlobalOptTxtPadR.val,(int)GlobalOptTxtPadT.val,(int)GlobalOptTxtPadB.val);
			VLGAnswer02.padding = new RectOffset((int)GlobalOptTxtPadL.val,(int)GlobalOptTxtPadR.val,(int)GlobalOptTxtPadT.val,(int)GlobalOptTxtPadB.val);
			VLGAnswer03.padding = new RectOffset((int)GlobalOptTxtPadL.val,(int)GlobalOptTxtPadR.val,(int)GlobalOptTxtPadT.val,(int)GlobalOptTxtPadB.val);

			// Changing line height
			TXTMainDialog.lineSpacing = GlobalOptLineHeight.val;
			TXTAnswer01.lineSpacing = GlobalOptLineHeight.val;
			TXTAnswer02.lineSpacing = GlobalOptLineHeight.val;
			TXTAnswer03.lineSpacing = GlobalOptLineHeight.val;

			// Changing font
			Font tmpFont = fontsFiles[ fontsChoices.IndexOf(GlobalOptFont.val) ];
			Color fontNewCol = new Color( GlobalOptTxtColor.colorPicker.currentColor.r, GlobalOptTxtColor.colorPicker.currentColor.g, GlobalOptTxtColor.colorPicker.currentColor.b, 1f );
			Color fontAnswersNewCol = new Color( GlobalOptAnswersTxtColor.colorPicker.currentColor.r, GlobalOptAnswersTxtColor.colorPicker.currentColor.g, GlobalOptAnswersTxtColor.colorPicker.currentColor.b, 1f );
			TXTMainDialog.font = tmpFont;
			TXTMainDialog.color = fontNewCol;
			TXTAnswer01.font = tmpFont;
			TXTAnswer01.color = fontAnswersNewCol;
			TXTAnswer02.font = tmpFont;
			TXTAnswer02.color = fontAnswersNewCol;
			TXTAnswer03.font = tmpFont;
			TXTAnswer03.color = fontAnswersNewCol;
			
			// Changing backgrounds
			
			// Background sprite
			// Image
			string bgValue = GlobalOptBg.val;
			if (bgValue != "Default" && backgroundsChoices.IndexOf(bgValue) != -1)
			{
				IMGMainDialog.sprite = backgroundsSprites[ backgroundsChoices.IndexOf(bgValue) - 1 ];
				IMGAnswer01.sprite = backgroundsSprites[ backgroundsChoices.IndexOf(bgValue) - 1 ];
				IMGAnswer02.sprite = backgroundsSprites[ backgroundsChoices.IndexOf(bgValue) - 1 ];
				IMGAnswer03.sprite = backgroundsSprites[ backgroundsChoices.IndexOf(bgValue) - 1 ];
			}
			else
			{
				IMGMainDialog.sprite = backgroundsSprites[ backgroundsChoices.IndexOf("basic_01") - 1 ];
				IMGAnswer01.sprite = backgroundsSprites[ backgroundsChoices.IndexOf("basic_01") - 1 ];
				IMGAnswer02.sprite = backgroundsSprites[ backgroundsChoices.IndexOf("basic_01") - 1 ];
				IMGAnswer03.sprite = backgroundsSprites[ backgroundsChoices.IndexOf("basic_01") - 1 ];
			}
			
			// Material value
			Material tmpMat = materialList[materialChoices.IndexOf(GlobalOptMaterial.val)];
			// I have to do that to instantiate materials
			Material matMainDialog = new Material(tmpMat);
			Material matAnswers = new Material(tmpMat);
			// Setting mat
			IMGMainDialog.material = matMainDialog;
			IMGAnswer01.material = matAnswers;
			IMGAnswer02.material = matAnswers;
			IMGAnswer03.material = matAnswers;
			
			// Material setup
			Color detMainDialogNewCol = new Color( GlobalOptMatDetColorMainDialog.colorPicker.currentColor.r, GlobalOptMatDetColorMainDialog.colorPicker.currentColor.g, GlobalOptMatDetColorMainDialog.colorPicker.currentColor.b, 1f );
			Color detAnswersNewCol = new Color( GlobalOptMatDetColorAnswers.colorPicker.currentColor.r, GlobalOptMatDetColorAnswers.colorPicker.currentColor.g, GlobalOptMatDetColorAnswers.colorPicker.currentColor.b, 1f );
			SetMaterialOptions(IMGMainDialog,detMainDialogNewCol);
			SetMaterialOptions(IMGAnswer01,detAnswersNewCol);
			SetMaterialOptions(IMGAnswer02,detAnswersNewCol);
			SetMaterialOptions(IMGAnswer03,detAnswersNewCol);

			// Background colors
			Color bgNewCol = new Color( GlobalOptBgColor.colorPicker.currentColor.r, GlobalOptBgColor.colorPicker.currentColor.g, GlobalOptBgColor.colorPicker.currentColor.b, GlobalOptBgOpacity.val );
			Color bgNewColOver = new Color( GlobalOptBgOverColor.colorPicker.currentColor.r, GlobalOptBgOverColor.colorPicker.currentColor.g, GlobalOptBgOverColor.colorPicker.currentColor.b, GlobalOptBgOverOpacity.val );
			Color bgNewColClick = new Color( GlobalOptBgClickColor.colorPicker.currentColor.r, GlobalOptBgClickColor.colorPicker.currentColor.g, GlobalOptBgClickColor.colorPicker.currentColor.b, GlobalOptBgClickOpacity.val );

			SetButtonColorBlock(BTNMainDialog, bgNewCol, bgNewColOver, bgNewColClick, GlobalOptBgTransitionSpeed.val);
			
			Color bgAnswersNewCol = new Color( GlobalOptAnswersBgColor.colorPicker.currentColor.r, GlobalOptAnswersBgColor.colorPicker.currentColor.g, GlobalOptAnswersBgColor.colorPicker.currentColor.b, GlobalOptAnswersBgOpacity.val );
			Color bgAnswersNewColOver = new Color( GlobalOptAnswersBgOverColor.colorPicker.currentColor.r, GlobalOptAnswersBgOverColor.colorPicker.currentColor.g, GlobalOptAnswersBgOverColor.colorPicker.currentColor.b, GlobalOptAnswersBgOverOpacity.val );
			Color bgAnswersNewColClick = new Color( GlobalOptAnswersBgClickColor.colorPicker.currentColor.r, GlobalOptAnswersBgClickColor.colorPicker.currentColor.g, GlobalOptAnswersBgClickColor.colorPicker.currentColor.b, GlobalOptAnswersBgClickOpacity.val );
			
			SetButtonColorBlock(BTNAnswer01, bgAnswersNewCol, bgAnswersNewColOver, bgAnswersNewColClick, GlobalOptBgTransitionSpeed.val);
			SetButtonColorBlock(BTNAnswer02, bgAnswersNewCol, bgAnswersNewColOver, bgAnswersNewColClick, GlobalOptBgTransitionSpeed.val);
			SetButtonColorBlock(BTNAnswer03, bgAnswersNewCol, bgAnswersNewColOver, bgAnswersNewColClick, GlobalOptBgTransitionSpeed.val);
			
			SetButtonColorBlock(BTNHideUI, bgNewCol, bgNewColOver, bgNewColClick, GlobalOptBgTransitionSpeed.val);
			SetButtonColorBlock(BTNShowUI, bgNewCol, bgNewColOver, bgNewColClick, GlobalOptBgTransitionSpeed.val);
		}

		protected void SetMaterialOptions( Image targetImg, Color newCol )
		{
			targetImg.material.SetColor("_DetailsColor",newCol);
			targetImg.material.SetFloat("_DetailsSpeedX", GlobalOptMatDetSSX.val);
			targetImg.material.SetFloat("_DetailsSpeedY", GlobalOptMatDetSSY.val);
			targetImg.material.SetFloat("_SubdetailsOpacity", GlobalOptMatSubDetOpacity.val);
			targetImg.material.SetFloat("_RefractionIntensity", GlobalOptMatRefraction.val);
		}

		protected void SetButtonColorBlock(Button targetBtn, Color normalColor, Color overColor, Color clickColor, float transitionSpeed )
		{
			ColorBlock tmpCb = targetBtn.colors;
			tmpCb.normalColor = normalColor;
			tmpCb.highlightedColor = overColor;
			tmpCb.pressedColor = clickColor;
			tmpCb.fadeDuration = transitionSpeed;
			targetBtn.colors = tmpCb;
		}

		protected void OnUIRotationDisabledChanged(bool enabled)
		{
			if( VAMSUIGO != null ) VAMSUIGO.transform.localRotation = Quaternion.Euler(0f,0f,0f);
		}
		
		protected void OnTestCurrentGroup()
		{
			OnStartDialogGroup(GroupSelector.val);
		}
		
		protected void OnEditDisplayedDialog()
		{
			GroupSelector.val = groupsIdsToNames.FirstOrDefault(el => el.Key == CurrentGroupID).Value;
			DialogSelector.val = ""+CurrentDialogID;
		}
		
		protected void OnToggleGroupId(string groupId)
		{
			if ( isInit == false ) return;

			// Updating the UI for the group
			UpdateLocalUIFromActiveGroup();
			
			// Creating the dialog choices
			dialogsChoices = new List<string>();
			for (var i = 0; i < activeGroup.GetDialogCount(); i++)
			{
				dialogsChoices.Add("" + i);				
			}
			DialogSelector.valNoCallback = "";
			DialogSelector.choices = dialogsChoices;
			DialogSelector.val = "0";
			
			// Updating the UI for the dialog
			UpdateLocalUIFromActiveDialog();
		}
		
		protected void OnToggleDialogId(string dialogId)
		{
			if ( isInit == false ) return;
			UpdateLocalUIFromActiveDialog();
		}

		protected void OnAddNewGroup()
		{
			int groupId = storyGroups.Count;
			int dialogId = 0;

			StoryGroup newGroup = new StoryGroup(this, groupId);
			newGroup.CreateNewDialog();
			storyGroups.Add(newGroup);
			RefreshSelectors(groupId, dialogId);
		}
		protected void OnDeleteCurrentGroup()
		{
			// We cannot delete the last dialog
			if (storyGroups.Count == 1)
			{
				SuperController.singleton.Alert("<b><color=green>VAMStory</color></b> needs to have at least one group.", () => {});
				return;
			}
			
			// Waiting for user confirmation
			SuperController.singleton.Alert("<b><color=#A90000FF>You're going to delete group " + GroupSelector.val + "\nand ALL THE DIALOGS inside of it.</color></b>\nAre you sure?", delegate {OnConfirmGroupDeletion();}, () => {});
		}

		protected void OnConfirmGroupDeletion()
		{
			// Saving a critical backup
			SaveBackup(1);
			
			// Clearing the "timers" pool (just to prevent errors)
			_timedTriggers.Clear();
			
			// Now delete
			int groupId = activeGroupId;
			
			storyGroups[groupId].Delete();
			storyGroups.RemoveAt(groupId);
			int grpCount = 0;
			foreach (StoryGroup grp in storyGroups)
			{
				grp.UpdateID(grpCount);
				grpCount++;
			}

			int selectedGroup = 0;
			if (storyGroups.Count > 1 && groupId > 0)
			{
				selectedGroup = groupId - 1;
			}

			RefreshSelectors(selectedGroup, 0);
		}
		
		protected void OnMoveCurrentGroup(bool downDirection = false)
		{
			if (_isMovingGrpOrDialog == true) return;
			if (downDirection == false && activeGroupId == 0) return;
			if (downDirection == true && activeGroupId == storyGroups.Count - 1)  return;
			_isMovingGrpOrDialog = true;
			
			StoryGroup currentGrp = activeGroup;
			
			int currentID = currentGrp.GroupID;
			int newGroupID = downDirection == false ? currentID - 1 : currentID + 1;

			StoryGroup previousGrp = storyGroups[newGroupID];

			// Updating ID of the previous action to properly synchronize
			// the new "positions" of the actions ( either scene wise or data wise )
			// Faking a high ID to ensure that the refresh of IDs will work
			// It will be properly resynced with the "RefreshAllActionsIDs" below
			previousGrp.UpdateID(100000); // Previous one (the place I'm moving to)
			currentGrp.UpdateID(newGroupID); // Current one (the one moving)
			
			// Updating the list
			storyGroups.RemoveAt(currentGrp.GroupID);
			storyGroups.Insert(newGroupID,currentGrp);
			storyGroups[currentID] = previousGrp;

			// Refreshing the UI and scene
			RefreshAllGroupsIDs();
			RefreshSelectors(newGroupID,int.Parse(DialogSelector.val));
			
			_isMovingGrpOrDialog = false;
		}
		
		protected void RefreshAllGroupsIDs()
		{
			int grpCount = 0;
			foreach (StoryGroup grp in storyGroups)
			{
				grp.UpdateID(grpCount);
				grpCount++;
			}
		}
		
		protected void OnAddNewDialog()
		{
			int groupId = activeGroupId;
			int dialogId = storyGroups[groupId].GetDialogCount();
			
			storyGroups[groupId].CreateNewDialog();
			RefreshSelectors(groupId, dialogId);
		}

		protected void OnDeleteCurrentDialog()
		{
			// Saving a critical backup
			SaveBackup(1);
			
			// Now delete
			int groupId = activeGroupId;
			int dialogId = int.Parse(DialogSelector.val);
			
			// We cannot delete the last dialog
			if (storyGroups[groupId].GetDialogCount() == 1)
			{
				SuperController.singleton.Alert("<b><color=green>Group " + groupId + "</color></b> needs to have at least one dialog line. Delete the group instead.", () => {});
				return;
			}
			
			// Waiting for user confirmation
			SuperController.singleton.Alert("<color=#A90000FF>Deleting dialog " + dialogId + " from group " + groupId + ":\n\n</color><size=30>" + storyGroups[groupId].GetDialogID(dialogId).DialogText.val + "</size>\n\nAre you sure?", delegate {OnConfirmDialogDeletion();}, () => {});
			

		}

		protected void OnConfirmDialogDeletion()
		{
			// Saving a critical backup
			SaveBackup(1);
			
			// Clearing the "timers" pool (just to prevent errors)
			_timedTriggers.Clear();
			
			// Now Delete
			int groupId = activeGroupId;
			int dialogId = int.Parse(DialogSelector.val);
			
			storyGroups[groupId].DeleteDialogID(dialogId);
			
			int selectedDialog = 0;
			if (storyGroups[groupId].GetDialogCount() > 1 && dialogId > 0)
			{
				selectedDialog = dialogId - 1;
			}
			
			RefreshSelectors(groupId, selectedDialog);
		}
		
		protected void OnMoveCurrentDialog(bool downDirection = false)
		{
			if (_isMovingGrpOrDialog == true) return;
			if (downDirection == false && activeDialog.DialogID == 0) return;
			if (downDirection == true && activeDialog.DialogID == activeGroup.GetDialogCount() - 1) return;
			_isMovingGrpOrDialog = true;
			
			StoryDialog currentDial = activeDialog;
			
			int currentID = currentDial.DialogID;
			int newDialogID = downDirection == false ? currentID - 1 : currentID + 1;
			activeGroup.MoveDialogID(currentID,downDirection);

			// Refreshing the UI
			RefreshSelectors(activeGroupId,newDialogID);
			
			_isMovingGrpOrDialog = false;
		}
		
		protected void OnCopyDialog()
		{
			if (CopySelector.val == "")
			{
				SuperController.singleton.Alert("Select a source dialog to copy first!", () => {});
				return;
			}
			
			string[] copyIds = CopySelector.val.Split('_');
			int sourceGroupId = groupsIdsToNames.FirstOrDefault(el => el.Value == copyIds[1]).Key;
			int sourceDialogId = int.Parse(copyIds[0]);
			int targetGroupId = activeGroupId;
			int targetDialogId = int.Parse(DialogSelector.val);

			if (sourceGroupId == targetGroupId && sourceDialogId == targetDialogId)
			{
				SuperController.singleton.Alert("You cannot copy a dialog on itself!", () => {});
			}
			else
			{
				StoryDialog sourceDial = storyGroups[sourceGroupId].GetDialogID(sourceDialogId);
				StoryDialog targetDial = storyGroups[targetGroupId].GetDialogID(targetDialogId);

				DialogTextInputFieldUI.text = sourceDial.DialogText.val;
				targetDial.DialogDelay.val = sourceDial.DialogDelay.val;
				targetDial.DialogAvatarUrl.val = sourceDial.DialogAvatarUrl.val;
				targetDial.DialogAvatarPosition.val = sourceDial.DialogAvatarPosition.val;
				targetDial.triggerOnShowDialog.RestoreFromJSON(sourceDial.triggerOnShowDialog.GetJSON());
				targetDial.triggerOnClickDialog.RestoreFromJSON(sourceDial.triggerOnClickDialog.GetJSON());
				
				// Updating dialog/group UI
				UpdateLocalUIFromActiveGroup();
				UpdateLocalUIFromActiveDialog();
			}
		}

		protected void OnTransferDialog()
		{
			if (CopySelector.val == "")
			{
				SuperController.singleton.Alert("Select a target dialog to transfer to first!", () => {});
				return;
			}
			
			string[] copyIds = CopySelector.val.Split('_');
			int sourceGroupId = activeGroupId;
			int sourceDialogId = int.Parse(DialogSelector.val);
			int targetGroupId = groupsIdsToNames.FirstOrDefault(el => el.Value == copyIds[1]).Key;
			int targetDialogId = int.Parse(copyIds[0]);

			if (sourceGroupId == targetGroupId && sourceDialogId == targetDialogId)
			{
				SuperController.singleton.Alert("You cannot transfer a dialog on itself!", () => {});
			}
			else
			{
				StoryDialog sourceDial = storyGroups[sourceGroupId].GetDialogID(sourceDialogId);
				StoryDialog targetDial = storyGroups[targetGroupId].GetDialogID(targetDialogId);

				targetDial.DialogText.val = sourceDial.DialogText.val;
				targetDial.DialogDelay.val = sourceDial.DialogDelay.val;
				targetDial.DialogAvatarUrl.val = sourceDial.DialogAvatarUrl.val;
				targetDial.DialogAvatarPosition.val = sourceDial.DialogAvatarPosition.val;
				targetDial.triggerOnShowDialog.RestoreFromJSON(sourceDial.triggerOnShowDialog.GetJSON());
				targetDial.triggerOnClickDialog.RestoreFromJSON(sourceDial.triggerOnClickDialog.GetJSON());
				
				// Updating dialog/group UI
				UpdateLocalUIFromActiveGroup();
				UpdateLocalUIFromActiveDialog();
				
				// Alert for the transfer (since you don't have any visual feedback)
				SuperController.singleton.Alert("<color=green>VAMStory</color>\n Datas transfered to Dialog ID " + targetDialogId + " (in " + copyIds[1] + ")", () => {});
			}
		}

		protected void OnStartDialogGroup(string startIdent)
		{
			if (!groupsIdsToNames.ContainsValue(startIdent))
			{
				logError("VAMStory: Dialog group cannot start, the group id: " + startIdent + " does not exists. (atom name: " + containingAtom.name + ")");
				logError("This might happen if you have an obsolete trigger referencing a group that has been renamed.");
				return;
			}
			
			StartDialogGroup.valNoCallback = ""; // Resetting so that the user can trigger several times the same group if needed
			CurrentGroupID = groupsIdsToNames.FirstOrDefault(el => el.Value == startIdent).Key;
			CurrentDialogID = 0;
			
			// Hiding
			ToggleStoryUI(false);
			
			// Moving the group
			storyGroups[CurrentGroupID].MoveGroup();
			
			// Executing group trigger
			storyGroups[CurrentGroupID].GroupTrigger.Trigger();
			AddTriggerToTimerPool(storyGroups[CurrentGroupID].GroupTrigger);
			
			// Checking for delay
			float delayDialog = storyGroups[CurrentGroupID].GetDialogID(CurrentDialogID).DialogDelay.val;
			if (delayDialog > 0f)
			{
				StartCoroutine(WaitForNextDialog(delayDialog));
			}
			else
			{
				SetupUIForCurrentDialogID();				
			}
		}
		
		private void OnUpdateDialogText()
		{
			activeDialog.DialogText.val = DialogTextUI.val;
		}
		
		private void OnSelectDialogAvatar()
		{
			SuperController.singleton.fileBrowserUI.defaultPath = lastFileExplorerPath;
			SuperController.singleton.fileBrowserUI.SetTitle("Select a texture");
			SuperController.singleton.fileBrowserUI.fileFormat = "png|jpg|tga|tiff";
			SuperController.singleton.fileBrowserUI.showFiles = true;
			SuperController.singleton.fileBrowserUI.showDirs = true;
			SuperController.singleton.fileBrowserUI.SetTextEntry(false);
			SuperController.singleton.fileBrowserUI.hideExtension = false;
			SuperController.singleton.fileBrowserUI.showInstallFolderInDirectoryList = false;
			List<ShortCut> shortCutList = new List<ShortCut>();
			ShortCut shortAp = new ShortCut();
			shortAp.package = "AddonPackages Filtered";
			shortAp.path = "AddonPackages";
			shortAp.displayName = "Filter: Custom";
			shortAp.packageFilter = "Custom";
			shortCutList.Add(shortAp);
			
			ShortCut shortAps = new ShortCut();
			shortAps.package = "AddonPackages Filtered";
			shortAps.path = "AddonPackages";
			shortAps.displayName = "Filter: Saves";
			shortAps.packageFilter = "Saves";
			shortCutList.Add(shortAps);
			
			ShortCut shortCutCustom = new ShortCut();
			shortCutCustom.path = "Custom\\";
			shortCutCustom.displayName = "Custom";
			shortCutCustom.flatten = false;
			shortCutList.Add(shortCutCustom);

			if (FileManagerSecure.DirectoryExists("Custom/Textures"))
			{
				ShortCut shortCutTex = new ShortCut();
				shortCutTex.path = "Custom\\Textures\\";
				shortCutTex.displayName = "Custom/Textures";
				shortCutTex.flatten = false;
				shortCutList.Add(shortCutTex);
			}
			
			ShortCut shortCutSaves = new ShortCut();
			shortCutSaves.path = "Saves\\";
			shortCutSaves.displayName = "Saves";
			shortCutSaves.flatten = false;
			shortCutList.Add(shortCutSaves);

			SuperController.singleton.fileBrowserUI.shortCuts = shortCutList;			
			SuperController.singleton.fileBrowserUI.Show((val) =>
			{
				if (val != "")
				{
					lastFileExplorerPath = val.Substring(0, val.LastIndexOfAny(new char[] {'/', '\\'}));
					DialogAvatarUrlUI.val = val;
					activeDialog.DialogAvatarUrl.val = val;
					// I'm forcing the load since the callback will not be triggered if
					// the same path is used ( a creator updating a png and reloading it throught the same URL )
					OnLoadAvatar(val);
				}
			});
		}

		private void OnLoadAvatar( string avatarUrl )
		{
			tmpTexture = GetResource(avatarUrl);
			tmpTexture.wrapMode = TextureWrapMode.Clamp;
			
			// Adding the new avatar if I don't have it in my Dict
			if (!DialogAvatars.ContainsKey(avatarUrl))
			{
				DialogAvatars.Add(avatarUrl, tmpTexture);
			}
			// Or updating
			else
			{
				Destroy(DialogAvatars[avatarUrl]);
				DialogAvatars[avatarUrl] = tmpTexture;
			}
		}

		private void PlayOnShowSound()
		{
			ASAudio.PlayOneShot(soundClips[soundChoices.IndexOf(GlobalOptSoundOnShow.val)], GlobalOptSoundOnShowVolume.val);
		}

		private void PlayOnClickSound()
		{
			ASAudio.PlayOneShot(soundClips[soundChoices.IndexOf(GlobalOptSoundOnClick.val)], GlobalOptSoundOnClickVolume.val);			
		}
		
		protected void OnSaveGroupPosition() {
			setButtonColor(positionSaveButton, savedPositionColor );
			StoryGroup grp = activeGroup;
			grp.GroupPosition.val = containingAtom.mainController.transform.position;
			grp.GroupRotation.val = containingAtom.mainController.transform.eulerAngles;
			SuperController.singleton.Alert("<b><color=green>VAMStory</color></b>\n\nPosition saved for Group " + grp.GroupID + ".", () => {});
		}
			
		protected void OnClearGroupPosition() {
			setButtonColor(positionSaveButton, unsavedPositionColor);
			StoryGroup grp = activeGroup;
			grp.GroupPosition.val = default(Vector3);
			grp.GroupRotation.val = default(Vector3);
			SuperController.singleton.Alert("<b><color=green>VAMStory</color></b>\n\nPosition cleared for Group " + grp.GroupID + ".", () => {});
		}
		
		private void OnAtomRename(string oldid, string newid)
		{
			GlobalOptStoryTrigger.SyncAtomNames();
			
			foreach( var grp in storyGroups ) {
				grp.GroupTrigger.SyncAtomNames();
				grp.GroupAnswer01Trigger.SyncAtomNames();
				grp.GroupAnswer02Trigger.SyncAtomNames();
				grp.GroupAnswer03Trigger.SyncAtomNames();
					
				foreach (StoryDialog dial in grp.GetDialogs())
				{
					dial.triggerOnShowDialog.SyncAtomNames();
					dial.triggerOnClickDialog.SyncAtomNames();
				}
			}
	    }
		
		// **************************************
		// Callbacks & similar ( gameplay )
		// **************************************
		protected void OnPlayStorySound(string soundName)
		{
			ASAudio.PlayOneShot(soundClips[soundChoices.IndexOf(soundName)], GlobalOptSoundOnClickVolume.val);
			A_PlayStorySound.valNoCallback = "";
		}
		
		protected void OnClickMainButton()
		{
			// Freeze ? no action
			if (isSceneFrozen) return;
			
			// Storing the current ID to prevent executing the next dialog in a case where the main button starts a new group
			int prevGroupID = CurrentGroupID;
			int prevDialogID = CurrentDialogID;
			
			// Playing Sound
			PlayOnClickSound();
			
			// Executing click trigger before switching to the next dialog
			storyGroups[CurrentGroupID].DoClickTriggerDialogID(CurrentDialogID);
			
			// Preventing next dialog if a new group has been started by the trigger
			if (CurrentGroupID != prevGroupID || CurrentDialogID != prevDialogID) return;
			
			// Checking for next dialog
			// If we haven't reached the end of the current group, let's continue
			if (CurrentDialogID + 1 < storyGroups[CurrentGroupID].GetDialogCount())
			{
				// Switching to next
				CurrentDialogID += 1;
				
				// Checking for delay
				float delayDialog = storyGroups[CurrentGroupID].GetDialogID(CurrentDialogID).DialogDelay.val;
				if (delayDialog > 0f)
				{
					ToggleStoryUI(false);
					StartCoroutine(WaitForNextDialog(delayDialog));
				}
				else
				{
					SetupUIForCurrentDialogID();				
				}
			// If we want to hide the story when clicking the last dialog of the group	
			}
			else if ( GlobalOptLastGroupDialogClose.val == true )
			{
				ToggleStoryUI(false);
			}

			StartCoroutine("UnselectButton"); // WTF (see routine)
		}

		protected void OnClickAnswer01Button()
		{
			// Freeze ? no action
			if (isSceneFrozen) return;
			
			// Playing Sound
			PlayOnClickSound();
			storyGroups[CurrentGroupID].DoClickTriggerAnswerID(1);
			
			StartCoroutine("UnselectButton"); // WTF (see routine)
		}
		
		protected void OnClickAnswer02Button()
		{
			// Freeze ? no action
			if (isSceneFrozen) return;
			
			// Playing Sound
			PlayOnClickSound();
			storyGroups[CurrentGroupID].DoClickTriggerAnswerID(2);
			
			StartCoroutine("UnselectButton"); // WTF (see routine)
		}
		
		protected void OnClickAnswer03Button()
		{
			// Freeze ? no action
			if (isSceneFrozen) return;
			
			// Playing Sound
			PlayOnClickSound();
			storyGroups[CurrentGroupID].DoClickTriggerAnswerID(3);
			
			StartCoroutine("UnselectButton"); // WTF (see routine)
		}

		protected void OnClickHideUIButton()
		{
			// Freeze ? no action
			if (isSceneFrozen) return;
			
			PlayOnClickSound();
			CGDialogPanel.alpha = 0f;
			CGDialogPanel.blocksRaycasts = false;
			CGDialogPanel.interactable = false;

			// If actions exists
			if ( CGStoryActionsMainPanel != null )
			{
				CGStoryActionsMainPanel.alpha = 0f;
				CGStoryActionsMainPanel.blocksRaycasts = false;
				CGStoryActionsMainPanel.interactable = false;
			}
			
			TRHideUIGrp.gameObject.SetActive(false);
			TRShowUIGrp.gameObject.SetActive(true);
		}
		
		protected void OnClickShowUIButton()
		{
			// Freeze ? no action
			if (isSceneFrozen) return;
			
			PlayOnShowSound();
			CGDialogPanel.alpha = 1f;
			CGDialogPanel.blocksRaycasts = true;
			CGDialogPanel.interactable = true;
			
			// If actions exists
			if ( CGStoryActionsMainPanel != null )
			{
				CGStoryActionsMainPanel.alpha = 1f;
				CGStoryActionsMainPanel.blocksRaycasts = true;
				CGStoryActionsMainPanel.interactable = true;
			}
			TRHideUIGrp.gameObject.SetActive(true);
			TRShowUIGrp.gameObject.SetActive(false);
		}
		
		// **************************
		// COROUTINES
		// **************************
		protected IEnumerator WaitForVAMSPlugins()
		{
			// Checking VAMExcite plugins every one second while loading or whatever
			while (_allVAMSReady == false)
			{
				// We're loading...
				// The reason behind that is that, a plugin looking for other atoms can load fast enough
				// to catch the storables list in the target atom PRIOR to the plugins storables being even present in the List 
				// which would result in a "false positive" of an Empty atom not owning a VAMS plugin
				if (_isSceneLoading)
				{
					yield return new WaitForSeconds(_vamsReadyCheckPeriod);					
				}
				else
				{
					bool allReady = true;
					foreach (Atom at in SuperController.singleton.GetAtoms())
					{
						if (at.type == "Empty" && at != containingAtom)
						{
							foreach (string stId in at.GetStorableIDs())
							{
								if (stId.Contains(".VAMStory"))
								{
									JSONStorable tmpVAMS = at.GetStorableByID(stId);
									if (tmpVAMS != null && tmpVAMS.GetBoolParamValue("VAMSValue_Ready") == false) allReady = false;
								}
							}
						}
					}
				
					if (allReady == false)
					{
						yield return new WaitForSeconds(_vamsReadyCheckPeriod);					
					}
					else
					{
						_allVAMSReady = true;
						yield break;
					}
				}
			}
		}
		
		public IEnumerator WaitForNextDialog(float waitTime )
		{
			yield return new WaitForSeconds(waitTime);
			SetupUIForCurrentDialogID();
		}
		
		public IEnumerator AutoBackup()
		{
			while (true)
			{
				yield return new WaitForSeconds(900);
				SaveBackup();
			}
		}
		
		// I don't know what the fuck is going on with Unity's UI, if you use the
		// event system immediately on the click delegate, the UI does not unfocus
		// This is the only frikkin alternative I found
		// THANK YOU NICE ENGINE OF THE HELL.
		public IEnumerator UnselectButton()
		{
			yield return new WaitForSeconds(0.1f);
			UnityEngine.EventSystems.EventSystem.current.SetSelectedGameObject(null);
		}

		// **************************
		// OVERRIDES SAVING/LOADING
		// **************************
		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			var json = base.GetJSON(includePhysical, includeAppearance, forceStore);
			if (includePhysical || forceStore)
			{
				needsStore = true;
				
				json[GlobalOptStoryTrigger.Name] = GlobalOptStoryTrigger.GetJSON(base.subScenePrefix);
				
				foreach( var grp in storyGroups ) {
					json[grp.GroupTrigger.Name] = grp.GroupTrigger.GetJSON(base.subScenePrefix);
					json[grp.GroupAnswer01Trigger.Name] = grp.GroupAnswer01Trigger.GetJSON(base.subScenePrefix);
					json[grp.GroupAnswer02Trigger.Name] = grp.GroupAnswer02Trigger.GetJSON(base.subScenePrefix);
					json[grp.GroupAnswer03Trigger.Name] = grp.GroupAnswer03Trigger.GetJSON(base.subScenePrefix);
					
					foreach (StoryDialog dial in grp.GetDialogs())
					{
						json[dial.triggerOnShowDialog.Name] = dial.triggerOnShowDialog.GetJSON(base.subScenePrefix);
						json[dial.triggerOnClickDialog.Name] = dial.triggerOnClickDialog.GetJSON(base.subScenePrefix);
					}
				}
			}
			return json;
		}
		
		public override void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, JSONArray presetAtoms = null, bool setMissingToDefault = true)
		{
			base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, presetAtoms);
			// Saving the original datas to load all dynamic storables
			dialogsDatas = jc;
		}
		
		public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
		{
			base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
			if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
			{
				GlobalOptStoryTrigger.RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
			}
		}
		
		// **************************
		// Local Tools
		// **************************
		private void logDebug( string debugText ) {
			SuperController.LogMessage( debugText );
		}
		
		private void logError( string debugText ) {
			SuperController.LogError( debugText );
		}

		private string GetCleanAndUniqueGroupName(int groupId)
		{
			bool nameExists = true;
			int newId = 0;
			string str = CleanString(GroupNameUIInputField.text);
			bool nameFound = false;
			while (nameFound == false)
			{
				str = str + ( newId > 0 ? ""+newId : "");
				int existsCount = 0;
				foreach (var idsNames in groupsIdsToNames)
				{
					if (idsNames.Key == groupId) continue; // we skip the same id (or the thing will try to find another unique name when stopping on it)
					if (idsNames.Value == str)
					{
						existsCount++;
					}
				}

				if (existsCount == 0) nameFound = true;

				newId++;
				// Security ?
				if (newId > 500)
				{
					logDebug("VAMStory: Did over 500 loops to find a unique group name, is there a problem?");
					break;
				}
			}
			
			return str;
		}
		
		private string GetCleanFilenameFromString(string str)
		{
			string tmp = CleanString(str);
			tmp = tmp.ToLower();
			return tmp;
		}
		
		private string CleanString(string str)
		{
			str = str.Trim();
			Regex reg1 = new Regex("[^a-zA-Z0-9\\s\\-]");
			Regex reg2 = new Regex("\\s+");
			str = reg1.Replace(str, "");
			str = reg2.Replace(str, "-");
			return str;
		}
		
		// **************************
		// Time to cleanup !
		// **************************
		void OnDestroy() {
			// Sync names
			SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;
			
			// Clearing actions
			ClearActionsUI();
			
			// Stop backup routine
			StopCoroutine("AutoBackup");
			
			// Removing the canvas from VAM's controller
			SuperController.singleton.RemoveCanvas(CANStoryUI);
			
			// Cleaning GameObjects
			Destroy(VAMSUIGO);
			
			// Cleaning triggers
			GlobalOptStoryTrigger.Remove();
			foreach( var grp in storyGroups ) {
				grp.GroupTrigger.Remove();
				grp.GroupAnswer01Trigger.Remove();
				grp.GroupAnswer02Trigger.Remove();
				grp.GroupAnswer03Trigger.Remove();
					
				foreach (StoryDialog dial in grp.GetDialogs())
				{
					dial.triggerOnShowDialog.Remove();
					dial.triggerOnClickDialog.Remove();
				}
			}

			// Unloading assetbundle
			AssetLoader.DoneWithAssetBundleFromFile(PLUGIN_PATH + UI_BUNDLE_PATH);
		}

		// ***********************************************************
		// EXTERNAL TOOL - Thank you great coders for your content!
		// ***********************************************************

		// *********** Chokapi Decal Maker START *********************
		public Texture2D GetResource(string path)
		{
			if (path == "" || path == null || !FileManagerSecure.FileExists(path) )
			{
				path = GetPackagePath(this) + "Custom/Scripts/VAMStory/default.png";
			}
			Texture2D texture;
			byte[] tmppng = FileManagerSecure.ReadAllBytes(path);
			texture = new Texture2D(1, 1, TextureFormat.RGBA32, false);
			texture.LoadImage(tmppng);
			return texture;
		}
		// *********** Chokapi Decal Maker END *********************
		
		// *********** MacGruber_Utils.cs START *********************
		// Get directory path where the plugin is located. Based on Alazi's & VAMDeluxe's method.
		public static string GetPluginPath(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
			string filename = self.manager.GetJSON()["plugins"][id].Value;
			return filename.Substring(0, filename.LastIndexOfAny(new char[] { '/', '\\' }));
		}
				
		// Get path prefix of the package that contains our plugin.
		public static string GetPackagePath(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
			string filename = self.manager.GetJSON()["plugins"][id].Value;
			int idx = filename.IndexOf(":/");
			if (idx >= 0)
				return filename.Substring(0, idx+2);
			else
				return string.Empty;
		}
				
		// Check if our plugin is running from inside a package
		public static bool IsInPackage(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
			string filename = self.manager.GetJSON()["plugins"][id].Value;
			return filename.IndexOf(":/") >= 0;
		}
		
		// ===========================================================================================
		// TriggerHandler implementation for easier handling of custom triggers.
		// Essentially call this in your plugin init code:
		//     StartCoroutine(SimpleTriggerHandler.LoadAssets());
		//
		// Credit to AcidBubbles for figuring out how to do custom triggers.
		//
		// hazmhox note : Edited a bit to allow the use of transitions ( original code coming from LogicBricks 4 )
		public class SimpleTriggerHandler : TriggerHandler
		{
			public static bool Loaded { get; private set; }

			private static SimpleTriggerHandler myInstance;

			private RectTransform myTriggerActionsPrefab;
			private RectTransform myTriggerActionMiniPrefab;
			private RectTransform myTriggerActionDiscretePrefab;
			private RectTransform myTriggerActionTransitionPrefab;

			public static SimpleTriggerHandler Instance {
				get {
					if (myInstance == null)
						myInstance = new SimpleTriggerHandler();
					return myInstance;
				}
			}

			public static void LoadAssets()
			{
				SuperController.singleton.StartCoroutine(Instance.LoadAssetsInternal());
			}

			private IEnumerator LoadAssetsInternal()
			{
				foreach (var x in LoadAsset("z_ui2", "TriggerActionsPanel", p => myTriggerActionsPrefab = p))
					yield return x;
				foreach (var x in LoadAsset("z_ui2", "TriggerActionMiniPanel", p => myTriggerActionMiniPrefab = p))
					yield return x;
				foreach (var x in LoadAsset("z_ui2", "TriggerActionDiscretePanel", p => myTriggerActionDiscretePrefab = p))
					yield return x;
				foreach (var x in LoadAsset("z_ui2", "TriggerActionTransitionPanel", p => myTriggerActionTransitionPrefab = p))
					yield return x;

				Loaded = true;
			}

			private IEnumerable LoadAsset(string assetBundleName, string assetName, Action<RectTransform> assign)
			{
				AssetBundleLoadAssetOperation request = AssetBundleManager.LoadAssetAsync(assetBundleName, assetName, typeof(GameObject));
				if (request == null)
					throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null request.");
				yield return request;
				GameObject go = request.GetAsset<GameObject>();
				if (go == null)
					throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null GameObject.");
				RectTransform prefab = go.GetComponent<RectTransform>();
				if (prefab == null)
					throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null RectTansform.");
				assign(prefab);
			}


			void TriggerHandler.RemoveTrigger(Trigger t)
			{
				// nothing to do
			}

			void TriggerHandler.DuplicateTrigger(Trigger t)
			{
				throw new NotImplementedException();
			}

			RectTransform TriggerHandler.CreateTriggerActionsUI()
			{
				return UnityEngine.Object.Instantiate(myTriggerActionsPrefab);
			}

			RectTransform TriggerHandler.CreateTriggerActionMiniUI()
			{
				return UnityEngine.Object.Instantiate(myTriggerActionMiniPrefab);
			}

			RectTransform TriggerHandler.CreateTriggerActionDiscreteUI()
			{
				return UnityEngine.Object.Instantiate(myTriggerActionDiscretePrefab);
			}

			RectTransform TriggerHandler.CreateTriggerActionTransitionUI()
			{
				return UnityEngine.Object.Instantiate(myTriggerActionTransitionPrefab);
			}

			void TriggerHandler.RemoveTriggerActionUI(RectTransform rt)
			{
				UnityEngine.Object.Destroy(rt?.gameObject);
			}
		}

		// Wrapper for easier handling of custom triggers.
		public class EventTrigger : Trigger
		{
			public string Name {
				get { return name; }
				set { name = value; myNeedInit = true; }
			}

			public MVRScript Owner {
				get; private set;
			}

			private string name;
			private bool myNeedInit = true;

			public EventTrigger(MVRScript owner, string name)
			{
				Name = name;
				Owner = owner;
				handler = SimpleTriggerHandler.Instance;
			}

			public void OpenPanelActionStart()
			{
				if (!SimpleTriggerHandler.Loaded)
				{
					SuperController.LogError("EventTrigger: You need to call SimpleTriggerHandler.LoadAssets() before use.");
					return;
				}

				triggerActionsParent = Owner.UITransform;
				InitTriggerUI();
				OpenTriggerActionsPanel();
				if (myNeedInit)
				{
					triggerActionsPanel.Find("Panel/Header Text").GetComponent<Text>().text = Name;
					triggerActionsPanel.Find("Content/Tab1/Label").GetComponent<Text>().text = "Event Actions";
					triggerActionsPanel.Find("Content/Tab2").gameObject.SetActive(false);
					triggerActionsPanel.Find("Content/Tab3").gameObject.SetActive(false);
					myNeedInit = false;
				}
			}
			
			public void OpenPanelTransition()
			{
				if (!SimpleTriggerHandler.Loaded)
				{
					SuperController.LogError("EventTrigger: You need to call SimpleTriggerHandler.LoadAssets() before use.");
					return;
				}

				triggerActionsParent = Owner.UITransform;
				InitTriggerUI();
				OpenTriggerActionsPanel();
				
				//getRoot(triggerActionsParent,0);
				if (myNeedInit)
				{
					triggerActionsPanel.Find("Panel/Header Text").GetComponent<Text>().text = Name;
					triggerActionsPanel.Find("Content/Tab2/Label").GetComponent<Text>().text = "Frequencies Actions";
					triggerActionsPanel.Find("Content/Tab1").gameObject.SetActive(false);
					triggerActionsPanel.Find("Content/Tab2").gameObject.SetActive(true);
					triggerActionsPanel.Find("Content/Tab2").GetComponent<Toggle>().isOn = true;					
					triggerActionsPanel.Find("Content/Tab3").gameObject.SetActive(false);
					myNeedInit = false;
				}
			}
	
			public void RestoreFromJSON(JSONClass jc, string subScenePrefix, bool isMerge, bool setMissingToDefault)
			{
				if (jc.HasKey(Name))
				{
					JSONClass tc = jc[Name].AsObject;
					if (tc != null)
						base.RestoreFromJSON(tc, subScenePrefix, isMerge);
				}
				else if (setMissingToDefault)
				{
					base.RestoreFromJSON(new JSONClass());
				}
			}

			public void Trigger()
			{
				active = true;
				active = false;
			}
			
			public void TriggerTransition(float val)
			{
				this.transitionInterpValue = val;
			}
		}
		// *********** MacGruber_Utils.cs END *********************
	}
}