using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using System;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.IO;
using MeshVR;
using MVR.FileManagementSecure;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;
using MVR;

// VAMStory Story Dialog
//
// Dialog / narration system for VAM

namespace VAMStoryPlugin
{
	public partial class VAMStory : MVRScript
    {
	    public class StoryDialog
	    {
		    private VAMStory _VAMS;

		    public int GroupID;
		    public int DialogID;

		    public JSONStorableString DialogText;
		    public JSONStorableFloat DialogDelay;
		    public EventTrigger triggerOnShowDialog;
		    public EventTrigger triggerOnClickDialog;
		    public JSONStorableStringChooser DialogAvatarPosition;
		    public JSONStorableUrl DialogAvatarUrl;

			protected UIDynamicButton triggerOnShowBtn;
			protected UIDynamicButton triggerOnClickBtn;
			protected UIDynamicButton dialogAvatarBtn;
			protected UIDynamicButton dialogRemoveAvatarBtn;

			protected InputField DialogTextInputField;
		    
		    private HSVColor hsvcDefault;
		    
		    protected JSONClass dialogsDatas { get; set; }
		    
		    public StoryDialog(VAMStory rootScript, int groupId, int dialogId)
		    {
			    _VAMS = rootScript;
			    dialogsDatas = _VAMS.dialogsDatas;

			    GroupID = groupId;
			    DialogID = dialogId;

			    hsvcDefault = HSVColorPicker.RGBToHSV(1f, 1f, 1f);
			    CreateDialogComponents();
		    }

		    protected void CreateDialogComponents()
		    {
			    // Temporary vars
			    UIDynamicTextField tmpTextField;
			    UIDynamicPopup tmpPopup;
			    UIDynamicSlider tmpSlider;
			    UIDynamicColorPicker tmpColor;
			    UIDynamicButton tmpButton;
			    UIDynamicToggle tmpToggle;
			    UIDynamic tmpSpacer;
			    
			    // Creating Dialog components
				
				// ******* DIALOG TEXT ***********
				DialogText = new JSONStorableString(SAVE_PREFIX + GroupID + "_" + DialogID + "-Text", "_default_");
				DialogText.valNoCallback = "Type some text...";
				
				// ******* DIALOG START DELAY ***********
				DialogDelay = new JSONStorableFloat(SAVE_PREFIX + GroupID + "_" + DialogID + "-Delay", 0.2f, 0f, 5f, false);

				// ******* TRIGGERS ***********
				triggerOnShowDialog = new EventTrigger(_VAMS, SAVE_PREFIX + GroupID + "_" + DialogID + "-TriggerOnShow");
				triggerOnClickDialog = new EventTrigger(_VAMS, SAVE_PREFIX + GroupID + "_" + DialogID + "-TriggerOnClick");

				// ******* AVATAR POSITION ***********
				DialogAvatarPosition = new JSONStorableStringChooser(SAVE_PREFIX + GroupID + "_" + DialogID + "-AvatarPosition", _VAMS.avatarPositionChoices, "Right", "Avatar position");
				
				// ******* AVATAR SELECTION ***********
				DialogAvatarUrl = new JSONStorableUrl(SAVE_PREFIX + GroupID + "_" + DialogID + "-AvatarUrl", "");
				DialogAvatarUrl.setCallbackFunction += (val) => { _VAMS.OnLoadAvatar(val); };

				// Registering everything
				RegisterStorables();
				
				// Restoring datas
				if (dialogsDatas != null)
				{
					DialogText.RestoreFromJSON( dialogsDatas, true, true, false);
					DialogDelay.RestoreFromJSON( dialogsDatas, true, true, false);
					DialogAvatarPosition.RestoreFromJSON( dialogsDatas, true, true, false);
					DialogAvatarUrl.RestoreFromJSON( dialogsDatas, true, true, false);
					triggerOnShowDialog.RestoreFromJSON( dialogsDatas, _VAMS.subScenePrefix, _VAMS.mergeRestore, true );
					triggerOnClickDialog.RestoreFromJSON( dialogsDatas, _VAMS.subScenePrefix, _VAMS.mergeRestore, true );
				}
		    }

		    public void DoClickTrigger()
		    {
			    triggerOnClickDialog.Trigger();
			    _VAMS.AddTriggerToTimerPool(triggerOnClickDialog);
		    }
		    
		    public void DoShowTrigger()
		    {
			    triggerOnShowDialog.Trigger();
			    _VAMS.AddTriggerToTimerPool(triggerOnShowDialog);
		    }

		    public void UpdateID(int newGroupId, int newDialogId)
		    {
			    GroupID = newGroupId;
			    DialogID = newDialogId;

			    // Updating dialog components (param names for the save)
			    UpdateComponentsIDs();
		    }

		    protected void UpdateComponentsIDs()
		    {
			    // Deregister
			    DeregisterStorables();
			    
			    // Updated
			    DialogText.name = SAVE_PREFIX + GroupID + "_" + DialogID + "-Text";
			    DialogDelay.name = SAVE_PREFIX + GroupID + "_" + DialogID + "-Delay";
			    DialogAvatarPosition.name = SAVE_PREFIX + GroupID + "_" + DialogID + "-AvatarPosition";
			    triggerOnShowDialog.Name = SAVE_PREFIX + GroupID + "_" + DialogID + "-TriggerOnShow";
			    triggerOnClickDialog.Name = SAVE_PREFIX + GroupID + "_" + DialogID + "-TriggerOnClick";
			    DialogAvatarUrl.name = SAVE_PREFIX + GroupID + "_" + DialogID + "-AvatarUrl";
			    
			    // Register
			    RegisterStorables();
		    }
		    
			public void Delete() {
				// Unregistering vars
				DeregisterStorables();
				
				// Removing triggers
				triggerOnShowDialog.Remove();
				triggerOnClickDialog.Remove();
				
				// Clearing data from the loaded data of the scene (to prevent re-restore after deleting)
				// This only happens when you load a scene and have data loaded from the json
				if (dialogsDatas != null)
				{
					dialogsDatas.Remove(SAVE_PREFIX + GroupID + "_" + DialogID + "-Text");
					dialogsDatas.Remove(SAVE_PREFIX + GroupID + "_" + DialogID + "-Delay");
					dialogsDatas.Remove(SAVE_PREFIX + GroupID + "_" + DialogID + "-AvatarPosition");
					dialogsDatas.Remove(SAVE_PREFIX + GroupID + "_" + DialogID + "-TriggerOnShow");
					dialogsDatas.Remove(SAVE_PREFIX + GroupID + "_" + DialogID + "-TriggerOnClick");
					dialogsDatas.Remove(SAVE_PREFIX + GroupID + "_" + DialogID + "-AvatarUrl");
				}
			}
			
			protected void DeregisterStorables()
			{
				_VAMS.DeregisterString(DialogText);
				_VAMS.DeregisterFloat(DialogDelay);
				_VAMS.DeregisterUrl(DialogAvatarUrl);
				_VAMS.DeregisterStringChooser(DialogAvatarPosition);
			}
			
			protected void RegisterStorables()
			{
				_VAMS.RegisterString(DialogText);
				_VAMS.RegisterFloat(DialogDelay);
				_VAMS.RegisterUrl(DialogAvatarUrl);
				_VAMS.RegisterStringChooser(DialogAvatarPosition);
			}
	    }
    }
}