using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using System;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.IO;
using MeshVR;
using MVR.FileManagementSecure;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;
using MVR;

// VAMStory Story Group
//
// Dialog / narration system for VAM

namespace VAMStoryPlugin
{
	public partial class VAMStory : MVRScript
    {
	    public class StoryGroup
	    {
		    private VAMStory _VAMS;

		    public int GroupID;

		    public JSONStorableString GroupName;
			public JSONStorableVector3 GroupPosition;
			public JSONStorableVector3 GroupRotation;
			public EventTrigger GroupTrigger;
			public JSONStorableBool GroupHideMainDialog;
		    public JSONStorableBool GroupAnswer01Enabled;
		    public JSONStorableString GroupAnswer01Txt;
		    public EventTrigger GroupAnswer01Trigger;
		    public JSONStorableBool GroupAnswer02Enabled;
		    public JSONStorableString GroupAnswer02Txt;
		    public EventTrigger GroupAnswer02Trigger;
		    public JSONStorableBool GroupAnswer03Enabled;
		    public JSONStorableString GroupAnswer03Txt;
		    public EventTrigger GroupAnswer03Trigger;
		    public JSONStorableStringChooser GroupStoryActions;
		    
		    protected List<UIDynamic> groupComponentsUI = new List<UIDynamic>();
		    protected List<StoryDialog> DialogList = new List<StoryDialog>();    
		    protected JSONClass dialogsDatas { get; set; }

		    public StoryGroup(VAMStory rootScript, int groupId)
		    {
			    _VAMS = rootScript;
			    dialogsDatas = _VAMS.dialogsDatas;
			    
			    GroupID = groupId;

			    CreateGroupComponents();
		    }

		    protected void CreateGroupComponents()
		    {
			    // Temporary vars
			    HSVColor hsvcDefault = HSVColorPicker.RGBToHSV(1f, 1f, 1f);
			    UIDynamicTextField tmpTextField;
			    UIDynamicPopup tmpPopup;
			    UIDynamicSlider tmpSlider;
			    UIDynamicColorPicker tmpColor;
			    UIDynamicToggle tmpToggle;
			    UIDynamic tmpSpacer;
			    string newDefaultAnswer = "Answer text...";
			    
				// ******* GROUP NAME ***********
				GroupName = new JSONStorableString(SAVE_PREFIX + GroupID + "-GroupName", "");
				GroupName.val = ""+GroupID;
				
				// ******* GROUP POSITION ***********
				GroupPosition = new JSONStorableVector3(SAVE_PREFIX + GroupID + "-Position", new Vector3(0f,0f,0f), new Vector3(0,0,0), new Vector3(0,0,0), false);
				GroupRotation = new JSONStorableVector3(SAVE_PREFIX + GroupID + "-Rotation", new Vector3(0f,0f,0f), new Vector3(0,0,0), new Vector3(0,0,0), false);

			    // ******* GROUP TRIGGER ***********
			    GroupTrigger = new EventTrigger(_VAMS, SAVE_PREFIX + GroupID + "-GroupTrigger");

			    // ******* MAIN DIALOG DISABLED ***********
			    GroupHideMainDialog = new JSONStorableBool(SAVE_PREFIX + GroupID + "-HideMainDialog", false);

			    // ******* ANSWER 01 ***********
			    GroupAnswer01Enabled = new JSONStorableBool(SAVE_PREFIX + GroupID + "-Answer01Enabled", false);
			    GroupAnswer01Txt = new JSONStorableString(SAVE_PREFIX + GroupID + "-Answer01Text", "_default_");
			    GroupAnswer01Txt.val = newDefaultAnswer;
			    GroupAnswer01Trigger = new EventTrigger(_VAMS, SAVE_PREFIX + GroupID + "-Answer01Trigger");

			    // ******* ANSWER 02 ***********
			    GroupAnswer02Enabled = new JSONStorableBool(SAVE_PREFIX + GroupID + "-Answer02Enabled", false);
			    GroupAnswer02Txt = new JSONStorableString(SAVE_PREFIX + GroupID + "-Answer02Text", "_default_");
			    GroupAnswer02Txt.val = newDefaultAnswer;
			    GroupAnswer02Trigger = new EventTrigger(_VAMS, SAVE_PREFIX + GroupID + "-Answer02Trigger");

			    // ******* ANSWER 03 ***********
			    GroupAnswer03Enabled = new JSONStorableBool(SAVE_PREFIX + GroupID + "-Answer03Enabled", false);
			    GroupAnswer03Txt = new JSONStorableString(SAVE_PREFIX + GroupID + "-Answer03Text", "_default_");
			    GroupAnswer03Txt.val = newDefaultAnswer;
			    GroupAnswer03Trigger = new EventTrigger(_VAMS, SAVE_PREFIX + GroupID + "-Answer03Trigger");

			    // ******* GROUP ACTIONS ***********
			    GroupStoryActions = new JSONStorableStringChooser(SAVE_PREFIX + GroupID + "-Actions", _VAMS.actionsChoices, "None", "Group Actions");

			    // Registering everything
			    RegisterStorables();
			    
			    // Restoring datas
			    if (dialogsDatas != null)
			    {
					// If you look at VAM's source code, the RestoreFromJSON is super odd
					// and the restore does not work properly for the Vector3
					// so I'm gonna do it manually :|
					// ---- Position
					if( dialogsDatas[GroupPosition.name] != null ) {
						Vector3 savedVectorPos = new Vector3( Single.Parse(dialogsDatas[GroupPosition.name][0]), Single.Parse(dialogsDatas[GroupPosition.name][1]), Single.Parse(dialogsDatas[GroupPosition.name][2]) );
				    	GroupPosition.valNoCallback = savedVectorPos;
					}
					
					// ---- Rotation
					if( dialogsDatas[GroupRotation.name] != null ) {
						Vector3 savedVectorRot = new Vector3( Single.Parse(dialogsDatas[GroupRotation.name][0]), Single.Parse(dialogsDatas[GroupRotation.name][1]), Single.Parse(dialogsDatas[GroupRotation.name][2]) );
						GroupRotation.valNoCallback = savedVectorRot;
					}

					// The other restore working properly
					GroupName.RestoreFromJSON( dialogsDatas, true, true, false );
					GroupTrigger.RestoreFromJSON( dialogsDatas, _VAMS.subScenePrefix, _VAMS.mergeRestore, true );
				    GroupAnswer01Enabled.RestoreFromJSON( dialogsDatas );
				    GroupAnswer01Txt.RestoreFromJSON( dialogsDatas, true, true, false );
				    GroupAnswer01Trigger.RestoreFromJSON( dialogsDatas, _VAMS.subScenePrefix, _VAMS.mergeRestore, true );
				    GroupAnswer02Enabled.RestoreFromJSON( dialogsDatas );
				    GroupAnswer02Txt.RestoreFromJSON( dialogsDatas, true, true, false );
				    GroupAnswer02Trigger.RestoreFromJSON( dialogsDatas, _VAMS.subScenePrefix, _VAMS.mergeRestore, true );
				    GroupAnswer03Enabled.RestoreFromJSON( dialogsDatas );
				    GroupAnswer03Txt.RestoreFromJSON( dialogsDatas, true, true, false );
				    GroupAnswer03Trigger.RestoreFromJSON( dialogsDatas, _VAMS.subScenePrefix, _VAMS.mergeRestore, true );
				    GroupStoryActions.RestoreFromJSON( dialogsDatas, true, true, false);
				    GroupHideMainDialog.RestoreFromJSON( dialogsDatas );
			    }
		    }
		    
		    public void UpdateID(int newGroupId)
		    {
			    GroupID = newGroupId;

			    // Updating dialogs
			    int dialCount = 0;
			    foreach( StoryDialog dial in DialogList )
			    {
				    dial.UpdateID(GroupID,dialCount);
				    dialCount++;
			    }
			    
			    // Updating group components (param names for the save)
			    UpdateComponentsIDs();
		    }
		    
		    protected void UpdateComponentsIDs()
		    {
			    // Deregister
			    DeregisterStorables();
			    
			    // Updated
			    GroupName.name = SAVE_PREFIX + GroupID + "-GroupName";
			    GroupPosition.name = SAVE_PREFIX + GroupID + "-Position";
			    GroupRotation.name = SAVE_PREFIX + GroupID + "-Rotation";
			    GroupTrigger.Name = SAVE_PREFIX + GroupID + "-GroupTrigger";
			    GroupAnswer01Enabled.name = SAVE_PREFIX + GroupID + "-Answer01Enabled";
			    GroupAnswer01Txt.name = SAVE_PREFIX + GroupID + "-Answer01Text";
			    GroupAnswer01Trigger.Name = SAVE_PREFIX + GroupID + "-Answer01Trigger";
			    GroupAnswer02Enabled.name = SAVE_PREFIX + GroupID + "-Answer02Enabled";
			    GroupAnswer02Txt.name = SAVE_PREFIX + GroupID + "-Answer02Text";
			    GroupAnswer02Trigger.Name = SAVE_PREFIX + GroupID + "-Answer02Trigger";
			    GroupAnswer03Enabled.name = SAVE_PREFIX + GroupID + "-Answer03Enabled";
			    GroupAnswer03Txt.name = SAVE_PREFIX + GroupID + "-Answer03Text";
			    GroupAnswer03Trigger.Name = SAVE_PREFIX + GroupID + "-Answer03Trigger";
			    GroupStoryActions.name = SAVE_PREFIX + GroupID + "-Actions";
			    GroupHideMainDialog.name = SAVE_PREFIX + GroupID + "-HideMainDialog";
			    
			    // Register
			    RegisterStorables();
		    }
		    
		    public void CreateNewDialog()
		    {
			    DialogList.Add(new StoryDialog(_VAMS, GroupID, DialogList.Count));
		    }

		    public void CreateDialogs(int count)
		    {
			    for (var i = 0; i < count; i++)
			    {
				    DialogList.Add(new StoryDialog(_VAMS, GroupID, i));
			    }
		    }

		    public List<StoryDialog> GetDialogs()
		    {
			    return DialogList;
		    }
		    
		    public StoryDialog GetDialogID(int dialogId)
		    {
			    return DialogList[dialogId];
		    }

			public void DeleteDialogID(int dialogId) {
				StoryDialog targetDial = GetDialogID(dialogId);
				targetDial.Delete();
				DialogList.RemoveAt(dialogId);
				RefreshAllDialogsIDs();
			}

			public void MoveDialogID(int dialogId, bool downDirection = false) {
				int currentID = dialogId;
				int newDialogID = downDirection ? dialogId + 1 : dialogId - 1;
				StoryDialog currentDial = DialogList[currentID];
				StoryDialog previousDial = DialogList[newDialogID];

				// Updating ID of the previous action to properly synchronize
				// the new "positions" of the actions ( either scene wise or data wise )
				// Faking a high ID to ensure that the refresh of IDs will work
				// It will be properly resynced with the "RefreshAllActionsIDs" below
				previousDial.UpdateID(GroupID,100000); // Previous one (the place I'm moving to)
				currentDial.UpdateID(GroupID,newDialogID); // Current one (the one moving)
			
				// Updating the list
				DialogList.RemoveAt(currentDial.DialogID);
				DialogList.Insert(newDialogID,currentDial);
				DialogList[currentID] = previousDial;
				
				RefreshAllDialogsIDs();
			}

			public void RefreshAllDialogsIDs()
			{
				int dialCount = 0;
				foreach( StoryDialog dial in DialogList )
				{
					dial.UpdateID(GroupID,dialCount);
					dialCount++;
				}
			}
		    
		    public int GetDialogCount()
		    {
			    return DialogList.Count;
		    }
		    
		    public bool IsLastDialogID( int dialogId )
		    {
			    return GetDialogCount() == dialogId + 1;
		    }

		    public bool hasSavedPosition()
		    {
			    return GroupPosition.val != default(Vector3);
		    }
		    
		    public void ToggleGroupUI( bool enabled )
		    {
			    foreach( UIDynamic uidElem in groupComponentsUI )
			    {
				    _VAMS.ToggleUIDynamic(enabled, uidElem);
			    }
		    }

		    public void DoClickTriggerAnswerID(int answerId)
		    {
			    switch (answerId)
			    {
				    case 1:
					    GroupAnswer01Trigger.Trigger();
					    _VAMS.AddTriggerToTimerPool(GroupAnswer01Trigger);
					    break;
					case 2:
						GroupAnswer02Trigger.Trigger();
						_VAMS.AddTriggerToTimerPool(GroupAnswer02Trigger);
						break;
					case 3:
						GroupAnswer03Trigger.Trigger();
						_VAMS.AddTriggerToTimerPool(GroupAnswer03Trigger);
						break;
					default:
						_VAMS.logDebug("VAMStory: you're trying to execute an answer trigger that does not exists.");
						break;
			    }
		    }
		    
		    public void DoClickTriggerDialogID(int dialogId)
		    {
			    DialogList[dialogId].DoClickTrigger();
		    }
		    
		    public void DoShowTriggerDialogID(int dialogId)
		    {
			    DialogList[dialogId].DoShowTrigger();
		    }

			public void MoveGroup() {
				// Don't move if we have the UI following the camera
				if (_VAMS.GlobalOptUIFollowCamera.val == true) return;
				
				// Only moving or rotating if we have a location saved, if not, we just stay were the atom is saved
				if(GroupPosition.val != default(Vector3)) _VAMS.containingAtom.mainController.MoveControl(GroupPosition.val);
				if(GroupRotation.val != default(Vector3)) _VAMS.containingAtom.mainController.RotateControl(GroupRotation.val);
			}
			
			public void Delete() {
				// Before deleting the group I need to clear the dialogs
				foreach( StoryDialog dial in DialogList )
				{
					dial.Delete();
				}

				DialogList.Clear();

				// Unregistering vars
				DeregisterStorables();

				// Removing triggers
				GroupTrigger.Remove();
				GroupAnswer01Trigger.Remove();
				GroupAnswer02Trigger.Remove();
				GroupAnswer03Trigger.Remove();

				// Clearing data from the loaded data of the scene (to prevent re-restore after deleting)
				// This only happens when you load a scene and have data loaded from the json
				if (dialogsDatas != null)
				{
					dialogsDatas.Remove(SAVE_PREFIX + GroupID + "-GroupName");
					dialogsDatas.Remove(SAVE_PREFIX + GroupID + "-Position");
					dialogsDatas.Remove(SAVE_PREFIX + GroupID + "-Rotation");
					dialogsDatas.Remove(SAVE_PREFIX + GroupID + "-GroupTrigger");
					dialogsDatas.Remove(SAVE_PREFIX + GroupID + "-Answer01Enabled");
					dialogsDatas.Remove(SAVE_PREFIX + GroupID + "-Answer01Text");
					dialogsDatas.Remove(SAVE_PREFIX + GroupID + "-Answer01Trigger");
					dialogsDatas.Remove(SAVE_PREFIX + GroupID + "-Answer02Enabled");
					dialogsDatas.Remove(SAVE_PREFIX + GroupID + "-Answer02Text");
					dialogsDatas.Remove(SAVE_PREFIX + GroupID + "-Answer02Trigger");
					dialogsDatas.Remove(SAVE_PREFIX + GroupID + "-Answer03Enabled");
					dialogsDatas.Remove(SAVE_PREFIX + GroupID + "-Answer03Text");
					dialogsDatas.Remove(SAVE_PREFIX + GroupID + "-Answer03Trigger");
					dialogsDatas.Remove(SAVE_PREFIX + GroupID + "-Actions");
					dialogsDatas.Remove(SAVE_PREFIX + GroupID + "-HideMainDialog");
				}
			}
			
			protected void DeregisterStorables()
			{
				_VAMS.DeregisterString(GroupName);
				
				_VAMS.DeregisterVector3(GroupPosition);
				_VAMS.DeregisterVector3(GroupRotation);
			
				_VAMS.DeregisterBool(GroupAnswer01Enabled);
				_VAMS.DeregisterString(GroupAnswer01Txt);
				_VAMS.DeregisterBool(GroupAnswer02Enabled);
				_VAMS.DeregisterString(GroupAnswer02Txt);
				_VAMS.DeregisterBool(GroupAnswer03Enabled);
				_VAMS.DeregisterString(GroupAnswer03Txt);
				_VAMS.DeregisterStringChooser(GroupStoryActions);
				_VAMS.DeregisterBool(GroupHideMainDialog);
			}
			
			protected void RegisterStorables()
			{
				_VAMS.RegisterString(GroupName);
				
				_VAMS.RegisterVector3(GroupPosition);
				_VAMS.RegisterVector3(GroupRotation);
			
				_VAMS.RegisterBool(GroupAnswer01Enabled);
				_VAMS.RegisterString(GroupAnswer01Txt);
				_VAMS.RegisterBool(GroupAnswer02Enabled);
				_VAMS.RegisterString(GroupAnswer02Txt);
				_VAMS.RegisterBool(GroupAnswer03Enabled);
				_VAMS.RegisterString(GroupAnswer03Txt);
				_VAMS.RegisterStringChooser(GroupStoryActions);
				_VAMS.RegisterBool(GroupHideMainDialog);
			}
	    }
    }
}