using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using System;
using System.Text;
using System.Text.RegularExpressions;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.IO;
using MeshVR;
using MVR.FileManagementSecure;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;
using MVR;

// VAMStory Actions
//
// Actions for VAMStory

namespace VAMStoryActionPlugin
{
	public partial class VAMStoryAction : MVRScript
    {
	    // Strings! (yes I know you like that)
	    private static string PLUGIN_PATH;
	    private static string BACKUP_PATH = "Saves\\PluginData\\vamstory\\backups";
	    private static string UI_BUNDLE_PATH = "/vamstory_assets.uibundle";
	    private static string SAVE_PREFIX = "Act_";
	    private static string BACKGROUNDS_PATH = "assets/vamstory/textures/backgrounds/";
	    private static string FONTS_PATH = "assets/vamstory/fonts/";
	    private static string SOUNDS_PATH = "assets/vamstory/sounds/";
	    private static string MATERIALS_PATH = "assets/vamstory/materials/materials_choices/";
	    
	    private string lastFileExplorerPath = "Custom";
	    
		// Game Objects, components, assets and assets lists
		protected Camera VAMSCamera;
		protected GameObject VAMSAUIGO;

		protected Transform TRStoryActionsUI;
		protected Transform TRAudio;
		protected Transform TRActionsMainPanel;
		protected Transform TRActionsPanel;
		protected Transform TRActionBtnModel;
		protected Transform TRActionTglModel;
		protected Transform TRActionTglBtnModel;
		protected Transform TRActionSliModel;
		protected Transform TRActionSliBigModel;
		protected Transform TRActionSpaModel;
		protected Transform TRActionSlider;
		protected Transform TRActionSliderBg;
		protected Transform TRActionSliderFill;
		protected Transform TRHideUIGrp;
		protected Transform TRHideUIBtn;
		protected Transform TRShowUIGrp;
		protected Transform TRShowUIBtn;

		protected Canvas CANStoryActionsUI;
		protected AudioSource ASAudio;
		protected CanvasGroup CGActionsPanel;
		protected GridLayoutGroup GLGActionsPanel;
		protected LayoutElement LEActionsPanel;
		protected CanvasGroup CGActionSlider;
		protected Slider SLActionSlider;
		protected Image IMGActionSliderBg;
		protected Image IMGActionSliderFill;
		protected Button BTNHideUI;
		protected RawImage IMGHideUI;
		protected Button BTNShowUI;
		protected RawImage IMGShowUI;

		protected Texture2D tmpTexture;
		protected List<string> backgroundsChoices = new List<String>();
		protected List<Sprite> backgroundsSprites = new List<Sprite>();
		protected List<string> fontsChoices = new List<String>();
		protected List<Font> fontsFiles = new List<Font>();
		protected List<string> soundChoices = new List<String>();
		protected List<AudioClip> soundClips = new List<AudioClip>();
		protected List<string> materialChoices = new List<String>() {"None"};
		protected List<Material> materialList = new List<Material>();
		
		// Actions
		protected JSONStorableAction showActionsUI;
		protected JSONStorableAction hideActionsUI;
		protected JSONStorableStringChooser A_PlayActionSound;
		protected JSONStorableFloat setSliderValue;
		protected JSONStorableAction A_ResetActionPanel;
		protected JSONStorableStringChooser A_SelectActionID;
		protected List<string> selectedActionChoices = new List<String>();
		protected JSONStorableAction A_ResetSelectedAction;
		protected JSONStorableAction A_HideSelectedAction;
		protected JSONStorableAction A_ShowSelectedAction;
		protected JSONStorableAction A_DisableSelectedAction;
		protected JSONStorableAction A_EnableSelectedAction;
		protected JSONStorableString A_ChangeLabelSelectedAction;
		protected JSONStorableString A_ChangeLabelOffSelectedAction;
		protected JSONStorableFloat A_SetSelectedActionSliderValue;
		
		// Static Storables
		protected JSONStorableBool GlobalOptEnableSlider;
		protected EventTrigger GlobalOptSliderTrigger;
		protected JSONStorableBool GlobalOptUIStartsHidden;
		protected JSONStorableBool GlobalOptUIRotationDisabled;
		protected JSONStorableBool GlobalOptUIConstraintToY;
		protected JSONStorableBool GlobalOptUIHideToggle;
		protected JSONStorableFloat GlobalOptUISize;
		protected JSONStorableFloat GlobalOptBtnAreaWidth;
		protected JSONStorableFloat GlobalOptBtnWidth;
		protected JSONStorableFloat GlobalOptBtnHeight;
		protected JSONStorableFloat GlobalOptBtnSpacing;
		protected JSONStorableStringChooser GlobalOptSoundOnClick;
		protected JSONStorableFloat GlobalOptSoundOnClickVolume;
		
		// Text options
		protected JSONStorableStringChooser GlobalOptUIAlignment;
		protected JSONStorableStringChooser GlobalOptTxtAlignment;
		protected JSONStorableFloat GlobalOptFontSize;
		protected JSONStorableFloat GlobalOptLineHeight;
		protected JSONStorableStringChooser GlobalOptFont;
		protected JSONStorableFloat GlobalOptTxtPadT;
		protected JSONStorableFloat GlobalOptTxtPadB;
		protected JSONStorableFloat GlobalOptTxtPadL;
		protected JSONStorableFloat GlobalOptTxtPadR;
		
		// Appearance
		protected JSONStorableStringChooser GlobalOptBg;
		protected JSONStorableColor GlobalOptBgColor;
		protected JSONStorableFloat GlobalOptBgOpacity;
		protected JSONStorableFloat GlobalOptBgTransitionSpeed;
		protected JSONStorableColor GlobalOptBgOverColor;
		protected JSONStorableFloat GlobalOptBgOverOpacity;
		protected JSONStorableColor GlobalOptBgClickColor;
		protected JSONStorableFloat GlobalOptBgClickOpacity;
		protected JSONStorableColor GlobalOptTxtColor;
		protected JSONStorableStringChooser GlobalOptMaterial;
		protected JSONStorableFloat GlobalOptMatDetSSX;
		protected JSONStorableFloat GlobalOptMatDetSSY;
		protected JSONStorableFloat GlobalOptMatSubDetOpacity;
		protected JSONStorableFloat GlobalOptMatRefraction;
		protected JSONStorableColor GlobalOptMatDetColor;

		// VAMS External values
		public JSONStorableBool VAMSValue_Ready;
		
		// The usual help section
		protected JSONStorableString helpText;
		
		// Action UI storables (or input fields)
		public JSONStorableString ActionTextUI;
		public JSONStorableString ActionTextOffUI;
		protected InputField ActionTextInputFieldUI;
		protected InputField ActionTextOffInputFieldUI;
		protected JSONStorableStringChooser ActionTypeUI;
		protected JSONStorableBool ActionToggleDefaultStateUI;
		protected JSONStorableBool ActionToggleNoExecAtStartUI;
		protected JSONStorableFloat ActionSliderDefaultValueUI;
		
		protected List<string> actionTypeChoices = new List<String>() {"Button","Toggle","Toggle Button","Slider","Slider Big","Spacer"};
		
		// Actions
		protected List<StoryAction> storyActions = new List<StoryAction>();
		protected StoryAction activeAction
		{
			get
			{
				return storyActions.ElementAtOrDefault(activeActionId);
			}
		}
		protected int activeActionId
		{
			get
			{
				return int.Parse(ActionSelector.val);
			}
		}
		
		protected List<EventTrigger> _timedTriggers = new List<EventTrigger>();
		
		// Plugin data (due to the fact that the UI is instantiated after init, I need to grab my data and restore manually
		protected JSONClass actionsDatas { get; set; }
		
		// Dynamic UI
		protected List<UIDynamic> globalControlsUIs = new List<UIDynamic>();
		protected List<UIDynamic> globalOptionsUIs = new List<UIDynamic>();
		protected List<UIDynamic> globalTextOptUIs = new List<UIDynamic>();
		protected List<UIDynamic> globalColorsOptUIs = new List<UIDynamic>();
		protected List<UIDynamic> helpUIs = new List<UIDynamic>();
		
		protected List<UIDynamic> actionsComponentsUI = new List<UIDynamic>();
		protected List<UIDynamic> actionsComponentsBtnUI = new List<UIDynamic>();
		protected List<UIDynamic> actionsComponentsTglUI = new List<UIDynamic>();
		protected List<UIDynamic> actionsComponentsTglBtnUI = new List<UIDynamic>();
		protected List<UIDynamic> actionsComponentsSliUI = new List<UIDynamic>();
		protected List<UIDynamic> actionsComponentsSpaUI = new List<UIDynamic>();
		protected List<UIDynamic> ignoreLayoutUI = new List<UIDynamic>();
		
		// Global UI components
		protected JSONStorableStringChooser RightUISelector;
		protected List<string> uisChoices = new List<string> {"Actions controls", "Global options", "Text options", "Appearance options", "Help"};
		protected JSONStorableStringChooser ActionSelector;
		protected List<string> actionsChoices;
		protected JSONStorableStringChooser CopySelector;
		protected List<string> copyChoices = new List<String>();
		protected List<string> alignmentChoices = new List<string> {"Left", "Center", "Right"};
		protected List<string> textAlignmentChoices = new List<string> {"UpperLeft", "UpperCenter", "UpperRight","MiddleLeft", "MiddleCenter", "MiddleRight","LowerLeft", "LowerCenter", "LowerRight"};

		protected UIDynamicTextField actionTitle;
		protected UIDynamicButton btnToggleOffTrigger;
		protected UIDynamicButton btnCopyToOn;
		
		protected UIDynamicTextField RightUIMainTitle;
		
		// Assetbundle
		private Request bundleRequest = null;

		// Init, security flags and values
		protected bool _isSceneLoading { get { return SuperController.singleton.isLoading; } }
		public bool isSceneFrozen { get { return SuperController.singleton.freezeAnimation; } }
		public bool isInEditMode { get { return SuperController.singleton.gameMode == SuperController.GameMode.Edit; } }
		
		private bool isBundleLoaded = false;
		private bool isInit = false;
		private bool _isMovingAction = false;
		private bool _allVAMSReady = false;
		private float _vamsReadyCheckPeriod = 0.5f;
		private int _selectedActionID = -1;

		public override void Init()
        {
            try
            {
                if (containingAtom.type != "Empty")
                {
                    logError("Please add VAMStoryActions on an Empty Atom");
                    return;
                }

                PLUGIN_PATH = GetPluginPath(this);

				VAMSCamera = SuperController.singleton.centerCameraTarget.targetCamera;
				
				// Backup folder
				FileManagerSecure.CreateDirectory(BACKUP_PATH);
				StartCoroutine("AutoBackup");
                
                // ******* CREATING MAIN TRIGGERS/ACTIONS ********
                // They are here to be displayed on TOP of every other JSONStorables
                JSONStorableAction fakeFuncUseBelow = new JSONStorableAction("- - - - Use these functions to control VAMStoryActions ↓ - - - - -", () => {});
                RegisterAction(fakeFuncUseBelow);
                
                showActionsUI = new JSONStorableAction("Show Actions UI", () => { ToggleActionPanelUI(true); });
                RegisterAction(showActionsUI);
                
                hideActionsUI = new JSONStorableAction("Hide Actions UI", () => { ToggleActionPanelUI(false); });
                RegisterAction(hideActionsUI);
                
                A_PlayActionSound = new JSONStorableStringChooser("Play Sound", soundChoices, "", "Play Sound") {isStorable=false,isRestorable=false};
                A_PlayActionSound.setCallbackFunction += (val) => { OnPlayActionSound( val ); };
                RegisterStringChooser(A_PlayActionSound);
                
                JSONStorableAction playActionsSound = new JSONStorableAction("Hide Actions UI", () => { ToggleActionPanelUI(false); });
                RegisterAction(hideActionsUI);
                
                setSliderValue = new JSONStorableFloat("Set Slider Value", 0.5f, 0f, 1f) {isStorable=false,isRestorable=false};
                setSliderValue.setCallbackFunction += (val) => { OnSetSliderValue(val); };
                RegisterFloat(setSliderValue);
                
                A_ResetActionPanel = new JSONStorableAction("Reset Action Panel", () => { OnResetActionPanel(); });
                RegisterAction(A_ResetActionPanel);
                
                A_SelectActionID = new JSONStorableStringChooser("Select Action ID", selectedActionChoices, "", "Select Action ID") {isStorable=false,isRestorable=false};
                A_SelectActionID.setCallbackFunction += (val) => { OnSelectActionID(); };
                RegisterStringChooser(A_SelectActionID);
                
                A_ResetSelectedAction = new JSONStorableAction("Reset selected action", () => { OnResetSelectedAction(); });
                RegisterAction(A_ResetSelectedAction);
                
                A_HideSelectedAction = new JSONStorableAction("Hide selected action", () => { OnToggleSelectedAction(false); });
                RegisterAction(A_HideSelectedAction);
                
                A_ShowSelectedAction = new JSONStorableAction("Show selected action", () => { OnToggleSelectedAction(true); });
                RegisterAction(A_ShowSelectedAction);
                
                A_DisableSelectedAction = new JSONStorableAction("Disable selected action", () => { OnDisableSelectedAction(false); });
                RegisterAction(A_DisableSelectedAction);
                
                A_EnableSelectedAction = new JSONStorableAction("Enable selected action", () => { OnDisableSelectedAction(true); });
                RegisterAction(A_EnableSelectedAction);

                A_ChangeLabelSelectedAction = new JSONStorableString("Change label for selected action", ""){isStorable=false,isRestorable=false};
                A_ChangeLabelSelectedAction.setCallbackFunction += (val) => { OnChangeLabelSelectedAction(val); };
                RegisterString(A_ChangeLabelSelectedAction);
                
                A_ChangeLabelOffSelectedAction = new JSONStorableString("Change label (off) for selected action", ""){isStorable=false,isRestorable=false};
                A_ChangeLabelOffSelectedAction.setCallbackFunction += (val) => { OnChangeLabelOffSelectedAction(val); };
                RegisterString(A_ChangeLabelOffSelectedAction);
                
                A_SetSelectedActionSliderValue = new JSONStorableFloat("Set selected action slider value", -1f, 0f, 1f) {isStorable=false,isRestorable=false};
                A_SetSelectedActionSliderValue.setJSONCallbackFunction += (val) => { OnSetSelectedActionSliderValue(); };
                RegisterFloat(A_SetSelectedActionSliderValue);
                A_SetSelectedActionSliderValue.min = 0f;
                
                JSONStorableAction fakeFuncReadExternalBelow = new JSONStorableAction("- - - - Vars below are meant to be read by external plugins ↓ - - - - -", () => {});
                RegisterAction(fakeFuncReadExternalBelow);
                
                VAMSValue_Ready = new JSONStorableBool("VAMSValue_Ready", false) {isStorable=false,isRestorable=false};
                RegisterBool(VAMSValue_Ready);

                JSONStorableAction fakeFuncGlobalOptions = new JSONStorableAction("- - - - Vars below can change global options during the scene ↓ - - - - -", () => {});
                RegisterAction(fakeFuncGlobalOptions);
                
                // ******* CREATING MAIN UI *********************************
                // Temporary vars
                HSVColor hsvcDefault = HSVColorPicker.RGBToHSV(1f, 1f, 1f);
                HSVColor hsvcDefaultOverClick = HSVColorPicker.RGBToHSV(0.55f, 0.55f, 0.55f);
                HSVColor hsvcTxtDefault = HSVColorPicker.RGBToHSV(0.2f, 0.2f, 0.2f);
                UIDynamicTextField tmpTextfield;
                UIDynamicPopup tmpPopup;
                UIDynamicSlider tmpSlider;
                UIDynamicColorPicker tmpColor;
                UIDynamicButton tmpButton;
                UIDynamicToggle tmpToggle;
                UIDynamic tmpSpacer;
                
                // ******* TITLE **********************
                createStaticDescriptionText("VAMStory Title","<color=#000><size=35><b>VAM STORY ACTIONS</b></size></color>",true,40,TextAnchor.MiddleRight);
                
				RightUISelector = new JSONStorableStringChooser("Right UI Selector", uisChoices, uisChoices.First(), "Show") {isStorable=false,isRestorable=false};
                RightUISelector.setCallbackFunction += (val) => { OnChangeRightUISelection(); };
                UIDynamicPopup RUISelsp = CreateScrollablePopup(RightUISelector, true);
                RUISelsp.labelWidth = 95f;

                tmpSpacer = CreateSpacer(true);
                setupSpacer(tmpSpacer, 25f);                                
                
                RightUIMainTitle = createStaticDescriptionText("Right UI Main Title","<color=#000><size=35><b>ACTIONS CONTROLS</b></size></color>",true,55,TextAnchor.MiddleRight);
				
                // ****** GLOBAL OPTIONS *********
                GlobalOptEnableSlider = new JSONStorableBool("Global_Opt_EnableSlider", false);
                tmpToggle = CreateToggle(GlobalOptEnableSlider,true);
                tmpToggle.label = "Enable Slider";
                RegisterBool(GlobalOptEnableSlider);
                GlobalOptEnableSlider.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                globalOptionsUIs.Add((UIDynamic) tmpToggle);
                
                GlobalOptSliderTrigger = new EventTrigger(this, "Global_Opt_SliderTrigger");
                GlobalOptSliderTrigger.active = true;
                
                tmpButton = CreateButton("Edit slider trigger",true);
                tmpButton.button.onClick.AddListener( () => { GlobalOptSliderTrigger.OpenPanelTransition(); } );
                setButtonColor(tmpButton, new Color(0.1f, 0.5f, 0.8f, 1f));
                setButtonTextColor(tmpButton, new Color(1f, 1f, 1f, 1f));
                globalOptionsUIs.Add((UIDynamic) tmpButton);
                
                GlobalOptUIStartsHidden = new JSONStorableBool("Global_Opt_UIStartsHidden", false);
                tmpToggle = CreateToggle(GlobalOptUIStartsHidden,true);
                tmpToggle.label = "UI starts hidden (at scene load)";
                RegisterBool(GlobalOptUIStartsHidden);
                globalOptionsUIs.Add((UIDynamic) tmpToggle);
                
				GlobalOptUIRotationDisabled = new JSONStorableBool("Global_Opt_UIRotationDisabled", false);
				tmpToggle = CreateToggle(GlobalOptUIRotationDisabled,true);
				tmpToggle.label = "UI rotation disabled";
				RegisterBool(GlobalOptUIRotationDisabled);
				GlobalOptUIRotationDisabled.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
				globalOptionsUIs.Add((UIDynamic) tmpToggle);
				
                GlobalOptUIConstraintToY = new JSONStorableBool("Global_Opt_UIConstraintToY", false);
                tmpToggle = CreateToggle(GlobalOptUIConstraintToY,true);
				tmpToggle.label = "UI rotation constrained to Y";
				RegisterBool(GlobalOptUIConstraintToY);
				globalOptionsUIs.Add((UIDynamic) tmpToggle);

				GlobalOptUIHideToggle = new JSONStorableBool("Global_Opt_UIHideToggle", false);
				tmpToggle = CreateToggle(GlobalOptUIHideToggle,true);
				tmpToggle.label = "UI hide toggle button";
				RegisterBool(GlobalOptUIHideToggle);
				GlobalOptUIHideToggle.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
				globalOptionsUIs.Add((UIDynamic) tmpToggle);
				
                GlobalOptUISize = new JSONStorableFloat("Global_Opt_UISize", 1f, 0.5f, 2f);
                tmpSlider = CreateSlider(GlobalOptUISize, true);
                tmpSlider.valueFormat = "F1";
                tmpSlider.label = "UI Size";
                RegisterFloat(GlobalOptUISize);
                GlobalOptUISize.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                globalOptionsUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptBtnAreaWidth = new JSONStorableFloat("Global_Opt_BtnAreaWidth", 1000f, 1000f, 2000f, false);
                tmpSlider = CreateSlider(GlobalOptBtnAreaWidth, true);
                tmpSlider.valueFormat = "F0";
                tmpSlider.label = "Actions Area Width";
                RegisterFloat(GlobalOptBtnAreaWidth);
                GlobalOptBtnAreaWidth.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                globalOptionsUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptBtnWidth = new JSONStorableFloat("Global_Opt_BtnWidth", 326f, 100f, 1000f);
                tmpSlider = CreateSlider(GlobalOptBtnWidth, true);
                tmpSlider.valueFormat = "F1";
                tmpSlider.label = "Actions Width";
                RegisterFloat(GlobalOptBtnWidth);
                GlobalOptBtnWidth.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                globalOptionsUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptBtnHeight = new JSONStorableFloat("Global_Opt_BtnHeight", 100f, 25f, 300f);
                tmpSlider = CreateSlider(GlobalOptBtnHeight, true);
                tmpSlider.valueFormat = "F1";
                tmpSlider.label = "Actions Height";
                RegisterFloat(GlobalOptBtnHeight);
                GlobalOptBtnHeight.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                globalOptionsUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptBtnSpacing = new JSONStorableFloat("Global_Opt_BtnSpacing", 10f, 0f, 500f);
                tmpSlider = CreateSlider(GlobalOptBtnSpacing, true);
                tmpSlider.valueFormat = "F1";
                tmpSlider.label = "Actions Spacing";
                RegisterFloat(GlobalOptBtnSpacing);
                GlobalOptBtnSpacing.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                globalOptionsUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptSoundOnClick = new JSONStorableStringChooser("Global_Opt_SndClick", soundChoices, "click_02", "Sound On Click");
                RegisterStringChooser(GlobalOptSoundOnClick);
                GlobalOptSoundOnClick.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpPopup = CreateScrollablePopup(GlobalOptSoundOnClick, true);
                tmpPopup.labelWidth = 120f;
                globalOptionsUIs.Add((UIDynamic) tmpPopup);

                GlobalOptSoundOnClickVolume = new JSONStorableFloat("Global_Opt_SndClickVol", 1f, 0f, 1f);
                tmpSlider = CreateSlider(GlobalOptSoundOnClickVolume, true);
                tmpSlider.valueFormat = "F3";
                tmpSlider.label = "On click sound volume";
                RegisterFloat(GlobalOptSoundOnClickVolume);
                globalOptionsUIs.Add((UIDynamic) tmpSlider);
                
                // Backups
                tmpSpacer = CreateSpacer(true);
                globalOptionsUIs.Add((UIDynamic) tmpSpacer);
                setupSpacer(tmpSpacer, 25f);                                

                tmpTextfield = createStaticDescriptionText("Backups Title","<color=#000><size=27><b>BACKUPS</b></size></color>",true,35,TextAnchor.MiddleRight);
                globalOptionsUIs.Add((UIDynamic) tmpTextfield);
                
                tmpButton = CreateButton("Restore backup",true);
                tmpButton.button.onClick.AddListener(() => { LoadBackup(); } );
                setButtonColor(tmpButton, new Color(0.1f, 0.5f, 0.7f, 1f));
                setButtonTextColor(tmpButton, new Color(1f, 1f, 1f, 1f));
                globalOptionsUIs.Add((UIDynamic) tmpButton);
                
                tmpButton = CreateButton("Save manual backup",true);
                tmpButton.button.onClick.AddListener(() => { SaveBackup(2); } );
                setButtonColor(tmpButton, new Color(0.1f, 0.7f, 0.4f, 1f));
                setButtonTextColor(tmpButton, new Color(1f, 1f, 1f, 1f));
                globalOptionsUIs.Add((UIDynamic) tmpButton);

                // ****** TEXT OPTIONS *********
                GlobalOptUIAlignment = new JSONStorableStringChooser("Global_Opt_UIAlignment", alignmentChoices, "Left", "UI Alignment");
                RegisterStringChooser(GlobalOptUIAlignment);
                GlobalOptUIAlignment.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpPopup = CreateScrollablePopup(GlobalOptUIAlignment, true);
                tmpPopup.labelWidth = 220f;
                globalTextOptUIs.Add((UIDynamic) tmpPopup);
                
                GlobalOptTxtAlignment = new JSONStorableStringChooser("Global_Opt_TxtAlignment", textAlignmentChoices, "MiddleCenter", "Text Alignment");
                RegisterStringChooser(GlobalOptTxtAlignment);
                GlobalOptTxtAlignment.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpPopup = CreateScrollablePopup(GlobalOptTxtAlignment, true);
                tmpPopup.labelWidth = 220f;
                globalTextOptUIs.Add((UIDynamic) tmpPopup);
                
                GlobalOptFontSize = new JSONStorableFloat("Global_Opt_FontSize", 40f, 30f, 60f, false);
                tmpSlider = CreateSlider(GlobalOptFontSize, true);
                tmpSlider.valueFormat = "F0";
                tmpSlider.label = "Font Size";
                RegisterFloat(GlobalOptFontSize);
                GlobalOptFontSize.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                globalTextOptUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptLineHeight = new JSONStorableFloat("Global_Opt_LineHeight", 1f, 0.1f, 10f);
                tmpSlider = CreateSlider(GlobalOptLineHeight, true);
                tmpSlider.valueFormat = "F2";
                tmpSlider.label = "Line Height";
                RegisterFloat(GlobalOptLineHeight);
                GlobalOptLineHeight.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                globalTextOptUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptFont = new JSONStorableStringChooser("Global_Opt_Font", fontsChoices, "arial", "Font");
                RegisterStringChooser(GlobalOptFont);
                GlobalOptFont.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpPopup = CreateScrollablePopup(GlobalOptFont, true);
                tmpPopup.labelWidth = 120f;
                globalTextOptUIs.Add((UIDynamic) tmpPopup);
                
                GlobalOptTxtPadT = new JSONStorableFloat("Global_Opt_TxtPadT", 10f, 0f, 150f, false);
                tmpSlider = CreateSlider(GlobalOptTxtPadT, true);
                tmpSlider.valueFormat = "F1";
                tmpSlider.label = "Text padding Top";
                RegisterFloat(GlobalOptTxtPadT);
                GlobalOptTxtPadT.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                globalTextOptUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptTxtPadB = new JSONStorableFloat("Global_Opt_TxtPadB", 10f, 0f, 150f, false);
                tmpSlider = CreateSlider(GlobalOptTxtPadB, true);
                tmpSlider.valueFormat = "F1";
                tmpSlider.label = "Text padding Bottom";
                RegisterFloat(GlobalOptTxtPadB);
                GlobalOptTxtPadB.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                globalTextOptUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptTxtPadL = new JSONStorableFloat("Global_Opt_TxtPadL", 10f, 0f, 150f, false);
                tmpSlider = CreateSlider(GlobalOptTxtPadL, true);
                tmpSlider.valueFormat = "F1";
                tmpSlider.label = "Text padding Left";
                RegisterFloat(GlobalOptTxtPadL);
                GlobalOptTxtPadL.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                globalTextOptUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptTxtPadR = new JSONStorableFloat("Global_Opt_TxtPadR", 10f, 0f, 150f, false);
                tmpSlider = CreateSlider(GlobalOptTxtPadR, true);
                tmpSlider.valueFormat = "F1";
                tmpSlider.label = "Text padding Right";
                RegisterFloat(GlobalOptTxtPadR);
                GlobalOptTxtPadR.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                globalTextOptUIs.Add((UIDynamic) tmpSlider);
                
                // ****** APPEARANCE OPTIONS *********
                GlobalOptBg = new JSONStorableStringChooser("Global_Opt_Bg", backgroundsChoices, "Default", "Style");
                GlobalOptBg.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpPopup = CreateScrollablePopup(GlobalOptBg, true);
                tmpPopup.labelWidth = 120f;
                RegisterStringChooser(GlobalOptBg);
                globalColorsOptUIs.Add((UIDynamic) tmpPopup);
                
                GlobalOptBgColor = new JSONStorableColor("Global_Opt_BgColor", hsvcDefault);
                GlobalOptBgColor.setJSONCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpColor = CreateColorPicker(GlobalOptBgColor, true);
                tmpColor.label = "Action background color";
                RegisterColor(GlobalOptBgColor);
                globalColorsOptUIs.Add((UIDynamic) tmpColor);
				
                GlobalOptBgOpacity = new JSONStorableFloat("Global_Opt_BgOpacity", 1f, 0f, 1f);
                GlobalOptBgOpacity.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpSlider = CreateSlider(GlobalOptBgOpacity, true);
                tmpSlider.label = "Action background opacity";
                RegisterFloat(GlobalOptBgOpacity);
                globalColorsOptUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptBgTransitionSpeed = new JSONStorableFloat("Global_Opt_BgTransitionSpeed", 0.1f, 0f, 0.5f);
                GlobalOptBgTransitionSpeed.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpSlider = CreateSlider(GlobalOptBgTransitionSpeed, true);
                tmpSlider.label = "Colors transition speed";
                RegisterFloat(GlobalOptBgTransitionSpeed);
                globalColorsOptUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptBgOverColor = new JSONStorableColor("Global_Opt_BgOverColor", hsvcDefaultOverClick);
                GlobalOptBgOverColor.setJSONCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpColor = CreateColorPicker(GlobalOptBgOverColor, true);
                tmpColor.label = "Action background color (over)";
                RegisterColor(GlobalOptBgOverColor);
                globalColorsOptUIs.Add((UIDynamic) tmpColor);
				
                GlobalOptBgOverOpacity = new JSONStorableFloat("Global_Opt_BgOverOpacity", 1f, 0f, 1f);
                GlobalOptBgOverOpacity.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpSlider = CreateSlider(GlobalOptBgOverOpacity, true);
                tmpSlider.label = "Action background opacity (over)";
                RegisterFloat(GlobalOptBgOverOpacity);
                globalColorsOptUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptBgClickColor = new JSONStorableColor("Global_Opt_BgClickColor", hsvcDefaultOverClick);
                GlobalOptBgClickColor.setJSONCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpColor = CreateColorPicker(GlobalOptBgClickColor, true);
                tmpColor.label = "Action background color (click)";
                RegisterColor(GlobalOptBgClickColor);
                globalColorsOptUIs.Add((UIDynamic) tmpColor);
				
                GlobalOptBgClickOpacity = new JSONStorableFloat("Global_Opt_BgClickOpacity", 1f, 0f, 1f);
                GlobalOptBgClickOpacity.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpSlider = CreateSlider(GlobalOptBgClickOpacity, true);
                tmpSlider.label = "Action background opacity (click)";
                RegisterFloat(GlobalOptBgClickOpacity);
                globalColorsOptUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptTxtColor = new JSONStorableColor("Global_Opt_TxtColor", hsvcTxtDefault);
                GlobalOptTxtColor.setJSONCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpColor = CreateColorPicker(GlobalOptTxtColor, true);
                tmpColor.label = "Action text Color";
                RegisterColor(GlobalOptTxtColor);
                globalColorsOptUIs.Add((UIDynamic) tmpColor);

                tmpSpacer = CreateSpacer(true);
                globalColorsOptUIs.Add((UIDynamic) tmpSpacer);
                setupSpacer(tmpSpacer, 25f); 
                
                GlobalOptMaterial = new JSONStorableStringChooser("Global_Opt_Material", materialChoices, "None", "Material");
                RegisterStringChooser(GlobalOptMaterial);
                GlobalOptMaterial.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpPopup = CreateScrollablePopup(GlobalOptMaterial, true);
                tmpPopup.labelWidth = 160f;
                globalColorsOptUIs.Add((UIDynamic) tmpPopup);
                
                GlobalOptMatDetSSX = new JSONStorableFloat("Global_Opt_MatDetSSX", 0.05f, -2f, 2f);
                tmpSlider = CreateSlider(GlobalOptMatDetSSX, true);
                tmpSlider.valueFormat = "F2";
                tmpSlider.label = "Details Scroll Speed X";
                RegisterFloat(GlobalOptMatDetSSX);
                GlobalOptMatDetSSX.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                globalColorsOptUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptMatDetSSY = new JSONStorableFloat("Global_Opt_MatDetSSY", 0.05f, -2f, 2f);
                tmpSlider = CreateSlider(GlobalOptMatDetSSY, true);
                tmpSlider.valueFormat = "F2";
                tmpSlider.label = "Details Scroll Speed Y";
                RegisterFloat(GlobalOptMatDetSSY);
                GlobalOptMatDetSSY.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                globalColorsOptUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptMatSubDetOpacity = new JSONStorableFloat("Global_Opt_MatSubDetOpacity", 0f, 0f, 1f);
                tmpSlider = CreateSlider(GlobalOptMatSubDetOpacity, true);
                tmpSlider.valueFormat = "F1";
                tmpSlider.label = "Sub-details Opacity";
                RegisterFloat(GlobalOptMatSubDetOpacity);
                GlobalOptMatSubDetOpacity.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                globalColorsOptUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptMatRefraction = new JSONStorableFloat("Global_Opt_MatRefraction", 0.5f, 0f, 1f);
                tmpSlider = CreateSlider(GlobalOptMatRefraction, true);
                tmpSlider.valueFormat = "F2";
                tmpSlider.label = "Details Refraction";
                RegisterFloat(GlobalOptMatRefraction);
                GlobalOptMatRefraction.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                globalColorsOptUIs.Add((UIDynamic) tmpSlider);

                GlobalOptMatDetColor = new JSONStorableColor("Global_Opt_MatDetColor", hsvcDefault);
                GlobalOptMatDetColor.setJSONCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpColor = CreateColorPicker(GlobalOptMatDetColor, true);
                tmpColor.label = "Details Color";
                RegisterColor(GlobalOptMatDetColor);
                globalColorsOptUIs.Add((UIDynamic) tmpColor);
                
				// Fake action	                
                JSONStorableAction fakeFuncDontUseBelow = new JSONStorableAction("- - - - DON'T Use these functions below ↓ - - - - -", () => {});
                RegisterAction(fakeFuncDontUseBelow);
                
                // ******* ACTIONS CONTROLS **********************

                // Actions choices
                actionsChoices = new List<string>();
                ActionSelector = new JSONStorableStringChooser("Action ID", actionsChoices, "", "Action ID") {isStorable=false,isRestorable=false};
                ActionSelector.setCallbackFunction += (val) => { OnToggleActionId(val); };
                UIDynamicPopup DSelsp = CreateScrollablePopup(ActionSelector, true);
                DSelsp.labelWidth = 250f;
                globalControlsUIs.Add((UIDynamic) DSelsp);
                
                UIDynamicButton newActionBtn = CreateButton("New action", true);
                newActionBtn.button.onClick.AddListener( () => { OnAddNewAction(); } );
                setButtonColor(newActionBtn, new Color(0.3f, 0.6f, 0.3f, 1f));
                setButtonTextColor(newActionBtn, new Color(1f, 1f, 1f, 1f));
                globalControlsUIs.Add((UIDynamic) newActionBtn);
                setupButtonWithLayout(newActionBtn, 190f);
                
                UIDynamicButton deleteActionBtn = CreateButton("Delete action", true);
                deleteActionBtn.button.onClick.AddListener( () => { OnDeleteCurrentAction(); } );
                setButtonColor(deleteActionBtn, new Color(0.6f, 0.3f, 0.3f, 1f));
                setButtonTextColor(deleteActionBtn, new Color(1f, 1f, 1f, 1f));
                globalControlsUIs.Add((UIDynamic) deleteActionBtn);
                ignoreLayoutUI.Add((UIDynamic) deleteActionBtn);
                setupButtonWithoutLayout(deleteActionBtn, 190f, new Vector2(210, -430));
                
                UIDynamicButton moveUpActionBtn = CreateButton("↑", true);
                moveUpActionBtn.button.onClick.AddListener( () => { OnMoveCurrentAction(false); } );
                setButtonColor(moveUpActionBtn, new Color(0.1f, 0.5f, 0.6f, 1f));
                setButtonTextColor(moveUpActionBtn, new Color(1f, 1f, 1f, 1f));
                globalControlsUIs.Add((UIDynamic) moveUpActionBtn);
                ignoreLayoutUI.Add((UIDynamic) moveUpActionBtn);
                setupButtonWithoutLayout(moveUpActionBtn, 50f, new Vector2(410, -430));
                
                UIDynamicButton moveDownActionBtn = CreateButton("↓", true);
                moveDownActionBtn.button.onClick.AddListener( () => { OnMoveCurrentAction(true); } );
                setButtonColor(moveDownActionBtn, new Color(0.1f, 0.5f, 0.6f, 1f));
                setButtonTextColor(moveDownActionBtn, new Color(1f, 1f, 1f, 1f));
                globalControlsUIs.Add((UIDynamic) moveDownActionBtn);
                ignoreLayoutUI.Add((UIDynamic) moveDownActionBtn);
                setupButtonWithoutLayout(moveDownActionBtn, 50f, new Vector2(470, -430));
                
                tmpTextfield = createStaticDescriptionText("Action copy Title","<color=#000><size=27><b>COPY ACTION</b></size></color>",true,35,TextAnchor.MiddleRight);
                globalControlsUIs.Add((UIDynamic) tmpTextfield);
                
                CopySelector = new JSONStorableStringChooser("Action Source", copyChoices, "", "Source/Target") {isStorable=false,isRestorable=false};
                UIDynamicPopup DSosp = CreateScrollablePopup(CopySelector, true);
                DSosp.labelWidth = 220f;
                globalControlsUIs.Add((UIDynamic) DSosp);
                
                UIDynamicButton copyActionBtn = CreateButton("Copy", true);
                copyActionBtn.button.onClick.AddListener(() => { OnCopyAction(); });
                setButtonColor(copyActionBtn, new Color(0.8f, 0.8f, 0.1f, 1f));
                globalControlsUIs.Add((UIDynamic) copyActionBtn);
                setupButtonWithLayout(copyActionBtn, 250f);
                
                UIDynamicButton transferActionBtn = CreateButton("Transfer", true);
                transferActionBtn.button.onClick.AddListener(() => { OnTransferAction(); });
                setButtonColor(transferActionBtn, new Color(0.8f, 0.5f, 0f, 1f));
                globalControlsUIs.Add((UIDynamic) transferActionBtn);
                ignoreLayoutUI.Add((UIDynamic) transferActionBtn);
                setupButtonWithoutLayout(transferActionBtn, 250f, new Vector2(270, -660));
                
                // ******* HELP **********************
                helpText = new JSONStorableString("Help",
	                "<size=30><color=#333>" +
	                "This plugin helps you create action panels for your scenes.\n\n" +
	                "<b>Formatting tips:</b>\n" +
	                "You can format the text of actions by using rich text snippets.\n" +
	                "<size=27>" +
	                "- <b>Bold:</b> <<i></i>b>Text</b<i></i>>\n" +
	                "- <b>Italic:</b> <<i></i>i>Text</i<i></i>>\n" +
	                "- <b>Size:</b> <<i></i>size=20>Text</size<i></i>>\n" +
	                "- <b>Color:</b> <<i></i>color=green>Text</color<i></i>>\n" +
	                "- <b>Color:</b> <<i></i>color=#00ffffff>Text</color<i></i>>\n" +
	                "</size>\n\n" +
	                "<b>PLUGIN HELP:</b>\n" +
	                "<b>VAMStoryActions</b> allows you to create <b>walls of buttons</b> in a couple of clicks.\n\n" +
	                "<color=#c81b0d><b>WARNING!: You need to keep VAMStoryActions and all other VAMS plugins atoms always on. See hub's documentation for more information.</b></color>\n\n" +
	                "<b>ACTIONS CONTROLS</b>\n" +
	                "This is where you navigate through your actions, as simple as that.\n\n" +
	                "<b>COPY/TRANSFER TRIGGER</b>\n" +
	                "This is the way to copy or transfer an action.\n\n" +
	                "Select an action in the Source/Target dropdown.\n\n" +
	                "Then click either <b>Copy</b> or <b>Transfer</b>.\n\n" +
	                "<b>Copy</b> will grab the values of the selected action in the dropdown and copy its content to the current action. <b>Transfer</b> will send the current action values to the selected action in the dropdown.\n\n" +
	                "<b>GLOBAL OPTIONS</b>\n" +
	                "This is where you can enable a trigger for the action panel.\n\n" +
	                "Click <b>Edit slider trigger</b> to set the actions bound to the slider.\n\n" +
	                "You can also configure here several appearance or behavior options for the panel ( if the UI rotates, its size, alignment etc... ).\n\n" +
	                "<b>BACKUPS</b>\n" +
	                "<b>VAMStoryActions</b> performs an automatic backup every 20 minutes only in edit mode.\n\n" +
	                "You can also restore a backup or save a manual backup here.\n\n" +
	                "You can use this as a <b>preset system</b> if you want..\n\n" +
	                "<b>APPEARANCE OPTIONS</b>\n" +
	                "This is where you can customize the colors and material appearance of the buttons.\n\n" +
	                "</color></size>"
                );
				
                tmpTextfield = CreateTextField(helpText, true);
                tmpTextfield.height = 850.0f;
                helpUIs.Add((UIDynamic) tmpTextfield);
                
                // Updating UI
                OnChangeRightUISelection();
                
                // Loading the assetbundle
                Request request = new AssetLoader.AssetBundleFromFileRequest {path = PLUGIN_PATH + UI_BUNDLE_PATH, callback = OnBundleLoaded};
                AssetLoader.QueueLoadAssetBundleFromFile(request);
                
                // Renaming callback
                SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;
            }
            catch(Exception e)
            {
                logError("VAMStoryActions - Exception caught: " + e);
            }
		}
		
		void Start()
		{
			
		}
		void Update() {
			// Base initialization after assetbundle load
			if (isInit == false && !_isSceneLoading && isBundleLoaded == true)
			{
				try
				{
					// Setting options and ready state for the suite once
					if (isInit == false && VAMSValue_Ready.val == false)
					{
						// Starting to wait for the VAMStory plugins
						StartCoroutine("WaitForVAMSPlugins");
						
						VAMSValue_Ready.val = true;						
					}
					// Firing triggers when the whole VAMS "suite" in the scene is ready
					else if ( _allVAMSReady )
					{
						isInit = true;
						
						// Everything is loaded, initialize the global options
						OnGlobalOptionsChanged();
						
						// Hide the UI based on the options
						ToggleActionPanelUI(!GlobalOptUIStartsHidden.val);
						
						// These actions are ready we can broadcast VAMStories to update their selectors
						SendBroadcastToVAMStories();
					}
				}
				catch(Exception e)
				{
					SuperController.LogError("VAMStory - Exception caught: " + e);
				}
			}
			
			// UI Update
			UpdateUIPosition();
			
			// Updating triggers with timers
			if (isInit == true && _timedTriggers.Count > 0)
			{
				foreach (EventTrigger trig in _timedTriggers)
				{
					trig.Update();
				}
			}
		}

		// **************************
		// Functions
		// **************************
		private void InitComponents()
		{
			if (VAMSAUIGO != null)
			{
				// Transforms init
				TRStoryActionsUI = getChildNamed(VAMSAUIGO.transform,"VAMStoryActionsUI");
				TRAudio = getChildNamed(VAMSAUIGO.transform,"VAMStoryActionsAudio");
				TRActionsMainPanel = getChildNamed(VAMSAUIGO.transform,"ActionsMainPanel");
				TRActionsPanel = getChildNamed(VAMSAUIGO.transform,"ActionsPanel");
				TRActionBtnModel = getChildNamed(VAMSAUIGO.transform,"ActionBtnModel");
				TRActionTglModel = getChildNamed(VAMSAUIGO.transform,"ActionTglModel");
				TRActionTglBtnModel = getChildNamed(VAMSAUIGO.transform,"ActionTglBtnModel");
				TRActionSliModel = getChildNamed(VAMSAUIGO.transform,"ActionSliModel");
				TRActionSliBigModel = getChildNamed(VAMSAUIGO.transform,"ActionSliBigModel");
				TRActionSpaModel = getChildNamed(VAMSAUIGO.transform,"ActionSpacerModel");
				TRActionSlider = getChildNamed(VAMSAUIGO.transform,"ActionSlider");
				TRActionSliderBg = getChildNamed(TRActionSlider,"Background");
				TRActionSliderFill = getChildNamed(TRActionSlider,"Fill");
				TRHideUIGrp = getChildNamed(VAMSAUIGO.transform,"HideGroup");
				TRHideUIBtn = getChildNamed(VAMSAUIGO.transform,"HideBtn");
				TRShowUIGrp = getChildNamed(VAMSAUIGO.transform,"ShowGroup");
				TRShowUIBtn = getChildNamed(VAMSAUIGO.transform,"ShowBtn");
				
				// Components init
				CANStoryActionsUI = TRStoryActionsUI.GetComponent<Canvas>();
				ASAudio = TRAudio.GetComponent<AudioSource>();
				CGActionsPanel = TRActionsPanel.GetComponent<CanvasGroup>();
				GLGActionsPanel = TRActionsPanel.GetComponent<GridLayoutGroup>();
				LEActionsPanel = TRActionsPanel.GetComponent<LayoutElement>();
				CGActionSlider = TRActionSlider.GetComponent<CanvasGroup>();
				SLActionSlider = TRActionSlider.GetComponent<Slider>();
				IMGActionSliderBg = TRActionSliderBg.GetComponent<Image>();
				IMGActionSliderFill = TRActionSliderFill.GetComponent<Image>();
				BTNHideUI = TRHideUIBtn.GetComponent<Button>();
				IMGHideUI = TRHideUIBtn.GetComponent<RawImage>();
				BTNShowUI = TRShowUIBtn.GetComponent<Button>();
				IMGShowUI = TRShowUIBtn.GetComponent<RawImage>();
				
				// Disabling the default models just in case
				TRActionBtnModel.gameObject.SetActive(false);
				TRActionTglModel.gameObject.SetActive(false);
				TRActionTglBtnModel.gameObject.SetActive(false);
				TRActionSliModel.gameObject.SetActive(false);
				TRActionSliBigModel.gameObject.SetActive(false);
				TRActionSpaModel.gameObject.SetActive(false);
				
				// Enabling or disabling the slider depending on the global options
				TRActionSlider.gameObject.SetActive(GlobalOptEnableSlider.val);
				
				// Adding Canvas to VAM's controller
				TRStoryActionsUI.gameObject.AddComponent<IgnoreCanvas>();
				SuperController.singleton.AddCanvas(CANStoryActionsUI);

				// Hiding toggle on UI
				TRShowUIGrp.gameObject.SetActive(false);
				
				// Slider action
				SLActionSlider.onValueChanged.AddListener(delegate { OnChangeSliderValue(); } );
				
				// Hide/Show buttons actions
				BTNHideUI.onClick.AddListener(delegate { OnClickHideUIButton(); } );
				BTNShowUI.onClick.AddListener(delegate { OnClickShowUIButton(); } );
			}
		}
		
		private void CreateDefaultAction()
		{
			StoryAction newAction = new StoryAction(this, 0);
			storyActions.Add(newAction);
			RefreshSelectors(0);
		}

		private GameObject CreateActionComponents( string actionType, int actionIndex )
		{
			GameObject newActionGO;
			if (actionType == "Button")
			{
				newActionGO = Instantiate(TRActionBtnModel.gameObject);	
			} else if (actionType == "Toggle")
			{
				newActionGO = Instantiate(TRActionTglModel.gameObject);	
			} else if (actionType == "Toggle Button")
			{
				newActionGO = Instantiate(TRActionTglBtnModel.gameObject);	
			}
			else if(actionType == "Slider")
			{
				newActionGO = Instantiate(TRActionSliModel.gameObject);	
			}
			else if(actionType == "Slider Big")
			{
				newActionGO = Instantiate(TRActionSliBigModel.gameObject);	
			}
			else
			{
				newActionGO = Instantiate(TRActionSpaModel.gameObject);
			}
			newActionGO.transform.SetParent(TRActionsPanel, false);
			newActionGO.transform.SetSiblingIndex(actionIndex);
			newActionGO.SetActive(true);
			return newActionGO;
		}

		private void LoadActions()
		{
			// The save system is pretty "awkward" as there are not a lot of possibilities to properly create a json file
			// that would help to restore properly the storables natively
			// So
			int actionsCount = 0;
			foreach (string paramKey in actionsDatas.Keys)
			{
				if (paramKey == "id") continue;// Skipping plugin id
				string[] paramSplit = paramKey.Split('-');

				if (paramSplit.Length != 2) continue; // Skipping anything that has not at least 2 entries on its param name
				
				// Only keeping "texts" (it's enough to know what I need to create the UI)
				if (paramSplit[1] == "Text")
				{
					string[] idsSplit = paramSplit[0].Split('_');
					if (idsSplit.Length == 2)
					{
						int splitId = int.Parse(idsSplit[1]);
						if (actionsCount < splitId) actionsCount = splitId;
					}
				}
			}

			// If we did find actions in the scene
			if (actionsCount > 0)
			{
				for (var i = 0; i <= actionsCount; i++)
				{
					StoryAction newAction = new StoryAction(this, i);
					storyActions.Add(newAction);
				}
				
				RefreshSelectors(0);
			}
			// If it's not the case, let's create the default action
			else
			{
				CreateDefaultAction();
			}
		}

		protected void CreateActionUI()
		{
			// Temporary vars
			HSVColor hsvcDefault = HSVColorPicker.RGBToHSV(1f, 1f, 1f);
			UIDynamicTextField tmpTextField;
			UIDynamicPopup tmpPopup;
			UIDynamicSlider tmpSlider;
			UIDynamicColorPicker tmpColor;
			UIDynamicButton tmpButton;
			UIDynamicToggle tmpToggle;
			UIDynamic tmpSpacer;
			    
			// Creating Action components
			// ******* ACTION TITLE (ID) ***********
			actionTitle = createStaticDescriptionText("ActionTitleUI","<color=#000><size=35><b>Action</b></size></color>",false,55,TextAnchor.MiddleLeft);
			actionsComponentsUI.Add((UIDynamic) actionTitle);
			
			// ******* ACTION TEXT LABEL ***********
			tmpTextField = createStaticDescriptionText("ActionTitleUI","<color=#000><size=30><b>Label</b></size></color>",false,35,TextAnchor.MiddleLeft);
			actionsComponentsUI.Add((UIDynamic) tmpTextField);
			
			// ******* ACTION TEXT ***********
			string newDefaultText = "Type label...";
			ActionTextUI = new JSONStorableString("ActionTextUI", "_default_"){isStorable=false,isRestorable=false};
			tmpTextField = CreateTextField(ActionTextUI);
			setupTextField(tmpTextField, 150f, false, false);
			ActionTextInputFieldUI = tmpTextField.UItext.gameObject.AddComponent<InputField>();
			ActionTextInputFieldUI.textComponent = tmpTextField.UItext;
			ActionTextInputFieldUI.lineType = InputField.LineType.MultiLineNewline;
			actionsComponentsUI.Add((UIDynamic) tmpTextField);
			ActionTextUI.valNoCallback = newDefaultText;
			ActionTextInputFieldUI.text = newDefaultText;				
			ActionTextInputFieldUI.onValueChanged.AddListener( delegate { OnActionTextChanged(); } );

			// ******* ACTION TEXT OFF LABEL ***********
			tmpTextField = createStaticDescriptionText("ActionTitleUI","<color=#000><size=30><b>Label (Off)</b></size></color>",false,35,TextAnchor.MiddleLeft);
			actionsComponentsTglBtnUI.Add((UIDynamic) tmpTextField);
			
			// ******* ACTION TEXT ***********
			string newDefaultTextOff = "Type label (off)...";
			ActionTextOffUI = new JSONStorableString("ActionTextOffUI", "_default_"){isStorable=false,isRestorable=false};
			tmpTextField = CreateTextField(ActionTextOffUI);
			setupTextField(tmpTextField, 150f, false, false);
			ActionTextOffInputFieldUI = tmpTextField.UItext.gameObject.AddComponent<InputField>();
			ActionTextOffInputFieldUI.textComponent = tmpTextField.UItext;
			ActionTextOffInputFieldUI.lineType = InputField.LineType.MultiLineNewline;
			actionsComponentsTglBtnUI.Add((UIDynamic) tmpTextField);
			ActionTextOffUI.valNoCallback = newDefaultTextOff;
			ActionTextOffInputFieldUI.text = newDefaultTextOff;				
			ActionTextOffInputFieldUI.onValueChanged.AddListener( delegate { OnActionTextOffChanged(); } );
			
			// ******* ACTION TYPE ***********
			ActionTypeUI = new JSONStorableStringChooser("Action Type", actionTypeChoices, actionTypeChoices.First(), "Action Type") {isStorable=false,isRestorable=false};
			ActionTypeUI.setCallbackFunction += (val) => { OnActionTypeChanged(); };
			tmpPopup = CreateScrollablePopup(ActionTypeUI, false);
			tmpPopup.labelWidth = 225f;
			actionsComponentsUI.Add((UIDynamic)tmpPopup);

			// ******* BUTTON ACTION COMPONENTS ***********
			tmpButton = CreateButton("On click trigger", false);
			actionsComponentsBtnUI.Add((UIDynamic)tmpButton);
			tmpButton.button.onClick.AddListener(() => { activeAction.triggerOnClickAction.OpenPanelActionStart(); } );

			// ******* TOGGLE ACTION COMPONENTS ***********
			tmpButton = CreateButton("Toggle On trigger", false);
			actionsComponentsTglUI.Add((UIDynamic)tmpButton);
			tmpButton.button.onClick.AddListener(() => { activeAction.triggerToggleOnAction.OpenPanelActionStart(); } );
			setupButtonWithLayout(tmpButton, 250f);
			
			btnToggleOffTrigger = CreateButton("Toggle Off trigger", false);
			actionsComponentsTglUI.Add((UIDynamic)btnToggleOffTrigger);
			btnToggleOffTrigger.button.onClick.AddListener(() => { activeAction.triggerToggleOffAction.OpenPanelActionStart(); } );
			ignoreLayoutUI.Add((UIDynamic) btnToggleOffTrigger);
			setupButtonWithoutLayout(btnToggleOffTrigger, 250f, new Vector2(270, -435));
			
			tmpButton = CreateButton("Copy to Off ↱", false);
			actionsComponentsTglUI.Add((UIDynamic)tmpButton);
			setButtonColor(tmpButton, new Color(0.8f, 0.8f, 0.1f, 1f));
			tmpButton.button.onClick.AddListener(() => { OnActionToggleCopyOnToOff(); } );
			setupButtonWithLayout(tmpButton, 250f);
			
			btnCopyToOn = CreateButton("↰ Copy to On Trigger", false);
			actionsComponentsTglUI.Add((UIDynamic)btnCopyToOn);
			setButtonColor(btnCopyToOn, new Color(0.8f, 0.8f, 0.1f, 1f));
			btnCopyToOn.button.onClick.AddListener(() => { OnActionToggleCopyOffToOn();  } );
			ignoreLayoutUI.Add((UIDynamic) btnCopyToOn);
			setupButtonWithoutLayout(btnCopyToOn, 250f, new Vector2(270, -500));
			
			ActionToggleDefaultStateUI = new JSONStorableBool("ActionToggleDefaultStateUI", false) {isStorable=false,isRestorable=false};
			tmpToggle = CreateToggle(ActionToggleDefaultStateUI,false);
			tmpToggle.label = "Default State";
			ActionToggleDefaultStateUI.setCallbackFunction += (val) => { OnToggleDefaultStateChanged(); };
			actionsComponentsTglUI.Add((UIDynamic) tmpToggle);
			
			ActionToggleNoExecAtStartUI = new JSONStorableBool("ActionToggleNoExecAtStartUI", false) {isStorable=false,isRestorable=false};
			tmpToggle = CreateToggle(ActionToggleNoExecAtStartUI,false);
			tmpToggle.label = "Do not execute triggers at start";
			ActionToggleNoExecAtStartUI.setCallbackFunction += (val) => { OnToggleNoExecAtStartChanged(); };
			actionsComponentsTglUI.Add((UIDynamic) tmpToggle);
			
			// ******* SLIDER ACTION COMPONENTS ***********
			tmpButton = CreateButton("Slider trigger", false);
			actionsComponentsSliUI.Add((UIDynamic)tmpButton);
			tmpButton.button.onClick.AddListener(() => { activeAction.triggerSliderAction.OpenPanelTransition(); } );
			
			ActionSliderDefaultValueUI = new JSONStorableFloat("ActionSliderDefaultValueUI", 0.3f, 0f, 1f) {isStorable=false,isRestorable=false};
			tmpSlider = CreateSlider(ActionSliderDefaultValueUI, false);
			tmpSlider.valueFormat = "F2";
			tmpSlider.label = "Slider default value";
			ActionSliderDefaultValueUI.setCallbackFunction += (val) => { OnSliderDefaultValueChanged(); };
			actionsComponentsSliUI.Add((UIDynamic) tmpSlider);

		}
		
		protected void ClearAllActionUIs()
		{
			foreach (var ActionItem in storyActions)
			{
				ActionItem.ToggleActionUI(false);			
			}
		}

		protected void RefreshSelectors( int actionId )
		{
			actionsChoices = new List<string>();
			copyChoices = new List<string>();
			copyChoices.Add(""); // The empty default one
			for (var i = 0; i < storyActions.Count; i++)
			{
				actionsChoices.Add("" + i);
				copyChoices.Add("" + i);
			}

			// Updating action selector
			ActionSelector.valNoCallback = "";
			ActionSelector.choices = actionsChoices;
			ActionSelector.val = "" + actionId;

			// Trigger for the action selection
			A_SelectActionID.valNoCallback = "";
			A_SelectActionID.choices = actionsChoices;

			CopySelector.choices = copyChoices;
		}
		
		protected void ToggleActionPanelUI(bool enabled)
		{
			if (TRStoryActionsUI != null)
			{
				TRStoryActionsUI.gameObject.SetActive(enabled);
			}
		}
		
		protected void ToggleActionUI(bool enabled, int actionId)
		{
			if (storyActions.ElementAtOrDefault(actionId) == null) return;
			storyActions[actionId].ToggleActionUI(enabled);
		}
		
		protected void ToggleUIDynamicGrp(List<UIDynamic> uiGrp,bool enabled)
		{
			foreach( UIDynamic uidElem in uiGrp )
			{
				ToggleUIDynamic(enabled, uidElem);
			}
		}
		
		protected void ToggleUIDynamic(bool enabled, UIDynamic target)
		{
			if (target == null) return;
			
			LayoutElement tmpLE = target.GetComponent<LayoutElement>();

			if (tmpLE != null)
			{
				if (enabled == true)
				{

					tmpLE.transform.localScale = new Vector3(1, 1, 1);
					if( !ignoreLayoutUI.Contains(target) ) tmpLE.ignoreLayout = false;
				}
				else
				{
					tmpLE.transform.localScale = new Vector3(0, 0, 0);
					tmpLE.ignoreLayout = true;
				}
			}
		}
		
		private void setupTextField( UIDynamicTextField target, float fieldHeight, bool disableBackground = true, bool disableScroll = true ) {
			if( disableBackground ) target.backgroundColor = new Color(1f, 1f, 1f, 0f);
			LayoutElement tfLayout = target.GetComponent<LayoutElement>();
			tfLayout.preferredHeight = tfLayout.minHeight = fieldHeight;
			target.height = fieldHeight;
			if( disableScroll ) disableScrollOnText(target);
		}

		private void setupSpacer(UIDynamic target, float newHeight)
		{
			LayoutElement tmpLE = target.GetComponent<LayoutElement>();
			tmpLE.minHeight = newHeight;
			tmpLE.preferredHeight = newHeight;
		}

		private void setupButtonWithLayout(UIDynamicButton target, float newSize)
		{
			ContentSizeFitter tmpCSF = target.button.transform.gameObject.AddComponent<ContentSizeFitter>();
			tmpCSF.horizontalFit = UnityEngine.UI.ContentSizeFitter.FitMode.MinSize;
			LayoutElement tmpLE = target.button.transform.gameObject.GetComponent<LayoutElement>();
			RectTransform tmpRectT = target.button.transform.GetComponent<RectTransform>();
			Rect tmpRect = tmpRectT.rect;
			tmpRectT.pivot = new Vector2(0f, 0.5f);
			tmpLE.minWidth = newSize;
			tmpLE.preferredWidth = newSize;
		}
		
		private void setupButtonWithoutLayout(UIDynamicButton target, float newSize, Vector2 newPosition)
		{
			ContentSizeFitter tmpCSF = target.button.transform.gameObject.AddComponent<ContentSizeFitter>();
			tmpCSF.horizontalFit = UnityEngine.UI.ContentSizeFitter.FitMode.MinSize;
			LayoutElement tmpLE = target.button.transform.gameObject.GetComponent<LayoutElement>();
			RectTransform tmpRectT = target.button.transform.GetComponent<RectTransform>();
			Rect tmpRect = tmpRectT.rect;
			tmpRectT.pivot = new Vector2(0f, 0.5f);
			tmpLE.minWidth = newSize;
			tmpLE.preferredWidth = newSize;
			tmpLE.ignoreLayout = true;
			tmpRectT.anchoredPosition = newPosition;
		}

		private void updateButtonPositionWithoutLayout(UIDynamicButton target, Vector2 newPosition)
		{
			RectTransform tmpRectT = target.button.transform.GetComponent<RectTransform>();
			tmpRectT.anchoredPosition = newPosition;
		}

		private void setButtonColor(UIDynamicButton targetBtn, Color newColor)
		{
			targetBtn.buttonColor = newColor;
		}
		
		private void setButtonTextColor(UIDynamicButton targetBtn, Color newColor)
		{
			targetBtn.textColor = newColor;
		}
		
		public UIDynamicTextField createStaticDescriptionText( string DescTitle, string DescText, bool rightSide, int fieldHeight, TextAnchor textAlignment = TextAnchor.UpperLeft ) {
			JSONStorableString staticDescString = new JSONStorableString(DescTitle,DescText) {isStorable=false,isRestorable=false};
			staticDescString.hidden = true;
			UIDynamicTextField staticDescStringField = CreateTextField(staticDescString, rightSide);
			staticDescStringField.backgroundColor = new Color(1f, 1f, 1f, 0f);
			staticDescStringField.UItext.alignment = textAlignment;
			LayoutElement sdsfLayout = staticDescStringField.GetComponent<LayoutElement>();
			sdsfLayout.preferredHeight = sdsfLayout.minHeight = fieldHeight;
			staticDescStringField.height = fieldHeight;
			disableScrollOnText(staticDescStringField);

			return staticDescStringField;
		}
		
		private static void disableScrollOnText(UIDynamicTextField target) {
			ScrollRect targetSR = target.UItext.transform.parent.transform.parent.transform.parent.GetComponent<ScrollRect>();
			if( targetSR != null ) {
				targetSR.horizontal = false;
				targetSR.vertical = false;
			}
		}
		
		private Transform getChildNamed( Transform parent, string childName ) {
			foreach( Transform child in parent )
			{
				if( child.name == childName ) {
					return child;
				} else {
					Transform childSearch = getChildNamed(child,childName);
					if (childSearch != null) return childSearch;
				}
			}
        
			return null;
		}

		protected void ToggleActionsUI(bool enabled)
		{
			if (TRStoryActionsUI != null)
			{
				TRStoryActionsUI.gameObject.SetActive(enabled);
			}
		}

		protected void UpdateUIPosition() {
			if (isInit == false || VAMSAUIGO == null) return;
			// UI Rotation
			if (GlobalOptUIRotationDisabled.val != true)
			{
				Vector3 lookPos = VAMSCamera.transform.position - VAMSAUIGO.transform.position;
				Quaternion lookRot = Quaternion.LookRotation(lookPos, Vector3.up);
				float eulerX = 0;
				if (GlobalOptUIConstraintToY.val == false)
				{
					eulerX = lookRot.eulerAngles.x;
				}

				float eulerY = lookRot.eulerAngles.y;
				Quaternion rotation = Quaternion.Euler(eulerX, eulerY, 0);
				VAMSAUIGO.transform.rotation = rotation;
			}
		}
		
		// Saving the datas
		// saveType: 0 = auto, 1 = critical, 2 = manual
		protected void SaveBackup( int saveType = 0 )
		{
			if (!isInEditMode) return;
			string fileName = "vamsa-" + GetCleanFilenameFromString(containingAtom.name);
			if (saveType == 0)
			{
				logDebug("VAMStoryActions: Saving automatic backup");
				string backupPath = CleanAndGetBackupPath( fileName, "auto" );
				SuperController.singleton.SaveJSON( this.GetJSON(), backupPath);
			} else if (saveType == 1)
			{
				string backupPath = CleanAndGetBackupPath( fileName, "critical" );
				SuperController.singleton.SaveJSON( this.GetJSON(), backupPath);
			} else if (saveType == 2)
			{
				string backupPath = BACKUP_PATH + "\\" + fileName + "-manual-" + DateTime.Now.ToString("yyyyMMdd-HHmmss") + ".json";
				SuperController.singleton.SaveJSON( this.GetJSON(), backupPath);
				logDebug("VAMStoryActions: Manual backup saved in:" + backupPath);
			}
		}

		protected string CleanAndGetBackupPath( string backupName, string backupSuffix )
		{
			// I'm keeping 5 files
			for (int i = 5; i > 0; i--)
			{
				string backupPath = BACKUP_PATH + "\\" + backupName + "-" + backupSuffix + "-" + i + ".json";
				if (FileManagerSecure.FileExists(backupPath) == true && i == 5)
				{
					FileManagerSecure.DeleteFile(backupPath);
				}
				else if(FileManagerSecure.FileExists(backupPath) == true)
				{
					FileManagerSecure.MoveFile(BACKUP_PATH + "\\" + backupName + "-" + backupSuffix + "-" + i + ".json", BACKUP_PATH + "\\" + backupName + "-" + backupSuffix + "-" + ( i + 1 ) + ".json",false);
				}
			}

			return BACKUP_PATH + "\\" + backupName + "-" + backupSuffix + "-1.json";
		}
		
		// Load backup
		protected void LoadBackup()
		{
			SuperController.singleton.fileBrowserUI.defaultPath = BACKUP_PATH;
			SuperController.singleton.fileBrowserUI.SetTitle("Select a backup");
			SuperController.singleton.fileBrowserUI.fileFormat = "json";
			SuperController.singleton.fileBrowserUI.showFiles = true;
			SuperController.singleton.fileBrowserUI.showDirs = true;
			SuperController.singleton.fileBrowserUI.SetTextEntry(false);
			SuperController.singleton.fileBrowserUI.hideExtension = false;
			SuperController.singleton.fileBrowserUI.showInstallFolderInDirectoryList = false;
			List<ShortCut> shortCutList = new List<ShortCut>();

			ShortCut shortCutCustom = new ShortCut();
			shortCutCustom.path = BACKUP_PATH;
			shortCutCustom.displayName = "VAMStoryActions Backups";
			shortCutCustom.flatten = false;
			shortCutList.Add(shortCutCustom);

			SuperController.singleton.fileBrowserUI.shortCuts = shortCutList;			
			SuperController.singleton.fileBrowserUI.Show((val) =>
			{
				if (val != "")
				{
					RestoreBackup(val);
				}
			});
		}

		protected void RestoreBackup(string backupPath)
		{
			// Loading the file
			string backupDatas = SuperController.singleton.ReadFileIntoString(backupPath);
			JSONClass backupJSON = JSON.Parse(backupDatas).AsObject;
			
			// Checking if the save is meant for VAMStory
			if ( !backupJSON.HasKey("id") || !backupJSON["id"].Value.Contains("_VAMStoryActionPlugin.VAMStoryAction") )
			{
				logError("VAMStoryActions: Backup not compatible with VAMStoryActions.");
				return;
			}
			
			logDebug("VAMStoryActions: Restoring " + backupPath);
			
			// Clearing all groups
			foreach (StoryAction act in storyActions)
			{
				act.Delete();
			}

			storyActions.Clear();
			
			// Restoring datas
			this.RestoreFromJSON(backupJSON);
			this.LateRestoreFromJSON(backupJSON);
			
			// Loading Dialogs
			LoadActions();
			
			// Refreshing UI
			RefreshSelectors(0);
		}

		// **************************************
		// Callbacks & similar ( when action UI value is changed )
		// **************************************
		protected void OnActionTextChanged()
		{
			ActionTextUI.val = ActionTextInputFieldUI.text;
			activeAction.ActionText.val = ActionTextInputFieldUI.text;

			activeAction.UpdateSceneComponents();
		}

		protected void OnActionTextOffChanged()
		{
			ActionTextOffUI.val = ActionTextOffInputFieldUI.text;
			activeAction.ActionTextOff.val = ActionTextOffInputFieldUI.text;

			activeAction.UpdateSceneComponents();
		}
		
		protected void OnActionTypeChanged()
		{
			OnActionTypeSelection();
			activeAction.ActionType.val = ActionTypeUI.val;
			activeAction.UpdateType();
			
			UpdateActionTypeUI();
		}

		protected void UpdateActionTypeUI()
		{
			if (ActionTypeUI.val == "Toggle")
			{
				updateButtonPositionWithoutLayout(btnToggleOffTrigger, new Vector2(270, -435));
				updateButtonPositionWithoutLayout(btnCopyToOn, new Vector2(270, -500));
			} else if (ActionTypeUI.val == "Toggle Button")
			{
				updateButtonPositionWithoutLayout(btnToggleOffTrigger, new Vector2(270, -650));
				updateButtonPositionWithoutLayout(btnCopyToOn, new Vector2(270, -715));
			}
		}
		
		protected void OnToggleDefaultStateChanged()
		{
			activeAction.ActionToggleDefaultState.val = ActionToggleDefaultStateUI.val;
		}

		protected void OnToggleNoExecAtStartChanged()
		{
			activeAction.ActionToggleNoExecAtStart.val = ActionToggleNoExecAtStartUI.val;
		}

		protected void OnActionToggleCopyOnToOff()
		{
			// Waiting for user confirmation
			SuperController.singleton.Alert("This will copy and overwrite\nall <b>ON</b> actions to <b>OFF</b> actions\nfor this toggle. \n\nAre you sure?", delegate {OnConfirmCopyOnToOff();}, () => {});
		}
		
		protected void OnActionToggleCopyOffToOn()
		{
			// Waiting for user confirmation
			SuperController.singleton.Alert("This will copy and overwrite\nall <b>OFF</b> actions to <b>ON</b> actions\nfor this toggle. \n\nAre you sure?", delegate {OnConfirmCopyOffToOn();}, () => {});
		}

		protected void OnConfirmCopyOnToOff()
		{
			activeAction.triggerToggleOffAction.RestoreFromJSON(activeAction.triggerToggleOnAction.GetJSON());
		}
		
		protected void OnConfirmCopyOffToOn()
		{
			activeAction.triggerToggleOnAction.RestoreFromJSON(activeAction.triggerToggleOffAction.GetJSON());
		}

		protected void OnSliderDefaultValueChanged()
		{
			activeAction.ActionSliderDefaultValue.val = ActionSliderDefaultValueUI.val;
		}
		
		protected void UpdateLocalUIFromActiveAction()
		{
			// Updating the local dialog UI based on dialog datas
			StoryAction act = activeAction;
			ActionTextUI.valNoCallback = act.ActionText.val;
			ActionTextOffUI.valNoCallback = act.ActionTextOff.val;
			ActionTypeUI.valNoCallback = act.ActionType.val;
			ActionTextInputFieldUI.text = act.ActionText.val;
			ActionTextOffInputFieldUI.text = act.ActionTextOff.val;
			ActionToggleDefaultStateUI.valNoCallback = act.ActionToggleDefaultState.val;
			ActionToggleNoExecAtStartUI.valNoCallback = act.ActionToggleNoExecAtStart.val;
			ActionSliderDefaultValueUI.valNoCallback = act.ActionSliderDefaultValue.val;
			
			// Changing the display of the action type
			// I'm doing this to prevent regenerating the scene components everytime
			OnActionTypeSelection();
			
			// Updating the title
			actionTitle.text = "<color=#000><size=35><b>ACTION " + ActionSelector.val + "</b></size></color>";
			
			// Updating the type UI if needed
			UpdateActionTypeUI();
		}
		
		// **************************************
		// Callbacks & similar ( internal )
		// **************************************
		private void OnBundleLoaded(Request aRequest) {
			try
			{
				// Loading triggers assets
				SimpleTriggerHandler.LoadAssets();
				
				// Reading my UI bundle
				bundleRequest = aRequest;
				GameObject tmp = bundleRequest.assetBundle.LoadAsset<GameObject>("Assets/VAMStory/prefabs/VAMStoryActions.prefab");
				VAMSAUIGO = Instantiate(tmp);
				
				// Parenting to my current Empty
				VAMSAUIGO.transform.SetParent(containingAtom.reParentObject.GetChild(0).transform, false);
				VAMSAUIGO.transform.localPosition = new Vector3(0, 0, 0);
				VAMSAUIGO.transform.localRotation = Quaternion.Euler(0f,0f,0f);
				
				// Grabbing all global assets
				backgroundsChoices = new List<string>();
				backgroundsChoices.Add("Default");
				fontsChoices = new List<string>();
				fontsChoices.Add("arial");
				soundChoices = new List<string>();
				materialChoices = new List<string>();
				materialChoices.Add("None");
				
				// Default assets
				fontsFiles.Add(Resources.GetBuiltinResource(typeof(Font), "Arial.ttf") as Font);
				materialList.Add( bundleRequest.assetBundle.LoadAsset<Material>("Assets/VAMStory/materials/M_VAMS_UI_default.mat") );
				
				// Other assets
				foreach (string bundleFileItem in bundleRequest.assetBundle.GetAllAssetNames())
				{
					if (bundleFileItem.Contains(BACKGROUNDS_PATH))
					{
						backgroundsChoices.Add(bundleFileItem.Replace(BACKGROUNDS_PATH, "").Replace(".png", ""));
						backgroundsSprites.Add( bundleRequest.assetBundle.LoadAsset<Sprite>(bundleFileItem) );
					}
					
					if (bundleFileItem.Contains(FONTS_PATH))
					{
						fontsChoices.Add(bundleFileItem.Replace(FONTS_PATH, "").Replace(".ttf", "").Replace(".otf", ""));
						fontsFiles.Add( bundleRequest.assetBundle.LoadAsset<Font>(bundleFileItem) );
					}
					
					if (bundleFileItem.Contains(SOUNDS_PATH))
					{
						soundChoices.Add(bundleFileItem.Replace(SOUNDS_PATH, "").Replace(".ogg", ""));
						soundClips.Add( bundleRequest.assetBundle.LoadAsset<AudioClip>(bundleFileItem) );
					}
					
					if (bundleFileItem.Contains(MATERIALS_PATH))
					{
						materialChoices.Add(bundleFileItem.Replace(MATERIALS_PATH, "").Replace(".mat", ""));
						materialList.Add( bundleRequest.assetBundle.LoadAsset<Material>(bundleFileItem) );
					}
				}
				
				// Settings global storable
				GlobalOptFont.choices = fontsChoices;
				GlobalOptFont.valNoCallback = fontsChoices.First();
				GlobalOptBg.choices = backgroundsChoices;
				GlobalOptBg.valNoCallback = backgroundsChoices.First();
				GlobalOptSoundOnClick.choices = soundChoices;
				GlobalOptSoundOnClick.valNoCallback = soundChoices.First();
				GlobalOptMaterial.choices = materialChoices;
				GlobalOptMaterial.valNoCallback = materialChoices.First();
				
				// Setting actions storables
				A_PlayActionSound.choices = soundChoices;
				
				// Since everything happens after loading the assets
				// I need to restore some of the global options that I just updated
				if (actionsDatas != null)
				{
					GlobalOptFont.RestoreFromJSON( actionsDatas );
					GlobalOptBg.RestoreFromJSON( actionsDatas );
					GlobalOptSoundOnClick.RestoreFromJSON( actionsDatas );
					GlobalOptMaterial.RestoreFromJSON( actionsDatas );
				}
				
				// Initializing UI
				CreateActionUI();
				
				// Initializing all UI components
				InitComponents();
					
				// If there is no data saved, let's create a default action
				if (actionsDatas == null)
				{
					CreateDefaultAction();
				}
				else
				{
					LoadActions();
				}
				
				// Updating action UI
				UpdateLocalUIFromActiveAction();
				OnActionTypeSelection();
				
				isBundleLoaded = true;
			}
			catch(Exception e)
			{
				SuperController.LogError("VAMStory - Exception caught: " + e);
			}
		}
		
		protected void OnActionTypeSelection()
		{
			if (ActionTypeUI.val == "Button")
			{
				ToggleUIDynamicGrp(actionsComponentsBtnUI, true);
				ToggleUIDynamicGrp(actionsComponentsTglUI, false);
				ToggleUIDynamicGrp(actionsComponentsTglBtnUI, false);
				ToggleUIDynamicGrp(actionsComponentsSliUI, false);
			} else if (ActionTypeUI.val == "Toggle")
			{
				ToggleUIDynamicGrp(actionsComponentsBtnUI, false);
				ToggleUIDynamicGrp(actionsComponentsTglUI, true);
				ToggleUIDynamicGrp(actionsComponentsTglBtnUI, false);
				ToggleUIDynamicGrp(actionsComponentsSliUI, false);
			} else if (ActionTypeUI.val == "Toggle Button")
			{
				ToggleUIDynamicGrp(actionsComponentsBtnUI, false);
				ToggleUIDynamicGrp(actionsComponentsTglUI, true);
				ToggleUIDynamicGrp(actionsComponentsTglBtnUI, true);
				ToggleUIDynamicGrp(actionsComponentsSliUI, false);
			} else if(ActionTypeUI.val == "Slider" || ActionTypeUI.val == "Slider Big") {
				ToggleUIDynamicGrp(actionsComponentsBtnUI, false);
				ToggleUIDynamicGrp(actionsComponentsTglUI, false);
				ToggleUIDynamicGrp(actionsComponentsTglBtnUI, false);
				ToggleUIDynamicGrp(actionsComponentsSliUI, true);
			} else if(ActionTypeUI.val == "Spacer") {
				ToggleUIDynamicGrp(actionsComponentsBtnUI, false);
				ToggleUIDynamicGrp(actionsComponentsTglUI, false);
				ToggleUIDynamicGrp(actionsComponentsTglBtnUI, false);
				ToggleUIDynamicGrp(actionsComponentsSliUI, false);
			}
		}
		
		protected void OnChangeRightUISelection()
		{
			if (RightUISelector.val == "Actions controls")
			{
				RightUIMainTitle.text = "<color=#000><size=35><b>ACTIONS CONTROLS</b></size></color>";
				
				ToggleUIDynamicGrp(globalControlsUIs, true);
				ToggleUIDynamicGrp(globalOptionsUIs, false);
				ToggleUIDynamicGrp(globalTextOptUIs, false);
				ToggleUIDynamicGrp(globalColorsOptUIs, false);
				ToggleUIDynamicGrp(helpUIs, false);
			} else if (RightUISelector.val == "Global options")
			{
				ToggleUIDynamicGrp(globalControlsUIs, false);
				ToggleUIDynamicGrp(globalOptionsUIs, true);
				ToggleUIDynamicGrp(globalTextOptUIs, false);
				ToggleUIDynamicGrp(globalColorsOptUIs, false);
				ToggleUIDynamicGrp(helpUIs, false);

				RightUIMainTitle.text = "<color=#000><size=35><b>GLOBAL OPTIONS</b></size></color>";
			} else if (RightUISelector.val == "Text options")
			{
				ToggleUIDynamicGrp(globalControlsUIs, false);
				ToggleUIDynamicGrp(globalOptionsUIs, false);
				ToggleUIDynamicGrp(globalTextOptUIs, true);
				ToggleUIDynamicGrp(globalColorsOptUIs, false);
				ToggleUIDynamicGrp(helpUIs, false);

				RightUIMainTitle.text = "<color=#000><size=35><b>TEXT OPTIONS</b></size></color>";
			} else if (RightUISelector.val == "Appearance options")
			{
				ToggleUIDynamicGrp(globalControlsUIs, false);
				ToggleUIDynamicGrp(globalOptionsUIs, false);
				ToggleUIDynamicGrp(globalTextOptUIs, false);
				ToggleUIDynamicGrp(globalColorsOptUIs, true);
				ToggleUIDynamicGrp(helpUIs, false);

				RightUIMainTitle.text = "<color=#000><size=35><b>APPEARANCE OPTIONS</b></size></color>";
			} else
			{
				ToggleUIDynamicGrp(globalControlsUIs, false);
				ToggleUIDynamicGrp(globalOptionsUIs, false);
				ToggleUIDynamicGrp(globalTextOptUIs, false);
				ToggleUIDynamicGrp(globalColorsOptUIs, false);
				ToggleUIDynamicGrp(helpUIs, true);
				
				RightUIMainTitle.text = "<color=#000><size=35><b>HELP</b></size></color>";
			}
		}
		
		protected void OnGlobalOptionsChanged()
		{
			if ( isInit == false ) return;
			
			// Hiding UI toggle
			TRHideUIGrp.gameObject.SetActive(!GlobalOptUIHideToggle.val);
			TRShowUIGrp.gameObject.SetActive(false);
			
			// Enabling/disabling slider
			TRActionSlider.gameObject.SetActive(GlobalOptEnableSlider.val);
			// Triggering the change if it is enabled
			if (GlobalOptEnableSlider.val == true)
			{
				OnChangeSliderValue();
			}
			
			// Changing UI size
			VAMSAUIGO.transform.localScale = new Vector3(GlobalOptUISize.val, GlobalOptUISize.val, GlobalOptUISize.val);
		
			// Changing rotation
			VAMSAUIGO.transform.localRotation = Quaternion.Euler(0f,0f,0f);
			
			// Changing buttons appearance
			foreach (StoryAction act in storyActions)
			{
				act.UpdateSceneComponents();
			}
			
			// UI Alignment
			if (GlobalOptUIAlignment.val == "Left")
			{
				GLGActionsPanel.childAlignment = TextAnchor.UpperLeft;
				GLGActionsPanel.startCorner = GridLayoutGroup.Corner.UpperLeft;
			}
			else if(GlobalOptUIAlignment.val == "Center")
			{
				GLGActionsPanel.childAlignment = TextAnchor.UpperCenter;
				GLGActionsPanel.startCorner = GridLayoutGroup.Corner.UpperLeft;
			}
			else
			{
				GLGActionsPanel.childAlignment = TextAnchor.UpperRight;
				GLGActionsPanel.startCorner = GridLayoutGroup.Corner.UpperRight;
			}

			// Buttons Size / Padding
			GLGActionsPanel.cellSize = new Vector2(GlobalOptBtnWidth.val, GlobalOptBtnHeight.val);
			GLGActionsPanel.spacing = new Vector2(GlobalOptBtnSpacing.val, GlobalOptBtnSpacing.val);
			GLGActionsPanel.padding = new RectOffset(0,0,(int)Mathf.Round(GlobalOptBtnSpacing.val),10);
			LEActionsPanel.minWidth = Mathf.Round(GlobalOptBtnAreaWidth.val);
			LEActionsPanel.preferredWidth = Mathf.Round(GlobalOptBtnAreaWidth.val);
			
			// Main Colors
			Color mainNewCol = new Color( GlobalOptBgColor.colorPicker.currentColor.r, GlobalOptBgColor.colorPicker.currentColor.g, GlobalOptBgColor.colorPicker.currentColor.b, Mathf.Clamp(GlobalOptBgOpacity.val,0.4f,1f) );
			Color sliderHandleNewCol = new Color( GlobalOptTxtColor.colorPicker.currentColor.r, GlobalOptTxtColor.colorPicker.currentColor.g, GlobalOptTxtColor.colorPicker.currentColor.b, Mathf.Clamp(GlobalOptBgOpacity.val,0.6f,1f) );
			
			// Slider
			IMGActionSliderBg.color = mainNewCol;
			IMGActionSliderFill.color = mainNewCol;
			
			// Hide/Show
			IMGHideUI.color = mainNewCol;
			IMGShowUI.color = mainNewCol;
		}
		
		protected void SetMaterialOptions( Image targetImg, Color newCol )
		{
			targetImg.material.SetColor("_DetailsColor",newCol);
			targetImg.material.SetFloat("_DetailsSpeedX", GlobalOptMatDetSSX.val);
			targetImg.material.SetFloat("_DetailsSpeedY", GlobalOptMatDetSSY.val);
			targetImg.material.SetFloat("_SubdetailsOpacity", GlobalOptMatSubDetOpacity.val);
			targetImg.material.SetFloat("_RefractionIntensity", GlobalOptMatRefraction.val);
		}
		
		protected void OnToggleActionId(string actionId)
		{
			if ( isInit == false ) return;
			// Updating the UI for the group
			UpdateLocalUIFromActiveAction();
		}
		
		protected void OnAddNewAction()
		{
			int actionId = storyActions.Count;
			StoryAction newAction = new StoryAction(this, actionId);
			storyActions.Add(newAction);
			RefreshSelectors(actionId);
		}

		protected void OnDeleteCurrentAction()
		{
			int actionId = int.Parse(ActionSelector.val);
			
			// We cannot delete the last action
			if (storyActions.Count == 1)
			{
				SuperController.singleton.Alert("<color=green>VAMStoryActions</color> needs to have at least one action.", () => {});
				return;
			}
			
			// Waiting for user confirmation
			SuperController.singleton.Alert("<color=#A90000FF>Deleting action " + actionId + ":\n\n</color><size=30>" + storyActions[actionId].ActionText.val + "</size>\n\nAre you sure?", delegate {OnConfirmActionDeletion();}, () => {});

		}

		protected void OnConfirmActionDeletion()
		{
			// Saving a critical backup
			SaveBackup(1);
			
			// Clearing the "timers" pool (just to prevent errors)
			_timedTriggers.Clear();
			
			// Now delete
			int actionId = int.Parse(ActionSelector.val);
			
			storyActions[actionId].Delete();
			storyActions.RemoveAt(actionId);
			RefreshAllActionsIDs();
			
			int selectedAction = 0;
			if (storyActions.Count > 1 && actionId > 0)
			{
				selectedAction = actionId - 1;
			}
			
			RefreshSelectors(selectedAction);
		}

		protected void RefreshAllActionsIDs()
		{
			int actCount = 0;
			foreach (StoryAction act in storyActions)
			{
				act.UpdateID(actCount);
				actCount++;
			}
		}
		
		protected void OnMoveCurrentAction(bool downDirection = false)
		{
			if (_isMovingAction == true) return;
			if (downDirection == false && activeAction.ActionID == 0) return;
			if (downDirection == true && activeAction.ActionID == storyActions.Count - 1) return;
			_isMovingAction = true;
			
			StoryAction currentAct = activeAction;
			
			int currentID = currentAct.ActionID;
			int newActionID = downDirection == false ? currentAct.ActionID - 1 : currentAct.ActionID + 1;

			StoryAction previousAct = storyActions[newActionID];

			// Updating ID of the previous action to properly synchronize
			// the new "positions" of the actions ( either scene wise or data wise )
			// Faking a high ID to ensure that the refresh of IDs will work
			// It will be properly resynced with the "RefreshAllActionsIDs" below
			previousAct.UpdateID(10000); // Previous one (the place I'm moving to)
			currentAct.UpdateID(newActionID); // Current one (the one moving)
			
			// Updating the list
			storyActions.RemoveAt(currentAct.ActionID);
			storyActions.Insert(newActionID,currentAct);
			storyActions[currentID] = previousAct;

			// Refreshing the UI and scene
			RefreshAllActionsIDs();
			RefreshSelectors(newActionID);
			
			_isMovingAction = false;
		}
		
		protected void OnCopyAction()
		{
			if (CopySelector.val == "")
			{
				SuperController.singleton.Alert("Select a source action to copy first!", () => {});
				return;
			}
			
			int sourceActionId = int.Parse(CopySelector.val);
			int targetActionId = int.Parse(ActionSelector.val);

			if (sourceActionId == targetActionId)
			{
				SuperController.singleton.Alert("You cannot copy an action on itself!", () => {});
			}
			else
			{
				StoryAction sourceAct = storyActions[sourceActionId];
				StoryAction targetAct = storyActions[targetActionId];

				// Copying all datas from the source action
				targetAct.CopyDatasFromAction(sourceAct);
				
				// I have to do this externally since the StoryAction is not "aware" of the UI text fields
				ActionTextInputFieldUI.text = sourceAct.ActionText.val;
				ActionTextOffInputFieldUI.text = sourceAct.ActionTextOff.val;
				
				// Updating dialog/group UI
				UpdateLocalUIFromActiveAction();
				
				// Updating scene
				targetAct.UpdateSceneComponents();
			}
		}

		protected void OnTransferAction()
		{
			if (CopySelector.val == "")
			{
				SuperController.singleton.Alert("Select a target action to transfer to first!", () => {});
				return;
			}
			
			int sourceActionId = int.Parse(ActionSelector.val);
			int targetActionId = int.Parse(CopySelector.val);

			if (sourceActionId == targetActionId)
			{
				SuperController.singleton.Alert("You cannot transfer an action on itself!", () => {});
			}
			else
			{
				StoryAction sourceAct = storyActions[sourceActionId];
				StoryAction targetAct = storyActions[targetActionId];

				// Copying all datas from the source action
				targetAct.CopyDatasFromAction(sourceAct);
				
				// I have to do this externally since the StoryAction is not "aware" of the UI text fields
				ActionTextInputFieldUI.text = sourceAct.ActionText.val;
				ActionTextOffInputFieldUI.text = sourceAct.ActionTextOff.val;

				// Updating dialog/group UI
				UpdateLocalUIFromActiveAction();
				
				// Updating scene
				targetAct.UpdateSceneComponents();
				
				// Alert for the transfer (since you don't have any visual feedback)
				SuperController.singleton.Alert("<color=green>VAMStoryActions</color>\n Datas transfered to Action ID " + targetActionId + "!", () => {});
			}
		}

		private void PlayOnClickSound()
		{
			ASAudio.PlayOneShot(soundClips[soundChoices.IndexOf(GlobalOptSoundOnClick.val)], GlobalOptSoundOnClickVolume.val);			
		}
		
		private void OnAtomRename(string oldid, string newid)
		{
			GlobalOptSliderTrigger.SyncAtomNames();
			
			foreach( var act in storyActions ) {
				act.triggerOnClickAction.SyncAtomNames();
				act.triggerToggleOnAction.SyncAtomNames();
				act.triggerToggleOffAction.SyncAtomNames();
				act.triggerSliderAction.SyncAtomNames();
			}
		}
		
		protected void AddTriggerToTimerPool(EventTrigger trig)
		{
			if (trig.HasActionsWithActiveTimers && !_timedTriggers.Contains(trig))
			{
				_timedTriggers.Add(trig);
			}
		}
		
		// **************************************
		// Callbacks & similar ( gameplay )
		// **************************************
		protected void OnPlayActionSound(string soundName)
		{
			ASAudio.PlayOneShot(soundClips[soundChoices.IndexOf(soundName)], GlobalOptSoundOnClickVolume.val);
			A_PlayActionSound.valNoCallback = "";
		}
		
		protected void OnSetSliderValue( float sliderVal )
		{
			SLActionSlider.value = sliderVal;
			setSliderValue.valNoCallback = -1f; // To force callback if it is called again
		}
		
		protected void OnChangeSliderValue()
		{
			GlobalOptSliderTrigger.TriggerTransition( SLActionSlider.value );
		}
		
		protected void OnResetActionPanel()
		{
			foreach (StoryAction act in storyActions)
			{
				act.ResetAction();
			}
		}
		
		protected void OnSelectActionID()
		{
			_selectedActionID = int.Parse(A_SelectActionID.val);
			A_SelectActionID.valNoCallback = "";
		}

		protected void OnResetSelectedAction()
		{
			if (_selectedActionID == -1)
			{
				if (isInEditMode) logDebug("VAMStoryActions: you have to select an action first.");
				return;
			}
			
			StoryAction tmpAct = storyActions.ElementAtOrDefault(_selectedActionID);
			if(tmpAct != null)
			{
				tmpAct.ResetAction();
			} else if (isInEditMode)
			{
				logDebug("VAMStoryActions: action ID " + _selectedActionID + " does not exists. (Reset Action)");
			}
		}
		
		protected void OnToggleSelectedAction(bool state)
		{
			if (_selectedActionID == -1)
			{
				if (isInEditMode) logDebug("VAMStoryActions: you have to select an action first.");
				return;
			}

			StoryAction tmpAct = storyActions.ElementAtOrDefault(_selectedActionID);
			if(tmpAct != null)
			{
				tmpAct.ToggleAction(state);
			} else if (isInEditMode)
			{
				logDebug("VAMStoryActions: action ID " + _selectedActionID + " does not exists. (Toggle action On/Off)");
			}
		}
		
		protected void OnDisableSelectedAction(bool state)
		{
			if (_selectedActionID == -1)
			{
				if (isInEditMode) logDebug("VAMStoryActions: you have to select an action first.");
				return;
			}

			StoryAction tmpAct = storyActions.ElementAtOrDefault(_selectedActionID);
			if(tmpAct != null)
			{
				tmpAct.DisableAction(state);
			} else if (isInEditMode)
			{
				logDebug("VAMStoryActions: action ID " + _selectedActionID + " does not exists. (Disable/Enable action)");
			}
		}

		protected void OnChangeLabelSelectedAction(string label)
		{
			A_ChangeLabelSelectedAction.valNoCallback = "";
			if (_selectedActionID == -1)
			{
				if (isInEditMode) logDebug("VAMStoryActions: you have to select an action first.");
				return;
			}
			
			StoryAction tmpAct = storyActions.ElementAtOrDefault(_selectedActionID);
			if(tmpAct != null)
			{
				tmpAct.ChangeLabelAction(label);
			} else if (isInEditMode)
			{
				logDebug("VAMStoryActions: action ID " + _selectedActionID + " does not exists. (Change label)");
			}
		}
		
		protected void OnChangeLabelOffSelectedAction(string label)
		{
			A_ChangeLabelOffSelectedAction.valNoCallback = "";
			if (_selectedActionID == -1)
			{
				if (isInEditMode) logDebug("VAMStoryActions: you have to select an action first.");
				return;
			}
			
			StoryAction tmpAct = storyActions.ElementAtOrDefault(_selectedActionID);
			if(tmpAct != null)
			{
				tmpAct.ChangeLabelOffAction(label);
			} else if (isInEditMode)
			{
				logDebug("VAMStoryActions: action ID " + _selectedActionID + " does not exists. (Change label)");
			}
		}
		
		protected void OnSetSelectedActionSliderValue()
		{
			if (_selectedActionID == -1)
			{
				if (isInEditMode) logDebug("VAMStoryActions: you have to select an action first.");
				return;
			}

			StoryAction tmpAct = storyActions.ElementAtOrDefault(_selectedActionID);
			if(tmpAct != null)
			{
				tmpAct.ChangeSliderPosition(A_SetSelectedActionSliderValue.val);
			} else if (isInEditMode)
			{
				logDebug("VAMStoryActions: action ID " + _selectedActionID + " does not exists. (Set slider value for action)");
			}
			
		}
		
		protected void OnClickHideUIButton()
		{
			// Freeze ? no action
			if (isSceneFrozen) return;
			
			PlayOnClickSound();
			
			CGActionsPanel.alpha = 0f;
			CGActionsPanel.blocksRaycasts = false;
			CGActionsPanel.interactable = false;
			
			CGActionSlider.alpha = 0f;
			CGActionSlider.blocksRaycasts = false;
			CGActionSlider.interactable = false;
			
			TRHideUIGrp.gameObject.SetActive(false);
			TRShowUIGrp.gameObject.SetActive(true);
		}
		
		protected void OnClickShowUIButton()
		{
			// Freeze ? no action
			if (isSceneFrozen) return;
			
			PlayOnClickSound();
			
			CGActionsPanel.alpha = 1f;
			CGActionsPanel.blocksRaycasts = true;
			CGActionsPanel.interactable = true;
			
			CGActionSlider.alpha = 1f;
			CGActionSlider.blocksRaycasts = true;
			CGActionSlider.interactable = true;
			
			TRHideUIGrp.gameObject.SetActive(true);
			TRShowUIGrp.gameObject.SetActive(false);
		}
		
		// **************************
		// COROUTINES
		// **************************
		protected IEnumerator WaitForVAMSPlugins()
		{
			// Checking VAMExcite plugins every one second while loading or whatever
			while (_allVAMSReady == false)
			{
				// We're loading...
				// The reason behind that is that, a plugin looking for other atoms can load fast enough
				// to catch the storables list in the target atom PRIOR to the plugins storables being even present in the List 
				// which would result in a "false positive" of an Empty atom not owning a VAMS plugin
				if (_isSceneLoading)
				{
					yield return new WaitForSeconds(_vamsReadyCheckPeriod);					
				}
				else
				{
					bool allReady = true;
					foreach (Atom at in SuperController.singleton.GetAtoms())
					{
						if (at.type == "Empty" && at != containingAtom)
						{
							foreach (string stId in at.GetStorableIDs())
							{
								if (stId.Contains(".VAMStory"))
								{
									JSONStorable tmpVAMS = at.GetStorableByID(stId);
									if (tmpVAMS != null && tmpVAMS.GetBoolParamValue("VAMSValue_Ready") == false) allReady = false;
								}
							}
						}
					}
					
					if (allReady == false)
					{
						yield return new WaitForSeconds(_vamsReadyCheckPeriod);					
					}
					else
					{
						_allVAMSReady = true;
						yield break;
					}
				}
			}
		}
		
		public IEnumerator AutoBackup()
		{
			while (true)
			{
				yield return new WaitForSeconds(1200);
				SaveBackup();
			}
		}
		
		// I don't know what the fuck is going on with Unity's UI, if you use the
		// event system immediately on the click delegate, the UI does not unfocus
		// This is the only frikkin alternative I found
		// THANK YOU NICE ENGINE OF THE HELL.
		public IEnumerator UnselectButton()
		{
			yield return new WaitForSeconds(0.1f);
			UnityEngine.EventSystems.EventSystem.current.SetSelectedGameObject(null);
		}
		
		// **************************
		// BROADCASTS SEND/RECEIVE
		// **************************
		public void SendBroadcastToVAMStories()
		{
			foreach(Atom at in SuperController.singleton.GetAtoms() ) {
				if( at.type == "Empty" ) {
					foreach (string stId in at.GetStorableIDs())
					{
						if (stId.Contains("_VAMStoryPlugin.VAMStory"))
						{
							at.BroadcastMessage("RefreshVAMStoryActionsSelectors");
						}
					}
				}
			}
		}

		public void ResetFromVAMStory()
		{
			// First, restoring options
			OnGlobalOptionsChanged();
			
			// The resetting the main actions panel dynamically modified by the grid layout group of VAMStory
			RectTransform panelRT = TRActionsMainPanel.gameObject.GetComponent<RectTransform>();
			panelRT.anchoredPosition = new Vector2(0, 0);
			panelRT.anchorMin = new Vector2(0, 0);
			panelRT.anchorMax = new Vector2(1, 1);
			panelRT.offsetMin = new Vector2(0, 0);
			panelRT.offsetMax = new Vector2(0, 0);
			panelRT.sizeDelta = new Vector2(0, 0);
		}

		// **************************
		// OVERRIDES SAVING/LOADING
		// **************************
		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			var json = base.GetJSON(includePhysical, includeAppearance, forceStore);
			if (includePhysical || forceStore)
			{
				needsStore = true;
				json[GlobalOptSliderTrigger.Name] = GlobalOptSliderTrigger.GetJSON(base.subScenePrefix);
				
				foreach( var act in storyActions ) {
					json[act.triggerOnClickAction.Name] = act.triggerOnClickAction.GetJSON(base.subScenePrefix);
					json[act.triggerToggleOnAction.Name] = act.triggerToggleOnAction.GetJSON(base.subScenePrefix);
					json[act.triggerToggleOffAction.Name] = act.triggerToggleOffAction.GetJSON(base.subScenePrefix);
					json[act.triggerSliderAction.Name] = act.triggerSliderAction.GetJSON(base.subScenePrefix);
				}
			}
			return json;
		}
		
		public override void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, JSONArray presetAtoms = null, bool setMissingToDefault = true)
		{
			base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, presetAtoms);
			// Saving the original datas to load all dynamic storables
			actionsDatas = jc;
		}
		
		public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
		{
			base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
			if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
			{
				GlobalOptSliderTrigger.RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
			}
		}
		
		// **************************
		// Local Tools
		// **************************
		private void logDebug( string debugText ) {
			SuperController.LogMessage( debugText );
		}
		
		private void logError( string debugText ) {
			SuperController.LogError( debugText );
		}
		
		private string GetCleanFilenameFromString(string str)
		{
			string tmp = CleanString(str);
			tmp = tmp.ToLower();
			return tmp;
		}
		
		private string CleanString(string str)
		{
			str = str.Trim();
			Regex reg1 = new Regex("[^a-zA-Z0-9\\s\\-]");
			Regex reg2 = new Regex("\\s+");
			str = reg1.Replace(str, "");
			str = reg2.Replace(str, "-");
			return str;
		}
		
		// **************************
		// Time to cleanup !
		// **************************
		void OnDestroy() {
			// Sync names
			SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;
			
			// Stop backup routine
			StopCoroutine("AutoBackup");
			
			// Removing the canvas from VAM's controller
			SuperController.singleton.RemoveCanvas(CANStoryActionsUI);
			
			// I have to clean the transform if it moved to a VAMStory too
			if (TRActionsMainPanel != null)
			{
				Destroy(TRActionsMainPanel.gameObject);
			}

			// Cleaning GameObjects
			Destroy(VAMSAUIGO);
			
			// Cleaning triggers
			GlobalOptSliderTrigger.active = false;
			GlobalOptSliderTrigger.Remove();
			foreach( var act in storyActions ) {
				act.triggerOnClickAction.Remove();
				act.triggerToggleOnAction.Remove();
				act.triggerToggleOffAction.Remove();
				act.triggerSliderAction.active = false;
				act.triggerSliderAction.Remove();
			}

			// Unloading assetbundle
			AssetLoader.DoneWithAssetBundleFromFile(PLUGIN_PATH + UI_BUNDLE_PATH);
		}

		// ***********************************************************
		// EXTERNAL TOOL - Thank you great coders for your content!
		// ***********************************************************

		// *********** Chokapi Decal Maker START *********************
		public Texture2D GetResource(string path)
		{
			if (path == "" || path == null || !FileManagerSecure.FileExists(path) )
			{
				path = GetPackagePath(this) + "Custom/Scripts/VAMStory/default.png";
			}
			Texture2D texture;
			byte[] tmppng = FileManagerSecure.ReadAllBytes(path);
			texture = new Texture2D(1, 1, TextureFormat.RGBA32, false);
			texture.LoadImage(tmppng);
			return texture;
		}
		// *********** Chokapi Decal Maker END *********************
		
		// *********** MacGruber_Utils.cs START *********************
		// Get directory path where the plugin is located. Based on Alazi's & VAMDeluxe's method.
		public static string GetPluginPath(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
			string filename = self.manager.GetJSON()["plugins"][id].Value;
			return filename.Substring(0, filename.LastIndexOfAny(new char[] { '/', '\\' }));
		}
				
		// Get path prefix of the package that contains our plugin.
		public static string GetPackagePath(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
			string filename = self.manager.GetJSON()["plugins"][id].Value;
			int idx = filename.IndexOf(":/");
			if (idx >= 0)
				return filename.Substring(0, idx+2);
			else
				return string.Empty;
		}
				
		// Check if our plugin is running from inside a package
		public static bool IsInPackage(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
			string filename = self.manager.GetJSON()["plugins"][id].Value;
			return filename.IndexOf(":/") >= 0;
		}
		
		// ===========================================================================================
		// TriggerHandler implementation for easier handling of custom triggers.
		// Essentially call this in your plugin init code:
		//     StartCoroutine(SimpleTriggerHandler.LoadAssets());
		//
		// Credit to AcidBubbles for figuring out how to do custom triggers.
		//
		// hazmhox note : Edited a bit to allow the use of transitions ( original code coming from LogicBricks 4 )
		public class SimpleTriggerHandler : TriggerHandler
		{
			public static bool Loaded { get; private set; }

			private static SimpleTriggerHandler myInstance;

			private RectTransform myTriggerActionsPrefab;
			private RectTransform myTriggerActionMiniPrefab;
			private RectTransform myTriggerActionDiscretePrefab;
			private RectTransform myTriggerActionTransitionPrefab;

			public static SimpleTriggerHandler Instance {
				get {
					if (myInstance == null)
						myInstance = new SimpleTriggerHandler();
					return myInstance;
				}
			}

			public static void LoadAssets()
			{
				SuperController.singleton.StartCoroutine(Instance.LoadAssetsInternal());
			}

			private IEnumerator LoadAssetsInternal()
			{
				foreach (var x in LoadAsset("z_ui2", "TriggerActionsPanel", p => myTriggerActionsPrefab = p))
					yield return x;
				foreach (var x in LoadAsset("z_ui2", "TriggerActionMiniPanel", p => myTriggerActionMiniPrefab = p))
					yield return x;
				foreach (var x in LoadAsset("z_ui2", "TriggerActionDiscretePanel", p => myTriggerActionDiscretePrefab = p))
					yield return x;
				foreach (var x in LoadAsset("z_ui2", "TriggerActionTransitionPanel", p => myTriggerActionTransitionPrefab = p))
					yield return x;

				Loaded = true;
			}

			private IEnumerable LoadAsset(string assetBundleName, string assetName, Action<RectTransform> assign)
			{
				AssetBundleLoadAssetOperation request = AssetBundleManager.LoadAssetAsync(assetBundleName, assetName, typeof(GameObject));
				if (request == null)
					throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null request.");
				yield return request;
				GameObject go = request.GetAsset<GameObject>();
				if (go == null)
					throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null GameObject.");
				RectTransform prefab = go.GetComponent<RectTransform>();
				if (prefab == null)
					throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null RectTansform.");
				assign(prefab);
			}


			void TriggerHandler.RemoveTrigger(Trigger t)
			{
				// nothing to do
			}

			void TriggerHandler.DuplicateTrigger(Trigger t)
			{
				throw new NotImplementedException();
			}

			RectTransform TriggerHandler.CreateTriggerActionsUI()
			{
				return UnityEngine.Object.Instantiate(myTriggerActionsPrefab);
			}

			RectTransform TriggerHandler.CreateTriggerActionMiniUI()
			{
				return UnityEngine.Object.Instantiate(myTriggerActionMiniPrefab);
			}

			RectTransform TriggerHandler.CreateTriggerActionDiscreteUI()
			{
				return UnityEngine.Object.Instantiate(myTriggerActionDiscretePrefab);
			}

			RectTransform TriggerHandler.CreateTriggerActionTransitionUI()
			{
				return UnityEngine.Object.Instantiate(myTriggerActionTransitionPrefab);
			}

			void TriggerHandler.RemoveTriggerActionUI(RectTransform rt)
			{
				UnityEngine.Object.Destroy(rt?.gameObject);
			}
		}

		// Wrapper for easier handling of custom triggers.
		public class EventTrigger : Trigger
		{
			public string Name {
				get { return name; }
				set { name = value; myNeedInit = true; }
			}

			public MVRScript Owner {
				get; private set;
			}

			private string name;
			private bool myNeedInit = true;

			public EventTrigger(MVRScript owner, string name)
			{
				Name = name;
				Owner = owner;
				handler = SimpleTriggerHandler.Instance;
			}

			public void OpenPanelActionStart()
			{
				if (!SimpleTriggerHandler.Loaded)
				{
					SuperController.LogError("EventTrigger: You need to call SimpleTriggerHandler.LoadAssets() before use.");
					return;
				}

				triggerActionsParent = Owner.UITransform;
				InitTriggerUI();
				OpenTriggerActionsPanel();
				if (myNeedInit)
				{
					triggerActionsPanel.Find("Panel/Header Text").GetComponent<Text>().text = Name;
					triggerActionsPanel.Find("Content/Tab1/Label").GetComponent<Text>().text = "Event Actions";
					triggerActionsPanel.Find("Content/Tab2").gameObject.SetActive(false);
					triggerActionsPanel.Find("Content/Tab3").gameObject.SetActive(false);
					myNeedInit = false;
				}
			}
			
			public void OpenPanelTransition()
			{
				if (!SimpleTriggerHandler.Loaded)
				{
					SuperController.LogError("EventTrigger: You need to call SimpleTriggerHandler.LoadAssets() before use.");
					return;
				}

				triggerActionsParent = Owner.UITransform;
				InitTriggerUI();
				OpenTriggerActionsPanel();
				
				//getRoot(triggerActionsParent,0);
				if (myNeedInit)
				{
					triggerActionsPanel.Find("Panel/Header Text").GetComponent<Text>().text = Name;
					triggerActionsPanel.Find("Content/Tab2/Label").GetComponent<Text>().text = "Frequencies Actions";
					triggerActionsPanel.Find("Content/Tab1").gameObject.SetActive(false);
					triggerActionsPanel.Find("Content/Tab2").gameObject.SetActive(true);
					triggerActionsPanel.Find("Content/Tab2").GetComponent<Toggle>().isOn = true;					
					triggerActionsPanel.Find("Content/Tab3").gameObject.SetActive(false);
					myNeedInit = false;
				}
			}
	
			public void RestoreFromJSON(JSONClass jc, string subScenePrefix, bool isMerge, bool setMissingToDefault)
			{
				if (jc.HasKey(Name))
				{
					JSONClass tc = jc[Name].AsObject;
					if (tc != null)
						base.RestoreFromJSON(tc, subScenePrefix, isMerge);
				}
				else if (setMissingToDefault)
				{
					base.RestoreFromJSON(new JSONClass());
				}
			}

			public void Trigger()
			{
				active = true;
				active = false;
			}
			
			public void TriggerTransition(float val)
			{
				this.transitionInterpValue = val;
			}
		}
		// *********** MacGruber_Utils.cs END *********************
	}
}