using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using System;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.IO;
using MeshVR;
using MVR.FileManagementSecure;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;
using MVR;

// VAMStory Actions Story Action
//
// Actions for VAMStory

namespace VAMStoryActionPlugin
{
	public partial class VAMStoryAction : MVRScript
    {
	    public class StoryAction
	    {
		    private VAMStoryAction _VAMSA;

		    public int ActionID;

		    public JSONStorableString ActionText;
		    public JSONStorableString ActionTextOff;
		    public JSONStorableString ActionType;
		    public EventTrigger triggerOnClickAction;
		    
		    public JSONStorableBool ActionToggleDefaultState;
		    public JSONStorableBool ActionToggleNoExecAtStart;
		    public EventTrigger triggerToggleOnAction;
		    public EventTrigger triggerToggleOffAction;
		    
		    public JSONStorableFloat ActionSliderDefaultValue;
		    public EventTrigger triggerSliderAction;
		    
			protected GameObject ActionGO;
			protected GameObject ActionObjGO;
			protected Button BTNAction;
			protected Button BTNActionOff;
			protected Image IMGAction;
			protected Image IMGActionOff;
			protected Text TXTAction;
			protected Text TXTActionOff;
			protected RectTransform RTAction;
			protected RectTransform RTActionOff;
			protected Rect REAction;
			
			protected Toggle TGActionToggle;
			protected Text TXTCheckmark;
			
			protected Slider SLActionSlider;
			protected Image IMGActionSliderBg;
			protected Image IMGActionSliderFill;
		    
		    protected List<UIDynamic> actionsComponentsUI = new List<UIDynamic>();
		    
		    protected JSONClass actionsDatas { get; set; }
		    private GameObject lastGO;
		    
		    public StoryAction(VAMStoryAction rootScript, int actionId)
		    {
			    _VAMSA = rootScript;
			    actionsDatas = _VAMSA.actionsDatas;

			    ActionID = actionId;
				    
			    CreateActionComponents();
		    }

		    protected void CreateActionComponents()
		    {
			    // Temporary vars
			    HSVColor hsvcDefault = HSVColorPicker.RGBToHSV(1f, 1f, 1f);
			    UIDynamicTextField tmpTextField;
			    UIDynamicPopup tmpPopup;
			    UIDynamicSlider tmpSlider;
			    UIDynamicColorPicker tmpColor;
			    UIDynamicButton tmpButton;
			    UIDynamicToggle tmpToggle;
			    UIDynamic tmpSpacer;
			    
			    // Creating Action components
		
				// ******* ACTION TEXT ***********
				ActionText = new JSONStorableString(SAVE_PREFIX + ActionID + "-Text", "_default_");
				ActionText.valNoCallback = "Type label...";
				
				// ******* ACTION TEXT OFF ***********
				ActionTextOff = new JSONStorableString(SAVE_PREFIX + ActionID + "-TextOff", "_default_");
				ActionTextOff.valNoCallback = "Type label (off)...";
				
				// ******* ACTION TYPE ***********
				ActionType = new JSONStorableString(SAVE_PREFIX + ActionID + "-Type", "_default_");
				ActionType.valNoCallback = "Button";

				// ******* TRIGGERS ***********
				triggerOnClickAction = new EventTrigger(_VAMSA, SAVE_PREFIX + ActionID + "-TriggerOnClick");
				triggerToggleOnAction = new EventTrigger(_VAMSA, SAVE_PREFIX + ActionID + "-TriggerToggleOn");
				triggerToggleOffAction = new EventTrigger(_VAMSA, SAVE_PREFIX + ActionID + "-TriggerToggleOff");
				triggerSliderAction = new EventTrigger(_VAMSA, SAVE_PREFIX + ActionID + "-TriggerSlider");
				triggerSliderAction.active = true;
				
				// ******* TOGGLE TYPE ***********
				ActionToggleDefaultState = new JSONStorableBool(SAVE_PREFIX + ActionID + "-ToggleDefaultState", false);
				ActionToggleNoExecAtStart = new JSONStorableBool(SAVE_PREFIX + ActionID + "-ToggleNoExecAtStart", false);
				
				// ******* SLIDER TYPE ***********
				ActionSliderDefaultValue = new JSONStorableFloat(SAVE_PREFIX + ActionID + "-SliderDefaultValue", 0.3f, 0f, 1f);

				// Registering everything
				RegisterStorables();
				
				// Restoring datas
				if (actionsDatas != null)
				{
					ActionText.RestoreFromJSON( actionsDatas, true, true, false);
					ActionTextOff.RestoreFromJSON( actionsDatas, true, true, false);
					ActionType.RestoreFromJSON( actionsDatas, true, true, false);
					ActionToggleDefaultState.RestoreFromJSON( actionsDatas, true, true, false);
					ActionToggleNoExecAtStart.RestoreFromJSON( actionsDatas, true, true, false);
					ActionSliderDefaultValue.RestoreFromJSON( actionsDatas, true, true, false);
					triggerOnClickAction.RestoreFromJSON( actionsDatas, _VAMSA.subScenePrefix, _VAMSA.mergeRestore, true );
					triggerToggleOnAction.RestoreFromJSON( actionsDatas, _VAMSA.subScenePrefix, _VAMSA.mergeRestore, true );
					triggerToggleOffAction.RestoreFromJSON( actionsDatas, _VAMSA.subScenePrefix, _VAMSA.mergeRestore, true );
					triggerSliderAction.RestoreFromJSON( actionsDatas, _VAMSA.subScenePrefix, _VAMSA.mergeRestore, true );
				}
				
				// Creating scene components
				ActionGO = _VAMSA.CreateActionComponents( ActionType.val, ActionID );
				
				// Setting name
				ActionGO.name = GetObjectName();
				
				// Grabbing the child object
				ActionObjGO = ActionGO.transform.GetChild(0).gameObject;

				// Initializing my global vars for the object
				InitObjectComponents();
				
				// Updating values
				UpdateSceneComponents();
				
				// When everything is ready executing toggle and slider default values (and triggers)
				if (ActionType.val == "Toggle")
				{
					if(!ActionToggleNoExecAtStart.val) ExecuteToggleTrigger( ActionToggleDefaultState.val );
				} else if (ActionType.val == "Toggle Button")
				{
					if (ActionToggleDefaultState.val == true)
					{
						if (!ActionToggleNoExecAtStart.val)
						{
							triggerToggleOnAction.Trigger();
							_VAMSA.AddTriggerToTimerPool(triggerToggleOnAction);
						}
						BTNAction.transform.gameObject.SetActive(false);
						BTNActionOff.transform.gameObject.SetActive(true);
					}
					else
					{
						if (!ActionToggleNoExecAtStart.val)
						{
							triggerToggleOffAction.Trigger();
							_VAMSA.AddTriggerToTimerPool(triggerToggleOffAction);
						}
						BTNAction.transform.gameObject.SetActive(true);
						BTNActionOff.transform.gameObject.SetActive(false);
					}
				} else if(ActionType.val == "Slider")
				{
					DoSliderTrigger();
				}
		    }
		    
		    public void ToggleActionUI( bool enabled )
		    {
			    foreach( UIDynamic uidElem in actionsComponentsUI )
			    {
				    _VAMSA.ToggleUIDynamic(enabled, uidElem);
			    }
		    }

		    public void ToggleAction(bool state)
		    {
			    ActionObjGO.SetActive(state);
		    }

		    public void ResetAction()
		    {
			    if (ActionType.val == "Toggle")
			    {
				    ChangeToggleState(ActionToggleDefaultState.val);
				    ExecuteToggleTrigger(ActionToggleDefaultState.val);
			    } else if (ActionType.val == "Toggle Button")
			    {
				    BTNAction.transform.gameObject.SetActive(!ActionToggleDefaultState.val);
				    BTNActionOff.transform.gameObject.SetActive(ActionToggleDefaultState.val);
				    if (ActionToggleDefaultState.val == true)
				    {
					    triggerToggleOnAction.Trigger();
					    _VAMSA.AddTriggerToTimerPool(triggerToggleOnAction);
				    }
				    else
				    {
					    triggerToggleOffAction.Trigger();
					    _VAMSA.AddTriggerToTimerPool(triggerToggleOffAction);
				    }
			    } else if (ActionType.val == "Slider")
			    {
				    ChangeSliderPosition(ActionSliderDefaultValue.val);
			    }
			    
			    // Resetting labels since we can change them at runtime
			    TXTAction.text = ActionText.val;
			    if(ActionType.val == "Toggle Button") TXTActionOff.text = ActionTextOff.val;
			    
			    // Re-enabling action visibility
			    DisableAction(true);
		    }
		    
		    public void DisableAction(bool state)
		    {
			    ActionGO.SetActive(state);
			    // enabling the object if we're trying to renable the button
			    // (preventing people using both and not understanding why the button stays invisible)
			    if (state) ToggleAction(true);
		    }
		    
		    public void ChangeLabelAction(string label)
		    {
			    if( TXTAction != null ) TXTAction.text = label;
		    }
		    
		    public void ChangeLabelOffAction(string label)
		    {
			    if( TXTActionOff != null ) TXTActionOff.text = label;
		    }
		    
		    public void DoClickTrigger()
		    {
			    // Freeze ? no action
			    if (_VAMSA.isSceneFrozen) return;
			    
			    _VAMSA.PlayOnClickSound();
			    triggerOnClickAction.Trigger();
			    _VAMSA.AddTriggerToTimerPool(triggerOnClickAction);

			    _VAMSA.StartCoroutine("UnselectButton"); // WTF (see routine)
		    }
		    
		    public void DoClickTriggerOn()
		    {
			    // Freeze ? no action
			    if (_VAMSA.isSceneFrozen) return;
			    
			    _VAMSA.PlayOnClickSound();
			    triggerToggleOnAction.Trigger();
			    _VAMSA.AddTriggerToTimerPool(triggerToggleOnAction);

			    BTNAction.transform.gameObject.SetActive(false);
			    BTNActionOff.transform.gameObject.SetActive(true);

			    _VAMSA.StartCoroutine("UnselectButton"); // WTF (see routine)
		    }
		    
		    public void DoClickTriggerOff()
		    {
			    // Freeze ? no action
			    if (_VAMSA.isSceneFrozen) return;
			    
			    _VAMSA.PlayOnClickSound();
			    triggerToggleOffAction.Trigger();
			    _VAMSA.AddTriggerToTimerPool(triggerToggleOffAction);
			    
			    BTNAction.transform.gameObject.SetActive(true);
			    BTNActionOff.transform.gameObject.SetActive(false);

			    _VAMSA.StartCoroutine("UnselectButton"); // WTF (see routine)
		    }
		    
		    public void DoToggleTrigger()
		    {
			    // Freeze ? no action
			    //if (_VAMSA.isSceneFrozen) return;
			    _VAMSA.PlayOnClickSound();
			    
			    bool tglValue = TGActionToggle.isOn;
			    ChangeToggleState(tglValue);
			    ExecuteToggleTrigger(tglValue);
		    }

		    public void ChangeToggleState(bool toggleEnabled)
		    {
			    if (TXTCheckmark == null) return;
			    if (toggleEnabled)
			    {
				    TXTCheckmark.text = "✓";
				    TGActionToggle.isOn = true;
			    }
			    else
			    {
				    TXTCheckmark.text = "";
				    TGActionToggle.isOn = false;
			    }
		    }
		    
		    protected void ExecuteToggleTrigger( bool tglEnabled )
		    {
			    if (tglEnabled)
			    {
				    triggerToggleOnAction.Trigger();
				    _VAMSA.AddTriggerToTimerPool(triggerToggleOnAction);
			    }
			    else
			    {
				    triggerToggleOffAction.Trigger();
				    _VAMSA.AddTriggerToTimerPool(triggerToggleOffAction);
			    }
		    }

		    public void ChangeSliderPosition(float sliderVal)
		    {
			    if (SLActionSlider == null) return;
			    SLActionSlider.value = sliderVal;
		    }
		    
		    public void DoSliderTrigger()
		    {
			    triggerSliderAction.TriggerTransition( SLActionSlider.value );
		    }

		    public void UpdateID(int newActionId)
		    {
			    ActionID = newActionId;
			    
				// Moving the object in the hierarchy (since the ID is the "position")
				ActionGO.transform.SetSiblingIndex(ActionID);
				
			    // Updating action components (param names for the save)
			    UpdateComponentsIDs();
		    }

		    public void UpdateType()
		    {
			    // Removing the current object
			    Destroy(ActionGO);
			    
			    // Creating the new one
			    ActionGO = _VAMSA.CreateActionComponents( ActionType.val, ActionID );
				
			    // Setting name
			    ActionGO.name = GetObjectName();
			    
			    // Grabbing the child object
			    ActionObjGO = ActionGO.transform.GetChild(0).gameObject;
			    
			    // Initializing my global vars for the object
			    InitObjectComponents();
				
			    // Updating values
			    UpdateSceneComponents();
		    }

		    protected string GetObjectName()
		    {
			    string objStr = "action-btn-" + ActionID;
			    if (ActionType.val == "Toggle")
			    {
				    objStr = "action-tgl-" + ActionID;
			    } else if (ActionType.val == "Toggle Button")
			    {
				    objStr = "action-tglbtn-" + ActionID;
			    } else if (ActionType.val == "Slider")
			    {
				    objStr = "action-sli-" + ActionID;
			    }

			    return objStr;
		    }

		    protected void InitObjectComponents()
		    {
			    // If it's a spacer, we don't need nothing
			    if (ActionType.val == "Spacer") return;
			    
			    // The rest needs some initialization
			    if (ActionType.val == "Button")
			    {
				    BTNAction = ActionObjGO.GetComponent<Button>();
				    IMGAction = ActionObjGO.GetComponent<Image>();
				    Transform tmpTR = _VAMSA.getChildNamed(ActionObjGO.transform, "Label");
				    TXTAction = tmpTR.GetComponent<Text>();
				    RTAction = tmpTR.GetComponent<RectTransform>();
				    REAction = RTAction.rect;

				    BTNAction.onClick.AddListener(delegate { DoClickTrigger(); } );				    
			    } else if (ActionType.val == "Toggle")
			    {
				    Transform tmpLabelTR = _VAMSA.getChildNamed(ActionObjGO.transform, "Label");
				    Transform tmpTglTR = _VAMSA.getChildNamed(ActionObjGO.transform, "Toggle");
				    Transform tmpChkTR = _VAMSA.getChildNamed(ActionObjGO.transform, "Checkmark");
				    
				    IMGAction = ActionObjGO.GetComponent<Image>();
				    TGActionToggle = tmpTglTR.GetComponent<Toggle>();
				    TXTAction = tmpLabelTR.GetComponent<Text>();
				    TXTCheckmark = tmpChkTR.GetComponent<Text>();
				    RTAction = tmpLabelTR.GetComponent<RectTransform>();
				    REAction = RTAction.rect;

				    ChangeToggleState(ActionToggleDefaultState.val);
				    TGActionToggle.onValueChanged.AddListener(delegate { DoToggleTrigger(); } );
			    }
			    else if (ActionType.val == "Toggle Button")
			    {
				    Transform tmpLabelTR = _VAMSA.getChildNamed(ActionObjGO.transform, "Label");
				    Transform tmpLabelOffTR = _VAMSA.getChildNamed(ActionObjGO.transform, "LabelOff");
				    Transform tmpBtnTR = _VAMSA.getChildNamed(ActionObjGO.transform, "BtnOn");
				    Transform tmpBtnOffTR = _VAMSA.getChildNamed(ActionObjGO.transform, "BtnOff");

				    BTNAction = tmpBtnTR.GetComponent<Button>();
				    BTNActionOff = tmpBtnOffTR.GetComponent<Button>();
				    IMGAction = tmpBtnTR.GetComponent<Image>();
				    IMGActionOff = tmpBtnOffTR.GetComponent<Image>();
				    TXTAction = tmpLabelTR.GetComponent<Text>();
				    TXTActionOff = tmpLabelOffTR.GetComponent<Text>();

				    RTAction = tmpLabelTR.GetComponent<RectTransform>();
				    RTActionOff = tmpLabelOffTR.GetComponent<RectTransform>();
				    REAction = RTAction.rect;

				    BTNAction.onClick.AddListener(delegate { DoClickTriggerOn(); } );
				    BTNActionOff.onClick.AddListener(delegate { DoClickTriggerOff(); } );			    
			    }
			    else if(ActionType.val == "Slider" || ActionType.val == "Slider Big")
			    {
				    Transform tmpLabelTR = _VAMSA.getChildNamed(ActionObjGO.transform, "Label");
				    Transform tmpSliTR = _VAMSA.getChildNamed(ActionObjGO.transform, "Slider");
				    
				    IMGAction = ActionObjGO.GetComponent<Image>();
				    SLActionSlider = tmpSliTR.GetComponent<Slider>();

				    TXTAction = tmpLabelTR.GetComponent<Text>();
				    RTAction = tmpLabelTR.GetComponent<RectTransform>();
				    REAction = RTAction.rect;

				    ChangeSliderPosition(ActionSliderDefaultValue.val);
				    SLActionSlider.onValueChanged.AddListener(delegate { DoSliderTrigger(); } );
				    DoSliderTrigger();
			    }

		    }
		    
		    public void UpdateSceneComponents()
		    {
			    if (ActionGO == null) return;
			    if (ActionType.val == "Spacer") return; 
			    
			    // Components
			    TXTAction.text = ActionText.val;
			    if(ActionType.val == "Toggle Button") TXTActionOff.text = ActionTextOff.val;
			    
				// Font size
				TXTAction.fontSize = (int)_VAMSA.GlobalOptFontSize.val;
				if(ActionType.val == "Toggle Button") TXTActionOff.fontSize = (int)_VAMSA.GlobalOptFontSize.val;
			
				// Line height
				TXTAction.lineSpacing = _VAMSA.GlobalOptLineHeight.val;
				if(ActionType.val == "Toggle Button") TXTActionOff.lineSpacing = _VAMSA.GlobalOptLineHeight.val;

			    // Padding
			    if (ActionType.val == "Button" || ActionType.val == "Toggle Button")
			    {
				    // Text alignment
				    TXTAction.alignment = (TextAnchor)Enum.Parse(typeof(TextAnchor),_VAMSA.GlobalOptTxtAlignment.val);
				    if(ActionType.val == "Toggle Button") TXTActionOff.alignment = (TextAnchor)Enum.Parse(typeof(TextAnchor),_VAMSA.GlobalOptTxtAlignment.val);
			    			    
				    RTAction.offsetMin = new Vector2(_VAMSA.GlobalOptTxtPadL.val, _VAMSA.GlobalOptTxtPadB.val);
				    RTAction.offsetMax = new Vector2(-_VAMSA.GlobalOptTxtPadR.val, -_VAMSA.GlobalOptTxtPadT.val);
				    if (ActionType.val == "Toggle Button")
				    {
					    RTActionOff.offsetMin = new Vector2(_VAMSA.GlobalOptTxtPadL.val, _VAMSA.GlobalOptTxtPadB.val);
					    RTActionOff.offsetMax = new Vector2(-_VAMSA.GlobalOptTxtPadR.val, -_VAMSA.GlobalOptTxtPadT.val);					    
				    }
			    }

			    // Font + color
			    Font tmpFont = _VAMSA.fontsFiles[ _VAMSA.fontsChoices.IndexOf(_VAMSA.GlobalOptFont.val) ];
			    Color fontNewCol = new Color( _VAMSA.GlobalOptTxtColor.colorPicker.currentColor.r, _VAMSA.GlobalOptTxtColor.colorPicker.currentColor.g, _VAMSA.GlobalOptTxtColor.colorPicker.currentColor.b, 1f );
			    TXTAction.font = tmpFont;
			    if (ActionType.val != "Slider Big")
			    {
				    TXTAction.color = fontNewCol;				    
			    }
			    else
			    {
				    TXTAction.color = new Color(0.3f,0.3f,0.3f,1f);
			    }
			    
			    if (ActionType.val == "Toggle Button")
			    {
				    TXTActionOff.font = tmpFont;
				    TXTActionOff.color = fontNewCol;
			    }

			    // Background sprite
			    // Grabbing material first

			    string bgValue = _VAMSA.GlobalOptBg.val;
			    if (bgValue != "Default" && _VAMSA.backgroundsChoices.IndexOf(bgValue) != -1)
			    {
				    IMGAction.sprite = _VAMSA.backgroundsSprites[ _VAMSA.backgroundsChoices.IndexOf(bgValue) - 1 ];
				    if(ActionType.val == "Toggle Button") IMGActionOff.sprite = _VAMSA.backgroundsSprites[ _VAMSA.backgroundsChoices.IndexOf(bgValue) - 1 ];
			    }
			    else
			    {
				    IMGAction.sprite = _VAMSA.backgroundsSprites[ _VAMSA.backgroundsChoices.IndexOf("basic_01") - 1 ];
				    if(ActionType.val == "Toggle Button") IMGActionOff.sprite = _VAMSA.backgroundsSprites[ _VAMSA.backgroundsChoices.IndexOf("basic_01") - 1 ];
			    }
			    
			    // Material value
			    Material tmpMat = _VAMSA.materialList[_VAMSA.materialChoices.IndexOf(_VAMSA.GlobalOptMaterial.val)];
			    // I have to do that to instantiate materials
			    Material matAction = new Material(tmpMat);
			    // Setting mat
			    IMGAction.material = matAction;
			    if(ActionType.val == "Toggle Button") IMGActionOff.material = matAction;
			    // Material setup
			    Color detActionNewCol = new Color( _VAMSA.GlobalOptMatDetColor.colorPicker.currentColor.r, _VAMSA.GlobalOptMatDetColor.colorPicker.currentColor.g, _VAMSA.GlobalOptMatDetColor.colorPicker.currentColor.b, 1f );
			    _VAMSA.SetMaterialOptions(IMGAction,detActionNewCol);
			    if(ActionType.val == "Toggle Button") _VAMSA.SetMaterialOptions(IMGActionOff,detActionNewCol);
			    
			    // Background color
			    Color bgNewCol = new Color( _VAMSA.GlobalOptBgColor.colorPicker.currentColor.r, _VAMSA.GlobalOptBgColor.colorPicker.currentColor.g, _VAMSA.GlobalOptBgColor.colorPicker.currentColor.b, _VAMSA.GlobalOptBgOpacity.val );
			    Color bgNewColOver = new Color( _VAMSA.GlobalOptBgOverColor.colorPicker.currentColor.r, _VAMSA.GlobalOptBgOverColor.colorPicker.currentColor.g, _VAMSA.GlobalOptBgOverColor.colorPicker.currentColor.b, _VAMSA.GlobalOptBgOverOpacity.val );
			    Color bgNewColClick = new Color( _VAMSA.GlobalOptBgClickColor.colorPicker.currentColor.r, _VAMSA.GlobalOptBgClickColor.colorPicker.currentColor.g, _VAMSA.GlobalOptBgClickColor.colorPicker.currentColor.b, _VAMSA.GlobalOptBgClickOpacity.val );

				// When it's a button
			    if (ActionType.val == "Button") {
				    ColorBlock tmpCb = BTNAction.colors;
				    tmpCb.normalColor = bgNewCol;
				    tmpCb.highlightedColor = bgNewColOver;
				    tmpCb.pressedColor = bgNewColClick;
				    tmpCb.fadeDuration = _VAMSA.GlobalOptBgTransitionSpeed.val;
				    BTNAction.colors = tmpCb;
			    } else if(ActionType.val == "Toggle Button")
			    {
				    ColorBlock tmpCb = BTNAction.colors;
				    tmpCb.normalColor = bgNewCol;
				    tmpCb.highlightedColor = bgNewColOver;
				    tmpCb.pressedColor = bgNewColClick;
				    tmpCb.fadeDuration = _VAMSA.GlobalOptBgTransitionSpeed.val;
				    BTNAction.colors = tmpCb;
				    
				    ColorBlock tmpCbOff = BTNActionOff.colors;
				    tmpCbOff.normalColor = bgNewCol;
				    tmpCbOff.highlightedColor = bgNewColOver;
				    tmpCbOff.pressedColor = bgNewColClick;
				    tmpCbOff.fadeDuration = _VAMSA.GlobalOptBgTransitionSpeed.val;
				    BTNActionOff.colors = tmpCbOff;
			    } else
			    {
				    IMGAction.color = bgNewCol;
			    }
		    }
		    
		    protected void UpdateComponentsIDs()
		    {
			    // Deregister
			    DeregisterStorables();

			    // Update
			    ActionText.name = SAVE_PREFIX + ActionID + "-Text";
			    ActionTextOff.name = SAVE_PREFIX + ActionID + "-TextOff";
			    ActionType.name = SAVE_PREFIX + ActionID + "-Type";
			    ActionToggleDefaultState.name = SAVE_PREFIX + ActionID + "-ToggleDefaultState";
			    ActionToggleNoExecAtStart.name = SAVE_PREFIX + ActionID + "-ToggleNoExecAtStart";
			    ActionSliderDefaultValue.name = SAVE_PREFIX + ActionID + "-SliderDefaultValue";
			    triggerOnClickAction.Name = SAVE_PREFIX + ActionID + "-TriggerOnClick";
			    triggerToggleOnAction.Name = SAVE_PREFIX + ActionID + "-TriggerToggleOn";
			    triggerToggleOffAction.Name = SAVE_PREFIX + ActionID + "-TriggerToggleOff";
			    triggerSliderAction.Name = SAVE_PREFIX + ActionID + "-TriggerSlider";
			    
			    // Register
			    RegisterStorables();
		    }

		    public void CopyDatasFromAction(StoryAction sourceAct)
		    {
			    ActionText.val = sourceAct.ActionText.val;
			    ActionTextOff.val = sourceAct.ActionTextOff.val;
			    ActionType.val = sourceAct.ActionType.val;
			    ActionToggleDefaultState.val = sourceAct.ActionToggleDefaultState.val;
			    ActionToggleNoExecAtStart.val = sourceAct.ActionToggleNoExecAtStart.val;
			    ActionSliderDefaultValue.val = sourceAct.ActionSliderDefaultValue.val;

			    triggerOnClickAction.RestoreFromJSON(sourceAct.triggerOnClickAction.GetJSON());
			    triggerToggleOnAction.RestoreFromJSON(sourceAct.triggerToggleOnAction.GetJSON());
			    triggerToggleOffAction.RestoreFromJSON(sourceAct.triggerToggleOffAction.GetJSON());
			    triggerSliderAction.RestoreFromJSON(sourceAct.triggerSliderAction.GetJSON());
		    }
		    
			public void Delete() {
				// Clearing gameobject
				Destroy(ActionGO);

				// Unregistering vars
				DeregisterStorables();

				// Removing triggers
				triggerOnClickAction.Remove();
				triggerToggleOnAction.Remove();
				triggerToggleOffAction.Remove();
				triggerSliderAction.Remove();
				
				// Clearing data from the loaded data of the scene (to prevent re-restore after deleting)
				// This only happens when you load a scene and have data loaded from the json
				if (actionsDatas != null)
				{
					actionsDatas.Remove(SAVE_PREFIX + ActionID + "-Text");
					actionsDatas.Remove(SAVE_PREFIX + ActionID + "-TextOff");
					actionsDatas.Remove(SAVE_PREFIX + ActionID + "-Type");
					actionsDatas.Remove(SAVE_PREFIX + ActionID + "-ToggleDefaultState");
					actionsDatas.Remove(SAVE_PREFIX + ActionID + "-SliderDefaultValue");
					actionsDatas.Remove(SAVE_PREFIX + ActionID + "-TriggerOnClick");
					actionsDatas.Remove(SAVE_PREFIX + ActionID + "-TriggerToggleOn");
					actionsDatas.Remove(SAVE_PREFIX + ActionID + "-TriggerToggleOff");
					actionsDatas.Remove(SAVE_PREFIX + ActionID + "-TriggerSlider");
				}
			}

			protected void DeregisterStorables()
			{
				_VAMSA.DeregisterString(ActionText);
				_VAMSA.DeregisterString(ActionTextOff);
				_VAMSA.DeregisterString(ActionType);
				_VAMSA.DeregisterBool(ActionToggleDefaultState);
				_VAMSA.DeregisterBool(ActionToggleNoExecAtStart);
				_VAMSA.DeregisterFloat(ActionSliderDefaultValue);
			}
			
			protected void RegisterStorables()
			{
				_VAMSA.RegisterString(ActionText);
				_VAMSA.RegisterString(ActionTextOff);
				_VAMSA.RegisterString(ActionType);
				_VAMSA.RegisterBool(ActionToggleDefaultState);
				_VAMSA.RegisterBool(ActionToggleNoExecAtStart);
				_VAMSA.RegisterFloat(ActionSliderDefaultValue);
			}
	    }
    }
}