using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using System;
using System.Text;
using System.Text.RegularExpressions;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.IO;
using MeshVR;
using MVR.FileManagementSecure;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;
using MVR;

// VAMStory Director
//
// Director for VAMStory

namespace VAMStoryDirectorPlugin
{
	public partial class VAMStoryDirector : MVRScript
    {
	    // Strings! (yes I know you like that)
	    private static string PLUGIN_PATH;
	    private static string BACKUP_PATH = "Saves\\PluginData\\vamstory\\backups";
	    private static string SAVE_PREFIX = "Dir_";
	    
		// Static Storables
		protected JSONStorableStringChooser GlobalOptOnLoadTrigger;

		// VAMS External values
		public JSONStorableBool VAMSValue_Ready;
		
		// The usual help section
		protected JSONStorableString helpText;
		
		// Action UI storables (or input fields)
		public JSONStorableString TriggerNameUI;
		protected InputField TriggerNameInputFieldUI;
		public JSONStorableFloat TriggerDelayUI;
		
		// Triggers
		protected List<StoryTrigger> storyTriggers = new List<StoryTrigger>();
		protected Dictionary<int, string> triggersIdsToNames = new Dictionary<int, string>();
		protected StoryTrigger activeTrigger
		{
			get
			{
				return storyTriggers.ElementAtOrDefault(activeTriggerId);
			}
		}
		protected int activeTriggerId
		{
			get
			{
				return triggersIdsToNames.FirstOrDefault(el => el.Value == TriggerSelector.val).Key;
			}
		}

		protected List<EventTrigger> _timedTriggers = new List<EventTrigger>();
		
		// Plugin data (due to the fact that the UI is instantiated after init, I need to grab my data and restore manually
		protected JSONClass directorDatas { get; set; }
		
		// Dynamic UI
		protected List<UIDynamic> globalControlsUIs = new List<UIDynamic>();
		protected List<UIDynamic> helpUIs = new List<UIDynamic>();
		
		protected List<UIDynamic> triggersComponentsUI = new List<UIDynamic>();
		protected List<UIDynamic> ignoreLayoutUI = new List<UIDynamic>();

		// Global UI components
		protected JSONStorableStringChooser RightUISelector;
		protected List<string> uisChoices = new List<string> {"Director controls", "Help"};
		protected JSONStorableStringChooser TriggerSelector;
		protected List<string> triggersChoices;
		protected List<string> triggersLoadChoices = new List<string> {"None"};
		protected JSONStorableStringChooser CopySelector;
		protected List<string> copyChoices = new List<String>();

		protected UIDynamicTextField triggerTitle;
		
		protected UIDynamicTextField RightUIMainTitle;
		
		// "Triggers / Actions"
		protected JSONStorableStringChooser ExecuteTrigger;
		protected List<string> executeTriggerChoices;
		
		protected JSONStorableStringChooser SetTriggerGateInactive;
		protected JSONStorableStringChooser SetTriggerGateActive;
		protected List<string> gateTriggerChoices;
		
		// Init, security flags and values
		protected bool _isSceneLoading { get { return SuperController.singleton.isLoading; } }
		public bool isSceneFrozen { get { return SuperController.singleton.freezeAnimation; } }
		public bool isInEditMode { get { return SuperController.singleton.gameMode == SuperController.GameMode.Edit; } }

		private bool isInit = false;
		private bool dataLoaded = false;
		private bool _allVAMSReady = false;
		private float _vamsReadyCheckPeriod = 0.5f;

		public override void Init()
        {
            try
            {
                if (containingAtom.type != "Empty")
                {
                    SuperController.LogError("Please add VAMStory on an Empty Atom");
                    return;
                }

                PLUGIN_PATH = GetPluginPath(this);

				// Backup folder
				FileManagerSecure.CreateDirectory(BACKUP_PATH);
				StartCoroutine("AutoBackup");
                
                // ******* CREATING MAIN TRIGGERS/ACTIONS ********
                // They are here to be displayed on TOP of every other JSONStorables
                JSONStorableAction fakeFuncUseBelow = new JSONStorableAction("- - - - Use these functions below ↓ - - - - -", () => {});
                RegisterAction(fakeFuncUseBelow);
                
                executeTriggerChoices = new List<string>();
                ExecuteTrigger = new JSONStorableStringChooser("Execute Trigger Name", executeTriggerChoices, "", "Execute Trigger Name") {isStorable=false,isRestorable=false};
                ExecuteTrigger.setCallbackFunction += (val) => { OnExecuteTriggerName(val); };
                RegisterStringChooser(ExecuteTrigger);
                
                gateTriggerChoices = new List<string>();
                SetTriggerGateInactive = new JSONStorableStringChooser("Set Gate Inactive for Trigger Name", gateTriggerChoices, "", "Set Gate Inactive for Trigger Name") {isStorable=false,isRestorable=false};
                SetTriggerGateInactive.setCallbackFunction += (val) => { OnSetGateForTriggerName(val, false); };
                RegisterStringChooser(SetTriggerGateInactive);
                
                SetTriggerGateActive = new JSONStorableStringChooser("Set Gate Active for Trigger Name", gateTriggerChoices, "", "Set Gate Active for Trigger Name") {isStorable=false,isRestorable=false};
                SetTriggerGateActive.setCallbackFunction += (val) => { OnSetGateForTriggerName(val, true); };
                RegisterStringChooser(SetTriggerGateActive);
                
                VAMSValue_Ready = new JSONStorableBool("VAMSValue_Ready", false) {isStorable=false,isRestorable=false};
                RegisterBool(VAMSValue_Ready);
                
                // Fake action	                
                JSONStorableAction fakeFuncDontUseBelow = new JSONStorableAction("- - - - DON'T Use these functions below ↓ - - - - -", () => {});
                RegisterAction(fakeFuncDontUseBelow);
                
                // ******* CREATING MAIN UI *********************************
                // Temporary vars
                UIDynamicTextField tmpTextfield;
                UIDynamicPopup tmpPopup;
                UIDynamicSlider tmpSlider;
                UIDynamicColorPicker tmpColor;
                UIDynamicButton tmpButton;
                UIDynamicToggle tmpToggle;
                UIDynamic tmpSpacer;
                
                // ******* TITLE **********************
                createStaticDescriptionText("VAMStory Title","<color=#000><size=35><b>VAM STORY DIRECTOR</b></size></color>",true,40,TextAnchor.MiddleRight);
                
				RightUISelector = new JSONStorableStringChooser("Right UI Selector", uisChoices, uisChoices.First(), "Show") {isStorable=false,isRestorable=false};
                RightUISelector.setCallbackFunction += (val) => { OnChangeRightUISelection(); };
                UIDynamicPopup RUISelsp = CreateScrollablePopup(RightUISelector, true);
                RUISelsp.labelWidth = 95f;

                tmpSpacer = CreateSpacer(true);
                setupSpacer(tmpSpacer, 20f);                                
                
                RightUIMainTitle = createStaticDescriptionText("Right UI Main Title","<color=#000><size=35><b>DIRECTOR CONTROLS</b></size></color>",true,55,TextAnchor.MiddleRight);
				
                tmpButton = CreateButton("Test current trigger",true);
                tmpButton.button.onClick.AddListener(() => { OnTestSelectedTrigger(); } );
                setButtonColor(tmpButton, new Color(0.1f, 0.5f, 0.8f, 1f));
                setButtonTextColor(tmpButton, new Color(1f, 1f, 1f, 1f));
                globalControlsUIs.Add((UIDynamic) tmpButton);
                
                // Trigger choices
                triggersChoices = new List<string>();
                TriggerSelector = new JSONStorableStringChooser("Trigger", triggersChoices, "", "Trigger") {isStorable=false,isRestorable=false};
                TriggerSelector.setCallbackFunction += (val) => { OnToggleTriggerId(val); };
                UIDynamicPopup DSelsp = CreateScrollablePopup(TriggerSelector, true);
                DSelsp.labelWidth = 250f;
                globalControlsUIs.Add((UIDynamic) DSelsp);
                
                UIDynamicButton newTriggerBtn = CreateButton("New trigger", true);
                newTriggerBtn.button.onClick.AddListener( () => { OnAddNewTrigger(); } );
                setButtonColor(newTriggerBtn, new Color(0.3f, 0.6f, 0.3f, 1f));
                setButtonTextColor(newTriggerBtn, new Color(1f, 1f, 1f, 1f));
                globalControlsUIs.Add((UIDynamic) newTriggerBtn);
                setupButtonWithLayout(newTriggerBtn, 250f);
                
                UIDynamicButton deleteTriggerBtn = CreateButton("Delete trigger", true);
                deleteTriggerBtn.button.onClick.AddListener( () => { OnDeleteCurrentTrigger(); } );
                setButtonColor(deleteTriggerBtn, new Color(0.6f, 0.3f, 0.3f, 1f));
                setButtonTextColor(deleteTriggerBtn, new Color(1f, 1f, 1f, 1f));
                globalControlsUIs.Add((UIDynamic) deleteTriggerBtn);
                ignoreLayoutUI.Add((UIDynamic) deleteTriggerBtn);
                setupButtonWithoutLayout(deleteTriggerBtn, 250f, new Vector2(270, -490));
                
                tmpTextfield = createStaticDescriptionText("Trigger copy Title","<color=#000><size=27><b>COPY TRIGGER</b></size></color>",true,35,TextAnchor.MiddleRight);
                globalControlsUIs.Add((UIDynamic) tmpTextfield);
                
                CopySelector = new JSONStorableStringChooser("Trigger Source", copyChoices, "", "Source/Target") {isStorable=false,isRestorable=false};
                UIDynamicPopup DSosp = CreateScrollablePopup(CopySelector, true);
                DSosp.labelWidth = 220f;
                globalControlsUIs.Add((UIDynamic) DSosp);
                
                UIDynamicButton copyTriggerBtn = CreateButton("Copy", true);
                copyTriggerBtn.button.onClick.AddListener(() => { OnCopyTrigger(); });
                setButtonColor(copyTriggerBtn, new Color(0.8f, 0.8f, 0.1f, 1f));
                globalControlsUIs.Add((UIDynamic) copyTriggerBtn);
                setupButtonWithLayout(copyTriggerBtn, 250f);
                
                UIDynamicButton transferTriggerBtn = CreateButton("Transfer", true);
                transferTriggerBtn.button.onClick.AddListener(() => { OnTransferTrigger(); });
                setButtonColor(transferTriggerBtn, new Color(0.8f, 0.5f, 0f, 1f));
                globalControlsUIs.Add((UIDynamic) transferTriggerBtn);
                ignoreLayoutUI.Add((UIDynamic) transferTriggerBtn);
                setupButtonWithoutLayout(transferTriggerBtn, 250f, new Vector2(270, -720));
                
                // Start trigger
                tmpSpacer = CreateSpacer(true);
                globalControlsUIs.Add((UIDynamic) tmpSpacer);
                setupSpacer(tmpSpacer, 15f);                                

                tmpTextfield = createStaticDescriptionText("Start trigger Title","<color=#000><size=27><b>START TRIGGER</b>\n select the trigger that has to be executed when the scene loads</size></color>",true,95,TextAnchor.MiddleRight);
                globalControlsUIs.Add((UIDynamic) tmpTextfield);
                
                GlobalOptOnLoadTrigger = new JSONStorableStringChooser("Global_Opt_LoadTrigger", triggersLoadChoices, "None", "Start trigger");
                RegisterStringChooser(GlobalOptOnLoadTrigger);
                tmpPopup = CreateScrollablePopup(GlobalOptOnLoadTrigger, true);
                tmpPopup.labelWidth = 220f;
                globalControlsUIs.Add((UIDynamic) tmpPopup);
                
                // Backups
                tmpSpacer = CreateSpacer(true);
                globalControlsUIs.Add((UIDynamic) tmpSpacer);
                setupSpacer(tmpSpacer, 15f);                                

                tmpTextfield = createStaticDescriptionText("Backups Title","<color=#000><size=27><b>BACKUPS</b></size></color>",true,35,TextAnchor.MiddleRight);
                globalControlsUIs.Add((UIDynamic) tmpTextfield);
                
                tmpButton = CreateButton("Restore backup",true);
                tmpButton.button.onClick.AddListener(() => { LoadBackup(); } );
                setButtonColor(tmpButton, new Color(0.1f, 0.5f, 0.7f, 1f));
                setButtonTextColor(tmpButton, new Color(1f, 1f, 1f, 1f));
                globalControlsUIs.Add((UIDynamic) tmpButton);
                
                tmpButton = CreateButton("Save manual backup",true);
                tmpButton.button.onClick.AddListener(() => { SaveBackup(2); } );
                setButtonColor(tmpButton, new Color(0.1f, 0.7f, 0.4f, 1f));
                setButtonTextColor(tmpButton, new Color(1f, 1f, 1f, 1f));
                globalControlsUIs.Add((UIDynamic) tmpButton);

                // ******* HELP **********************
                helpText = new JSONStorableString("Help",
	                "<size=30><color=#333>" +
	                "This plugin helps you control your scene with triggers.\n\n" +
	                "<b>PLUGIN HELP:</b>\n" +
	                "<b>VAMStoryDirector</b> is a simple way to control the scene flow by using triggers all in one place.\n\n" +
	                "You can create how many triggers as you want and call them from any external source ( a button, Timeline, VAMStory... ).\n\n" +
	                "<color=#c81b0d><b>WARNING!: You need to keep VAMStoryDirector and all other VAMS plugins atoms always on. See hub's documentation for more information.</b></color>\n\n" +
	                "<b>TRIGGER CONTROLS</b>\n" +
	                "This is where you navigate through your triggers, as simple as that.\n\n" +
	                "<b>Test current trigger</b> will execute the currently selected trigger. It will take the delay you've configured, so just click once.\n\n" +
	                "<b>COPY/TRANSFER TRIGGER</b>\n" +
	                "This is the way to copy or transfer a trigger inside the director. (delay and trigger actions only)\n\n" +
	                "Select a trigger in the Source/Target dropdown.\n\n" +
	                "Then click either <b>Copy</b> or <b>Transfer</b>.\n\n" +
	                "<b>Copy</b> will grab the values of the selected trigger in the dropdown and copy its content to the current trigger. <b>Transfer</b> will send the current trigger values to the selected trigger in the dropdown.\n\n" +
	                "<b>START TRIGGER</b>\n" +
	                "Here, select one of the triggers of your director to be executed when the scene is started.\n\n" +
	                "<b>BACKUPS</b>\n" +
	                "<b>VAMStoryDirector</b> performs an automatic backup every 30 minutes only in edit mode.\n\n" +
	                "You can also restore a backup or save a manual backup here.\n\n" +
	                "You can use this as a <b>preset system</b> if you want..\n\n" +
	                "</color></size>"
                );
				
                tmpTextfield = CreateTextField(helpText, true);
                tmpTextfield.height = 850.0f;
                helpUIs.Add((UIDynamic) tmpTextfield);

                // Renaming callback
                SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;
                
                // Updating UI
                OnChangeRightUISelection();
                
                // Loading triggers assets
                SimpleTriggerHandler.LoadAssets();
                
                // Initializing UI
                CreateTriggerUI();
            }
            catch(Exception e)
            {
                SuperController.LogError("VAMStory - Exception caught: " + e);
            }
		}
		
		void Start()
		{

		}
		
		void Update() {
			// Base initialization after assetbundle load
			if (isInit == false && !_isSceneLoading)
			{
				try
				{
					// Setting options and ready state for the suite once
					if (isInit == false && VAMSValue_Ready.val == false)
					{
						// Starting to wait for the VAMStory plugins
						StartCoroutine("WaitForVAMSPlugins");
						
						// If the restore process has not happened (scene loading)
						// then we create the default dataset
						if (dataLoaded == false)
						{
							dataLoaded = true;
							CreateDefaultTrigger();
						}

						VAMSValue_Ready.val = true;						
					}
					// Firing triggers when the whole VAMS "suite" in the scene is ready
					else if ( _allVAMSReady )
					{
						isInit = true;
						
						// Firing the selected trigger at load
						if ( GlobalOptOnLoadTrigger.val != "None" )
						{
							int startTriggerId = triggersIdsToNames.FirstOrDefault(el => el.Value == GlobalOptOnLoadTrigger.val).Key;
							StoryTrigger startTrigger = storyTriggers.ElementAtOrDefault(startTriggerId);
							float trigDelay = startTrigger.TriggerDelay.val;
							if (trigDelay > 0f)
							{
								StartCoroutine(WaitForNextTrigger(startTriggerId));
							}
							else
							{
								DoExecuteTriggerId(startTriggerId);
							}
						}
					}
				}
				catch(Exception e)
				{
					SuperController.LogError("VAMStory - Exception caught: " + e);
				}
			}

			// Updating triggers with timers
			if (isInit == true && _timedTriggers.Count > 0)
			{
				foreach (EventTrigger trig in _timedTriggers)
				{
					trig.Update();
				}
			}
		}

		// **************************
		// Functions
		// **************************

		private void CreateDefaultTrigger()
		{
			StoryTrigger newTrigger = new StoryTrigger(this, 0);
			storyTriggers.Add(newTrigger);
			RefreshSelectors(0);
		}

		private void LoadTriggers()
		{
			// The save system is pretty "awkward" as there are not a lot of possibilities to properly create a json file
			// that would help to restore properly the storables natively
			// So
			int triggersCount = 0;
			foreach (string paramKey in directorDatas.Keys)
			{
				if (paramKey == "id") continue;// Skipping plugin id
				string[] paramSplit = paramKey.Split('-');

				if (paramSplit.Length != 2) continue; // Skipping anything that has not at least 2 entries on its param name
				
				// Only keeping "texts" (it's enough to know what I need to create the UI)
				if (paramSplit[1] == "Name")
				{
					string[] idsSplit = paramSplit[0].Split('_');
					if (idsSplit.Length == 2)
					{
						int splitId = int.Parse(idsSplit[1]);
						if (triggersCount < splitId) triggersCount = splitId;
					}
				}
			}

			// If we did find triggers in the scene
			if (triggersCount > 0)
			{
				for (var i = 0; i <= triggersCount; i++)
				{
					StoryTrigger newTrigger = new StoryTrigger(this, i);
					storyTriggers.Add(newTrigger);
				}
				
				RefreshSelectors(0);
			}
			// If it's not the case, let's create the default trigger
			else
			{
				CreateDefaultTrigger();
			}
		}

		protected void CreateTriggerUI()
		{
			// Temporary vars
		    UIDynamicTextField tmpTextfield;
		    UIDynamicPopup tmpPopup;
		    UIDynamicSlider tmpSlider;
		    UIDynamicColorPicker tmpColor;
		    UIDynamicButton tmpButton;
		    UIDynamicToggle tmpToggle;
		    UIDynamic tmpSpacer;
		    
		    // Creating Trigger components
			// ******* TRIGGER TITLE (ID) ***********
			triggerTitle = createStaticDescriptionText("TriggerTitleUI","<color=#000><size=35><b>Trigger</b></size></color>",false,55,TextAnchor.MiddleLeft);
			triggersComponentsUI.Add((UIDynamic) triggerTitle);
			
			// ******* TRIGGER TEXT ***********
			TriggerNameUI = new JSONStorableString("TriggerNameUI", "_default_"){isStorable=false,isRestorable=false};
			tmpTextfield = CreateTextField(TriggerNameUI);
			setupTextField(tmpTextfield, 150f, false, false);
			TriggerNameInputFieldUI = tmpTextfield.UItext.gameObject.AddComponent<InputField>();
			TriggerNameInputFieldUI.textComponent = tmpTextfield.UItext;
			TriggerNameInputFieldUI.lineType = InputField.LineType.MultiLineNewline;
			triggersComponentsUI.Add((UIDynamic) tmpTextfield);
			TriggerNameUI.valNoCallback = ""+activeTriggerId;
			TriggerNameInputFieldUI.text = ""+activeTriggerId;				
			TriggerNameInputFieldUI.onValueChanged.AddListener( delegate { OnTriggerNameChanged(); } );
			TriggerNameInputFieldUI.onEndEdit.AddListener(delegate { OnTriggerNameEdited(); } );

			// ******* TRIGGERS ***********
			tmpTextfield = createStaticDescriptionText("GlobalTrigger Title","<color=#000><size=25><b>ALWAYS TRIGGERS</b></size></color>",false,35,TextAnchor.MiddleLeft);
			triggersComponentsUI.Add((UIDynamic) tmpTextfield);
			
			tmpButton = CreateButton("Edit Global trigger", false);
			triggersComponentsUI.Add((UIDynamic)tmpButton);
			tmpButton.button.onClick.AddListener(() => { activeTrigger.triggerAction.OpenPanelActionStart(); } );
			
			tmpTextfield = createStaticDescriptionText("DesktopTrigger Title","<color=#000><size=25><b>TRIGGERS ONLY IN DESKTOP MODE</b></size></color>",false,35,TextAnchor.MiddleLeft);
			triggersComponentsUI.Add((UIDynamic) tmpTextfield);
			
			tmpButton = CreateButton("Edit Desktop trigger", false);
			triggersComponentsUI.Add((UIDynamic)tmpButton);
			tmpButton.button.onClick.AddListener(() => { activeTrigger.triggerActionDesktop.OpenPanelActionStart(); } );
			
			tmpTextfield = createStaticDescriptionText("VRTrigger Title","<color=#000><size=25><b>TRIGGERS ONLY IN VR MODE</b></size></color>",false,35,TextAnchor.MiddleLeft);
			triggersComponentsUI.Add((UIDynamic) tmpTextfield);
			
			tmpButton = CreateButton("Edit VR trigger", false);
			triggersComponentsUI.Add((UIDynamic)tmpButton);
			tmpButton.button.onClick.AddListener(() => { activeTrigger.triggerActionVR.OpenPanelActionStart(); } );
			
			tmpTextfield = createStaticDescriptionText("GateTrigger Title","<color=#000><size=25><b>GATE TRIGGER</b></size></color>",false,35,TextAnchor.MiddleLeft);
			triggersComponentsUI.Add((UIDynamic) tmpTextfield);
			
			tmpButton = CreateButton("Check state", false);
			tmpButton.button.onClick.AddListener(() => { OnCheckTriggerGateState(); });
			setButtonColor(tmpButton, new Color(0.15f, 0.5f, 0.5f, 1f));
			setButtonTextColor(tmpButton, new Color(1f, 1f, 1f, 1f));
			triggersComponentsUI.Add((UIDynamic) tmpButton);
			ignoreLayoutUI.Add((UIDynamic) tmpButton);
			setupButtonWithoutLayout(tmpButton, 170f, new Vector2(365, -610));
			
			tmpButton = CreateButton("Edit Gate Inactive trigger", false);
			triggersComponentsUI.Add((UIDynamic)tmpButton);
			tmpButton.button.onClick.AddListener(() => { activeTrigger.triggerActionGateInactive.OpenPanelActionStart(); } );
			
			tmpButton = CreateButton("Edit Gate Active trigger", false);
			triggersComponentsUI.Add((UIDynamic)tmpButton);
			tmpButton.button.onClick.AddListener(() => { activeTrigger.triggerActionGateActive.OpenPanelActionStart(); } );
			
			// ******* TRIGGER START DELAY ***********
			TriggerDelayUI = new JSONStorableFloat("TriggerDelayUI", 0f, 0f, 5f, false){isStorable=false,isRestorable=false};
			TriggerDelayUI.setCallbackFunction += (val) => { OnTriggerDelayChanged(); };
			tmpSlider = CreateSlider(TriggerDelayUI);
			tmpSlider.valueFormat = "F1";
			tmpSlider.label = "Trigger delay";
			triggersComponentsUI.Add((UIDynamic) tmpSlider);
		}

		protected void ClearAllTriggerUIs()
		{
			foreach (var TriggerItem in storyTriggers)
			{
				TriggerItem.ToggleTriggerUI(false);			
			}
		}

		protected void RefreshSelectors( int triggerId )
		{
			triggersChoices = new List<string>();
			triggersLoadChoices = new List<string>();
			triggersLoadChoices.Add("None"); // The empty default one
			copyChoices = new List<string>();
			copyChoices.Add(""); // The empty default one
			triggersIdsToNames.Clear();

			foreach (var trig in storyTriggers)
			{
				triggersChoices.Add(trig.TriggerName.val);
				triggersLoadChoices.Add(trig.TriggerName.val);
				copyChoices.Add(trig.TriggerName.val);
				triggersIdsToNames.Add(trig.TriggerID, trig.TriggerName.val);
			}

			// Updating trigger selector
			TriggerSelector.valNoCallback = "";
			TriggerSelector.choices = triggersChoices;
			TriggerSelector.val = triggersIdsToNames[triggerId];

			string oldVal = GlobalOptOnLoadTrigger.val; // doing this to reset the slider :/
			GlobalOptOnLoadTrigger.valNoCallback = "None";
			GlobalOptOnLoadTrigger.choices = triggersLoadChoices;
			GlobalOptOnLoadTrigger.val = triggersIdsToNames.ContainsValue(oldVal) ? oldVal : "None";
			
			CopySelector.choices = copyChoices;
			
			// Also updating the trigger system to execute the triggers
			ExecuteTrigger.valNoCallback = "";
			SetTriggerGateInactive.valNoCallback = "";
			SetTriggerGateActive.valNoCallback = "";
			ExecuteTrigger.choices = triggersChoices;
			SetTriggerGateInactive.choices = triggersChoices;
			SetTriggerGateActive.choices = triggersChoices;
		}
		
		protected void ToggleTriggerUI(bool enabled, string triggerName)
		{
			int triggerId = triggersIdsToNames.FirstOrDefault(el => el.Value == triggerName).Key;
			if (storyTriggers.ElementAtOrDefault(triggerId) == null) return;
			storyTriggers[triggerId].ToggleTriggerUI(enabled);
		}
		
		protected void ToggleUIDynamic(bool enabled, UIDynamic target)
		{
			if (target == null) return;
			
			LayoutElement tmpLE = target.GetComponent<LayoutElement>();

			if (tmpLE != null)
			{
				if (enabled == true)
				{

					tmpLE.transform.localScale = new Vector3(1, 1, 1);
					if( !ignoreLayoutUI.Contains(target) ) tmpLE.ignoreLayout = false;
				}
				else
				{
					tmpLE.transform.localScale = new Vector3(0, 0, 0);
					tmpLE.ignoreLayout = true;
				}
			}
		}
		
		private void setupTextField( UIDynamicTextField target, float fieldHeight, bool disableBackground = true, bool disableScroll = true ) {
			if( disableBackground ) target.backgroundColor = new Color(1f, 1f, 1f, 0f);
			LayoutElement tfLayout = target.GetComponent<LayoutElement>();
			tfLayout.preferredHeight = tfLayout.minHeight = fieldHeight;
			target.height = fieldHeight;
			if( disableScroll ) disableScrollOnText(target);
		}

		private void setupSpacer(UIDynamic target, float newHeight)
		{
			LayoutElement tmpLE = target.GetComponent<LayoutElement>();
			tmpLE.minHeight = newHeight;
			tmpLE.preferredHeight = newHeight;
		}

		private void setupButtonWithLayout(UIDynamicButton target, float newSize)
		{
			ContentSizeFitter tmpCSF = target.button.transform.gameObject.AddComponent<ContentSizeFitter>();
			tmpCSF.horizontalFit = UnityEngine.UI.ContentSizeFitter.FitMode.MinSize;
			LayoutElement tmpLE = target.button.transform.gameObject.GetComponent<LayoutElement>();
			RectTransform tmpRectT = target.button.transform.GetComponent<RectTransform>();
			Rect tmpRect = tmpRectT.rect;
			tmpRectT.pivot = new Vector2(0f, 0.5f);
			tmpLE.minWidth = newSize;
			tmpLE.preferredWidth = newSize;
		}
		
		private void setupButtonWithoutLayout(UIDynamicButton target, float newSize, Vector2 newPosition)
		{
			ContentSizeFitter tmpCSF = target.button.transform.gameObject.AddComponent<ContentSizeFitter>();
			tmpCSF.horizontalFit = UnityEngine.UI.ContentSizeFitter.FitMode.MinSize;
			LayoutElement tmpLE = target.button.transform.gameObject.GetComponent<LayoutElement>();
			RectTransform tmpRectT = target.button.transform.GetComponent<RectTransform>();
			Rect tmpRect = tmpRectT.rect;
			tmpRectT.pivot = new Vector2(0f, 0.5f);
			tmpLE.minWidth = newSize;
			tmpLE.preferredWidth = newSize;
			tmpLE.ignoreLayout = true;
			tmpRectT.anchoredPosition = newPosition;
		}

		private void setButtonColor(UIDynamicButton targetBtn, Color newColor)
		{
			targetBtn.buttonColor = newColor;
		}
		
		private void setButtonTextColor(UIDynamicButton targetBtn, Color newColor)
		{
			targetBtn.textColor = newColor;
		}
		
		public UIDynamicTextField createStaticDescriptionText( string DescTitle, string DescText, bool rightSide, int fieldHeight, TextAnchor textAlignment = TextAnchor.UpperLeft ) {
			JSONStorableString staticDescString = new JSONStorableString(DescTitle,DescText) {isStorable=false,isRestorable=false};
			staticDescString.hidden = true;
			UIDynamicTextField staticDescStringField = CreateTextField(staticDescString, rightSide);
			staticDescStringField.backgroundColor = new Color(1f, 1f, 1f, 0f);
			staticDescStringField.UItext.alignment = textAlignment;
			LayoutElement sdsfLayout = staticDescStringField.GetComponent<LayoutElement>();
			sdsfLayout.preferredHeight = sdsfLayout.minHeight = fieldHeight;
			staticDescStringField.height = fieldHeight;
			disableScrollOnText(staticDescStringField);

			return staticDescStringField;
		}
		
		private static void disableScrollOnText(UIDynamicTextField target) {
			ScrollRect targetSR = target.UItext.transform.parent.transform.parent.transform.parent.GetComponent<ScrollRect>();
			if( targetSR != null ) {
				targetSR.horizontal = false;
				targetSR.vertical = false;
			}
		}
		
		private Transform getChildNamed( Transform parent, string childName ) {
			foreach( Transform child in parent )
			{
				if( child.name == childName ) {
					return child;
				} else {
					Transform childSearch = getChildNamed(child,childName);
					if (childSearch != null) return childSearch;
				}
			}
        
			return null;
		}
		
		// Saving the datas
		// saveType: 0 = auto, 1 = critical, 2 = manual
		protected void SaveBackup( int saveType = 0 )
		{
			if (SuperController.singleton.gameMode != SuperController.GameMode.Edit) return;
			string fileName = "vamsd-" + GetCleanFilenameFromString(containingAtom.name);
			if (saveType == 0)
			{
				logDebug("VAMStoryDirector: Saving automatic backup");
				string backupPath = CleanAndGetBackupPath( fileName, "auto" );
				SuperController.singleton.SaveJSON( this.GetJSON(), backupPath);
			} else if (saveType == 1)
			{
				string backupPath = CleanAndGetBackupPath( fileName, "critical" );
				SuperController.singleton.SaveJSON( this.GetJSON(), backupPath);
			} else if (saveType == 2)
			{
				string backupPath = BACKUP_PATH + "\\" + fileName + "-manual-" + DateTime.Now.ToString("yyyyMMdd-HHmmss") + ".json";
				SuperController.singleton.SaveJSON( this.GetJSON(), backupPath);
				logDebug("VAMStoryDirector: Manual backup saved in:" + backupPath);
			}
		}

		protected string CleanAndGetBackupPath( string backupName, string backupSuffix )
		{
			// I'm keeping 5 files
			for (int i = 5; i > 0; i--)
			{
				string backupPath = BACKUP_PATH + "\\" + backupName + "-" + backupSuffix + "-" + i + ".json";
				if (FileManagerSecure.FileExists(backupPath) == true && i == 5)
				{
					FileManagerSecure.DeleteFile(backupPath);
				}
				else if(FileManagerSecure.FileExists(backupPath) == true)
				{
					FileManagerSecure.MoveFile(BACKUP_PATH + "\\" + backupName + "-" + backupSuffix + "-" + i + ".json", BACKUP_PATH + "\\" + backupName + "-" + backupSuffix + "-" + ( i + 1 ) + ".json",false);
				}
			}

			return BACKUP_PATH + "\\" + backupName + "-" + backupSuffix + "-1.json";
		}
		
		// Load backup
		protected void LoadBackup()
		{
			SuperController.singleton.fileBrowserUI.defaultPath = BACKUP_PATH;
			SuperController.singleton.fileBrowserUI.SetTitle("Select a backup");
			SuperController.singleton.fileBrowserUI.fileFormat = "json";
			SuperController.singleton.fileBrowserUI.showFiles = true;
			SuperController.singleton.fileBrowserUI.showDirs = true;
			SuperController.singleton.fileBrowserUI.SetTextEntry(false);
			SuperController.singleton.fileBrowserUI.hideExtension = false;
			SuperController.singleton.fileBrowserUI.showInstallFolderInDirectoryList = false;
			List<ShortCut> shortCutList = new List<ShortCut>();

			ShortCut shortCutCustom = new ShortCut();
			shortCutCustom.path = BACKUP_PATH;
			shortCutCustom.displayName = "VAMStoryDirector Backups";
			shortCutCustom.flatten = false;
			shortCutList.Add(shortCutCustom);

			SuperController.singleton.fileBrowserUI.shortCuts = shortCutList;			
			SuperController.singleton.fileBrowserUI.Show((val) =>
			{
				if (val != "")
				{
					RestoreBackup(val);
				}
			});
		}

		protected void RestoreBackup(string backupPath)
		{
			// Loading the file
			string backupDatas = SuperController.singleton.ReadFileIntoString(backupPath);
			JSONClass backupJSON = JSON.Parse(backupDatas).AsObject;
			
			// Checking if the save is meant for VAMStory
			if ( !backupJSON.HasKey("id") || !backupJSON["id"].Value.Contains("_VAMStoryDirectorPlugin.VAMStoryDirector") )
			{
				logError("VAMStoryDirector: Backup not compatible with VAMStoryDirector.");
				return;
			}
			
			logDebug("VAMStoryDirector: Restoring " + backupPath);
			
			// Clearing all groups
			foreach (StoryTrigger act in storyTriggers)
			{
				act.Delete();
			}

			storyTriggers.Clear();
			
			// Restoring datas
			this.RestoreFromJSON(backupJSON);
			
			// Loading Dialogs
			LoadTriggers();
			
			// Refreshing UI
			RefreshSelectors(0);
		}
		
		// **************************************
		// Callbacks & similar ( when action UI value is changed )
		// **************************************
		protected void OnTriggerNameChanged()
		{
			TriggerNameUI.val = TriggerNameInputFieldUI.text;
			activeTrigger.TriggerName.val = TriggerNameUI.val;
		}
		
		protected void OnTriggerNameEdited()
		{
			string newName = GetCleanAndUniqueTriggerName();
			TriggerNameInputFieldUI.text = newName;
			UpdateTriggersIdsToNameDict();
			RefreshSelectors(triggersIdsToNames.FirstOrDefault(el => el.Value == newName).Key);
		}
		
		protected void OnTriggerDelayChanged()
		{
			activeTrigger.TriggerDelay.val = TriggerDelayUI.val;
		}
		
		protected void UpdateTriggersIdsToNameDict()
		{
			StoryTrigger tri = activeTrigger;
			if (!triggersIdsToNames.ContainsKey(tri.TriggerID))
			{
				triggersIdsToNames.Add(tri.TriggerID,TriggerNameUI.val);
			}
			else
			{
				triggersIdsToNames[tri.TriggerID] = TriggerNameUI.val;
			}
		}
		
		protected void UpdateLocalUIFromActiveTrigger()
		{
			// Updating the local dialog UI based on dialog datas
			StoryTrigger tri = activeTrigger;
			TriggerNameUI.valNoCallback = tri.TriggerName.val;
			TriggerNameInputFieldUI.text = tri.TriggerName.val;
			TriggerDelayUI.valNoCallback = tri.TriggerDelay.val;
		}

		protected void AddTriggerToTimerPool(EventTrigger trig)
		{
			if (trig.HasActionsWithActiveTimers && !_timedTriggers.Contains(trig))
			{
				_timedTriggers.Add(trig);
			}
		}
		
		protected void RemoveTriggerFromTimerPool(EventTrigger trig)
		{
			if (_timedTriggers.Contains(trig))
			{
				_timedTriggers.Remove(trig);
			}
		}
		
		// **************************************
		// Callbacks & similar ( internal )
		// **************************************
		protected void OnChangeRightUISelection()
		{
			if (RightUISelector.val == "Director controls")
			{
				RightUIMainTitle.text = "<color=#000><size=35><b>TRIGGERS CONTROLS</b></size></color>";
				
				foreach( UIDynamic uidElem in globalControlsUIs )
				{
					ToggleUIDynamic(true, uidElem);
				}

				foreach( UIDynamic uidElem in helpUIs )
				{
					ToggleUIDynamic(false, uidElem);
				}
			} else
			{
				foreach( UIDynamic uidElem in globalControlsUIs )
				{
					ToggleUIDynamic(false, uidElem);
				}

				foreach( UIDynamic uidElem in helpUIs )
				{
					ToggleUIDynamic(true, uidElem);
				}
				
				RightUIMainTitle.text = "<color=#000><size=35><b>HELP</b></size></color>";
			}
		}
		
		protected void OnToggleTriggerId(string triggerName)
		{
			if ( isInit == false ) return;
			
			// Updating trigger UI
			UpdateLocalUIFromActiveTrigger();
		}
		
		protected void OnAddNewTrigger()
		{
			int triggerId = storyTriggers.Count;
			StoryTrigger newTrigger = new StoryTrigger(this, triggerId);
			storyTriggers.Add(newTrigger);
			RefreshSelectors(triggerId);
		}

		protected void OnDeleteCurrentTrigger()
		{
			int triggerId = activeTriggerId;
			
			// We cannot delete the last trigger
			if (storyTriggers.Count == 1)
			{
				SuperController.singleton.Alert("<color=green>VAMStoryDirector</color> needs to have at least one trigger.", () => {});
				return;
			}
			
			// Waiting for user confirmation
			SuperController.singleton.Alert("<color=#A90000FF>Deleting trigger " + triggerId + ":\n\n</color><size=30>" + storyTriggers[triggerId].TriggerName.val + "</size>\n\nAre you sure?", delegate {OnConfirmTriggerDeletion();}, () => {});

		}

		protected void OnConfirmTriggerDeletion()
		{
			// Saving a critical backup
			SaveBackup(1);
			
			// Clearing the "timers" pool (just to prevent errors)
			_timedTriggers.Clear();
			
			// Now delete
			int triggerId = activeTriggerId;
			
			storyTriggers[triggerId].Delete();
			storyTriggers.RemoveAt(triggerId);
			int actCount = 0;
			foreach (StoryTrigger act in storyTriggers)
			{
				act.UpdateID(actCount);
				actCount++;
			}
			
			int selectedTrigger = 0;
			if (storyTriggers.Count > 1 && triggerId > 0)
			{
				selectedTrigger = triggerId - 1;
			}
			
			RefreshSelectors(selectedTrigger);
		}

		protected void OnCopyTrigger()
		{
			if (CopySelector.val == "")
			{
				SuperController.singleton.Alert("Select a source trigger to copy first!", () => {});
				return;
			}
			
			int sourceTriggerId = triggersIdsToNames.FirstOrDefault(el => el.Value == CopySelector.val).Key;
			int targetTriggerId = triggersIdsToNames.FirstOrDefault(el => el.Value == TriggerSelector.val).Key;

			if (sourceTriggerId == targetTriggerId)
			{
				SuperController.singleton.Alert("You cannot copy a trigger on itself!", () => {});
			}
			else
			{
				StoryTrigger sourceTri = storyTriggers[sourceTriggerId];
				StoryTrigger targetTri = storyTriggers[targetTriggerId];

				targetTri.TriggerDelay.val = sourceTri.TriggerDelay.val;
				targetTri.triggerAction.RestoreFromJSON(sourceTri.triggerAction.GetJSON());
				targetTri.triggerActionDesktop.RestoreFromJSON(sourceTri.triggerActionDesktop.GetJSON());
				targetTri.triggerActionVR.RestoreFromJSON(sourceTri.triggerActionVR.GetJSON());
				targetTri.triggerActionGateInactive.RestoreFromJSON(sourceTri.triggerActionGateInactive.GetJSON());
				targetTri.triggerActionGateActive.RestoreFromJSON(sourceTri.triggerActionGateActive.GetJSON());
				
				// Updating trigger UI
				UpdateLocalUIFromActiveTrigger();
			}
		}

		protected void OnTransferTrigger()
		{
			if (CopySelector.val == "")
			{
				SuperController.singleton.Alert("Select a target trigger to transfer to first", () => {});
				return;
			}
			
			int sourceTriggerId = triggersIdsToNames.FirstOrDefault(el => el.Value == TriggerSelector.val).Key;
			int targetTriggerId = triggersIdsToNames.FirstOrDefault(el => el.Value == CopySelector.val).Key;

			if (sourceTriggerId == targetTriggerId)
			{
				SuperController.singleton.Alert("You cannot transfer a trigger on itself", () => {});
			}
			else
			{
				StoryTrigger sourceTri = storyTriggers[sourceTriggerId];
				StoryTrigger targetTri = storyTriggers[targetTriggerId];

				targetTri.TriggerDelay.val = sourceTri.TriggerDelay.val;
				targetTri.triggerAction.RestoreFromJSON(sourceTri.triggerAction.GetJSON());
				targetTri.triggerActionDesktop.RestoreFromJSON(sourceTri.triggerActionDesktop.GetJSON());
				targetTri.triggerActionVR.RestoreFromJSON(sourceTri.triggerActionVR.GetJSON());
				targetTri.triggerActionGateInactive.RestoreFromJSON(sourceTri.triggerActionGateInactive.GetJSON());
				targetTri.triggerActionGateActive.RestoreFromJSON(sourceTri.triggerActionGateActive.GetJSON());
				
				SuperController.singleton.Alert("<color=green>VAMStoryDirector</color>\n Datas transfered to Trigger " + targetTri.TriggerName.val, () => {});
			}
		}
		
		private void OnAtomRename(string oldid, string newid)
		{
			foreach( var tri in storyTriggers ) {
				tri.triggerAction.SyncAtomNames();
				tri.triggerActionDesktop.SyncAtomNames();
				tri.triggerActionVR.SyncAtomNames();
				tri.triggerActionGateInactive.SyncAtomNames();
				tri.triggerActionGateActive.SyncAtomNames();
			}
		}

		protected void OnExecuteTriggerName(string triggerName)
		{
			ExecuteTrigger.valNoCallback = "";  // Resetting so that the user can trigger several times the same group if needed
			var triggerEntry = triggersIdsToNames.Where(el => el.Value == triggerName).Select(el => (KeyValuePair<int,string>?) el).FirstOrDefault();
			if (triggerEntry != null)
			{
				int startTriggerId = triggerEntry.Value.Key;
				StoryTrigger startTrigger = storyTriggers.ElementAtOrDefault(startTriggerId);
				if (startTrigger != null)
				{
					float trigDelay = startTrigger.TriggerDelay.val;
					if (trigDelay > 0f)
					{
						StartCoroutine(WaitForNextTrigger(startTrigger.TriggerID));
					}
					else
					{
						DoExecuteTriggerId(startTrigger.TriggerID);
					}
				}
			}
			else if(isInEditMode)
			{
				logDebug("VAMStoryDirector: trigger named \"" + triggerName + "\" does not exists or has been deleted. (Execute Trigger Name)");
			}
		}

		protected void DoExecuteTriggerId(int triggerId)
		{
			// Executing triggers
			
			// Global trigger
			storyTriggers[triggerId].triggerAction.Trigger();
			AddTriggerToTimerPool(storyTriggers[triggerId].triggerAction);
			
			// VR trigger
			if (SuperController.singleton.isOVR || SuperController.singleton.isOpenVR)
			{
				storyTriggers[triggerId].triggerActionVR.Trigger();
				AddTriggerToTimerPool(storyTriggers[triggerId].triggerActionVR);
			}
			// Desktop trigger
			else
			{
				storyTriggers[triggerId].triggerActionDesktop.Trigger();
				AddTriggerToTimerPool(storyTriggers[triggerId].triggerActionDesktop);
			}
			
			// Gate trigger
			if ( storyTriggers[triggerId].TriggerGateActive == false )
			{
				storyTriggers[triggerId].triggerActionGateInactive.Trigger();
				AddTriggerToTimerPool(storyTriggers[triggerId].triggerActionGateInactive);
				RemoveTriggerFromTimerPool(storyTriggers[triggerId].triggerActionGateActive); // Just in case, removing the "active" version
			}
			else
			{
				storyTriggers[triggerId].triggerActionGateActive.Trigger();
				AddTriggerToTimerPool(storyTriggers[triggerId].triggerActionGateActive);
				RemoveTriggerFromTimerPool(storyTriggers[triggerId].triggerActionGateInactive); // Just in case, removing the "inactive" version
			}
		}

		protected void OnTestSelectedTrigger()
		{
			OnExecuteTriggerName(TriggerSelector.val);
		}

		protected void OnSetGateForTriggerName(string triggerName, bool state)
		{
			SetTriggerGateActive.valNoCallback = "";  // Resetting so that the user can trigger several times the same group if needed
			SetTriggerGateInactive.valNoCallback = "";  // Resetting so that the user can trigger several times the same group if needed
			var triggerEntry = triggersIdsToNames.Where(el => el.Value == triggerName).Select(el => (KeyValuePair<int,string>?) el).FirstOrDefault();
			if (triggerEntry != null)
			{
				int startTriggerId = triggerEntry.Value.Key;
				StoryTrigger startTrigger = storyTriggers.ElementAtOrDefault(startTriggerId);
				if (startTrigger != null)
				{
					startTrigger.TriggerGateActive = state;
				}
			}
			else if(isInEditMode)
			{
				logDebug("VAMStoryDirector: trigger named \"" + triggerName + "\" does not exists or has been deleted. (Set Gate Inactive for Trigger Name)");
			}
		}

		protected void OnCheckTriggerGateState()
		{
			SuperController.singleton.Alert("<color=green>VAMStoryDirector</color>\nTrigger \"" + activeTrigger.TriggerName.val + "\"\n gate is " + (activeTrigger.TriggerGateActive == true ? "active" : "inactive") + ".", () => {});
		}
		
		
		// **************************************
		// Callbacks & similar ( gameplay )
		// **************************************

		
		// **************************
		// COROUTINES
		// **************************
		protected IEnumerator WaitForVAMSPlugins()
		{
			// Checking VAMExcite plugins every one second while loading or whatever
			while (_allVAMSReady == false)
			{
				// We're loading...
				// The reason behind that is that, a plugin looking for other atoms can load fast enough
				// to catch the storables list in the target atom PRIOR to the plugins storables being even present in the List 
				// which would result in a "false positive" of an Empty atom not owning a VAMS plugin
				if (_isSceneLoading)
				{
					yield return new WaitForSeconds(_vamsReadyCheckPeriod);					
				}
				else
				{
					bool allReady = true;
					foreach (Atom at in SuperController.singleton.GetAtoms())
					{
						if (at.type == "Empty" && at != containingAtom)
						{
							foreach (string stId in at.GetStorableIDs())
							{
								if (stId.Contains(".VAMStory"))
								{
									JSONStorable tmpVAMS = at.GetStorableByID(stId);
									if (tmpVAMS != null && tmpVAMS.GetBoolParamValue("VAMSValue_Ready") == false) allReady = false;
								}
							}
						}
					}

					if (allReady == false)
					{
						yield return new WaitForSeconds(_vamsReadyCheckPeriod);
					}
					else
					{
						_allVAMSReady = true;
						yield break;
					}
				}
			}
		}
		
		public IEnumerator WaitForNextTrigger(int triggerId)
		{
			yield return new WaitForSeconds(storyTriggers[triggerId].TriggerDelay.val);
			DoExecuteTriggerId(triggerId);
		}
		
		public IEnumerator AutoBackup()
		{
			while (true)
			{
				yield return new WaitForSeconds(1800);
				SaveBackup();
			}
		}
		
		// **************************
		// BROADCASTS SEND/RECEIVE
		// **************************


		// **************************
		// OVERRIDES SAVING/LOADING
		// **************************
		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			var json = base.GetJSON(includePhysical, includeAppearance, forceStore);
			if (includePhysical || forceStore)
			{
				needsStore = true;
				//json[GlobalOptSliderTrigger.Name] = GlobalOptSliderTrigger.GetJSON(base.subScenePrefix);
				
				foreach( var act in storyTriggers ) {
					json[act.triggerAction.Name] = act.triggerAction.GetJSON(base.subScenePrefix);
					json[act.triggerActionDesktop.Name] = act.triggerActionDesktop.GetJSON(base.subScenePrefix);
					json[act.triggerActionVR.Name] = act.triggerActionVR.GetJSON(base.subScenePrefix);
					json[act.triggerActionGateInactive.Name] = act.triggerActionGateInactive.GetJSON(base.subScenePrefix);
					json[act.triggerActionGateActive.Name] = act.triggerActionGateActive.GetJSON(base.subScenePrefix);
				}
			}
			return json;
		}
		
		public override void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, JSONArray presetAtoms = null, bool setMissingToDefault = true)
		{
			base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, presetAtoms);
			// Saving the original datas to load all dynamic storables
			directorDatas = jc;
			
			// We flag the restore process as occured for the init
			dataLoaded = true;
			
			// Loading the datas
			LoadTriggers();
			
			// Updating trigger UI
			UpdateLocalUIFromActiveTrigger();
		}
		
		// **************************
		// Local Tools
		// **************************
		private void logDebug( string debugText ) {
			SuperController.LogMessage( debugText );
		}
		
		private void logError( string debugText ) {
			SuperController.LogError( debugText );
		}
		
		private string GetCleanAndUniqueTriggerName()
		{
			bool nameExists = true;
			int triggerId = activeTrigger.TriggerID;
			int newId = 0;
			string str = CleanString(TriggerNameInputFieldUI.text);
			bool nameFound = false;
			while (nameFound == false)
			{
				str = str + ( newId > 0 ? ""+newId : "");
				int existsCount = 0;
				foreach (var idsNames in triggersIdsToNames)
				{
					if (idsNames.Key == triggerId) continue; // we skip the same id (or the thing will try to find another unique name when stopping on it)
					if (idsNames.Value == str)
					{
						existsCount++;
					}
				}

				if (existsCount == 0) nameFound = true;

				newId++;
				// Security ?
				if (newId > 500)
				{
					logDebug("VAMStoryDirector: Did over 500 loops to find a unique trigger name, is there a problem?");
					break;
				}
			}
			
			return str;
		}
		
		private string GetCleanFilenameFromString(string str)
		{
			string tmp = CleanString(str);
			tmp = tmp.ToLower();
			return tmp;
		}
		
		private string CleanString(string str)
		{
			str = str.Trim();
			Regex reg1 = new Regex("[^a-zA-Z0-9\\s\\-]");
			Regex reg2 = new Regex("\\s+");
			str = reg1.Replace(str, "");
			str = reg2.Replace(str, "-");
			return str;
		}

		// **************************
		// Time to cleanup !
		// **************************
		void OnDestroy() {
			// Sync names
			SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;
			
			// Stop backup routine
			StopCoroutine("AutoBackup");
			
			// Cleaning triggers
			foreach( var tri in storyTriggers ) {
				tri.triggerAction.Remove();
				tri.triggerActionDesktop.Remove();
				tri.triggerActionVR.Remove();
				tri.triggerActionGateInactive.Remove();
				tri.triggerActionGateActive.Remove();
			}
		}

		// ***********************************************************
		// EXTERNAL TOOL - Thank you great coders for your content!
		// ***********************************************************

		// *********** Chokapi Decal Maker START *********************
		public Texture2D GetResource(string path)
		{
			if (path == "" || path == null || !FileManagerSecure.FileExists(path) )
			{
				path = GetPackagePath(this) + "Custom/Scripts/VAMStory/default.png";
			}
			Texture2D texture;
			byte[] tmppng = FileManagerSecure.ReadAllBytes(path);
			texture = new Texture2D(1, 1, TextureFormat.RGBA32, false);
			texture.LoadImage(tmppng);
			return texture;
		}
		// *********** Chokapi Decal Maker END *********************
		
		// *********** MacGruber_Utils.cs START *********************
		// Get directory path where the plugin is located. Based on Alazi's & VAMDeluxe's method.
		public static string GetPluginPath(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
			string filename = self.manager.GetJSON()["plugins"][id].Value;
			return filename.Substring(0, filename.LastIndexOfAny(new char[] { '/', '\\' }));
		}
				
		// Get path prefix of the package that contains our plugin.
		public static string GetPackagePath(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
			string filename = self.manager.GetJSON()["plugins"][id].Value;
			int idx = filename.IndexOf(":/");
			if (idx >= 0)
				return filename.Substring(0, idx+2);
			else
				return string.Empty;
		}
				
		// Check if our plugin is running from inside a package
		public static bool IsInPackage(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
			string filename = self.manager.GetJSON()["plugins"][id].Value;
			return filename.IndexOf(":/") >= 0;
		}
		
		// ===========================================================================================
		// TriggerHandler implementation for easier handling of custom triggers.
		// Essentially call this in your plugin init code:
		//     StartCoroutine(SimpleTriggerHandler.LoadAssets());
		//
		// Credit to AcidBubbles for figuring out how to do custom triggers.
		//
		// hazmhox note : Edited a bit to allow the use of transitions ( original code coming from LogicBricks 4 )
		public class SimpleTriggerHandler : TriggerHandler
		{
			public static bool Loaded { get; private set; }

			private static SimpleTriggerHandler myInstance;

			private RectTransform myTriggerActionsPrefab;
			private RectTransform myTriggerActionMiniPrefab;
			private RectTransform myTriggerActionDiscretePrefab;
			private RectTransform myTriggerActionTransitionPrefab;

			public static SimpleTriggerHandler Instance {
				get {
					if (myInstance == null)
						myInstance = new SimpleTriggerHandler();
					return myInstance;
				}
			}

			public static void LoadAssets()
			{
				SuperController.singleton.StartCoroutine(Instance.LoadAssetsInternal());
			}

			private IEnumerator LoadAssetsInternal()
			{
				foreach (var x in LoadAsset("z_ui2", "TriggerActionsPanel", p => myTriggerActionsPrefab = p))
					yield return x;
				foreach (var x in LoadAsset("z_ui2", "TriggerActionMiniPanel", p => myTriggerActionMiniPrefab = p))
					yield return x;
				foreach (var x in LoadAsset("z_ui2", "TriggerActionDiscretePanel", p => myTriggerActionDiscretePrefab = p))
					yield return x;
				foreach (var x in LoadAsset("z_ui2", "TriggerActionTransitionPanel", p => myTriggerActionTransitionPrefab = p))
					yield return x;

				Loaded = true;
			}

			private IEnumerable LoadAsset(string assetBundleName, string assetName, Action<RectTransform> assign)
			{
				AssetBundleLoadAssetOperation request = AssetBundleManager.LoadAssetAsync(assetBundleName, assetName, typeof(GameObject));
				if (request == null)
					throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null request.");
				yield return request;
				GameObject go = request.GetAsset<GameObject>();
				if (go == null)
					throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null GameObject.");
				RectTransform prefab = go.GetComponent<RectTransform>();
				if (prefab == null)
					throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null RectTansform.");
				assign(prefab);
			}


			void TriggerHandler.RemoveTrigger(Trigger t)
			{
				// nothing to do
			}

			void TriggerHandler.DuplicateTrigger(Trigger t)
			{
				throw new NotImplementedException();
			}

			RectTransform TriggerHandler.CreateTriggerActionsUI()
			{
				return UnityEngine.Object.Instantiate(myTriggerActionsPrefab);
			}

			RectTransform TriggerHandler.CreateTriggerActionMiniUI()
			{
				return UnityEngine.Object.Instantiate(myTriggerActionMiniPrefab);
			}

			RectTransform TriggerHandler.CreateTriggerActionDiscreteUI()
			{
				return UnityEngine.Object.Instantiate(myTriggerActionDiscretePrefab);
			}

			RectTransform TriggerHandler.CreateTriggerActionTransitionUI()
			{
				return UnityEngine.Object.Instantiate(myTriggerActionTransitionPrefab);
			}

			void TriggerHandler.RemoveTriggerActionUI(RectTransform rt)
			{
				UnityEngine.Object.Destroy(rt?.gameObject);
			}
		}

		// Wrapper for easier handling of custom triggers.
		public class EventTrigger : Trigger
		{
			public string Name {
				get { return name; }
				set { name = value; myNeedInit = true; }
			}

			public MVRScript Owner {
				get; private set;
			}

			private string name;
			private bool myNeedInit = true;

			public EventTrigger(MVRScript owner, string name)
			{
				Name = name;
				Owner = owner;
				handler = SimpleTriggerHandler.Instance;
			}

			public void OpenPanelActionStart()
			{
				if (!SimpleTriggerHandler.Loaded)
				{
					SuperController.LogError("EventTrigger: You need to call SimpleTriggerHandler.LoadAssets() before use.");
					return;
				}

				triggerActionsParent = Owner.UITransform;
				InitTriggerUI();
				OpenTriggerActionsPanel();
				if (myNeedInit)
				{
					triggerActionsPanel.Find("Panel/Header Text").GetComponent<Text>().text = Name;
					triggerActionsPanel.Find("Content/Tab1/Label").GetComponent<Text>().text = "Event Actions";
					triggerActionsPanel.Find("Content/Tab2").gameObject.SetActive(false);
					triggerActionsPanel.Find("Content/Tab3").gameObject.SetActive(false);
					myNeedInit = false;
				}
			}
			
			public void OpenPanelTransition()
			{
				if (!SimpleTriggerHandler.Loaded)
				{
					SuperController.LogError("EventTrigger: You need to call SimpleTriggerHandler.LoadAssets() before use.");
					return;
				}

				triggerActionsParent = Owner.UITransform;
				InitTriggerUI();
				OpenTriggerActionsPanel();
				
				//getRoot(triggerActionsParent,0);
				if (myNeedInit)
				{
					triggerActionsPanel.Find("Panel/Header Text").GetComponent<Text>().text = Name;
					triggerActionsPanel.Find("Content/Tab2/Label").GetComponent<Text>().text = "Frequencies Actions";
					triggerActionsPanel.Find("Content/Tab1").gameObject.SetActive(false);
					triggerActionsPanel.Find("Content/Tab2").gameObject.SetActive(true);
					triggerActionsPanel.Find("Content/Tab2").GetComponent<Toggle>().isOn = true;					
					triggerActionsPanel.Find("Content/Tab3").gameObject.SetActive(false);
					myNeedInit = false;
				}
			}
	
			public void RestoreFromJSON(JSONClass jc, string subScenePrefix, bool isMerge, bool setMissingToDefault)
			{
				if (jc.HasKey(Name))
				{
					JSONClass tc = jc[Name].AsObject;
					if (tc != null)
						base.RestoreFromJSON(tc, subScenePrefix, isMerge);
				}
				else if (setMissingToDefault)
				{
					base.RestoreFromJSON(new JSONClass());
				}
			}

			public void Trigger()
			{
				active = true;
				active = false;
			}
			
			public void TriggerTransition(float val)
			{
				this.transitionInterpValue = val;
			}
		}
		// *********** MacGruber_Utils.cs END *********************
	}
}