using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using System;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.IO;
using MeshVR;
using MVR.FileManagementSecure;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;
using MVR;

// VAMStory Director Story Trigger
//
// Director for VAMStory

namespace VAMStoryDirectorPlugin
{
	public partial class VAMStoryDirector : MVRScript
    {
	    public class StoryTrigger
	    {
		    private VAMStoryDirector _VAMSD;

		    public int TriggerID;
		    public bool TriggerGateActive = false;

		    public JSONStorableString TriggerName;
		    public EventTrigger triggerAction;
		    public EventTrigger triggerActionDesktop;
		    public EventTrigger triggerActionVR;
		    public EventTrigger triggerActionGateInactive;
		    public EventTrigger triggerActionGateActive;
		    public JSONStorableFloat TriggerDelay;
			
			protected InputField TriggerNameInputField;
			
		    protected List<UIDynamic> triggersComponentsUI = new List<UIDynamic>();

		    protected JSONClass directorDatas { get; set; }
		    
		    public StoryTrigger(VAMStoryDirector rootScript, int triggerId)
		    {
			    _VAMSD = rootScript;
			    directorDatas = _VAMSD.directorDatas;

			    TriggerID = triggerId;
				    
			    CreateTriggerComponents();
		    }

		    protected void CreateTriggerComponents()
		    {
			    // Temporary vars
			    UIDynamicTextField tmpTextfield;
			    UIDynamicPopup tmpPopup;
			    UIDynamicSlider tmpSlider;
			    UIDynamicColorPicker tmpColor;
			    UIDynamicButton tmpButton;
			    UIDynamicToggle tmpToggle;
			    UIDynamic tmpSpacer;
			    
			    // Creating Trigger components
			    // ******* TRIGGER TEXT ***********
				TriggerName = new JSONStorableString(SAVE_PREFIX + TriggerID + "-Name", "_default_");
				TriggerName.val = "" + TriggerID;

				// ******* TRIGGERS ***********
				triggerAction = new EventTrigger(_VAMSD, SAVE_PREFIX + TriggerID + "-Trigger");
				triggerActionDesktop = new EventTrigger(_VAMSD, SAVE_PREFIX + TriggerID + "-TriggerDesktop");
				triggerActionVR = new EventTrigger(_VAMSD, SAVE_PREFIX + TriggerID + "-TriggerVR");
				triggerActionGateInactive = new EventTrigger(_VAMSD, SAVE_PREFIX + TriggerID + "-TriggerGateInactive");
				triggerActionGateActive = new EventTrigger(_VAMSD, SAVE_PREFIX + TriggerID + "-TriggerGateActive");
				
				// ******* TRIGGER START DELAY ***********
				TriggerDelay = new JSONStorableFloat(SAVE_PREFIX + TriggerID + "-Delay", 0f, 0f, 5f, false);
				
				// Registering everything
				RegisterStorables();
				
				// Restoring datas
				if (directorDatas != null)
				{
					TriggerName.RestoreFromJSON( directorDatas, true, true, false);
					triggerAction.RestoreFromJSON( directorDatas, _VAMSD.subScenePrefix, _VAMSD.mergeRestore, true );
					triggerActionDesktop.RestoreFromJSON( directorDatas, _VAMSD.subScenePrefix, _VAMSD.mergeRestore, true );
					triggerActionVR.RestoreFromJSON( directorDatas, _VAMSD.subScenePrefix, _VAMSD.mergeRestore, true );
					triggerActionGateInactive.RestoreFromJSON( directorDatas, _VAMSD.subScenePrefix, _VAMSD.mergeRestore, true );
					triggerActionGateActive.RestoreFromJSON( directorDatas, _VAMSD.subScenePrefix, _VAMSD.mergeRestore, true );
					TriggerDelay.RestoreFromJSON( directorDatas, true, true, false);
				}
		    }

		    public void UpdateTriggerName(string val)
		    {
			    TriggerNameInputField.text = val;
		    }
		    
		    public void ToggleTriggerUI( bool enabled )
		    {
			    foreach( UIDynamic uidElem in triggersComponentsUI )
			    {
				    _VAMSD.ToggleUIDynamic(enabled, uidElem);
			    }
		    }

		    public void UpdateID(int newTriggerId)
		    {
			    TriggerID = newTriggerId;

			    // Updating trigger components (param names for the save)
			    UpdateComponentsIDs();
		    }
		    
		    protected void UpdateComponentsIDs()
		    {
			    // Deregister
			    DeregisterStorables();

			    // Update
			    TriggerName.name = SAVE_PREFIX + TriggerID + "-Name";
			    triggerAction.Name = SAVE_PREFIX + TriggerID + "-Trigger";
			    triggerActionDesktop.Name = SAVE_PREFIX + TriggerID + "-TriggerDesktop";
			    triggerActionVR.Name = SAVE_PREFIX + TriggerID + "-TriggerVR";
			    triggerActionGateInactive.Name = SAVE_PREFIX + TriggerID + "-TriggerGateInactive";
			    triggerActionGateActive.Name = SAVE_PREFIX + TriggerID + "-TriggerGateActive";
			    TriggerDelay.name = SAVE_PREFIX + TriggerID + "-Delay";
			    
			    // Register
			    RegisterStorables();
		    }
		    
			public void Delete() {
				// Unregistering vars
				DeregisterStorables();

				// Removing triggers
				triggerAction.Remove();
				triggerActionDesktop.Remove();
				triggerActionVR.Remove();
				triggerActionGateInactive.Remove();
				triggerActionGateActive.Remove();
				
				// Clearing data from the loaded data of the scene (to prevent re-restore after deleting)
				// This only happens when you load a scene and have data loaded from the json
				if (directorDatas != null)
				{
					directorDatas.Remove(SAVE_PREFIX + TriggerID + "-Name");
					directorDatas.Remove(SAVE_PREFIX + TriggerID + "-Trigger");
					directorDatas.Remove(SAVE_PREFIX + TriggerID + "-TriggerDesktop");
					directorDatas.Remove(SAVE_PREFIX + TriggerID + "-TriggerVR");
					directorDatas.Remove(SAVE_PREFIX + TriggerID + "-TriggerGateInactive");
					directorDatas.Remove(SAVE_PREFIX + TriggerID + "-TriggerGateActive");
				}
			}

			protected void DeregisterStorables()
			{
				_VAMSD.DeregisterString(TriggerName);
				_VAMSD.DeregisterFloat(TriggerDelay);
			}
			
			protected void RegisterStorables()
			{
				_VAMSD.RegisterString(TriggerName);
				_VAMSD.RegisterFloat(TriggerDelay);
			}
	    }
    }
}