﻿using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using UnityEngine.Events;
using System;
using System.Text;
using System.Collections;
using System.Collections.Generic;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;
using SimpleJSON;

namespace VAMStory
{
    public static class Utils
    {
	    public static Transform GetChildNamed( Transform parent, string childName ) {
		    foreach( Transform child in parent )
		    {
			    if( child.name == childName ) {
				    return child;
			    } else {
				    Transform childSearch = GetChildNamed(child,childName);
				    if (childSearch != null) return childSearch;
			    }
		    }
        
		    return null;
	    }
	    
	    public static UIDynamicTextField CreateStaticDescriptionText(MVRScript script, string DescTitle, string DescText, bool rightSide, int fieldHeight, TextAnchor textAlignment = TextAnchor.UpperLeft, bool disableBackground = true, bool disableScroll = true ) {
		    JSONStorableString staticDescString = new JSONStorableString(DescTitle,DescText) {isStorable=false,isRestorable=false};
		    staticDescString.hidden = true;
		    UIDynamicTextField staticDescStringField = script.CreateTextField(staticDescString, rightSide);
		    if( disableBackground ) staticDescStringField.backgroundColor = new Color(1f, 1f, 1f, 0f);
		    staticDescStringField.UItext.alignment = textAlignment;
		    LayoutElement sdsfLayout = staticDescStringField.GetComponent<LayoutElement>();
		    sdsfLayout.preferredHeight = sdsfLayout.minHeight = fieldHeight;
		    staticDescStringField.height = fieldHeight;
		    if( disableScroll ) Utils.DisableScrollOnText(staticDescStringField);
		    return staticDescStringField;
	    }
	    
	    public static void SetupTextField( UIDynamicTextField target, float fieldHeight, bool disableBackground = false, bool disableScroll = true ) {
			if( disableBackground ) target.backgroundColor = new Color(1f, 1f, 1f, 0f);
			LayoutElement tfLayout = target.GetComponent<LayoutElement>();
			tfLayout.preferredHeight = tfLayout.minHeight = fieldHeight;
			target.height = fieldHeight;
			if( disableScroll ) DisableScrollOnText(target);
		}
	    
	    public static void DisableScrollOnText(UIDynamicTextField target) {
		    ScrollRect targetSR = target.UItext.transform.parent.transform.parent.transform.parent.GetComponent<ScrollRect>();
		    if( targetSR != null ) {
			    targetSR.horizontal = false;
			    targetSR.vertical = false;
		    }
	    }
	    
	    public static void ScrollToTop(UIDynamicTextField target) {
		    ScrollRect targetSR = target.UItext.transform.parent.transform.parent.transform.parent.GetComponent<ScrollRect>();
		    if( targetSR != null ) {
			    targetSR.normalizedPosition = new Vector2(0, 1);
		    }
	    }

	    public static void SetupSpacer(UIDynamic target, float newHeight)
		{
			LayoutElement tmpLE = target.GetComponent<LayoutElement>();
			tmpLE.minHeight = newHeight;
			tmpLE.preferredHeight = newHeight;
		}

	    public static void SetupButtonWithLayout(UIDynamicButton target, float newSize)
		{
			ContentSizeFitter tmpCSF = target.button.transform.gameObject.AddComponent<ContentSizeFitter>();
			tmpCSF.horizontalFit = UnityEngine.UI.ContentSizeFitter.FitMode.MinSize;
			LayoutElement tmpLE = target.button.transform.gameObject.GetComponent<LayoutElement>();
			RectTransform tmpRectT = target.button.transform.GetComponent<RectTransform>();
			Rect tmpRect = tmpRectT.rect;
			tmpRectT.pivot = new Vector2(0f, 0.5f);
			tmpLE.minWidth = newSize;
			tmpLE.preferredWidth = newSize;
		}
		
	    public static void SetupButtonWithoutLayout(UIDynamicButton target, float newSize, Vector2 newPosition)
		{
			ContentSizeFitter tmpCSF = target.button.transform.gameObject.AddComponent<ContentSizeFitter>();
			tmpCSF.horizontalFit = UnityEngine.UI.ContentSizeFitter.FitMode.MinSize;
			LayoutElement tmpLE = target.button.transform.gameObject.GetComponent<LayoutElement>();
			RectTransform tmpRectT = target.button.transform.GetComponent<RectTransform>();
			Rect tmpRect = tmpRectT.rect;
			tmpRectT.pivot = new Vector2(0f, 0.5f);
			tmpLE.minWidth = newSize;
			tmpLE.preferredWidth = newSize;
			tmpLE.ignoreLayout = true;
			tmpRectT.anchoredPosition = newPosition;
		}

	    public static void SetButtonColor(UIDynamicButton targetBtn, Color newColor)
		{
			targetBtn.buttonColor = newColor;
		}
		
	    public static void SetButtonTextColor(UIDynamicButton targetBtn, Color newColor)
		{
			targetBtn.textColor = newColor;
		}
		
	    public static void UpdateButtonPositionWithoutLayout(UIDynamicButton target, Vector2 newPosition)
		{
			RectTransform tmpRectT = target.button.transform.GetComponent<RectTransform>();
			tmpRectT.anchoredPosition = newPosition;
		}
		
		// ***********************************************************
		// EXTERNAL TOOLS - Thank you great coders for your content!
		// ***********************************************************
		
		// *********** MacGruber_Utils.cs START *********************
		// Get directory path where the plugin is located. Based on Alazi's & VAMDeluxe's method.
		public static string GetPluginPath(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
			string filename = self.manager.GetJSON()["plugins"][id].Value;
			return filename.Substring(0, filename.LastIndexOfAny(new char[] { '/', '\\' }));
		}
				
		// Get path prefix of the package that contains our plugin.
		public static string GetPackagePath(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
			string filename = self.manager.GetJSON()["plugins"][id].Value;
			int idx = filename.IndexOf(":/");
			if (idx >= 0)
				return filename.Substring(0, idx+2);
			else
				return string.Empty;
		}
				
		// Check if our plugin is running from inside a package
		public static bool IsInPackage(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
			string filename = self.manager.GetJSON()["plugins"][id].Value;
			return filename.IndexOf(":/") >= 0;
		}	
		
		// Thanks : https://forum.unity.com/threads/re-map-a-number-from-one-range-to-another.119437/
		public static float Remap(float source, float sourceFrom, float sourceTo, float targetFrom, float targetTo)
		{
			return targetFrom + (source-sourceFrom)*(targetTo-targetFrom)/(sourceTo-sourceFrom);
		}
	}
    
    class PathExt
    {
        //  Because we can't use System.IO.Path ........

        private static readonly char DirectorySeparatorChar = '\\';
        private const string DirectorySeparatorCharAsString = "\\";
        private static readonly char AltDirectorySeparatorChar = '/';
        private static readonly char VolumeSeparatorChar = ':';

        private static void CheckInvalidPathChars(string path, bool checkAdditional = false)
        {
            if (path == null)
                throw new ArgumentNullException("path");
        }

        public static String GetFileName(String path)
        {
            if (path != null)
            {
                CheckInvalidPathChars(path);

                int length = path.Length;
                for (int i = length; --i >= 0;)
                {
                    char ch = path[i];
                    if (ch == DirectorySeparatorChar || ch == AltDirectorySeparatorChar || ch == VolumeSeparatorChar)
                        return path.Substring(i + 1, length - i - 1);

                }
            }
            return path;
        }

        public static String GetFileNameWithoutExtension(String path)
        {
            path = GetFileName(path);
            if (path != null)
            {
                int i;
                if ((i = path.LastIndexOf('.')) == -1)
                    return path; // No path extension found
                else
                    return path.Substring(0, i);
            }
            return null;
        }

        public static String GetExtension(String path)
        {
            if (path == null)
                return null;

            CheckInvalidPathChars(path);
            int length = path.Length;
            for (int i = length; --i >= 0;)
            {
                char ch = path[i];
                if (ch == '.')
                {
                    if (i != length - 1)
                        return path.Substring(i, length - i);
                    else
                        return String.Empty;
                }
                if (ch == DirectorySeparatorChar || ch == AltDirectorySeparatorChar || ch == VolumeSeparatorChar)
                    break;
            }
            return String.Empty;
        }

        public string FirstLetterToUpper(string str)
        {
            if (str == null)
                return null;

            if (str.Length > 1)
                return char.ToUpper(str[0]) + str.Substring(1);

            return str.ToUpper();
        }
    }
}
