using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using System;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.IO;
using MeshVR;
using MVR.FileManagementSecure;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;
using MVR;

// VAMStory - VAMSColoredBox v1.0
//
// A handy colored box to start your scene properly

namespace VAMStory
{
    public class VAMSFovController : MVRScript
    {	
		// Static string
		private static string PLUGIN_PATH;
		private static string PLUGIN_ROOT_PATH;
		private static string UI_BUNDLE_PATH = "/vamstory_assets.uibundle";
		private static string EXCEPTION_PREFIX = "VAMSColoredBox - Exception caught: ";

		// JSONStorables
		protected JSONStorableColor BoxColor;
		
		// Game Objects and components
		private GameObject VAMSCBGO;
		private Transform TRColoredBox;
		private Renderer RENColoredBox;
		
		// Assetbundle
		private Request bundleRequest = null;
		
		// Misc

			
		public override void Init()
        {			 
            try
            {
	            PLUGIN_PATH = Utils.GetPluginPath(this);
	            PLUGIN_ROOT_PATH = PLUGIN_PATH.Replace("/tools", "");
	            
				
				Utils.CreateStaticDescriptionText(this,"VAMSColoredBoxTitle","<color=#000><size=35><b>VAMSColoredBox</b></size></color>",true,40,TextAnchor.MiddleRight);
				Utils.CreateStaticDescriptionText(this,"Right UI Main Title","<color=#000><size=28>A handy colored box to start your scene properly</size></color>",true,135,TextAnchor.MiddleRight);

				// Temporary vars
				HSVColor hsvcDefault = HSVColorPicker.RGBToHSV(0f, 0f, 0f);
				UIDynamicTextField tmpTextfield;
				UIDynamicPopup tmpPopup;
				UIDynamicSlider tmpSlider;
				UIDynamicColorPicker tmpColor;
				UIDynamicButton tmpButton;
				UIDynamicToggle tmpToggle;
				UIDynamic tmpSpacer;
				
				BoxColor = new JSONStorableColor("Box_Color", hsvcDefault);
				BoxColor.setJSONCallbackFunction += (val) => { OnBoxColorChanged(); };
				tmpColor = CreateColorPicker(BoxColor, true);
				tmpColor.label = "Box Color";
				RegisterColor(BoxColor);
				
				// Loading the assetbundle
				Request request = new AssetLoader.AssetBundleFromFileRequest {path = PLUGIN_ROOT_PATH + UI_BUNDLE_PATH, callback = OnBundleLoaded};
				AssetLoader.QueueLoadAssetBundleFromFile(request);
				
            }
            catch(Exception e)
            {
                SuperController.LogError(EXCEPTION_PREFIX + e);
            }
		}
		
		void Start()
        {
			try
            {

			}
            catch(Exception e)
            {
                SuperController.LogError(EXCEPTION_PREFIX + e);
            }
        }
		
		void Update() {

		}
		
		// **************************
		// Callbacks & similar
		// **************************
		private void OnBundleLoaded(Request aRequest)
		{
			try
			{
				// Reading bundle
				bundleRequest = aRequest;
				GameObject tmp = bundleRequest.assetBundle.LoadAsset<GameObject>("Assets/VAMStory/prefabs/VAMStoryColoredBox.prefab");
				VAMSCBGO = Instantiate(tmp);
				
				// Parenting to my current Empty
				VAMSCBGO.transform.SetParent(containingAtom.reParentObject.GetChild(0).transform, false);
				VAMSCBGO.transform.localPosition = new Vector3(0, 0, 0);
				
				// Getting colored box for the material
				TRColoredBox = Utils.GetChildNamed(VAMSCBGO.transform, "colored_box");
				RENColoredBox = TRColoredBox.GetComponent<Renderer>();
				
				// Everything is ok, let's change the saved color
				OnBoxColorChanged();
			}
			catch(Exception e)
			{
				SuperController.LogError(EXCEPTION_PREFIX + e);
			}
		}

		private void OnBoxColorChanged()
		{
			if (RENColoredBox != null)
			{
				// Changing color
				Color newColor = new Color(BoxColor.colorPicker.currentColor.r, BoxColor.colorPicker.currentColor.g, BoxColor.colorPicker.currentColor.b, 1f);
				RENColoredBox.material.SetColor("_Color", newColor);
			}
		}

		// **************************
		// Local Tools
		// **************************
		private void logDebug( string debugText ) {
			SuperController.LogMessage( debugText );
		}
				
		// **************************
		// Time to cleanup !
		// **************************
		void OnDestroy()
		{
			// Cleaning GameObjects
			Destroy(VAMSCBGO);

			// Unloading assetbundle
			AssetLoader.DoneWithAssetBundleFromFile(PLUGIN_ROOT_PATH + UI_BUNDLE_PATH);
		}
    }
}