using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using System;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.IO;
using MeshVR;

// VAMStory - VAMSFovController v1.0
//
// Controls the FOV of the main camera

namespace VAMStory
{
    public class VAMSFovController : MVRScript
    {	
		// Static string
		public static string PLUGIN_PATH;
		public static string EXCEPTION_PREFIX = "VAMSFovController - Exception caught: ";

		// JSONStorables
		public JSONStorableFloat FOVValue;
		public JSONStorableFloat A_ForceFovValue;
		public JSONStorableAction A_DisableFov;
		public JSONStorableAction A_EnableFov;
		
		// Game Objects and components
		
		// Misc
		protected float _originalFOV = 0f;
		protected bool _fovEnabled = true;
			
		public override void Init()
        {			 
            try
            {
	            PLUGIN_PATH = Utils.GetPluginPath(this);
	            
	            _originalFOV = SuperController.singleton.monitorCameraFOV;
                FOVValue = new JSONStorableFloat("FOV Value", _originalFOV, FOVValueCallback, 1f, 100f){ isStorable = false };
				CreateSlider(FOVValue);
				
				Utils.CreateStaticDescriptionText(this,"VAMSFovControllerTitle","<color=#000><size=35><b>VAMSFovController</b></size></color>",true,40,TextAnchor.MiddleRight);
				Utils.CreateStaticDescriptionText(this,"Right UI Main Title","<color=#000><size=28>Allows you to control the FOV of the main camera. Can be used to enforce a FOV for a desktop scene, or animate FOV for a cinematic scene.</size></color>",true,135,TextAnchor.MiddleRight);

				A_ForceFovValue = new JSONStorableFloat("Force FOV Value", _originalFOV, ForceFOVValueCallback, 1f, 100f){ isStorable = false };
				A_DisableFov = new JSONStorableAction("Disable FOV", () => { _fovEnabled = false; });
				A_EnableFov = new JSONStorableAction("Enable FOV", () => { _fovEnabled = true; });
				
				RegisterFloat(FOVValue);
				RegisterFloat(A_ForceFovValue);
				RegisterAction(A_DisableFov);
				RegisterAction(A_EnableFov);
            }
            catch(Exception e)
            {
                SuperController.LogError(EXCEPTION_PREFIX + e);
            }
		}
		
		void Start()
        {
			try
            {

			}
            catch(Exception e)
            {
                SuperController.LogError(EXCEPTION_PREFIX + e);
            }
        }
		
		void Update() {

		}

		protected void SetFovValue(float fovVal)
		{
			SuperController.singleton.monitorCameraFOV = fovVal;
		}
		
		// **************************
		// Callbacks & similar
		// **************************
		private void FOVValueCallback(float FOVValue)
		{
			if (_fovEnabled == false) return;
			SetFovValue(FOVValue);
		}

		private void ForceFOVValueCallback(float FOVValue)
		{
			SetFovValue(FOVValue);
		}
		
		// **************************
		// Local Tools
		// **************************
		private void logDebug( string debugText ) {
			SuperController.LogMessage( debugText );
		}
				
		// **************************
		// Time to cleanup !
		// **************************
		void OnDestroy()
		{
			SuperController.singleton.monitorCameraFOV = _originalFOV;
		}

	}
}