﻿using UnityEngine;
using UnityEngine.UI;
using System;
using System.Text;
using System.Text.RegularExpressions;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.IO;
using MeshVR;
using MVR.FileManagementSecure;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;
using MVR;

// VAMStory - VAMSVaTrigger v1.0
//
// Helps trigger VA for person atoms

namespace VAMStory
{
    public class VAMSVATrigger : MVRScript
    {
	    public static string EXCEPTION_PREFIX = "VATrigger - Exception caught: ";
	    
	    /** VA **/
	    private string lastFileExplorerPath = "Custom";
	    private Dictionary<string, string> _vaFiles = new Dictionary<string, string>();
	    private Dictionary<string, AudioClip> _vaAudioclips = new Dictionary<string, AudioClip>();
	    private JSONStorableString VAFolder;
	    protected List<string> vaChoices = new List<string>();
	    
	    protected Color btnGreenBgColor = new Color(0.4f, 0.7f, 0.3f, 1f);
	    protected Color btnGreenTxtColor = new Color(1f, 1f, 1f, 1f);
	    protected Color btnRedBgColor = new Color(0.7f, 0.3f, 0.3f, 1f);
	    protected Color btnRedTxtColor = new Color(1f, 1f, 1f, 1f);
	    
	    public JSONStorableStringChooser A_PlayVAImmediate;
	    
	    public AudioSourceControl headAudioSource;

	    protected bool _isSceneLoading { get { return SuperController.singleton.isLoading; } }
	    protected bool _isSceneFrozen { get { return SuperController.singleton.freezeAnimation; } }
	    protected bool _isVoiceInitialized = false;
	    protected bool _isCorrectAtom = false;

		public override void Init()
        {

			if ( containingAtom.type != "Person" )
			{
				logError("VAMSVATrigger can only be used on a character.");
				return;
			}

			_isCorrectAtom = true;
			
            try
            {
	            headAudioSource = containingAtom.GetStorableByID("HeadAudioSource") as AudioSourceControl;

	            /* *** Temporary UIDynamic *** */
	            UIDynamicTextField tmpTextfield;
	            UIDynamicPopup tmpPopup;
	            UIDynamicSlider tmpSlider;
	            UIDynamicButton tmpButton;
	            UIDynamicToggle tmpToggle;
	            UIDynamic tmpSpacer;
	            
	            /* ********************************* */
	            /* **** VA                      **** */
	            /* ********************************* */
	            Utils.CreateStaticDescriptionText(this,"VAMM2Title","<color=#000><size=35><b>VAMSVATrigger</b></size></color>",false,40,TextAnchor.MiddleLeft);
	            string pluginDesc =
		            "<color=#000><size=30>Allows you to trigger <b>voice acting</b> on a character.\n\n" +
		            "Select a folder containing all voice samples for this character. And start the voice samples using the <b>Play VA Immediate</b> trigger from the plugin." +
		            "</size></color>";
	            Utils.CreateStaticDescriptionText(this,"Right UI Main Title",pluginDesc,false,590,TextAnchor.MiddleLeft);
	            
	            tmpTextfield = Utils.CreateStaticDescriptionText(this,"Global options VA","<color=#000><size=28><b>VA (Voice acting)</b></size></color>\n<size=24>Folder containing all voice lines\nfor the character.</size>",true,95,TextAnchor.MiddleRight);
			
	            VAFolder = new JSONStorableString("VAFolder", "");
	            tmpTextfield = CreateTextField(VAFolder,true);
	            Utils.SetupTextField(tmpTextfield, 100f, false, false);

	            tmpButton = CreateButton("Select folder", true);
	            tmpButton.button.onClick.AddListener(() => { OnSelectVAFolder(); } );
				
	            tmpButton = CreateButton("Sync folder", true);
	            tmpButton.button.onClick.AddListener(() => { OnSyncVAFolder(); } );
	            Utils.SetButtonColor(tmpButton, btnGreenBgColor);
	            Utils.SetButtonTextColor(tmpButton, btnGreenTxtColor);
				
	            tmpButton = CreateButton("Clear VA", true);
	            tmpButton.button.onClick.AddListener(() => { OnClearVAFolder(); } );
	            Utils.SetButtonColor(tmpButton, btnRedBgColor);
	            Utils.SetButtonTextColor(tmpButton, btnRedTxtColor);
				
	            A_PlayVAImmediate = new JSONStorableStringChooser("Play VA Immediate", vaChoices, "", "Play VA Immediate", A_PlayVAImmediateCallback){ isStorable = false, isRestorable=false };
	            
	            RegisterStringChooser(A_PlayVAImmediate);
	            
	            RegisterString(VAFolder);

			}
            catch(Exception e)
            {
                SuperController.LogError("VATrigger - Exception caught: " + e);
            }
		}
		
		void Start() { }
		
		void Update() {
			if (_isSceneLoading || _isSceneFrozen || !_isCorrectAtom) return;
			
			if (!_isSceneLoading && !_isVoiceInitialized)
			{
				// Creating / loading VA Folder
				LoadVA();
					
				// We've done everything! great!
				_isVoiceInitialized = true;
			}
		}

		// **************************
		// FUNCTIONS (Audio related)
		// **************************
		protected void LoadVA()
		{
			// Clearing first
			UnloadAndClearVA();
			
			// Then loading
			foreach (var vaFile in _vaFiles)
			{
				LoadVAAudioClip(vaFile.Key, vaFile.Value);
			}

			UpdateVAChoices();
		}

		protected void UnloadAndClearVA()
		{
			if (_vaAudioclips.Count == 0) return;
			foreach (var clip in _vaAudioclips)
			{
				Destroy(clip.Value);
			}
			_vaAudioclips.Clear();
		}

		protected void LoadVAAudioClip(string vaName, string filePath)
		{
			string nPath = SuperController.singleton.NormalizeLoadPath(filePath);
			if (!_vaAudioclips.ContainsKey(vaName))
			{
				URLAudioClip uac = (URLAudioClip)URLAudioClipManager.singleton.GetClip(nPath);
				if(uac == null) uac = (URLAudioClip)URLAudioClipManager.singleton.QueueClip(nPath);
				if (uac != null)
				{
					StartCoroutine(AddNamedAudioClipToPool(uac,vaName));
				}
			}
		}

		protected bool IsVAFilenameValid(string fileName)
		{
			return Regex.IsMatch(fileName, @"^[0-9a-zA-Z\-_. ]{1,32}$");;
		}
		
		protected void ImportVAFolder( bool clearVA = true )
		{
			if (VAFolder.val != "")
			{
				if (FileManagerSecure.DirectoryExists(VAFolder.val))
				{
					int oldCount = _vaFiles.Count;
					
					// Looking for the files paths
					int errorCount = 0;
					Dictionary<string, string> audioPaths = new Dictionary<String, String>();
					foreach (string file in FileManagerSecure.GetFiles(VAFolder.val,"*.mp3") )
					{
						audioPaths.Add(PathExt.GetFileNameWithoutExtension(file),file);
						if (!IsVAFilenameValid(PathExt.GetFileNameWithoutExtension(file))) errorCount++;
					}
					
					foreach (string file in FileManagerSecure.GetFiles(VAFolder.val,"*.ogg") )
					{
						audioPaths.Add(PathExt.GetFileNameWithoutExtension(file),file);
						if (!IsVAFilenameValid(PathExt.GetFileNameWithoutExtension(file))) errorCount++;
					}
					
					foreach (string file in FileManagerSecure.GetFiles(VAFolder.val,"*.wav") )
					{
						audioPaths.Add(PathExt.GetFileNameWithoutExtension(file),file);
						if (!IsVAFilenameValid(PathExt.GetFileNameWithoutExtension(file))) errorCount++;
					}
					
					// Checking if an error happened in the paths, and stopping the process if it's the case
					if (errorCount>0)
					{
						// Cleaning to be sure
						if (clearVA)
						{
							VAFolder.val = "";
							_vaFiles.Clear();
							UnloadAndClearVA();							
						}
						// Alerting the creator
						SuperController.singleton.Alert("<color=orange>VAMMoan\nVA folder synchronization</color>\nImport failed: "+errorCount+" file"+(errorCount>1?"s are":" is")+" not named properly.\n\n<size=29>Your VA filenames have to follow these rules:\n\nAny letter uppercase or lowercase\nAny number\n - _ . allowed\nWhite spaces allowed\nNo special characters\n1 to 32 characters long</size>", () => {});
						return;
					}
					
					// Loading the files
					_vaFiles.Clear();
					foreach (var audioFile in audioPaths )
					{
						string nPath = SuperController.singleton.NormalizeLoadPath(audioFile.Value);
						if (!_vaFiles.ContainsKey(audioFile.Key))
						{
							_vaFiles.Add(audioFile.Key, nPath);
						}
					}
					
					// End message
					int newCount = _vaFiles.Count;
					int filesAdded = newCount - oldCount;
					string fileChanges = "";
					if (filesAdded >= 0)
					{
						fileChanges = filesAdded + " new file" + (filesAdded > 1 ? "s" : "") + " added.";
					}
					else
					{
						fileChanges = Mathf.Abs(filesAdded) + " file" + (Mathf.Abs(filesAdded) > 1 ? "s" : "") + " removed.";
					}
					SuperController.singleton.Alert("<b><color=orange>VAMMoan\nVA folder synchronization</color></b>\nTotal VA files: " + newCount + ".\n" + fileChanges, () => {});
				}
				else
				{
					logError(EXCEPTION_PREFIX + "VA folder not found (" + VAFolder.val + ")");
				}
			}

			LoadVA();
		}

		protected void UpdateVAChoices()
		{
			vaChoices = new List<string>();
			foreach (var vaFile in _vaFiles)
			{
				vaChoices.Add(vaFile.Key);
			}
			A_PlayVAImmediate.choices = vaChoices;
		}
		
		protected void PlayVAClipImmediate(string clipName)
		{
			AudioClip imClip = null;
			if (_vaAudioclips.ContainsKey(clipName))
			{
				imClip = _vaAudioclips[clipName];
			}
			
			if (imClip == null)
			{
				logError(EXCEPTION_PREFIX + "Voice line not found (" + clipName + ")");
				return;
			}
			
	
			StopAndClearHeadAudioSource();
			PlayAudio(imClip);
		}
		
		protected void StopAndClearHeadAudioSource() {
			// Trying to prevent as much as I can the "pop" when clearing the audio source
			float currentVolume = headAudioSource.GetFloatParamValue("volume");
			headAudioSource.SetFloatParamValue("volume",0f);
			headAudioSource.CallAction("ClearQueue");
			headAudioSource.Stop();
			headAudioSource.SetFloatParamValue("volume",currentVolume);
		}
		
		protected void PlayAudio(AudioClip clip)
		{
			if (clip != null)
			{
				headAudioSource.audioSource.clip = clip;
				headAudioSource.audioSource.Play();
			}
		}
		
		// **************************
		// Callbacks
		// **************************
		protected void OnSelectVAFolder()
		{
			SuperController.singleton.directoryBrowserUI.defaultPath = lastFileExplorerPath;
			SuperController.singleton.directoryBrowserUI.SetTitle("Select a folder containing your voice lines");
			SuperController.singleton.directoryBrowserUI.keepOpen = false;
			SuperController.singleton.directoryBrowserUI.SetTextEntry(true);
			SuperController.singleton.directoryBrowserUI.showInstallFolderInDirectoryList = false;

			List<ShortCut> shortCutList = new List<ShortCut>();
			ShortCut shortAp = new ShortCut();
			shortAp.package = "AddonPackages Filtered";
			shortAp.path = "AddonPackages";
			shortAp.displayName = "Filter: Custom";
			shortAp.packageFilter = "Custom";
			shortCutList.Add(shortAp);
			
			ShortCut shortAp2 = new ShortCut();
			shortAp2.package = "AddonPackages Filtered";
			shortAp2.path = "AddonPackages";
			shortAp2.displayName = "Filter: Saves";
			shortAp2.packageFilter = "Saves";
			shortCutList.Add(shortAp2);
			
			ShortCut shortCutCustom = new ShortCut();
			shortCutCustom.path = "Custom\\";
			shortCutCustom.displayName = "Custom";
			shortCutCustom.flatten = false;
			shortCutList.Add(shortCutCustom);
			
			ShortCut shortCutSaves = new ShortCut();
			shortCutSaves.path = "Saves\\";
			shortCutSaves.displayName = "Saves";
			shortCutSaves.flatten = false;
			shortCutList.Add(shortCutSaves);
			
			SuperController.singleton.directoryBrowserUI.shortCuts = shortCutList;
			SuperController.singleton.directoryBrowserUI.Show((val) =>
			{
				if (val != "")
				{
					lastFileExplorerPath = val.Substring(0, val.LastIndexOfAny(new char[] {'/', '\\'}));
					VAFolder.val = SuperController.singleton.NormalizeLoadPath(val);
					ImportVAFolder();
				}
			});
		}

		protected void OnSyncVAFolder()
		{
			ImportVAFolder(false);
		}

		protected void OnClearVAFolder()
		{
			VAFolder.val = "";
			_vaFiles.Clear();
			UnloadAndClearVA();
			SuperController.singleton.Alert("<b><color=orange>VAMMoan\nVA clearing</color></b>\nAll VA files have been removed.", () => {});
		}
		
		protected void A_PlayVAImmediateCallback(string val)
		{
			PlayVAClipImmediate(val);
			A_PlayVAImmediate.valNoCallback = "";
		}

		// **************************************
		// Coroutines
		// **************************************
		protected IEnumerator AddNamedAudioClipToPool(URLAudioClip uac, string clipFilename)
		{
			// Waiting for URLAudioClip to be ready
			while (!uac.ready)
			{
				yield return new WaitForSeconds(0.1f);
			}

			// Creating my own audioclip to save to the pool
			AudioClip tmpAc = AudioClip.Create("va-" + clipFilename, uac.sourceClip.samples, uac.sourceClip.channels, uac.sourceClip.frequency, false);
			float[] copyAc = new float[uac.sourceClip.samples * uac.sourceClip.channels];
			uac.sourceClip.GetData(copyAc, 0);
			tmpAc.SetData(copyAc, 0);
			
			// Adding to the pool and removing the "Scene audio" clip
			_vaAudioclips.Add(clipFilename, tmpAc);
			URLAudioClipManager.singleton.RemoveClip(uac);
		}

		// **************************
		// OVERRIDES SAVING/LOADING
		// **************************
		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			var json = base.GetJSON(includePhysical, includeAppearance, forceStore);
			if (includePhysical || forceStore)
			{
				needsStore = true;
				
				// Storing VA files ( used to ensure that dependencies are properly caught by the package builder )
				JSONClass VAJson = new JSONClass();
				foreach (var vaFile in _vaFiles)
				{
					VAJson[vaFile.Key] = SuperController.singleton.NormalizeSavePath(vaFile.Value);
				}

				json["VAFiles"] = VAJson;
				json["VAFolder"] = SuperController.singleton.NormalizeSavePath(json["VAFolder"].Value);
			}
			return json;
		}
		
		public override void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, JSONArray presetAtoms = null, bool setMissingToDefault = true)
		{
			base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, presetAtoms);
			
			// Restoring the VA Files
			JSONClass vaFiles = jc["VAFiles"].AsObject;
			foreach (string key in vaFiles.Keys)
			{
				_vaFiles.Add(key, vaFiles[key]);
			}
		}
		
		// **************************
		// Local Tools
		// **************************
		private void logDebug( string debugText ) {
			SuperController.LogMessage( debugText );
		}
		
		private void logError( string debugText ) {
			SuperController.LogError( debugText );
		}
		

		
		// **************************
		// Time to cleanup !
		// **************************
		void OnDestroy() {
			// Cleaning modified values

		}

	}
}