﻿using UnityEngine;
using UnityEngine.UI;
using System;
using System.Text;
using System.Text.RegularExpressions;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.IO;
using MeshVR;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;
using MVR;
// VAMStory - VAMSDistance v1.0
//
// Check the distance between two things and execute triggers when crossing thresholds

namespace VAMStory
{
    public class VAMVFXEditor : MVRScript
    {
	    // Static string
	    public static string PLUGIN_PATH;
	    public static string EXCEPTION_PREFIX = "VAMSVFXEditor - Exception caught: ";
	    
	    // JSONStorables
		public JSONStorableFloat VFXScale;
		public JSONStorableFloat VFXParticleOpacity;
		public JSONStorableFloat VFXParticleSize;
		public JSONStorableColor VFXParticleColor;

		// Lists
		protected List<ParticleSystem> _prtSystems = new List<ParticleSystem>();
		
		// Gameobjects, transforms...
		protected Transform _TRFirstChild;
		
		// Security and initialization
		private bool editorInitialized = false;
		private int initLoops = 0;
		private int maxInitLoops = 36000;
		private bool cuaIsValid = false;

		// Flags
		protected bool _isInit = false;
			
		public override void Init()
        {

			try
            {
	            PLUGIN_PATH = Utils.GetPluginPath(this);

	            
	            /* *** Temporary UIDynamic *** */
	            UIDynamicTextField tmpTextfield;
	            UIDynamicPopup tmpPopup;
	            UIDynamicSlider tmpSlider;
	            UIDynamicButton tmpButton;
	            UIDynamicToggle tmpToggle;
	            UIDynamicColorPicker tmpColor;
	            UIDynamic tmpSpacer;
	            
	            HSVColor hsvcDefault = HSVColorPicker.RGBToHSV(1f, 1f, 1f);
	            
	            /* UI */
	            VFXScale = new JSONStorableFloat("VFXScale", 1f, 0.001f, 1f, false);
	            VFXScale.setCallbackFunction += (val) => { OnChangeVFX(); };
	            tmpSlider = CreateSlider(VFXScale, false);
	            tmpSlider.valueFormat = "F3";
	            tmpSlider.label = "VFX Scale";
	            
				VFXParticleOpacity = new JSONStorableFloat("VFXParticleOpacity", 0.5f, 0f, 1f, true);
				VFXParticleOpacity.setCallbackFunction += (val) => { OnChangeVFX(); };
	            tmpSlider = CreateSlider(VFXParticleOpacity, false);
	            tmpSlider.valueFormat = "F3";
	            tmpSlider.label = "Particles Opacity";
	            
	            VFXParticleSize = new JSONStorableFloat("VFXParticleSize", 0.5f, 0f, 2f, false);
	            VFXParticleSize.setCallbackFunction += (val) => { OnChangeVFX(); };
	            tmpSlider = CreateSlider(VFXParticleSize, false);
	            tmpSlider.valueFormat = "F3";
	            tmpSlider.label = "Particles Size";
	            
	            VFXParticleColor = new JSONStorableColor("VFXParticleColor", hsvcDefault);
	            VFXParticleColor.setJSONCallbackFunction += (val) => { OnChangeVFX(); };
	            tmpColor = CreateColorPicker(VFXParticleColor, false);
	            tmpColor.label = "Particles color";

	            Utils.CreateStaticDescriptionText(this,"VAMSDistanceTitle","<color=#000><size=35><b>VAMSVFXEditor</b></size></color>",true,40,TextAnchor.MiddleRight);
	            string pluginDesc =
		            "<color=#000><size=30>Allows you to edit common VFX (particle systems) available on the hub.\n\n" +
		            "The system is not fail-proof since the plugin can't know how the particle systems are made. You might need to test different settings to get the result you want... and sometimes it might not work for some particle systems.\n\n" +
		            "You can change the scale of the whole system, and the opacity, size and color of the particles." +
		            "</size></color>";
	            Utils.CreateStaticDescriptionText(this,"Right UI Main Title",pluginDesc,true,590,TextAnchor.MiddleRight);

	            /*Utils.CreateStaticDescriptionText(this,"VAMM2Title","<color=#000><size=35><b>VAMSDistance</b></size></color>",true,40,TextAnchor.MiddleRight);
	            string pluginDesc =
		            "<color=#000><size=30>Checks the distance between this object and another. Select the target and set triggers when the thresholds are crossed.\n\n" +
		            "By default the plugin checks the distance between this object and the player (the main camera), but you can change the reference object to compare the distance between this object and any other object in the scene.\n\n" +
		            "The threshold transition trigger allows you to create a smooth transition between <b>Transition lower threshold</b> and <b>threshold</b> value." +
		            "</size></color>";
	            Utils.CreateStaticDescriptionText(this,"Right UI Main Title",pluginDesc,true,590,TextAnchor.MiddleRight);*/
	            
	            /* ***************************************** */
	            /* **** REGISTERING STORABLES/ACTIONS   **** */
	            /* ***************************************** */
	            RegisterFloat(VFXScale);
	            RegisterFloat(VFXParticleOpacity);
	            RegisterFloat(VFXParticleSize);
	            RegisterColor(VFXParticleColor);

            }
            catch(Exception e)
            {
                SuperController.LogError(EXCEPTION_PREFIX + e);
            }
		}
		
		void Start() { }
		
		void Update()
		{
			// If the editor is not initialized, we're gonna try to do it
			// I'm putting 3600 tries ( at 60fps it is 1min )
			if( initLoops < maxInitLoops && editorInitialized == false ) {
				initLoops++;
				// If our CUA is loading and everything is ok, we can remove the loading UI and Init the final UI
				if (InitVFX() == true)
				{
					cuaIsValid = true; // Our CUA is initialized, we consider it as valid
					StartCoroutine("CheckCUAStatus"); // We're gonna check if the CUA is valid every 5 secs

					OnChangeVFX();
				}
			}
			
			// Triggering an error if we have reached our limit of tries
			if( initLoops == maxInitLoops && editorInitialized == false ) {
				initLoops++; // To avoid spamming the log with the error
				SuperController.LogError(EXCEPTION_PREFIX + "Could not initialize the VFX... The CUA might not be valid OR is not a VFX / Particle system.");
			}
		}

		private bool InitVFX()
		{
			if( editorInitialized == true ) return true;
			if (containingAtom.reParentObject.GetChild(0).GetChild(0).childCount == 0) return false;
			try
			{
				_TRFirstChild = containingAtom.reParentObject.GetChild(0).GetChild(0).GetChild(0);
				if (_TRFirstChild != null)
				{
					GetAllParticleSystems(_TRFirstChild);

					if (_prtSystems.Count > 0)
					{
						editorInitialized = true;
						return true;						
					}
					else
					{
						initLoops = maxInitLoops;
					}
				}
			}
			catch(Exception e)
			{
				SuperController.LogError(EXCEPTION_PREFIX + e);
				return false;
			}

			return false;
		}

		// **************************
		// Functions
		// **************************
		protected void GetAllParticleSystems( Transform parent )
		{
			ParticleSystem tmpPrtSys = parent.GetComponent<ParticleSystem>();
			if (tmpPrtSys != null)
			{
				_prtSystems.Add(tmpPrtSys);
			}
			
			foreach (Transform tr in parent)
			{
				tmpPrtSys = tr.GetComponent<ParticleSystem>();
				if (tmpPrtSys != null)
				{
					_prtSystems.Add(tmpPrtSys);
				}

				GetAllParticleSystems(tr);
			}
		}
		
		// **************************
		// Callbacks & similar
		// **************************
		protected void OnChangeVFX()
		{
			if (!editorInitialized) return;
			_TRFirstChild.localScale = new Vector3(VFXScale.val, VFXScale.val, VFXScale.val);
			if (_prtSystems.Count > 0)
			{
				foreach (ParticleSystem ps in _prtSystems)
				{
					ps.transform.localScale = new Vector3(VFXScale.val, VFXScale.val, VFXScale.val);
					var prtMain = ps.main;
					Color currentColor = prtMain.startColor.color;
					currentColor.r = VFXParticleColor.colorPicker.currentColor.r;
					currentColor.g = VFXParticleColor.colorPicker.currentColor.g;
					currentColor.b = VFXParticleColor.colorPicker.currentColor.b;
					currentColor.a = VFXParticleOpacity.val;
					prtMain.startColor = currentColor;
					prtMain.startSize = VFXParticleSize.val;
				}
			}
		}
		
		// **************************
		// COROUTINES
		// **************************
		
		// Routine checking every 5secs if our CUA is still valid, this will invalidate the UI if the user change the CUA in the asset tab
		private IEnumerator CheckCUAStatus()
		{
			while (cuaIsValid)
			{
				// Our object root is null, it means our CUA is invalid
				if (_TRFirstChild == null)
				{
					// We invalidate it so that the coroutine stops by itself, remove the UI and show the UI to alert the user
					cuaIsValid = false;
				}
				yield return new WaitForSeconds(5.0f);
			}
		}

		
		// **************************
		// Local Tools
		// **************************
		private void logDebug( string debugText ) {
			SuperController.LogMessage( debugText );
		}
		
		private void logError( string debugText ) {
			SuperController.LogError( debugText );
		}
		

		
		// **************************
		// Time to cleanup !
		// **************************
		void OnDestroy()
		{

		}
	}
}